/* Generated by Cython 0.28.5 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "name": "pyrobuf_list",
        "sources": [
            "pyrobuf/src\\pyrobuf_list.pyx"
        ]
    },
    "module_name": "pyrobuf_list"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_28_5"
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (0 && PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0; // PyThread_create_key reports success always
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif // TSS (Thread Specific Storage) API
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pyrobuf_list
#define __PYX_HAVE_API__pyrobuf_list
/* Early includes */
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include "pythread.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "pyrobuf\\src\\pyrobuf_list.pyx",
  "stringsource",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_12pyrobuf_list_NullListener;
struct __pyx_obj_12pyrobuf_list_TypedList;
struct __pyx_obj_12pyrobuf_list_BytesList;
struct __pyx_obj_12pyrobuf_list_StringList;
struct __pyx_obj_12pyrobuf_list_DoubleList;
struct __pyx_obj_12pyrobuf_list_FloatList;
struct __pyx_obj_12pyrobuf_list_IntList;
struct __pyx_obj_12pyrobuf_list_Int32List;
struct __pyx_obj_12pyrobuf_list_Uint32List;
struct __pyx_obj_12pyrobuf_list_Int64List;
struct __pyx_obj_12pyrobuf_list_Uint64List;
struct __pyx_obj_12pyrobuf_list_CharList;
struct __pyx_obj_12pyrobuf_list___pyx_scope_struct____iter__;
struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_1___iter__;
struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_2___iter__;
struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_3___iter__;
struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_4___iter__;
struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_5___iter__;
struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_6___iter__;
struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_7___iter__;

/* "pyrobuf_list.pxd":6
 * 
 * 
 * cdef class NullListener:             # <<<<<<<<<<<<<<
 *     cpdef void _Modified(self)
 * 
 */
struct __pyx_obj_12pyrobuf_list_NullListener {
  PyObject_HEAD
  struct __pyx_vtabstruct_12pyrobuf_list_NullListener *__pyx_vtab;
};


/* "pyrobuf_list.pxd":12
 * 
 * 
 * cdef class TypedList(list):             # <<<<<<<<<<<<<<
 * 
 *     cdef type _list_type
 */
struct __pyx_obj_12pyrobuf_list_TypedList {
  PyListObject __pyx_base;
  PyTypeObject *_list_type;
  PyObject *_listener;
};


/* "pyrobuf_list.pxd":18
 * 
 * 
 * cdef class BytesList(list):             # <<<<<<<<<<<<<<
 * 
 *     cdef PyObject *_listener
 */
struct __pyx_obj_12pyrobuf_list_BytesList {
  PyListObject __pyx_base;
  PyObject *_listener;
};


/* "pyrobuf_list.pxd":23
 * 
 * 
 * cdef class StringList(list):             # <<<<<<<<<<<<<<
 * 
 *     cdef PyObject *_listener
 */
struct __pyx_obj_12pyrobuf_list_StringList {
  PyListObject __pyx_base;
  PyObject *_listener;
};


/* "pyrobuf_list.pxd":32
 * 
 * 
 * cdef class DoubleList:             # <<<<<<<<<<<<<<
 * 
 *     cdef double *_data
 */
struct __pyx_obj_12pyrobuf_list_DoubleList {
  PyObject_HEAD
  struct __pyx_vtabstruct_12pyrobuf_list_DoubleList *__pyx_vtab;
  double *_data;
  size_t _n_items;
  size_t _size;
  PyObject *_listener;
  Py_ssize_t shape[1];
  Py_ssize_t strides[1];
};


/* "pyrobuf_list.pxd":53
 * 
 * 
 * cdef class FloatList:             # <<<<<<<<<<<<<<
 * 
 *     cdef float *_data
 */
struct __pyx_obj_12pyrobuf_list_FloatList {
  PyObject_HEAD
  struct __pyx_vtabstruct_12pyrobuf_list_FloatList *__pyx_vtab;
  float *_data;
  size_t _n_items;
  size_t _size;
  PyObject *_listener;
  Py_ssize_t shape[1];
  Py_ssize_t strides[1];
};


/* "pyrobuf_list.pxd":74
 * 
 * 
 * cdef class IntList:             # <<<<<<<<<<<<<<
 * 
 *     cdef int *_data
 */
struct __pyx_obj_12pyrobuf_list_IntList {
  PyObject_HEAD
  struct __pyx_vtabstruct_12pyrobuf_list_IntList *__pyx_vtab;
  int *_data;
  size_t _n_items;
  size_t _size;
  PyObject *_listener;
  Py_ssize_t shape[1];
  Py_ssize_t strides[1];
};


/* "pyrobuf_list.pxd":95
 * 
 * 
 * cdef class Int32List:             # <<<<<<<<<<<<<<
 * 
 *     cdef int32_t *_data
 */
struct __pyx_obj_12pyrobuf_list_Int32List {
  PyObject_HEAD
  struct __pyx_vtabstruct_12pyrobuf_list_Int32List *__pyx_vtab;
  int32_t *_data;
  size_t _n_items;
  size_t _size;
  PyObject *_listener;
  Py_ssize_t shape[1];
  Py_ssize_t strides[1];
};


/* "pyrobuf_list.pxd":116
 * 
 * 
 * cdef class Uint32List:             # <<<<<<<<<<<<<<
 * 
 *     cdef uint32_t *_data
 */
struct __pyx_obj_12pyrobuf_list_Uint32List {
  PyObject_HEAD
  struct __pyx_vtabstruct_12pyrobuf_list_Uint32List *__pyx_vtab;
  uint32_t *_data;
  size_t _n_items;
  size_t _size;
  PyObject *_listener;
  Py_ssize_t shape[1];
  Py_ssize_t strides[1];
};


/* "pyrobuf_list.pxd":137
 * 
 * 
 * cdef class Int64List:             # <<<<<<<<<<<<<<
 * 
 *     cdef int64_t *_data
 */
struct __pyx_obj_12pyrobuf_list_Int64List {
  PyObject_HEAD
  struct __pyx_vtabstruct_12pyrobuf_list_Int64List *__pyx_vtab;
  int64_t *_data;
  size_t _n_items;
  size_t _size;
  PyObject *_listener;
  Py_ssize_t shape[1];
  Py_ssize_t strides[1];
};


/* "pyrobuf_list.pxd":158
 * 
 * 
 * cdef class Uint64List:             # <<<<<<<<<<<<<<
 * 
 *     cdef uint64_t *_data
 */
struct __pyx_obj_12pyrobuf_list_Uint64List {
  PyObject_HEAD
  struct __pyx_vtabstruct_12pyrobuf_list_Uint64List *__pyx_vtab;
  uint64_t *_data;
  size_t _n_items;
  size_t _size;
  PyObject *_listener;
  Py_ssize_t shape[1];
  Py_ssize_t strides[1];
};


/* "pyrobuf_list.pxd":179
 * 
 * 
 * cdef class CharList:             # <<<<<<<<<<<<<<
 * 
 *     cdef char *_data
 */
struct __pyx_obj_12pyrobuf_list_CharList {
  PyObject_HEAD
  struct __pyx_vtabstruct_12pyrobuf_list_CharList *__pyx_vtab;
  char *_data;
  size_t _n_items;
  size_t _size;
  PyObject *_listener;
  Py_ssize_t shape[1];
  Py_ssize_t strides[1];
};


/* "pyrobuf_list.pyx":291
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */
struct __pyx_obj_12pyrobuf_list___pyx_scope_struct____iter__ {
  PyObject_HEAD
  size_t __pyx_v_idx;
  struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self;
};


/* "pyrobuf_list.pyx":501
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */
struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_1___iter__ {
  PyObject_HEAD
  size_t __pyx_v_idx;
  struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self;
};


/* "pyrobuf_list.pyx":711
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */
struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_2___iter__ {
  PyObject_HEAD
  size_t __pyx_v_idx;
  struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self;
};


/* "pyrobuf_list.pyx":921
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */
struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_3___iter__ {
  PyObject_HEAD
  size_t __pyx_v_idx;
  struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self;
};


/* "pyrobuf_list.pyx":1131
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */
struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_4___iter__ {
  PyObject_HEAD
  size_t __pyx_v_idx;
  struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self;
};


/* "pyrobuf_list.pyx":1341
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */
struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_5___iter__ {
  PyObject_HEAD
  size_t __pyx_v_idx;
  struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self;
};


/* "pyrobuf_list.pyx":1551
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */
struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_6___iter__ {
  PyObject_HEAD
  size_t __pyx_v_idx;
  struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self;
};


/* "pyrobuf_list.pyx":1761
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */
struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_7___iter__ {
  PyObject_HEAD
  size_t __pyx_v_idx;
  struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self;
};



/* "pyrobuf_list.pyx":7
 * 
 * 
 * cdef class NullListener:             # <<<<<<<<<<<<<<
 *     property _is_present_in_parent:
 *         def __get__(self):
 */

struct __pyx_vtabstruct_12pyrobuf_list_NullListener {
  void (*_Modified)(struct __pyx_obj_12pyrobuf_list_NullListener *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12pyrobuf_list_NullListener *__pyx_vtabptr_12pyrobuf_list_NullListener;


/* "pyrobuf_list.pyx":241
 * 
 * 
 * cdef class DoubleList:             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):
 */

struct __pyx_vtabstruct_12pyrobuf_list_DoubleList {
  void (*_append)(struct __pyx_obj_12pyrobuf_list_DoubleList *, double);
  PyObject *(*_equal)(struct __pyx_obj_12pyrobuf_list_DoubleList *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*append)(struct __pyx_obj_12pyrobuf_list_DoubleList *, double, int __pyx_skip_dispatch);
  PyObject *(*extend)(struct __pyx_obj_12pyrobuf_list_DoubleList *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*insert)(struct __pyx_obj_12pyrobuf_list_DoubleList *, int, double, int __pyx_skip_dispatch);
  PyObject *(*pop)(struct __pyx_obj_12pyrobuf_list_DoubleList *, int __pyx_skip_dispatch);
  PyObject *(*remove)(struct __pyx_obj_12pyrobuf_list_DoubleList *, double, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12pyrobuf_list_DoubleList *__pyx_vtabptr_12pyrobuf_list_DoubleList;


/* "pyrobuf_list.pyx":451
 * 
 * 
 * cdef class FloatList:             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):
 */

struct __pyx_vtabstruct_12pyrobuf_list_FloatList {
  void (*_append)(struct __pyx_obj_12pyrobuf_list_FloatList *, float);
  PyObject *(*_equal)(struct __pyx_obj_12pyrobuf_list_FloatList *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*append)(struct __pyx_obj_12pyrobuf_list_FloatList *, float, int __pyx_skip_dispatch);
  PyObject *(*extend)(struct __pyx_obj_12pyrobuf_list_FloatList *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*insert)(struct __pyx_obj_12pyrobuf_list_FloatList *, int, float, int __pyx_skip_dispatch);
  PyObject *(*pop)(struct __pyx_obj_12pyrobuf_list_FloatList *, int __pyx_skip_dispatch);
  PyObject *(*remove)(struct __pyx_obj_12pyrobuf_list_FloatList *, float, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12pyrobuf_list_FloatList *__pyx_vtabptr_12pyrobuf_list_FloatList;


/* "pyrobuf_list.pyx":661
 * 
 * 
 * cdef class IntList:             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):
 */

struct __pyx_vtabstruct_12pyrobuf_list_IntList {
  void (*_append)(struct __pyx_obj_12pyrobuf_list_IntList *, int);
  PyObject *(*_equal)(struct __pyx_obj_12pyrobuf_list_IntList *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*append)(struct __pyx_obj_12pyrobuf_list_IntList *, int, int __pyx_skip_dispatch);
  PyObject *(*extend)(struct __pyx_obj_12pyrobuf_list_IntList *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*insert)(struct __pyx_obj_12pyrobuf_list_IntList *, int, int, int __pyx_skip_dispatch);
  PyObject *(*pop)(struct __pyx_obj_12pyrobuf_list_IntList *, int __pyx_skip_dispatch);
  PyObject *(*remove)(struct __pyx_obj_12pyrobuf_list_IntList *, int, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12pyrobuf_list_IntList *__pyx_vtabptr_12pyrobuf_list_IntList;


/* "pyrobuf_list.pyx":871
 * 
 * 
 * cdef class Int32List:             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):
 */

struct __pyx_vtabstruct_12pyrobuf_list_Int32List {
  void (*_append)(struct __pyx_obj_12pyrobuf_list_Int32List *, int32_t);
  PyObject *(*_equal)(struct __pyx_obj_12pyrobuf_list_Int32List *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*append)(struct __pyx_obj_12pyrobuf_list_Int32List *, int32_t, int __pyx_skip_dispatch);
  PyObject *(*extend)(struct __pyx_obj_12pyrobuf_list_Int32List *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*insert)(struct __pyx_obj_12pyrobuf_list_Int32List *, int, int32_t, int __pyx_skip_dispatch);
  PyObject *(*pop)(struct __pyx_obj_12pyrobuf_list_Int32List *, int __pyx_skip_dispatch);
  PyObject *(*remove)(struct __pyx_obj_12pyrobuf_list_Int32List *, int32_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12pyrobuf_list_Int32List *__pyx_vtabptr_12pyrobuf_list_Int32List;


/* "pyrobuf_list.pyx":1081
 * 
 * 
 * cdef class Uint32List:             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):
 */

struct __pyx_vtabstruct_12pyrobuf_list_Uint32List {
  void (*_append)(struct __pyx_obj_12pyrobuf_list_Uint32List *, uint32_t);
  PyObject *(*_equal)(struct __pyx_obj_12pyrobuf_list_Uint32List *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*append)(struct __pyx_obj_12pyrobuf_list_Uint32List *, uint32_t, int __pyx_skip_dispatch);
  PyObject *(*extend)(struct __pyx_obj_12pyrobuf_list_Uint32List *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*insert)(struct __pyx_obj_12pyrobuf_list_Uint32List *, int, uint32_t, int __pyx_skip_dispatch);
  PyObject *(*pop)(struct __pyx_obj_12pyrobuf_list_Uint32List *, int __pyx_skip_dispatch);
  PyObject *(*remove)(struct __pyx_obj_12pyrobuf_list_Uint32List *, uint32_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12pyrobuf_list_Uint32List *__pyx_vtabptr_12pyrobuf_list_Uint32List;


/* "pyrobuf_list.pyx":1291
 * 
 * 
 * cdef class Int64List:             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):
 */

struct __pyx_vtabstruct_12pyrobuf_list_Int64List {
  void (*_append)(struct __pyx_obj_12pyrobuf_list_Int64List *, int64_t);
  PyObject *(*_equal)(struct __pyx_obj_12pyrobuf_list_Int64List *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*append)(struct __pyx_obj_12pyrobuf_list_Int64List *, int64_t, int __pyx_skip_dispatch);
  PyObject *(*extend)(struct __pyx_obj_12pyrobuf_list_Int64List *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*insert)(struct __pyx_obj_12pyrobuf_list_Int64List *, int, int64_t, int __pyx_skip_dispatch);
  PyObject *(*pop)(struct __pyx_obj_12pyrobuf_list_Int64List *, int __pyx_skip_dispatch);
  PyObject *(*remove)(struct __pyx_obj_12pyrobuf_list_Int64List *, int64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12pyrobuf_list_Int64List *__pyx_vtabptr_12pyrobuf_list_Int64List;


/* "pyrobuf_list.pyx":1501
 * 
 * 
 * cdef class Uint64List:             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):
 */

struct __pyx_vtabstruct_12pyrobuf_list_Uint64List {
  void (*_append)(struct __pyx_obj_12pyrobuf_list_Uint64List *, uint64_t);
  PyObject *(*_equal)(struct __pyx_obj_12pyrobuf_list_Uint64List *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*append)(struct __pyx_obj_12pyrobuf_list_Uint64List *, uint64_t, int __pyx_skip_dispatch);
  PyObject *(*extend)(struct __pyx_obj_12pyrobuf_list_Uint64List *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*insert)(struct __pyx_obj_12pyrobuf_list_Uint64List *, int, uint64_t, int __pyx_skip_dispatch);
  PyObject *(*pop)(struct __pyx_obj_12pyrobuf_list_Uint64List *, int __pyx_skip_dispatch);
  PyObject *(*remove)(struct __pyx_obj_12pyrobuf_list_Uint64List *, uint64_t, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12pyrobuf_list_Uint64List *__pyx_vtabptr_12pyrobuf_list_Uint64List;


/* "pyrobuf_list.pyx":1711
 * 
 * 
 * cdef class CharList:             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):
 */

struct __pyx_vtabstruct_12pyrobuf_list_CharList {
  void (*_append)(struct __pyx_obj_12pyrobuf_list_CharList *, char);
  PyObject *(*_equal)(struct __pyx_obj_12pyrobuf_list_CharList *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*append)(struct __pyx_obj_12pyrobuf_list_CharList *, char, int __pyx_skip_dispatch);
  PyObject *(*extend)(struct __pyx_obj_12pyrobuf_list_CharList *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*insert)(struct __pyx_obj_12pyrobuf_list_CharList *, int, char, int __pyx_skip_dispatch);
  PyObject *(*pop)(struct __pyx_obj_12pyrobuf_list_CharList *, int __pyx_skip_dispatch);
  PyObject *(*remove)(struct __pyx_obj_12pyrobuf_list_CharList *, char, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_12pyrobuf_list_CharList *__pyx_vtabptr_12pyrobuf_list_CharList;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* IncludeStringH.proto */
#include <string.h>

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int32_t(int32_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_char(char value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int32_t __Pyx_PyInt_As_int32_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint64_t __Pyx_PyInt_As_uint64_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->exc_type, &(self)->exc_value, &(self)->exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(self);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->exc_type, (self)->exc_value, (self)->exc_traceback);\
    (self)->exc_type = (self)->exc_value = (self)->exc_traceback = NULL;\
    }
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__pyx_CoroutineObject *self);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* VoidPtrExport.proto */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig);

/* PyIdentifierFromString.proto */
#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

/* ModuleImport.proto */
static PyObject *__Pyx_ImportModule(const char *name);

/* TypeImport.proto */
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static void __pyx_f_12pyrobuf_list_12NullListener__Modified(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_NullListener *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_10DoubleList__equal(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_10DoubleList_append(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, double __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_10DoubleList_extend(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, PyObject *__pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_10DoubleList_insert(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, int __pyx_v_i, double __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_10DoubleList_pop(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_10DoubleList_remove(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, double __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_12pyrobuf_list_10DoubleList__append(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, double __pyx_v_x); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_9FloatList__equal(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_9FloatList_append(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, float __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_9FloatList_extend(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, PyObject *__pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_9FloatList_insert(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, int __pyx_v_i, float __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_9FloatList_pop(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_9FloatList_remove(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, float __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_12pyrobuf_list_9FloatList__append(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, float __pyx_v_x); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_7IntList__equal(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_7IntList_append(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_7IntList_extend(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, PyObject *__pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_7IntList_insert(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_i, int __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_7IntList_pop(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_7IntList_remove(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_12pyrobuf_list_7IntList__append(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_x); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int32List__equal(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int32List_append(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int32_t __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int32List_extend(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, PyObject *__pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int32List_insert(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int __pyx_v_i, int32_t __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int32List_pop(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int32List_remove(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int32_t __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_12pyrobuf_list_9Int32List__append(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int32_t __pyx_v_x); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint32List__equal(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint32List_append(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, uint32_t __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint32List_extend(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, PyObject *__pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint32List_insert(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, int __pyx_v_i, uint32_t __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint32List_pop(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint32List_remove(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, uint32_t __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_12pyrobuf_list_10Uint32List__append(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, uint32_t __pyx_v_x); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int64List__equal(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int64List_append(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int64_t __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int64List_extend(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, PyObject *__pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int64List_insert(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int __pyx_v_i, int64_t __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int64List_pop(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int64List_remove(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int64_t __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_12pyrobuf_list_9Int64List__append(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int64_t __pyx_v_x); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint64List__equal(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint64List_append(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, uint64_t __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint64List_extend(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, PyObject *__pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint64List_insert(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, int __pyx_v_i, uint64_t __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint64List_pop(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint64List_remove(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, uint64_t __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_12pyrobuf_list_10Uint64List__append(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, uint64_t __pyx_v_x); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_8CharList__equal(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_8CharList_append(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, char __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_8CharList_extend(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, PyObject *__pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_8CharList_insert(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, int __pyx_v_i, char __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_8CharList_pop(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_12pyrobuf_list_8CharList_remove(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, char __pyx_v_x, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_12pyrobuf_list_8CharList__append(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, char __pyx_v_x); /* proto*/

/* Module declarations from 'libc.stdint' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython.version' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'pyrobuf_list' */
static PyTypeObject *__pyx_ptype_12pyrobuf_list_NullListener = 0;
static PyTypeObject *__pyx_ptype_12pyrobuf_list_TypedList = 0;
static PyTypeObject *__pyx_ptype_12pyrobuf_list_BytesList = 0;
static PyTypeObject *__pyx_ptype_12pyrobuf_list_StringList = 0;
static PyTypeObject *__pyx_ptype_12pyrobuf_list_DoubleList = 0;
static PyTypeObject *__pyx_ptype_12pyrobuf_list_FloatList = 0;
static PyTypeObject *__pyx_ptype_12pyrobuf_list_IntList = 0;
static PyTypeObject *__pyx_ptype_12pyrobuf_list_Int32List = 0;
static PyTypeObject *__pyx_ptype_12pyrobuf_list_Uint32List = 0;
static PyTypeObject *__pyx_ptype_12pyrobuf_list_Int64List = 0;
static PyTypeObject *__pyx_ptype_12pyrobuf_list_Uint64List = 0;
static PyTypeObject *__pyx_ptype_12pyrobuf_list_CharList = 0;
static PyTypeObject *__pyx_ptype_12pyrobuf_list___pyx_scope_struct____iter__ = 0;
static PyTypeObject *__pyx_ptype_12pyrobuf_list___pyx_scope_struct_1___iter__ = 0;
static PyTypeObject *__pyx_ptype_12pyrobuf_list___pyx_scope_struct_2___iter__ = 0;
static PyTypeObject *__pyx_ptype_12pyrobuf_list___pyx_scope_struct_3___iter__ = 0;
static PyTypeObject *__pyx_ptype_12pyrobuf_list___pyx_scope_struct_4___iter__ = 0;
static PyTypeObject *__pyx_ptype_12pyrobuf_list___pyx_scope_struct_5___iter__ = 0;
static PyTypeObject *__pyx_ptype_12pyrobuf_list___pyx_scope_struct_6___iter__ = 0;
static PyTypeObject *__pyx_ptype_12pyrobuf_list___pyx_scope_struct_7___iter__ = 0;
__PYX_EXTERN_C DL_EXPORT(struct __pyx_obj_12pyrobuf_list_NullListener) *null_listener;
static PyObject *__pyx_f_12pyrobuf_list___pyx_unpickle_NullListener__set_state(struct __pyx_obj_12pyrobuf_list_NullListener *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "pyrobuf_list"
extern int __pyx_module_is_main_pyrobuf_list;
int __pyx_module_is_main_pyrobuf_list = 0;

/* Implementation of 'pyrobuf_list' */
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_super;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_ValueError;
static const char __pyx_k_i[] = "i";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_iter[] = "__iter__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_equal[] = "_equal";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_utf_8[] = "utf-8";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_extend[] = "extend";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_insert[] = "insert";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_remove[] = "remove";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_setitem[] = "__setitem__";
static const char __pyx_k_Modified[] = "_Modified";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_listener[] = "listener";
static const char __pyx_k_pyx_capi[] = "__pyx_capi__";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_MergeFrom[] = "MergeFrom";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_list_type[] = "list_type";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_pyrobuf_list[] = "pyrobuf_list";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_null_listener[] = "null_listener";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_type_mismatch[] = "type mismatch";
static const char __pyx_k_x_not_in_list[] = "x not in list";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_IntList___iter[] = "IntList.__iter__";
static const char __pyx_k_CharList___iter[] = "CharList.__iter__";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_FloatList___iter[] = "FloatList.__iter__";
static const char __pyx_k_Int32List___iter[] = "Int32List.__iter__";
static const char __pyx_k_Int64List___iter[] = "Int64List.__iter__";
static const char __pyx_k_DoubleList___iter[] = "DoubleList.__iter__";
static const char __pyx_k_Uint32List___iter[] = "Uint32List.__iter__";
static const char __pyx_k_Uint64List___iter[] = "Uint64List.__iter__";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_pop_from_empty_list[] = "pop from empty list";
static const char __pyx_k_is_present_in_parent[] = "_is_present_in_parent";
static const char __pyx_k_list_index_out_of_range[] = "list index out of range";
static const char __pyx_k_pyx_unpickle_NullListener[] = "__pyx_unpickle_NullListener";
static const char __pyx_k_r_has_type_s_but_expected_one_o[] = "%r has type %s, but expected one of: (%s,)";
static const char __pyx_k_Incompatible_checksums_s_vs_0xd4[] = "Incompatible checksums (%s vs 0xd41d8cd = ())";
static const char __pyx_k_Operation_not_supported_for_list[] = "Operation not supported for lists";
static const char __pyx_k_list_indices_must_be_integers_no[] = "list indices must be integers, not %s";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_self__listener_cannot_be_convert[] = "self._listener cannot be converted to a Python object for pickling";
static const char __pyx_k_r_has_type_s_but_expected_one_o_2[] = "%r has type %s, but expected one of: (%s, %s)";
static PyObject *__pyx_n_s_AssertionError;
static PyObject *__pyx_n_s_CharList___iter;
static PyObject *__pyx_n_s_DoubleList___iter;
static PyObject *__pyx_n_s_FloatList___iter;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xd4;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_n_s_Int32List___iter;
static PyObject *__pyx_n_s_Int64List___iter;
static PyObject *__pyx_n_s_IntList___iter;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_n_s_MergeFrom;
static PyObject *__pyx_n_s_Modified;
static PyObject *__pyx_kp_s_Operation_not_supported_for_list;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_Uint32List___iter;
static PyObject *__pyx_n_s_Uint64List___iter;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_decode;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_equal;
static PyObject *__pyx_n_s_extend;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_insert;
static PyObject *__pyx_n_s_is_present_in_parent;
static PyObject *__pyx_n_s_iter;
static PyObject *__pyx_kp_s_list_index_out_of_range;
static PyObject *__pyx_kp_s_list_indices_must_be_integers_no;
static PyObject *__pyx_n_s_list_type;
static PyObject *__pyx_n_s_listener;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_null_listener;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pop;
static PyObject *__pyx_kp_s_pop_from_empty_list;
static PyObject *__pyx_n_s_pyrobuf_list;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_capi;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_NullListener;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_kp_s_r_has_type_s_but_expected_one_o;
static PyObject *__pyx_kp_s_r_has_type_s_but_expected_one_o_2;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_remove;
static PyObject *__pyx_kp_s_self__listener_cannot_be_convert;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_setitem;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_super;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_kp_s_type_mismatch;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_kp_s_utf_8;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_kp_s_x_not_in_list;
static PyObject *__pyx_n_s_zip;
static PyObject *__pyx_pf_12pyrobuf_list_12NullListener_21_is_present_in_parent___get__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_NullListener *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_12NullListener__Modified(struct __pyx_obj_12pyrobuf_list_NullListener *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_12NullListener_2__reduce_cython__(struct __pyx_obj_12pyrobuf_list_NullListener *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_12NullListener_4__setstate_cython__(struct __pyx_obj_12pyrobuf_list_NullListener *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12pyrobuf_list_9TypedList___init__(struct __pyx_obj_12pyrobuf_list_TypedList *__pyx_v_self, PyTypeObject *__pyx_v_list_type, PyObject *__pyx_v_listener); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9TypedList_9list_type___get__(struct __pyx_obj_12pyrobuf_list_TypedList *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_9TypedList_2__setitem__(struct __pyx_obj_12pyrobuf_list_TypedList *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9TypedList_4add(struct __pyx_obj_12pyrobuf_list_TypedList *__pyx_v_self, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9TypedList_6append(struct __pyx_obj_12pyrobuf_list_TypedList *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9TypedList_8extend(struct __pyx_obj_12pyrobuf_list_TypedList *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9TypedList_10insert(struct __pyx_obj_12pyrobuf_list_TypedList *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9TypedList_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_TypedList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9TypedList_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_TypedList *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12pyrobuf_list_9BytesList___init__(struct __pyx_obj_12pyrobuf_list_BytesList *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static int __pyx_pf_12pyrobuf_list_9BytesList_2__setitem__(struct __pyx_obj_12pyrobuf_list_BytesList *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9BytesList_4append(struct __pyx_obj_12pyrobuf_list_BytesList *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9BytesList_6extend(struct __pyx_obj_12pyrobuf_list_BytesList *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9BytesList_8insert(struct __pyx_obj_12pyrobuf_list_BytesList *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9BytesList_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_BytesList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9BytesList_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_BytesList *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12pyrobuf_list_10StringList___init__(struct __pyx_obj_12pyrobuf_list_StringList *__pyx_v_self, PyObject *__pyx_v_listener); /* proto */
static int __pyx_pf_12pyrobuf_list_10StringList_2__setitem__(struct __pyx_obj_12pyrobuf_list_StringList *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10StringList_4append(struct __pyx_obj_12pyrobuf_list_StringList *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10StringList_6extend(struct __pyx_obj_12pyrobuf_list_StringList *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10StringList_8insert(struct __pyx_obj_12pyrobuf_list_StringList *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10StringList_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_StringList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10StringList_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_StringList *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12pyrobuf_list_10DoubleList___cinit__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, size_t __pyx_v_size, PyObject *__pyx_v_listener); /* proto */
static void __pyx_pf_12pyrobuf_list_10DoubleList_2__dealloc__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_10DoubleList_4__contains__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, double __pyx_v_x); /* proto */
static int __pyx_pf_12pyrobuf_list_10DoubleList_6__delitem__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, int __pyx_v_i); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_8__getitem__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, int __pyx_v_i); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_10__iter__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_12pyrobuf_list_10DoubleList_13__len__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_15__repr__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_10DoubleList_17__setitem__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, int __pyx_v_i, double __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_19__str__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_10DoubleList_21__getbuffer__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags); /* proto */
static void __pyx_pf_12pyrobuf_list_10DoubleList_23__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_25__richcmp__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_27__neq__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_29_equal(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_31append(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, double __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_33extend(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_35insert(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, int __pyx_v_i, double __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_37pop(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_39remove(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, double __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_41__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_43__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12pyrobuf_list_9FloatList___cinit__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, size_t __pyx_v_size, PyObject *__pyx_v_listener); /* proto */
static void __pyx_pf_12pyrobuf_list_9FloatList_2__dealloc__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_9FloatList_4__contains__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, float __pyx_v_x); /* proto */
static int __pyx_pf_12pyrobuf_list_9FloatList_6__delitem__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, int __pyx_v_i); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_8__getitem__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, int __pyx_v_i); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_10__iter__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_12pyrobuf_list_9FloatList_13__len__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_15__repr__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_9FloatList_17__setitem__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, int __pyx_v_i, float __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_19__str__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_9FloatList_21__getbuffer__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags); /* proto */
static void __pyx_pf_12pyrobuf_list_9FloatList_23__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_25__richcmp__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_27__neq__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_29_equal(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_31append(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, float __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_33extend(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_35insert(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, int __pyx_v_i, float __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_37pop(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_39remove(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, float __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_41__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_43__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12pyrobuf_list_7IntList___cinit__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, size_t __pyx_v_size, PyObject *__pyx_v_listener); /* proto */
static void __pyx_pf_12pyrobuf_list_7IntList_2__dealloc__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_7IntList_4__contains__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_x); /* proto */
static int __pyx_pf_12pyrobuf_list_7IntList_6__delitem__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_i); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_7IntList_8__getitem__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_i); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_7IntList_10__iter__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_12pyrobuf_list_7IntList_13__len__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_7IntList_15__repr__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_7IntList_17__setitem__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_i, int __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_7IntList_19__str__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_7IntList_21__getbuffer__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags); /* proto */
static void __pyx_pf_12pyrobuf_list_7IntList_23__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_7IntList_25__richcmp__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_7IntList_27__neq__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_7IntList_29_equal(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_7IntList_31append(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_7IntList_33extend(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_7IntList_35insert(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_i, int __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_7IntList_37pop(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_7IntList_39remove(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_7IntList_41__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_7IntList_43__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12pyrobuf_list_9Int32List___cinit__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, size_t __pyx_v_size, PyObject *__pyx_v_listener); /* proto */
static void __pyx_pf_12pyrobuf_list_9Int32List_2__dealloc__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_9Int32List_4__contains__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int32_t __pyx_v_x); /* proto */
static int __pyx_pf_12pyrobuf_list_9Int32List_6__delitem__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int __pyx_v_i); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_8__getitem__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int __pyx_v_i); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_10__iter__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_12pyrobuf_list_9Int32List_13__len__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_15__repr__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_9Int32List_17__setitem__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int __pyx_v_i, int32_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_19__str__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_9Int32List_21__getbuffer__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags); /* proto */
static void __pyx_pf_12pyrobuf_list_9Int32List_23__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_25__richcmp__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_27__neq__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_29_equal(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_31append(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int32_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_33extend(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_35insert(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int __pyx_v_i, int32_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_37pop(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_39remove(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int32_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_41__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_43__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12pyrobuf_list_10Uint32List___cinit__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, size_t __pyx_v_size, PyObject *__pyx_v_listener); /* proto */
static void __pyx_pf_12pyrobuf_list_10Uint32List_2__dealloc__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_10Uint32List_4__contains__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, uint32_t __pyx_v_x); /* proto */
static int __pyx_pf_12pyrobuf_list_10Uint32List_6__delitem__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, int __pyx_v_i); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_8__getitem__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, int __pyx_v_i); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_10__iter__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_12pyrobuf_list_10Uint32List_13__len__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_15__repr__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_10Uint32List_17__setitem__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, int __pyx_v_i, uint32_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_19__str__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_10Uint32List_21__getbuffer__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags); /* proto */
static void __pyx_pf_12pyrobuf_list_10Uint32List_23__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_25__richcmp__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_27__neq__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_29_equal(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_31append(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, uint32_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_33extend(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_35insert(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, int __pyx_v_i, uint32_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_37pop(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_39remove(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, uint32_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_41__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_43__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12pyrobuf_list_9Int64List___cinit__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, size_t __pyx_v_size, PyObject *__pyx_v_listener); /* proto */
static void __pyx_pf_12pyrobuf_list_9Int64List_2__dealloc__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_9Int64List_4__contains__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int64_t __pyx_v_x); /* proto */
static int __pyx_pf_12pyrobuf_list_9Int64List_6__delitem__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int __pyx_v_i); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_8__getitem__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int __pyx_v_i); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_10__iter__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_12pyrobuf_list_9Int64List_13__len__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_15__repr__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_9Int64List_17__setitem__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int __pyx_v_i, int64_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_19__str__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_9Int64List_21__getbuffer__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags); /* proto */
static void __pyx_pf_12pyrobuf_list_9Int64List_23__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_25__richcmp__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_27__neq__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_29_equal(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_31append(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int64_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_33extend(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_35insert(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int __pyx_v_i, int64_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_37pop(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_39remove(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int64_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_41__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_43__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12pyrobuf_list_10Uint64List___cinit__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, size_t __pyx_v_size, PyObject *__pyx_v_listener); /* proto */
static void __pyx_pf_12pyrobuf_list_10Uint64List_2__dealloc__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_10Uint64List_4__contains__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, uint64_t __pyx_v_x); /* proto */
static int __pyx_pf_12pyrobuf_list_10Uint64List_6__delitem__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, int __pyx_v_i); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_8__getitem__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, int __pyx_v_i); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_10__iter__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_12pyrobuf_list_10Uint64List_13__len__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_15__repr__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_10Uint64List_17__setitem__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, int __pyx_v_i, uint64_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_19__str__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_10Uint64List_21__getbuffer__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags); /* proto */
static void __pyx_pf_12pyrobuf_list_10Uint64List_23__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_25__richcmp__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_27__neq__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_29_equal(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_31append(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, uint64_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_33extend(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_35insert(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, int __pyx_v_i, uint64_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_37pop(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_39remove(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, uint64_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_41__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_43__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_12pyrobuf_list_8CharList___cinit__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, size_t __pyx_v_size, PyObject *__pyx_v_listener); /* proto */
static void __pyx_pf_12pyrobuf_list_8CharList_2__dealloc__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_8CharList_4__contains__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, char __pyx_v_x); /* proto */
static int __pyx_pf_12pyrobuf_list_8CharList_6__delitem__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, int __pyx_v_i); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_8CharList_8__getitem__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, int __pyx_v_i); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_8CharList_10__iter__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_12pyrobuf_list_8CharList_13__len__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_8CharList_15__repr__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_8CharList_17__setitem__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, int __pyx_v_i, char __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_8CharList_19__str__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self); /* proto */
static int __pyx_pf_12pyrobuf_list_8CharList_21__getbuffer__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags); /* proto */
static void __pyx_pf_12pyrobuf_list_8CharList_23__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_8CharList_25__richcmp__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_8CharList_27__neq__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_8CharList_29_equal(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_8CharList_31append(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, char __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_8CharList_33extend(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_8CharList_35insert(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, int __pyx_v_i, char __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_8CharList_37pop(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_8CharList_39remove(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, char __pyx_v_x); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_8CharList_41__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list_8CharList_43__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_12pyrobuf_list___pyx_unpickle_NullListener(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_12pyrobuf_list_NullListener(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12pyrobuf_list_TypedList(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12pyrobuf_list_BytesList(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12pyrobuf_list_StringList(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12pyrobuf_list_DoubleList(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12pyrobuf_list_FloatList(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12pyrobuf_list_IntList(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12pyrobuf_list_Int32List(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12pyrobuf_list_Uint32List(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12pyrobuf_list_Int64List(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12pyrobuf_list_Uint64List(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12pyrobuf_list_CharList(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12pyrobuf_list___pyx_scope_struct____iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_1___iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_2___iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_3___iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_4___iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_5___iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_6___iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_7___iter__(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_222419149;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__48;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__50;
static PyObject *__pyx_tuple__51;
static PyObject *__pyx_tuple__52;
static PyObject *__pyx_tuple__53;
static PyObject *__pyx_tuple__54;
static PyObject *__pyx_tuple__55;
static PyObject *__pyx_tuple__56;
static PyObject *__pyx_tuple__57;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_tuple__59;
static PyObject *__pyx_tuple__60;
static PyObject *__pyx_tuple__61;
static PyObject *__pyx_tuple__62;
static PyObject *__pyx_tuple__63;
static PyObject *__pyx_tuple__64;
static PyObject *__pyx_tuple__65;
static PyObject *__pyx_tuple__66;
static PyObject *__pyx_tuple__67;
static PyObject *__pyx_tuple__68;
static PyObject *__pyx_tuple__69;
static PyObject *__pyx_tuple__70;
static PyObject *__pyx_tuple__71;
static PyObject *__pyx_tuple__72;
static PyObject *__pyx_tuple__73;
static PyObject *__pyx_tuple__74;
static PyObject *__pyx_tuple__75;
static PyObject *__pyx_tuple__76;
static PyObject *__pyx_tuple__77;
static PyObject *__pyx_tuple__78;
static PyObject *__pyx_tuple__79;
static PyObject *__pyx_tuple__80;
static PyObject *__pyx_tuple__81;
static PyObject *__pyx_tuple__82;
static PyObject *__pyx_tuple__83;
static PyObject *__pyx_tuple__84;
static PyObject *__pyx_tuple__85;
static PyObject *__pyx_tuple__86;
static PyObject *__pyx_tuple__87;
static PyObject *__pyx_codeobj__88;
/* Late includes */
struct __pyx_obj_12pyrobuf_list_NullListener *null_listener = 0;

/* "pyrobuf_list.pyx":9
 * cdef class NullListener:
 *     property _is_present_in_parent:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return True
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_12NullListener_21_is_present_in_parent_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_12NullListener_21_is_present_in_parent_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_12NullListener_21_is_present_in_parent___get__(((struct __pyx_obj_12pyrobuf_list_NullListener *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_12NullListener_21_is_present_in_parent___get__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_NullListener *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyrobuf_list.pyx":10
 *     property _is_present_in_parent:
 *         def __get__(self):
 *             return True             # <<<<<<<<<<<<<<
 * 
 *     cpdef void _Modified(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":9
 * cdef class NullListener:
 *     property _is_present_in_parent:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return True
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":12
 *             return True
 * 
 *     cpdef void _Modified(self):             # <<<<<<<<<<<<<<
 *         return
 * 
 */

static PyObject *__pyx_pw_12pyrobuf_list_12NullListener_1_Modified(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static void __pyx_f_12pyrobuf_list_12NullListener__Modified(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_NullListener *__pyx_v_self, int __pyx_skip_dispatch) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_Modified", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_Modified); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_12NullListener_1_Modified)) {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 12, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 12, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":13
 * 
 *     cpdef void _Modified(self):
 *         return             # <<<<<<<<<<<<<<
 * 
 * cdef public NullListener null_listener = NullListener()
 */
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":12
 *             return True
 * 
 *     cpdef void _Modified(self):             # <<<<<<<<<<<<<<
 *         return
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("pyrobuf_list.NullListener._Modified", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_12NullListener_1_Modified(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_12NullListener_1_Modified(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_Modified (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_12NullListener__Modified(((struct __pyx_obj_12pyrobuf_list_NullListener *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_12NullListener__Modified(struct __pyx_obj_12pyrobuf_list_NullListener *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_Modified", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_void_to_None(__pyx_f_12pyrobuf_list_12NullListener__Modified(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.NullListener._Modified", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef bint use_setstate
 *     state = ()
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_12NullListener_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_12NullListener_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_12NullListener_2__reduce_cython__(((struct __pyx_obj_12pyrobuf_list_NullListener *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_12NullListener_2__reduce_cython__(struct __pyx_obj_12pyrobuf_list_NullListener *__pyx_v_self) {
  int __pyx_v_use_setstate;
  PyObject *__pyx_v_state = NULL;
  PyObject *__pyx_v__dict = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":4
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":5
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":6
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":7
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":5
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":9
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_NullListener, (type(self), 0xd41d8cd, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":10
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_NullListener, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":11
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_NullListener, (type(self), 0xd41d8cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_NullListener, (type(self), 0xd41d8cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_pyx_unpickle_NullListener); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":10
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_NullListener, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":13
 *         return __pyx_unpickle_NullListener, (type(self), 0xd41d8cd, None), state
 *     else:
 *         return __pyx_unpickle_NullListener, (type(self), 0xd41d8cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_NullListener__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_pyx_unpickle_NullListener); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef bint use_setstate
 *     state = ()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.NullListener.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":14
 *     else:
 *         return __pyx_unpickle_NullListener, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NullListener__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_12NullListener_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_12NullListener_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_12NullListener_4__setstate_cython__(((struct __pyx_obj_12pyrobuf_list_NullListener *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_12NullListener_4__setstate_cython__(struct __pyx_obj_12pyrobuf_list_NullListener *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":15
 *         return __pyx_unpickle_NullListener, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_NullListener__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 15, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_12pyrobuf_list___pyx_unpickle_NullListener__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":14
 *     else:
 *         return __pyx_unpickle_NullListener, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NullListener__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.NullListener.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":20
 * cdef class TypedList(list):
 * 
 *     def __init__(self, type list_type, object listener=None):             # <<<<<<<<<<<<<<
 *         self._list_type = list_type
 *         if listener is None:
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_9TypedList_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12pyrobuf_list_9TypedList_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyTypeObject *__pyx_v_list_type = 0;
  PyObject *__pyx_v_listener = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_list_type,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_list_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_listener);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 20, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_list_type = ((PyTypeObject*)values[0]);
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 20, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.TypedList.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_list_type), (&PyType_Type), 1, "list_type", 1))) __PYX_ERR(0, 20, __pyx_L1_error)
  __pyx_r = __pyx_pf_12pyrobuf_list_9TypedList___init__(((struct __pyx_obj_12pyrobuf_list_TypedList *)__pyx_v_self), __pyx_v_list_type, __pyx_v_listener);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_9TypedList___init__(struct __pyx_obj_12pyrobuf_list_TypedList *__pyx_v_self, PyTypeObject *__pyx_v_list_type, PyObject *__pyx_v_listener) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyrobuf_list.pyx":21
 * 
 *     def __init__(self, type list_type, object listener=None):
 *         self._list_type = list_type             # <<<<<<<<<<<<<<
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_list_type));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_list_type));
  __Pyx_GOTREF(__pyx_v_self->_list_type);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_list_type));
  __pyx_v_self->_list_type = __pyx_v_list_type;

  /* "pyrobuf_list.pyx":22
 *     def __init__(self, type list_type, object listener=None):
 *         self._list_type = list_type
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
  __pyx_t_1 = (__pyx_v_listener == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":23
 *         self._list_type = list_type
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener             # <<<<<<<<<<<<<<
 *         else:
 *             self._listener = <PyObject *>listener
 */
    __pyx_v_self->_listener = ((PyObject *)null_listener);

    /* "pyrobuf_list.pyx":22
 *     def __init__(self, type list_type, object listener=None):
 *         self._list_type = list_type
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
    goto __pyx_L3;
  }

  /* "pyrobuf_list.pyx":25
 *             self._listener = <PyObject *>null_listener
 *         else:
 *             self._listener = <PyObject *>listener             # <<<<<<<<<<<<<<
 * 
 *     property list_type:
 */
  /*else*/ {
    __pyx_v_self->_listener = ((PyObject *)__pyx_v_listener);
  }
  __pyx_L3:;

  /* "pyrobuf_list.pyx":20
 * cdef class TypedList(list):
 * 
 *     def __init__(self, type list_type, object listener=None):             # <<<<<<<<<<<<<<
 *         self._list_type = list_type
 *         if listener is None:
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":28
 * 
 *     property list_type:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._list_type
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9TypedList_9list_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9TypedList_9list_type_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9TypedList_9list_type___get__(((struct __pyx_obj_12pyrobuf_list_TypedList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9TypedList_9list_type___get__(struct __pyx_obj_12pyrobuf_list_TypedList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "pyrobuf_list.pyx":29
 *     property list_type:
 *         def __get__(self):
 *             return self._list_type             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, i, x):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_list_type));
  __pyx_r = ((PyObject *)__pyx_v_self->_list_type);
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":28
 * 
 *     property list_type:
 *         def __get__(self):             # <<<<<<<<<<<<<<
 *             return self._list_type
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":31
 *             return self._list_type
 * 
 *     def __setitem__(self, i, x):             # <<<<<<<<<<<<<<
 *         try:
 *             assert isinstance(x, self._list_type)
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_9TypedList_3__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_x); /*proto*/
static int __pyx_pw_12pyrobuf_list_9TypedList_3__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_x) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9TypedList_2__setitem__(((struct __pyx_obj_12pyrobuf_list_TypedList *)__pyx_v_self), ((PyObject *)__pyx_v_i), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_9TypedList_2__setitem__(struct __pyx_obj_12pyrobuf_list_TypedList *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_x) {
  PyObject *__pyx_v_listener = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "pyrobuf_list.pyx":32
 * 
 *     def __setitem__(self, i, x):
 *         try:             # <<<<<<<<<<<<<<
 *             assert isinstance(x, self._list_type)
 *         except AssertionError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "pyrobuf_list.pyx":33
 *     def __setitem__(self, i, x):
 *         try:
 *             assert isinstance(x, self._list_type)             # <<<<<<<<<<<<<<
 *         except AssertionError:
 *             raise Exception("type mismatch")
 */
      #ifndef CYTHON_WITHOUT_ASSERTIONS
      if (unlikely(!Py_OptimizeFlag)) {
        __pyx_t_4 = ((PyObject *)__pyx_v_self->_list_type);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_x, __pyx_t_4); 
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!(__pyx_t_5 != 0))) {
          PyErr_SetNone(PyExc_AssertionError);
          __PYX_ERR(0, 33, __pyx_L3_error)
        }
      }
      #endif

      /* "pyrobuf_list.pyx":32
 * 
 *     def __setitem__(self, i, x):
 *         try:             # <<<<<<<<<<<<<<
 *             assert isinstance(x, self._list_type)
 *         except AssertionError:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "pyrobuf_list.pyx":34
 *         try:
 *             assert isinstance(x, self._list_type)
 *         except AssertionError:             # <<<<<<<<<<<<<<
 *             raise Exception("type mismatch")
 * 
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AssertionError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("pyrobuf_list.TypedList.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 34, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_8);

      /* "pyrobuf_list.pyx":35
 *             assert isinstance(x, self._list_type)
 *         except AssertionError:
 *             raise Exception("type mismatch")             # <<<<<<<<<<<<<<
 * 
 *         super(TypedList, self).__setitem__(i, x)
 */
      __pyx_t_9 = __Pyx_PyObject_Call(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_tuple_, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 35, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 35, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "pyrobuf_list.pyx":32
 * 
 *     def __setitem__(self, i, x):
 *         try:             # <<<<<<<<<<<<<<
 *             assert isinstance(x, self._list_type)
 *         except AssertionError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "pyrobuf_list.pyx":37
 *             raise Exception("type mismatch")
 * 
 *         super(TypedList, self).__setitem__(i, x)             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(((PyObject *)__pyx_ptype_12pyrobuf_list_TypedList));
  __Pyx_GIVEREF(((PyObject *)__pyx_ptype_12pyrobuf_list_TypedList));
  PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)__pyx_ptype_12pyrobuf_list_TypedList));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_7, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_setitem); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_7)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_i, __pyx_v_x};
    __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 37, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_8);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_i, __pyx_v_x};
    __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 37, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_8);
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_6, __pyx_v_i);
    __Pyx_INCREF(__pyx_v_x);
    __Pyx_GIVEREF(__pyx_v_x);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_6, __pyx_v_x);
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_9, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pyrobuf_list.pyx":38
 * 
 *         super(TypedList, self).__setitem__(i, x)
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_8 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_8);
  __pyx_v_listener = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "pyrobuf_list.pyx":39
 *         super(TypedList, self).__setitem__(i, x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_10 = ((!__pyx_t_5) != 0);
  if (__pyx_t_10) {

    /* "pyrobuf_list.pyx":40
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 40, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    if (__pyx_t_9) {
      __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 40, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    } else {
      __pyx_t_8 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 40, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "pyrobuf_list.pyx":39
 *         super(TypedList, self).__setitem__(i, x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":31
 *             return self._list_type
 * 
 *     def __setitem__(self, i, x):             # <<<<<<<<<<<<<<
 *         try:
 *             assert isinstance(x, self._list_type)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pyrobuf_list.TypedList.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":44
 * 
 * 
 *     def add(self, **kwargs):             # <<<<<<<<<<<<<<
 *         elt = self._list_type(**kwargs)
 *         super(TypedList, self).append(elt)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9TypedList_5add(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9TypedList_5add(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("add", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return NULL;}
  if (__pyx_kwds && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "add", 1))) return NULL;
  if (__pyx_kwds) {
    __pyx_v_kwargs = PyDict_Copy(__pyx_kwds); if (unlikely(!__pyx_v_kwargs)) return NULL;
    __Pyx_GOTREF(__pyx_v_kwargs);
  } else {
    __pyx_v_kwargs = NULL;
  }
  __pyx_r = __pyx_pf_12pyrobuf_list_9TypedList_4add(((struct __pyx_obj_12pyrobuf_list_TypedList *)__pyx_v_self), __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9TypedList_4add(struct __pyx_obj_12pyrobuf_list_TypedList *__pyx_v_self, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_elt = NULL;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("add", 0);

  /* "pyrobuf_list.pyx":45
 * 
 *     def add(self, **kwargs):
 *         elt = self._list_type(**kwargs)             # <<<<<<<<<<<<<<
 *         super(TypedList, self).append(elt)
 *         listener = <object>self._listener
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_v_self->_list_type), __pyx_empty_tuple, __pyx_v_kwargs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_elt = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":46
 *     def add(self, **kwargs):
 *         elt = self._list_type(**kwargs)
 *         super(TypedList, self).append(elt)             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_ptype_12pyrobuf_list_TypedList));
  __Pyx_GIVEREF(((PyObject *)__pyx_ptype_12pyrobuf_list_TypedList));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_ptype_12pyrobuf_list_TypedList));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_Append(__pyx_t_2, __pyx_v_elt); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyrobuf_list.pyx":47
 *         elt = self._list_type(**kwargs)
 *         super(TypedList, self).append(elt)
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_listener = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "pyrobuf_list.pyx":48
 *         super(TypedList, self).append(elt)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 *         return elt
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (__pyx_t_5) {

    /* "pyrobuf_list.pyx":49
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 *         return elt
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    if (__pyx_t_6) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyrobuf_list.pyx":48
 *         super(TypedList, self).append(elt)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 *         return elt
 */
  }

  /* "pyrobuf_list.pyx":50
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 *         return elt             # <<<<<<<<<<<<<<
 * 
 *     def append(self, x):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_elt);
  __pyx_r = __pyx_v_elt;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":44
 * 
 * 
 *     def add(self, **kwargs):             # <<<<<<<<<<<<<<
 *         elt = self._list_type(**kwargs)
 *         super(TypedList, self).append(elt)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.TypedList.add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_elt);
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":52
 *         return elt
 * 
 *     def append(self, x):             # <<<<<<<<<<<<<<
 *         try:
 *             assert isinstance(x, self._list_type)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9TypedList_7append(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9TypedList_7append(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9TypedList_6append(((struct __pyx_obj_12pyrobuf_list_TypedList *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9TypedList_6append(struct __pyx_obj_12pyrobuf_list_TypedList *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("append", 0);

  /* "pyrobuf_list.pyx":53
 * 
 *     def append(self, x):
 *         try:             # <<<<<<<<<<<<<<
 *             assert isinstance(x, self._list_type)
 *         except AssertionError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "pyrobuf_list.pyx":54
 *     def append(self, x):
 *         try:
 *             assert isinstance(x, self._list_type)             # <<<<<<<<<<<<<<
 *         except AssertionError:
 *             raise Exception("type mismatch")
 */
      #ifndef CYTHON_WITHOUT_ASSERTIONS
      if (unlikely(!Py_OptimizeFlag)) {
        __pyx_t_4 = ((PyObject *)__pyx_v_self->_list_type);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_x, __pyx_t_4); 
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!(__pyx_t_5 != 0))) {
          PyErr_SetNone(PyExc_AssertionError);
          __PYX_ERR(0, 54, __pyx_L3_error)
        }
      }
      #endif

      /* "pyrobuf_list.pyx":53
 * 
 *     def append(self, x):
 *         try:             # <<<<<<<<<<<<<<
 *             assert isinstance(x, self._list_type)
 *         except AssertionError:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "pyrobuf_list.pyx":55
 *         try:
 *             assert isinstance(x, self._list_type)
 *         except AssertionError:             # <<<<<<<<<<<<<<
 *             raise Exception("type mismatch")
 * 
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AssertionError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("pyrobuf_list.TypedList.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 55, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_8);

      /* "pyrobuf_list.pyx":56
 *             assert isinstance(x, self._list_type)
 *         except AssertionError:
 *             raise Exception("type mismatch")             # <<<<<<<<<<<<<<
 * 
 *         super(TypedList, self).append(x)
 */
      __pyx_t_9 = __Pyx_PyObject_Call(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 56, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 56, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "pyrobuf_list.pyx":53
 * 
 *     def append(self, x):
 *         try:             # <<<<<<<<<<<<<<
 *             assert isinstance(x, self._list_type)
 *         except AssertionError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "pyrobuf_list.pyx":58
 *             raise Exception("type mismatch")
 * 
 *         super(TypedList, self).append(x)             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(((PyObject *)__pyx_ptype_12pyrobuf_list_TypedList));
  __Pyx_GIVEREF(((PyObject *)__pyx_ptype_12pyrobuf_list_TypedList));
  PyTuple_SET_ITEM(__pyx_t_8, 0, ((PyObject *)__pyx_ptype_12pyrobuf_list_TypedList));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_8, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_10 = __Pyx_PyObject_Append(__pyx_t_7, __pyx_v_x); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 58, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pyrobuf_list.pyx":59
 * 
 *         super(TypedList, self).append(x)
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_7 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_7);
  __pyx_v_listener = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "pyrobuf_list.pyx":60
 *         super(TypedList, self).append(x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_11 = ((!__pyx_t_5) != 0);
  if (__pyx_t_11) {

    /* "pyrobuf_list.pyx":61
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def extend(self, x):
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 61, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 61, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 61, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "pyrobuf_list.pyx":60
 *         super(TypedList, self).append(x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":52
 *         return elt
 * 
 *     def append(self, x):             # <<<<<<<<<<<<<<
 *         try:
 *             assert isinstance(x, self._list_type)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pyrobuf_list.TypedList.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":63
 *             listener._Modified()
 * 
 *     def extend(self, x):             # <<<<<<<<<<<<<<
 *         for i in x:
 *             elt = self._list_type()
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9TypedList_9extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9TypedList_9extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extend (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9TypedList_8extend(((struct __pyx_obj_12pyrobuf_list_TypedList *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9TypedList_8extend(struct __pyx_obj_12pyrobuf_list_TypedList *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_elt = NULL;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("extend", 0);

  /* "pyrobuf_list.pyx":64
 * 
 *     def extend(self, x):
 *         for i in x:             # <<<<<<<<<<<<<<
 *             elt = self._list_type()
 *             elt.MergeFrom(i)
 */
  if (likely(PyList_CheckExact(__pyx_v_x)) || PyTuple_CheckExact(__pyx_v_x)) {
    __pyx_t_1 = __pyx_v_x; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 64, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 64, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 64, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 64, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 64, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 64, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 64, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "pyrobuf_list.pyx":65
 *     def extend(self, x):
 *         for i in x:
 *             elt = self._list_type()             # <<<<<<<<<<<<<<
 *             elt.MergeFrom(i)
 *             super(TypedList, self).append(elt)
 */
    __pyx_t_4 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_v_self->_list_type)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 65, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_elt, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "pyrobuf_list.pyx":66
 *         for i in x:
 *             elt = self._list_type()
 *             elt.MergeFrom(i)             # <<<<<<<<<<<<<<
 *             super(TypedList, self).append(elt)
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_elt, __pyx_n_s_MergeFrom); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 66, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_i};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 66, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_4);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_i};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 66, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_4);
      } else
      #endif
      {
        __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 66, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_v_i);
        __Pyx_GIVEREF(__pyx_v_i);
        PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_i);
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 66, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "pyrobuf_list.pyx":67
 *             elt = self._list_type()
 *             elt.MergeFrom(i)
 *             super(TypedList, self).append(elt)             # <<<<<<<<<<<<<<
 * 
 *         listener = <object>self._listener
 */
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)__pyx_ptype_12pyrobuf_list_TypedList));
    __Pyx_GIVEREF(((PyObject *)__pyx_ptype_12pyrobuf_list_TypedList));
    PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)__pyx_ptype_12pyrobuf_list_TypedList));
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)__pyx_v_self));
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_8 = __Pyx_PyObject_Append(__pyx_t_5, __pyx_v_elt); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 67, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "pyrobuf_list.pyx":64
 * 
 *     def extend(self, x):
 *         for i in x:             # <<<<<<<<<<<<<<
 *             elt = self._list_type()
 *             elt.MergeFrom(i)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":69
 *             super(TypedList, self).append(elt)
 * 
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":70
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = ((!__pyx_t_9) != 0);
  if (__pyx_t_10) {

    /* "pyrobuf_list.pyx":71
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def insert(self, i, x):
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":70
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":63
 *             listener._Modified()
 * 
 *     def extend(self, x):             # <<<<<<<<<<<<<<
 *         for i in x:
 *             elt = self._list_type()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyrobuf_list.TypedList.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_elt);
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":73
 *             listener._Modified()
 * 
 *     def insert(self, i, x):             # <<<<<<<<<<<<<<
 *         try:
 *             assert isinstance(x, self._list_type)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9TypedList_11insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9TypedList_11insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_v_x = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("insert (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_i,&__pyx_n_s_x,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, 1); __PYX_ERR(0, 73, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "insert") < 0)) __PYX_ERR(0, 73, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_i = values[0];
    __pyx_v_x = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 73, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.TypedList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9TypedList_10insert(((struct __pyx_obj_12pyrobuf_list_TypedList *)__pyx_v_self), __pyx_v_i, __pyx_v_x);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9TypedList_10insert(struct __pyx_obj_12pyrobuf_list_TypedList *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_x) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("insert", 0);

  /* "pyrobuf_list.pyx":74
 * 
 *     def insert(self, i, x):
 *         try:             # <<<<<<<<<<<<<<
 *             assert isinstance(x, self._list_type)
 *         except AssertionError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "pyrobuf_list.pyx":75
 *     def insert(self, i, x):
 *         try:
 *             assert isinstance(x, self._list_type)             # <<<<<<<<<<<<<<
 *         except AssertionError:
 *             raise Exception("type mismatch")
 */
      #ifndef CYTHON_WITHOUT_ASSERTIONS
      if (unlikely(!Py_OptimizeFlag)) {
        __pyx_t_4 = ((PyObject *)__pyx_v_self->_list_type);
        __Pyx_INCREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_x, __pyx_t_4); 
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!(__pyx_t_5 != 0))) {
          PyErr_SetNone(PyExc_AssertionError);
          __PYX_ERR(0, 75, __pyx_L3_error)
        }
      }
      #endif

      /* "pyrobuf_list.pyx":74
 * 
 *     def insert(self, i, x):
 *         try:             # <<<<<<<<<<<<<<
 *             assert isinstance(x, self._list_type)
 *         except AssertionError:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "pyrobuf_list.pyx":76
 *         try:
 *             assert isinstance(x, self._list_type)
 *         except AssertionError:             # <<<<<<<<<<<<<<
 *             raise Exception("type mismatch")
 * 
 */
    __pyx_t_6 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AssertionError);
    if (__pyx_t_6) {
      __Pyx_AddTraceback("pyrobuf_list.TypedList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_7, &__pyx_t_8) < 0) __PYX_ERR(0, 76, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_8);

      /* "pyrobuf_list.pyx":77
 *             assert isinstance(x, self._list_type)
 *         except AssertionError:
 *             raise Exception("type mismatch")             # <<<<<<<<<<<<<<
 * 
 *         super(TypedList, self).insert(i, x)
 */
      __pyx_t_9 = __Pyx_PyObject_Call(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 77, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_Raise(__pyx_t_9, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __PYX_ERR(0, 77, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "pyrobuf_list.pyx":74
 * 
 *     def insert(self, i, x):
 *         try:             # <<<<<<<<<<<<<<
 *             assert isinstance(x, self._list_type)
 *         except AssertionError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "pyrobuf_list.pyx":79
 *             raise Exception("type mismatch")
 * 
 *         super(TypedList, self).insert(i, x)             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(((PyObject *)__pyx_ptype_12pyrobuf_list_TypedList));
  __Pyx_GIVEREF(((PyObject *)__pyx_ptype_12pyrobuf_list_TypedList));
  PyTuple_SET_ITEM(__pyx_t_7, 0, ((PyObject *)__pyx_ptype_12pyrobuf_list_TypedList));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_7, 1, ((PyObject *)__pyx_v_self));
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_insert); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_7, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_7)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_i, __pyx_v_x};
    __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_8);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_i, __pyx_v_x};
    __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_8);
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_i);
    __Pyx_GIVEREF(__pyx_v_i);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_6, __pyx_v_i);
    __Pyx_INCREF(__pyx_v_x);
    __Pyx_GIVEREF(__pyx_v_x);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_6, __pyx_v_x);
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_9, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 79, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pyrobuf_list.pyx":80
 * 
 *         super(TypedList, self).insert(i, x)
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_8 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_8);
  __pyx_v_listener = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "pyrobuf_list.pyx":81
 *         super(TypedList, self).insert(i, x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_10 = ((!__pyx_t_5) != 0);
  if (__pyx_t_10) {

    /* "pyrobuf_list.pyx":82
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    if (__pyx_t_9) {
      __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 82, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    } else {
      __pyx_t_8 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 82, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "pyrobuf_list.pyx":81
 *         super(TypedList, self).insert(i, x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":73
 *             listener._Modified()
 * 
 *     def insert(self, i, x):             # <<<<<<<<<<<<<<
 *         try:
 *             assert isinstance(x, self._list_type)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pyrobuf_list.TypedList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9TypedList_13__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9TypedList_13__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9TypedList_12__reduce_cython__(((struct __pyx_obj_12pyrobuf_list_TypedList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9TypedList_12__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_TypedList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.TypedList.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9TypedList_15__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9TypedList_15__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9TypedList_14__setstate_cython__(((struct __pyx_obj_12pyrobuf_list_TypedList *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9TypedList_14__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_TypedList *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.TypedList.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":87
 * cdef class BytesList(list):
 * 
 *     def __init__(self, object listener=None):             # <<<<<<<<<<<<<<
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_9BytesList_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12pyrobuf_list_9BytesList_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_listener = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_listener,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_listener);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 87, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_listener = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 87, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.BytesList.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9BytesList___init__(((struct __pyx_obj_12pyrobuf_list_BytesList *)__pyx_v_self), __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_9BytesList___init__(struct __pyx_obj_12pyrobuf_list_BytesList *__pyx_v_self, PyObject *__pyx_v_listener) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyrobuf_list.pyx":88
 * 
 *     def __init__(self, object listener=None):
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
  __pyx_t_1 = (__pyx_v_listener == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":89
 *     def __init__(self, object listener=None):
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener             # <<<<<<<<<<<<<<
 *         else:
 *             self._listener = <PyObject *>listener
 */
    __pyx_v_self->_listener = ((PyObject *)null_listener);

    /* "pyrobuf_list.pyx":88
 * 
 *     def __init__(self, object listener=None):
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
    goto __pyx_L3;
  }

  /* "pyrobuf_list.pyx":91
 *             self._listener = <PyObject *>null_listener
 *         else:
 *             self._listener = <PyObject *>listener             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, i, x):
 */
  /*else*/ {
    __pyx_v_self->_listener = ((PyObject *)__pyx_v_listener);
  }
  __pyx_L3:;

  /* "pyrobuf_list.pyx":87
 * cdef class BytesList(list):
 * 
 *     def __init__(self, object listener=None):             # <<<<<<<<<<<<<<
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":93
 *             self._listener = <PyObject *>listener
 * 
 *     def __setitem__(self, i, x):             # <<<<<<<<<<<<<<
 *         cdef list values
 * 
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_9BytesList_3__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_x); /*proto*/
static int __pyx_pw_12pyrobuf_list_9BytesList_3__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_x) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9BytesList_2__setitem__(((struct __pyx_obj_12pyrobuf_list_BytesList *)__pyx_v_self), ((PyObject *)__pyx_v_i), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_9BytesList_2__setitem__(struct __pyx_obj_12pyrobuf_list_BytesList *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_x) {
  PyObject *__pyx_v_values = 0;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_listener = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "pyrobuf_list.pyx":96
 *         cdef list values
 * 
 *         if isinstance(i, int):             # <<<<<<<<<<<<<<
 *             if not isinstance(x, bytes):
 *                 raise TypeError("%r has type %s, but expected one of: (%s,)" % (x, type(x), bytes))
 */
  __pyx_t_1 = PyInt_Check(__pyx_v_i); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":97
 * 
 *         if isinstance(i, int):
 *             if not isinstance(x, bytes):             # <<<<<<<<<<<<<<
 *                 raise TypeError("%r has type %s, but expected one of: (%s,)" % (x, type(x), bytes))
 * 
 */
    __pyx_t_2 = PyBytes_Check(__pyx_v_x); 
    __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "pyrobuf_list.pyx":98
 *         if isinstance(i, int):
 *             if not isinstance(x, bytes):
 *                 raise TypeError("%r has type %s, but expected one of: (%s,)" % (x, type(x), bytes))             # <<<<<<<<<<<<<<
 * 
 *             list.__setitem__(self, i, x)
 */
      __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 98, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_v_x);
      __Pyx_GIVEREF(__pyx_v_x);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_x);
      __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_x)));
      __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_x)));
      PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)Py_TYPE(__pyx_v_x)));
      __Pyx_INCREF(((PyObject *)(&PyBytes_Type)));
      __Pyx_GIVEREF(((PyObject *)(&PyBytes_Type)));
      PyTuple_SET_ITEM(__pyx_t_3, 2, ((PyObject *)(&PyBytes_Type)));
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_r_has_type_s_but_expected_one_o, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 98, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 98, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 98, __pyx_L1_error)

      /* "pyrobuf_list.pyx":97
 * 
 *         if isinstance(i, int):
 *             if not isinstance(x, bytes):             # <<<<<<<<<<<<<<
 *                 raise TypeError("%r has type %s, but expected one of: (%s,)" % (x, type(x), bytes))
 * 
 */
    }

    /* "pyrobuf_list.pyx":100
 *                 raise TypeError("%r has type %s, but expected one of: (%s,)" % (x, type(x), bytes))
 * 
 *             list.__setitem__(self, i, x)             # <<<<<<<<<<<<<<
 *         elif isinstance(i, slice):
 *             values = list()
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyList_Type)), __pyx_n_s_setitem); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 100, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[4] = {__pyx_t_5, ((PyObject *)__pyx_v_self), __pyx_v_i, __pyx_v_x};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 100, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[4] = {__pyx_t_5, ((PyObject *)__pyx_v_self), __pyx_v_i, __pyx_v_x};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 100, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_self));
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_i);
      __Pyx_INCREF(__pyx_v_x);
      __Pyx_GIVEREF(__pyx_v_x);
      PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_v_x);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":96
 *         cdef list values
 * 
 *         if isinstance(i, int):             # <<<<<<<<<<<<<<
 *             if not isinstance(x, bytes):
 *                 raise TypeError("%r has type %s, but expected one of: (%s,)" % (x, type(x), bytes))
 */
    goto __pyx_L3;
  }

  /* "pyrobuf_list.pyx":101
 * 
 *             list.__setitem__(self, i, x)
 *         elif isinstance(i, slice):             # <<<<<<<<<<<<<<
 *             values = list()
 * 
 */
  __pyx_t_1 = PySlice_Check(__pyx_v_i); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (likely(__pyx_t_2)) {

    /* "pyrobuf_list.pyx":102
 *             list.__setitem__(self, i, x)
 *         elif isinstance(i, slice):
 *             values = list()             # <<<<<<<<<<<<<<
 * 
 *             for val in x:
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_values = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":104
 *             values = list()
 * 
 *             for val in x:             # <<<<<<<<<<<<<<
 *                 if isinstance(val, bytes):
 *                     values.append(val)
 */
    if (likely(PyList_CheckExact(__pyx_v_x)) || PyTuple_CheckExact(__pyx_v_x)) {
      __pyx_t_3 = __pyx_v_x; __Pyx_INCREF(__pyx_t_3); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 104, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_4); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 104, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 104, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8); __Pyx_INCREF(__pyx_t_4); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 104, __pyx_L1_error)
          #else
          __pyx_t_4 = PySequence_ITEM(__pyx_t_3, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 104, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_9(__pyx_t_3);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 104, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "pyrobuf_list.pyx":105
 * 
 *             for val in x:
 *                 if isinstance(val, bytes):             # <<<<<<<<<<<<<<
 *                     values.append(val)
 *                 else:
 */
      __pyx_t_2 = PyBytes_Check(__pyx_v_val); 
      __pyx_t_1 = (__pyx_t_2 != 0);
      if (likely(__pyx_t_1)) {

        /* "pyrobuf_list.pyx":106
 *             for val in x:
 *                 if isinstance(val, bytes):
 *                     values.append(val)             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise TypeError("%r has type %s, but expected one of: (%s,)" % (val, type(val), bytes))
 */
        __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_values, __pyx_v_val); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 106, __pyx_L1_error)

        /* "pyrobuf_list.pyx":105
 * 
 *             for val in x:
 *                 if isinstance(val, bytes):             # <<<<<<<<<<<<<<
 *                     values.append(val)
 *                 else:
 */
        goto __pyx_L7;
      }

      /* "pyrobuf_list.pyx":108
 *                     values.append(val)
 *                 else:
 *                     raise TypeError("%r has type %s, but expected one of: (%s,)" % (val, type(val), bytes))             # <<<<<<<<<<<<<<
 * 
 *             list.__setitem__(self, i, values)
 */
      /*else*/ {
        __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 108, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_v_val);
        __Pyx_GIVEREF(__pyx_v_val);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_val);
        __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_val)));
        __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_val)));
        PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)Py_TYPE(__pyx_v_val)));
        __Pyx_INCREF(((PyObject *)(&PyBytes_Type)));
        __Pyx_GIVEREF(((PyObject *)(&PyBytes_Type)));
        PyTuple_SET_ITEM(__pyx_t_4, 2, ((PyObject *)(&PyBytes_Type)));
        __pyx_t_7 = __Pyx_PyString_Format(__pyx_kp_s_r_has_type_s_but_expected_one_o, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 108, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 108, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 108, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "pyrobuf_list.pyx":104
 *             values = list()
 * 
 *             for val in x:             # <<<<<<<<<<<<<<
 *                 if isinstance(val, bytes):
 *                     values.append(val)
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":110
 *                     raise TypeError("%r has type %s, but expected one of: (%s,)" % (val, type(val), bytes))
 * 
 *             list.__setitem__(self, i, values)             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError("list indices must be integers, not %s" % type(i).__name__)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyList_Type)), __pyx_n_s_setitem); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[4] = {__pyx_t_7, ((PyObject *)__pyx_v_self), __pyx_v_i, __pyx_v_values};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[4] = {__pyx_t_7, ((PyObject *)__pyx_v_self), __pyx_v_i, __pyx_v_values};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_self));
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_v_i);
      __Pyx_INCREF(__pyx_v_values);
      __Pyx_GIVEREF(__pyx_v_values);
      PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_v_values);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":101
 * 
 *             list.__setitem__(self, i, x)
 *         elif isinstance(i, slice):             # <<<<<<<<<<<<<<
 *             values = list()
 * 
 */
    goto __pyx_L3;
  }

  /* "pyrobuf_list.pyx":112
 *             list.__setitem__(self, i, values)
 *         else:
 *             raise TypeError("list indices must be integers, not %s" % type(i).__name__)             # <<<<<<<<<<<<<<
 * 
 *         listener = <object>self._listener
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_v_i)), __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_list_indices_must_be_integers_no, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 112, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 112, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "pyrobuf_list.pyx":114
 *             raise TypeError("list indices must be integers, not %s" % type(i).__name__)
 * 
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":115
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":116
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 116, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 116, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 116, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":115
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":93
 *             self._listener = <PyObject *>listener
 * 
 *     def __setitem__(self, i, x):             # <<<<<<<<<<<<<<
 *         cdef list values
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyrobuf_list.BytesList.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_values);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":120
 * 
 * 
 *     def append(self, x):             # <<<<<<<<<<<<<<
 *         if not isinstance(x, bytes):
 *             raise TypeError("%r has type %s, but expected one of: (%s,)" % (x, type(x), bytes))
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9BytesList_5append(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9BytesList_5append(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9BytesList_4append(((struct __pyx_obj_12pyrobuf_list_BytesList *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9BytesList_4append(struct __pyx_obj_12pyrobuf_list_BytesList *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("append", 0);

  /* "pyrobuf_list.pyx":121
 * 
 *     def append(self, x):
 *         if not isinstance(x, bytes):             # <<<<<<<<<<<<<<
 *             raise TypeError("%r has type %s, but expected one of: (%s,)" % (x, type(x), bytes))
 * 
 */
  __pyx_t_1 = PyBytes_Check(__pyx_v_x); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "pyrobuf_list.pyx":122
 *     def append(self, x):
 *         if not isinstance(x, bytes):
 *             raise TypeError("%r has type %s, but expected one of: (%s,)" % (x, type(x), bytes))             # <<<<<<<<<<<<<<
 * 
 *         list.append(self, x)
 */
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_x);
    __Pyx_GIVEREF(__pyx_v_x);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_x);
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_x)));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_x)));
    PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)Py_TYPE(__pyx_v_x)));
    __Pyx_INCREF(((PyObject *)(&PyBytes_Type)));
    __Pyx_GIVEREF(((PyObject *)(&PyBytes_Type)));
    PyTuple_SET_ITEM(__pyx_t_3, 2, ((PyObject *)(&PyBytes_Type)));
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_r_has_type_s_but_expected_one_o, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 122, __pyx_L1_error)

    /* "pyrobuf_list.pyx":121
 * 
 *     def append(self, x):
 *         if not isinstance(x, bytes):             # <<<<<<<<<<<<<<
 *             raise TypeError("%r has type %s, but expected one of: (%s,)" % (x, type(x), bytes))
 * 
 */
  }

  /* "pyrobuf_list.pyx":124
 *             raise TypeError("%r has type %s, but expected one of: (%s,)" % (x, type(x), bytes))
 * 
 *         list.append(self, x)             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  if (unlikely(((PyObject *)__pyx_v_self) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "descriptor '%s' requires a '%s' object but received a 'NoneType'", "append", "list");
    __PYX_ERR(0, 124, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyList_Append(((PyObject*)__pyx_v_self), __pyx_v_x); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 124, __pyx_L1_error)

  /* "pyrobuf_list.pyx":125
 * 
 *         list.append(self, x)
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":126
 *         list.append(self, x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = ((!__pyx_t_2) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":127
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def extend(self, x):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":126
 *         list.append(self, x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":120
 * 
 * 
 *     def append(self, x):             # <<<<<<<<<<<<<<
 *         if not isinstance(x, bytes):
 *             raise TypeError("%r has type %s, but expected one of: (%s,)" % (x, type(x), bytes))
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.BytesList.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":129
 *             listener._Modified()
 * 
 *     def extend(self, x):             # <<<<<<<<<<<<<<
 *         for val in x:
 *             if not isinstance(val, bytes):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9BytesList_7extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9BytesList_7extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extend (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9BytesList_6extend(((struct __pyx_obj_12pyrobuf_list_BytesList *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9BytesList_6extend(struct __pyx_obj_12pyrobuf_list_BytesList *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("extend", 0);

  /* "pyrobuf_list.pyx":130
 * 
 *     def extend(self, x):
 *         for val in x:             # <<<<<<<<<<<<<<
 *             if not isinstance(val, bytes):
 *                 raise TypeError("%r has type %s, but expected one of: (%s,)" % (val, type(val), bytes))
 */
  if (likely(PyList_CheckExact(__pyx_v_x)) || PyTuple_CheckExact(__pyx_v_x)) {
    __pyx_t_1 = __pyx_v_x; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 130, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 130, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 130, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 130, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 130, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 130, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "pyrobuf_list.pyx":131
 *     def extend(self, x):
 *         for val in x:
 *             if not isinstance(val, bytes):             # <<<<<<<<<<<<<<
 *                 raise TypeError("%r has type %s, but expected one of: (%s,)" % (val, type(val), bytes))
 *             list.append(self, val)
 */
    __pyx_t_5 = PyBytes_Check(__pyx_v_val); 
    __pyx_t_6 = ((!(__pyx_t_5 != 0)) != 0);
    if (unlikely(__pyx_t_6)) {

      /* "pyrobuf_list.pyx":132
 *         for val in x:
 *             if not isinstance(val, bytes):
 *                 raise TypeError("%r has type %s, but expected one of: (%s,)" % (val, type(val), bytes))             # <<<<<<<<<<<<<<
 *             list.append(self, val)
 * 
 */
      __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 132, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_v_val);
      __Pyx_GIVEREF(__pyx_v_val);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_val);
      __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_val)));
      __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_val)));
      PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)Py_TYPE(__pyx_v_val)));
      __Pyx_INCREF(((PyObject *)(&PyBytes_Type)));
      __Pyx_GIVEREF(((PyObject *)(&PyBytes_Type)));
      PyTuple_SET_ITEM(__pyx_t_4, 2, ((PyObject *)(&PyBytes_Type)));
      __pyx_t_7 = __Pyx_PyString_Format(__pyx_kp_s_r_has_type_s_but_expected_one_o, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 132, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 132, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 132, __pyx_L1_error)

      /* "pyrobuf_list.pyx":131
 *     def extend(self, x):
 *         for val in x:
 *             if not isinstance(val, bytes):             # <<<<<<<<<<<<<<
 *                 raise TypeError("%r has type %s, but expected one of: (%s,)" % (val, type(val), bytes))
 *             list.append(self, val)
 */
    }

    /* "pyrobuf_list.pyx":133
 *             if not isinstance(val, bytes):
 *                 raise TypeError("%r has type %s, but expected one of: (%s,)" % (val, type(val), bytes))
 *             list.append(self, val)             # <<<<<<<<<<<<<<
 * 
 *         listener = <object>self._listener
 */
    if (unlikely(((PyObject *)__pyx_v_self) == Py_None)) {
      PyErr_Format(PyExc_TypeError, "descriptor '%s' requires a '%s' object but received a 'NoneType'", "append", "list");
      __PYX_ERR(0, 133, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_PyList_Append(((PyObject*)__pyx_v_self), __pyx_v_val); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 133, __pyx_L1_error)

    /* "pyrobuf_list.pyx":130
 * 
 *     def extend(self, x):
 *         for val in x:             # <<<<<<<<<<<<<<
 *             if not isinstance(val, bytes):
 *                 raise TypeError("%r has type %s, but expected one of: (%s,)" % (val, type(val), bytes))
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":135
 *             list.append(self, val)
 * 
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":136
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = ((!__pyx_t_6) != 0);
  if (__pyx_t_5) {

    /* "pyrobuf_list.pyx":137
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def insert(self, i, x):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_7) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 137, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":136
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":129
 *             listener._Modified()
 * 
 *     def extend(self, x):             # <<<<<<<<<<<<<<
 *         for val in x:
 *             if not isinstance(val, bytes):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyrobuf_list.BytesList.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":139
 *             listener._Modified()
 * 
 *     def insert(self, i, x):             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(x, bytes):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9BytesList_9insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9BytesList_9insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_v_x = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("insert (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_i,&__pyx_n_s_x,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, 1); __PYX_ERR(0, 139, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "insert") < 0)) __PYX_ERR(0, 139, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_i = values[0];
    __pyx_v_x = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 139, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.BytesList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9BytesList_8insert(((struct __pyx_obj_12pyrobuf_list_BytesList *)__pyx_v_self), __pyx_v_i, __pyx_v_x);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9BytesList_8insert(struct __pyx_obj_12pyrobuf_list_BytesList *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_x) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("insert", 0);

  /* "pyrobuf_list.pyx":141
 *     def insert(self, i, x):
 * 
 *         if not isinstance(x, bytes):             # <<<<<<<<<<<<<<
 *             raise TypeError("%r has type %s, but expected one of: (%s,)" % (x, type(x), bytes))
 * 
 */
  __pyx_t_1 = PyBytes_Check(__pyx_v_x); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "pyrobuf_list.pyx":142
 * 
 *         if not isinstance(x, bytes):
 *             raise TypeError("%r has type %s, but expected one of: (%s,)" % (x, type(x), bytes))             # <<<<<<<<<<<<<<
 * 
 *         list.insert(self, i, x)
 */
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_x);
    __Pyx_GIVEREF(__pyx_v_x);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_x);
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_x)));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_x)));
    PyTuple_SET_ITEM(__pyx_t_3, 1, ((PyObject *)Py_TYPE(__pyx_v_x)));
    __Pyx_INCREF(((PyObject *)(&PyBytes_Type)));
    __Pyx_GIVEREF(((PyObject *)(&PyBytes_Type)));
    PyTuple_SET_ITEM(__pyx_t_3, 2, ((PyObject *)(&PyBytes_Type)));
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_r_has_type_s_but_expected_one_o, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 142, __pyx_L1_error)

    /* "pyrobuf_list.pyx":141
 *     def insert(self, i, x):
 * 
 *         if not isinstance(x, bytes):             # <<<<<<<<<<<<<<
 *             raise TypeError("%r has type %s, but expected one of: (%s,)" % (x, type(x), bytes))
 * 
 */
  }

  /* "pyrobuf_list.pyx":144
 *             raise TypeError("%r has type %s, but expected one of: (%s,)" % (x, type(x), bytes))
 * 
 *         list.insert(self, i, x)             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  if (unlikely(((PyObject *)__pyx_v_self) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "descriptor '%s' requires a '%s' object but received a 'NoneType'", "insert", "list");
    __PYX_ERR(0, 144, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 144, __pyx_L1_error)
  __pyx_t_6 = PyList_Insert(((PyObject*)__pyx_v_self), __pyx_t_5, __pyx_v_x); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 144, __pyx_L1_error)

  /* "pyrobuf_list.pyx":145
 * 
 *         list.insert(self, i, x)
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":146
 *         list.insert(self, i, x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_1 = ((!__pyx_t_2) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":147
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 147, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_7) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 147, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 147, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":146
 *         list.insert(self, i, x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":139
 *             listener._Modified()
 * 
 *     def insert(self, i, x):             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(x, bytes):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyrobuf_list.BytesList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9BytesList_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9BytesList_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9BytesList_10__reduce_cython__(((struct __pyx_obj_12pyrobuf_list_BytesList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9BytesList_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_BytesList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.BytesList.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9BytesList_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9BytesList_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9BytesList_12__setstate_cython__(((struct __pyx_obj_12pyrobuf_list_BytesList *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9BytesList_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_BytesList *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.BytesList.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":152
 * cdef class StringList(list):
 * 
 *     def __init__(self, object listener=None):             # <<<<<<<<<<<<<<
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_10StringList_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12pyrobuf_list_10StringList_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_listener = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_listener,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_listener);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 152, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_listener = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 152, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.StringList.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10StringList___init__(((struct __pyx_obj_12pyrobuf_list_StringList *)__pyx_v_self), __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_10StringList___init__(struct __pyx_obj_12pyrobuf_list_StringList *__pyx_v_self, PyObject *__pyx_v_listener) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "pyrobuf_list.pyx":153
 * 
 *     def __init__(self, object listener=None):
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
  __pyx_t_1 = (__pyx_v_listener == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":154
 *     def __init__(self, object listener=None):
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener             # <<<<<<<<<<<<<<
 *         else:
 *             self._listener = <PyObject *>listener
 */
    __pyx_v_self->_listener = ((PyObject *)null_listener);

    /* "pyrobuf_list.pyx":153
 * 
 *     def __init__(self, object listener=None):
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
    goto __pyx_L3;
  }

  /* "pyrobuf_list.pyx":156
 *             self._listener = <PyObject *>null_listener
 *         else:
 *             self._listener = <PyObject *>listener             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, i, x):
 */
  /*else*/ {
    __pyx_v_self->_listener = ((PyObject *)__pyx_v_listener);
  }
  __pyx_L3:;

  /* "pyrobuf_list.pyx":152
 * cdef class StringList(list):
 * 
 *     def __init__(self, object listener=None):             # <<<<<<<<<<<<<<
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":158
 *             self._listener = <PyObject *>listener
 * 
 *     def __setitem__(self, i, x):             # <<<<<<<<<<<<<<
 *         cdef list values
 *         cdef str value
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_10StringList_3__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_x); /*proto*/
static int __pyx_pw_12pyrobuf_list_10StringList_3__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_x) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10StringList_2__setitem__(((struct __pyx_obj_12pyrobuf_list_StringList *)__pyx_v_self), ((PyObject *)__pyx_v_i), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_10StringList_2__setitem__(struct __pyx_obj_12pyrobuf_list_StringList *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_x) {
  PyObject *__pyx_v_values = 0;
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_listener = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "pyrobuf_list.pyx":162
 *         cdef str value
 * 
 *         if isinstance(i, int):             # <<<<<<<<<<<<<<
 *             if isinstance(x, str):
 *                 value = x
 */
  __pyx_t_1 = PyInt_Check(__pyx_v_i); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":163
 * 
 *         if isinstance(i, int):
 *             if isinstance(x, str):             # <<<<<<<<<<<<<<
 *                 value = x
 *             elif isinstance(x, bytes):
 */
    __pyx_t_2 = PyString_Check(__pyx_v_x); 
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "pyrobuf_list.pyx":164
 *         if isinstance(i, int):
 *             if isinstance(x, str):
 *                 value = x             # <<<<<<<<<<<<<<
 *             elif isinstance(x, bytes):
 *                 value = x.decode('utf-8')
 */
      if (!(likely(PyString_CheckExact(__pyx_v_x))||((__pyx_v_x) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_x)->tp_name), 0))) __PYX_ERR(0, 164, __pyx_L1_error)
      __pyx_t_3 = __pyx_v_x;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_v_value = ((PyObject*)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "pyrobuf_list.pyx":163
 * 
 *         if isinstance(i, int):
 *             if isinstance(x, str):             # <<<<<<<<<<<<<<
 *                 value = x
 *             elif isinstance(x, bytes):
 */
      goto __pyx_L4;
    }

    /* "pyrobuf_list.pyx":165
 *             if isinstance(x, str):
 *                 value = x
 *             elif isinstance(x, bytes):             # <<<<<<<<<<<<<<
 *                 value = x.decode('utf-8')
 *             else:
 */
    __pyx_t_1 = PyBytes_Check(__pyx_v_x); 
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (likely(__pyx_t_2)) {

      /* "pyrobuf_list.pyx":166
 *                 value = x
 *             elif isinstance(x, bytes):
 *                 value = x.decode('utf-8')             # <<<<<<<<<<<<<<
 *             else:
 *                 raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (x, type(x), bytes, str))
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_decode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 166, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 166, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (!(likely(PyString_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(0, 166, __pyx_L1_error)
      __pyx_v_value = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "pyrobuf_list.pyx":165
 *             if isinstance(x, str):
 *                 value = x
 *             elif isinstance(x, bytes):             # <<<<<<<<<<<<<<
 *                 value = x.decode('utf-8')
 *             else:
 */
      goto __pyx_L4;
    }

    /* "pyrobuf_list.pyx":168
 *                 value = x.decode('utf-8')
 *             else:
 *                 raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (x, type(x), bytes, str))             # <<<<<<<<<<<<<<
 * 
 *             list.__setitem__(self, i, value)
 */
    /*else*/ {
      __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_v_x);
      __Pyx_GIVEREF(__pyx_v_x);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_x);
      __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_x)));
      __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_x)));
      PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)Py_TYPE(__pyx_v_x)));
      __Pyx_INCREF(((PyObject *)(&PyBytes_Type)));
      __Pyx_GIVEREF(((PyObject *)(&PyBytes_Type)));
      PyTuple_SET_ITEM(__pyx_t_4, 2, ((PyObject *)(&PyBytes_Type)));
      __Pyx_INCREF(((PyObject *)(&PyString_Type)));
      __Pyx_GIVEREF(((PyObject *)(&PyString_Type)));
      PyTuple_SET_ITEM(__pyx_t_4, 3, ((PyObject *)(&PyString_Type)));
      __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_r_has_type_s_but_expected_one_o_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 168, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 168, __pyx_L1_error)
    }
    __pyx_L4:;

    /* "pyrobuf_list.pyx":170
 *                 raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (x, type(x), bytes, str))
 * 
 *             list.__setitem__(self, i, value)             # <<<<<<<<<<<<<<
 *         elif isinstance(i, slice):
 *             values = list()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyList_Type)), __pyx_n_s_setitem); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[4] = {__pyx_t_5, ((PyObject *)__pyx_v_self), __pyx_v_i, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 170, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[4] = {__pyx_t_5, ((PyObject *)__pyx_v_self), __pyx_v_i, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 170, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_self));
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_i);
      __Pyx_INCREF(__pyx_v_value);
      __Pyx_GIVEREF(__pyx_v_value);
      PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_6, __pyx_v_value);
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "pyrobuf_list.pyx":162
 *         cdef str value
 * 
 *         if isinstance(i, int):             # <<<<<<<<<<<<<<
 *             if isinstance(x, str):
 *                 value = x
 */
    goto __pyx_L3;
  }

  /* "pyrobuf_list.pyx":171
 * 
 *             list.__setitem__(self, i, value)
 *         elif isinstance(i, slice):             # <<<<<<<<<<<<<<
 *             values = list()
 * 
 */
  __pyx_t_2 = PySlice_Check(__pyx_v_i); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (likely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":172
 *             list.__setitem__(self, i, value)
 *         elif isinstance(i, slice):
 *             values = list()             # <<<<<<<<<<<<<<
 * 
 *             for val in x:
 */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 172, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_values = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "pyrobuf_list.pyx":174
 *             values = list()
 * 
 *             for val in x:             # <<<<<<<<<<<<<<
 *                 if isinstance(val, str):
 *                     values.append(val)
 */
    if (likely(PyList_CheckExact(__pyx_v_x)) || PyTuple_CheckExact(__pyx_v_x)) {
      __pyx_t_4 = __pyx_v_x; __Pyx_INCREF(__pyx_t_4); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_8 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_x); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 174, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 174, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 174, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 174, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 174, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 174, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_9(__pyx_t_4);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 174, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "pyrobuf_list.pyx":175
 * 
 *             for val in x:
 *                 if isinstance(val, str):             # <<<<<<<<<<<<<<
 *                     values.append(val)
 *                 elif isinstance(val, bytes):
 */
      __pyx_t_1 = PyString_Check(__pyx_v_val); 
      __pyx_t_2 = (__pyx_t_1 != 0);
      if (__pyx_t_2) {

        /* "pyrobuf_list.pyx":176
 *             for val in x:
 *                 if isinstance(val, str):
 *                     values.append(val)             # <<<<<<<<<<<<<<
 *                 elif isinstance(val, bytes):
 *                     values.append(val.decode('utf-8'))
 */
        __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_values, __pyx_v_val); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 176, __pyx_L1_error)

        /* "pyrobuf_list.pyx":175
 * 
 *             for val in x:
 *                 if isinstance(val, str):             # <<<<<<<<<<<<<<
 *                     values.append(val)
 *                 elif isinstance(val, bytes):
 */
        goto __pyx_L7;
      }

      /* "pyrobuf_list.pyx":177
 *                 if isinstance(val, str):
 *                     values.append(val)
 *                 elif isinstance(val, bytes):             # <<<<<<<<<<<<<<
 *                     values.append(val.decode('utf-8'))
 *                 else:
 */
      __pyx_t_2 = PyBytes_Check(__pyx_v_val); 
      __pyx_t_1 = (__pyx_t_2 != 0);
      if (likely(__pyx_t_1)) {

        /* "pyrobuf_list.pyx":178
 *                     values.append(val)
 *                 elif isinstance(val, bytes):
 *                     values.append(val.decode('utf-8'))             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (val, type(val), bytes, str))
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_decode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 178, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 178, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_values, __pyx_t_7); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 178, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "pyrobuf_list.pyx":177
 *                 if isinstance(val, str):
 *                     values.append(val)
 *                 elif isinstance(val, bytes):             # <<<<<<<<<<<<<<
 *                     values.append(val.decode('utf-8'))
 *                 else:
 */
        goto __pyx_L7;
      }

      /* "pyrobuf_list.pyx":180
 *                     values.append(val.decode('utf-8'))
 *                 else:
 *                     raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (val, type(val), bytes, str))             # <<<<<<<<<<<<<<
 * 
 *             list.__setitem__(self, i, values)
 */
      /*else*/ {
        __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 180, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_v_val);
        __Pyx_GIVEREF(__pyx_v_val);
        PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_val);
        __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_val)));
        __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_val)));
        PyTuple_SET_ITEM(__pyx_t_7, 1, ((PyObject *)Py_TYPE(__pyx_v_val)));
        __Pyx_INCREF(((PyObject *)(&PyBytes_Type)));
        __Pyx_GIVEREF(((PyObject *)(&PyBytes_Type)));
        PyTuple_SET_ITEM(__pyx_t_7, 2, ((PyObject *)(&PyBytes_Type)));
        __Pyx_INCREF(((PyObject *)(&PyString_Type)));
        __Pyx_GIVEREF(((PyObject *)(&PyString_Type)));
        PyTuple_SET_ITEM(__pyx_t_7, 3, ((PyObject *)(&PyString_Type)));
        __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_r_has_type_s_but_expected_one_o_2, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 180, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 180, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_7, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __PYX_ERR(0, 180, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "pyrobuf_list.pyx":174
 *             values = list()
 * 
 *             for val in x:             # <<<<<<<<<<<<<<
 *                 if isinstance(val, str):
 *                     values.append(val)
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "pyrobuf_list.pyx":182
 *                     raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (val, type(val), bytes, str))
 * 
 *             list.__setitem__(self, i, values)             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError("list indices must be integers, not %s" % type(i).__name__)
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)(&PyList_Type)), __pyx_n_s_setitem); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 182, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = NULL;
    __pyx_t_6 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_6 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[4] = {__pyx_t_3, ((PyObject *)__pyx_v_self), __pyx_v_i, __pyx_v_values};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 182, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[4] = {__pyx_t_3, ((PyObject *)__pyx_v_self), __pyx_v_i, __pyx_v_values};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_6, 3+__pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 182, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(3+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 182, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (__pyx_t_3) {
        __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
      }
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_self));
      __Pyx_INCREF(__pyx_v_i);
      __Pyx_GIVEREF(__pyx_v_i);
      PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_v_i);
      __Pyx_INCREF(__pyx_v_values);
      __Pyx_GIVEREF(__pyx_v_values);
      PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_6, __pyx_v_values);
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 182, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "pyrobuf_list.pyx":171
 * 
 *             list.__setitem__(self, i, value)
 *         elif isinstance(i, slice):             # <<<<<<<<<<<<<<
 *             values = list()
 * 
 */
    goto __pyx_L3;
  }

  /* "pyrobuf_list.pyx":184
 *             list.__setitem__(self, i, values)
 *         else:
 *             raise TypeError("list indices must be integers, not %s" % type(i).__name__)             # <<<<<<<<<<<<<<
 * 
 *         listener = <object>self._listener
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)Py_TYPE(__pyx_v_i)), __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_PyString_Format(__pyx_kp_s_list_indices_must_be_integers_no, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 184, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "pyrobuf_list.pyx":186
 *             raise TypeError("list indices must be integers, not %s" % type(i).__name__)
 * 
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_4 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_listener = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "pyrobuf_list.pyx":187
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":188
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 188, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 188, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "pyrobuf_list.pyx":187
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":158
 *             self._listener = <PyObject *>listener
 * 
 *     def __setitem__(self, i, x):             # <<<<<<<<<<<<<<
 *         cdef list values
 *         cdef str value
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyrobuf_list.StringList.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_values);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":192
 * 
 * 
 *     def append(self, x):             # <<<<<<<<<<<<<<
 *         cdef str value
 *         if isinstance(x, str):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10StringList_5append(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10StringList_5append(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10StringList_4append(((struct __pyx_obj_12pyrobuf_list_StringList *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10StringList_4append(struct __pyx_obj_12pyrobuf_list_StringList *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("append", 0);

  /* "pyrobuf_list.pyx":194
 *     def append(self, x):
 *         cdef str value
 *         if isinstance(x, str):             # <<<<<<<<<<<<<<
 *             value = x
 *         elif isinstance(x, bytes):
 */
  __pyx_t_1 = PyString_Check(__pyx_v_x); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":195
 *         cdef str value
 *         if isinstance(x, str):
 *             value = x             # <<<<<<<<<<<<<<
 *         elif isinstance(x, bytes):
 *             value = x.decode('utf-8')
 */
    if (!(likely(PyString_CheckExact(__pyx_v_x))||((__pyx_v_x) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_x)->tp_name), 0))) __PYX_ERR(0, 195, __pyx_L1_error)
    __pyx_t_3 = __pyx_v_x;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_value = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":194
 *     def append(self, x):
 *         cdef str value
 *         if isinstance(x, str):             # <<<<<<<<<<<<<<
 *             value = x
 *         elif isinstance(x, bytes):
 */
    goto __pyx_L3;
  }

  /* "pyrobuf_list.pyx":196
 *         if isinstance(x, str):
 *             value = x
 *         elif isinstance(x, bytes):             # <<<<<<<<<<<<<<
 *             value = x.decode('utf-8')
 *         else:
 */
  __pyx_t_2 = PyBytes_Check(__pyx_v_x); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (likely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":197
 *             value = x
 *         elif isinstance(x, bytes):
 *             value = x.decode('utf-8')             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (x, type(x), bytes, str))
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_decode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!(likely(PyString_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(0, 197, __pyx_L1_error)
    __pyx_v_value = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "pyrobuf_list.pyx":196
 *         if isinstance(x, str):
 *             value = x
 *         elif isinstance(x, bytes):             # <<<<<<<<<<<<<<
 *             value = x.decode('utf-8')
 *         else:
 */
    goto __pyx_L3;
  }

  /* "pyrobuf_list.pyx":199
 *             value = x.decode('utf-8')
 *         else:
 *             raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (x, type(x), bytes, str))             # <<<<<<<<<<<<<<
 * 
 *         list.append(self, value)
 */
  /*else*/ {
    __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 199, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_x);
    __Pyx_GIVEREF(__pyx_v_x);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_x);
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_x)));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_x)));
    PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)Py_TYPE(__pyx_v_x)));
    __Pyx_INCREF(((PyObject *)(&PyBytes_Type)));
    __Pyx_GIVEREF(((PyObject *)(&PyBytes_Type)));
    PyTuple_SET_ITEM(__pyx_t_4, 2, ((PyObject *)(&PyBytes_Type)));
    __Pyx_INCREF(((PyObject *)(&PyString_Type)));
    __Pyx_GIVEREF(((PyObject *)(&PyString_Type)));
    PyTuple_SET_ITEM(__pyx_t_4, 3, ((PyObject *)(&PyString_Type)));
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_r_has_type_s_but_expected_one_o_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 199, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 199, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 199, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "pyrobuf_list.pyx":201
 *             raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (x, type(x), bytes, str))
 * 
 *         list.append(self, value)             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  if (unlikely(((PyObject *)__pyx_v_self) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "descriptor '%s' requires a '%s' object but received a 'NoneType'", "append", "list");
    __PYX_ERR(0, 201, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyList_Append(((PyObject*)__pyx_v_self), __pyx_v_value); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 201, __pyx_L1_error)

  /* "pyrobuf_list.pyx":202
 * 
 *         list.append(self, value)
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_4 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_listener = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "pyrobuf_list.pyx":203
 *         list.append(self, value)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":204
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def extend(self, x):
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (__pyx_t_6) {
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 204, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 204, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "pyrobuf_list.pyx":203
 *         list.append(self, value)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":192
 * 
 * 
 *     def append(self, x):             # <<<<<<<<<<<<<<
 *         cdef str value
 *         if isinstance(x, str):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.StringList.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":206
 *             listener._Modified()
 * 
 *     def extend(self, x):             # <<<<<<<<<<<<<<
 *         cdef str value
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10StringList_7extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10StringList_7extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extend (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10StringList_6extend(((struct __pyx_obj_12pyrobuf_list_StringList *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10StringList_6extend(struct __pyx_obj_12pyrobuf_list_StringList *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("extend", 0);

  /* "pyrobuf_list.pyx":209
 *         cdef str value
 * 
 *         for val in x:             # <<<<<<<<<<<<<<
 *             if isinstance(val, bytes):
 *                 value = val.decode('utf-8')
 */
  if (likely(PyList_CheckExact(__pyx_v_x)) || PyTuple_CheckExact(__pyx_v_x)) {
    __pyx_t_1 = __pyx_v_x; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 209, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 209, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 209, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 209, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 209, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 209, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "pyrobuf_list.pyx":210
 * 
 *         for val in x:
 *             if isinstance(val, bytes):             # <<<<<<<<<<<<<<
 *                 value = val.decode('utf-8')
 *             elif isinstance(val, str):
 */
    __pyx_t_5 = PyBytes_Check(__pyx_v_val); 
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {

      /* "pyrobuf_list.pyx":211
 *         for val in x:
 *             if isinstance(val, bytes):
 *                 value = val.decode('utf-8')             # <<<<<<<<<<<<<<
 *             elif isinstance(val, str):
 *                 value = val
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_decode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 211, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 211, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!(likely(PyString_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_7)->tp_name), 0))) __PYX_ERR(0, 211, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_value, ((PyObject*)__pyx_t_7));
      __pyx_t_7 = 0;

      /* "pyrobuf_list.pyx":210
 * 
 *         for val in x:
 *             if isinstance(val, bytes):             # <<<<<<<<<<<<<<
 *                 value = val.decode('utf-8')
 *             elif isinstance(val, str):
 */
      goto __pyx_L5;
    }

    /* "pyrobuf_list.pyx":212
 *             if isinstance(val, bytes):
 *                 value = val.decode('utf-8')
 *             elif isinstance(val, str):             # <<<<<<<<<<<<<<
 *                 value = val
 *             else:
 */
    __pyx_t_6 = PyString_Check(__pyx_v_val); 
    __pyx_t_5 = (__pyx_t_6 != 0);
    if (likely(__pyx_t_5)) {

      /* "pyrobuf_list.pyx":213
 *                 value = val.decode('utf-8')
 *             elif isinstance(val, str):
 *                 value = val             # <<<<<<<<<<<<<<
 *             else:
 *                 raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (val, type(val), bytes, str))
 */
      if (!(likely(PyString_CheckExact(__pyx_v_val))||((__pyx_v_val) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_val)->tp_name), 0))) __PYX_ERR(0, 213, __pyx_L1_error)
      __pyx_t_7 = __pyx_v_val;
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_XDECREF_SET(__pyx_v_value, ((PyObject*)__pyx_t_7));
      __pyx_t_7 = 0;

      /* "pyrobuf_list.pyx":212
 *             if isinstance(val, bytes):
 *                 value = val.decode('utf-8')
 *             elif isinstance(val, str):             # <<<<<<<<<<<<<<
 *                 value = val
 *             else:
 */
      goto __pyx_L5;
    }

    /* "pyrobuf_list.pyx":215
 *                 value = val
 *             else:
 *                 raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (val, type(val), bytes, str))             # <<<<<<<<<<<<<<
 *             list.append(self, value)
 * 
 */
    /*else*/ {
      __pyx_t_7 = PyTuple_New(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 215, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_v_val);
      __Pyx_GIVEREF(__pyx_v_val);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_v_val);
      __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_val)));
      __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_val)));
      PyTuple_SET_ITEM(__pyx_t_7, 1, ((PyObject *)Py_TYPE(__pyx_v_val)));
      __Pyx_INCREF(((PyObject *)(&PyBytes_Type)));
      __Pyx_GIVEREF(((PyObject *)(&PyBytes_Type)));
      PyTuple_SET_ITEM(__pyx_t_7, 2, ((PyObject *)(&PyBytes_Type)));
      __Pyx_INCREF(((PyObject *)(&PyString_Type)));
      __Pyx_GIVEREF(((PyObject *)(&PyString_Type)));
      PyTuple_SET_ITEM(__pyx_t_7, 3, ((PyObject *)(&PyString_Type)));
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_r_has_type_s_but_expected_one_o_2, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 215, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 215, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_7, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __PYX_ERR(0, 215, __pyx_L1_error)
    }
    __pyx_L5:;

    /* "pyrobuf_list.pyx":216
 *             else:
 *                 raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (val, type(val), bytes, str))
 *             list.append(self, value)             # <<<<<<<<<<<<<<
 * 
 *         listener = <object>self._listener
 */
    if (unlikely(((PyObject *)__pyx_v_self) == Py_None)) {
      PyErr_Format(PyExc_TypeError, "descriptor '%s' requires a '%s' object but received a 'NoneType'", "append", "list");
      __PYX_ERR(0, 216, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_PyList_Append(((PyObject*)__pyx_v_self), __pyx_v_value); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 216, __pyx_L1_error)

    /* "pyrobuf_list.pyx":209
 *         cdef str value
 * 
 *         for val in x:             # <<<<<<<<<<<<<<
 *             if isinstance(val, bytes):
 *                 value = val.decode('utf-8')
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":218
 *             list.append(self, value)
 * 
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":219
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = ((!__pyx_t_5) != 0);
  if (__pyx_t_6) {

    /* "pyrobuf_list.pyx":220
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def insert(self, i, x):
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 220, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 220, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":219
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":206
 *             listener._Modified()
 * 
 *     def extend(self, x):             # <<<<<<<<<<<<<<
 *         cdef str value
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyrobuf_list.StringList.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":222
 *             listener._Modified()
 * 
 *     def insert(self, i, x):             # <<<<<<<<<<<<<<
 *         cdef str value
 *         if isinstance(x, bytes):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10StringList_9insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10StringList_9insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_i = 0;
  PyObject *__pyx_v_x = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("insert (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_i,&__pyx_n_s_x,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, 1); __PYX_ERR(0, 222, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "insert") < 0)) __PYX_ERR(0, 222, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_i = values[0];
    __pyx_v_x = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 222, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.StringList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10StringList_8insert(((struct __pyx_obj_12pyrobuf_list_StringList *)__pyx_v_self), __pyx_v_i, __pyx_v_x);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10StringList_8insert(struct __pyx_obj_12pyrobuf_list_StringList *__pyx_v_self, PyObject *__pyx_v_i, PyObject *__pyx_v_x) {
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("insert", 0);

  /* "pyrobuf_list.pyx":224
 *     def insert(self, i, x):
 *         cdef str value
 *         if isinstance(x, bytes):             # <<<<<<<<<<<<<<
 *             value = x.decode('utf-8')
 *         elif isinstance(x, str):
 */
  __pyx_t_1 = PyBytes_Check(__pyx_v_x); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":225
 *         cdef str value
 *         if isinstance(x, bytes):
 *             value = x.decode('utf-8')             # <<<<<<<<<<<<<<
 *         elif isinstance(x, str):
 *             value = x
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_decode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!(likely(PyString_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(0, 225, __pyx_L1_error)
    __pyx_v_value = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "pyrobuf_list.pyx":224
 *     def insert(self, i, x):
 *         cdef str value
 *         if isinstance(x, bytes):             # <<<<<<<<<<<<<<
 *             value = x.decode('utf-8')
 *         elif isinstance(x, str):
 */
    goto __pyx_L3;
  }

  /* "pyrobuf_list.pyx":226
 *         if isinstance(x, bytes):
 *             value = x.decode('utf-8')
 *         elif isinstance(x, str):             # <<<<<<<<<<<<<<
 *             value = x
 *         else:
 */
  __pyx_t_2 = PyString_Check(__pyx_v_x); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (likely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":227
 *             value = x.decode('utf-8')
 *         elif isinstance(x, str):
 *             value = x             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (x, type(x), bytes, str))
 */
    if (!(likely(PyString_CheckExact(__pyx_v_x))||((__pyx_v_x) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_x)->tp_name), 0))) __PYX_ERR(0, 227, __pyx_L1_error)
    __pyx_t_4 = __pyx_v_x;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_v_value = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "pyrobuf_list.pyx":226
 *         if isinstance(x, bytes):
 *             value = x.decode('utf-8')
 *         elif isinstance(x, str):             # <<<<<<<<<<<<<<
 *             value = x
 *         else:
 */
    goto __pyx_L3;
  }

  /* "pyrobuf_list.pyx":229
 *             value = x
 *         else:
 *             raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (x, type(x), bytes, str))             # <<<<<<<<<<<<<<
 * 
 *         list.insert(self, i, value)
 */
  /*else*/ {
    __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_x);
    __Pyx_GIVEREF(__pyx_v_x);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_x);
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_x)));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_x)));
    PyTuple_SET_ITEM(__pyx_t_4, 1, ((PyObject *)Py_TYPE(__pyx_v_x)));
    __Pyx_INCREF(((PyObject *)(&PyBytes_Type)));
    __Pyx_GIVEREF(((PyObject *)(&PyBytes_Type)));
    PyTuple_SET_ITEM(__pyx_t_4, 2, ((PyObject *)(&PyBytes_Type)));
    __Pyx_INCREF(((PyObject *)(&PyString_Type)));
    __Pyx_GIVEREF(((PyObject *)(&PyString_Type)));
    PyTuple_SET_ITEM(__pyx_t_4, 3, ((PyObject *)(&PyString_Type)));
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_r_has_type_s_but_expected_one_o_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 229, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "pyrobuf_list.pyx":231
 *             raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (x, type(x), bytes, str))
 * 
 *         list.insert(self, i, value)             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  if (unlikely(((PyObject *)__pyx_v_self) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "descriptor '%s' requires a '%s' object but received a 'NoneType'", "insert", "list");
    __PYX_ERR(0, 231, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_5 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 231, __pyx_L1_error)
  __pyx_t_6 = PyList_Insert(((PyObject*)__pyx_v_self), __pyx_t_5, __pyx_v_value); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 231, __pyx_L1_error)

  /* "pyrobuf_list.pyx":232
 * 
 *         list.insert(self, i, value)
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_4 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_listener = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "pyrobuf_list.pyx":233
 *         list.insert(self, i, value)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":234
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (__pyx_t_7) {
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 234, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else {
      __pyx_t_4 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 234, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "pyrobuf_list.pyx":233
 *         list.insert(self, i, value)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":222
 *             listener._Modified()
 * 
 *     def insert(self, i, x):             # <<<<<<<<<<<<<<
 *         cdef str value
 *         if isinstance(x, bytes):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("pyrobuf_list.StringList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10StringList_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10StringList_11__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10StringList_10__reduce_cython__(((struct __pyx_obj_12pyrobuf_list_StringList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10StringList_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_StringList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.StringList.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10StringList_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10StringList_13__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10StringList_12__setstate_cython__(((struct __pyx_obj_12pyrobuf_list_StringList *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10StringList_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_StringList *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.StringList.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":243
 * cdef class DoubleList:
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):             # <<<<<<<<<<<<<<
 *         self._data = <double *>PyMem_Malloc(size * sizeof(double))
 *         if not self._data:
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_10DoubleList_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12pyrobuf_list_10DoubleList_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_size;
  PyObject *__pyx_v_listener = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_size,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_listener);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 243, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_size = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_size == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 243, __pyx_L3_error)
    } else {
      __pyx_v_size = ((size_t)16);
    }
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 243, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList___cinit__(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self), __pyx_v_size, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_10DoubleList___cinit__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, size_t __pyx_v_size, PyObject *__pyx_v_listener) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "pyrobuf_list.pyx":244
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <double *>PyMem_Malloc(size * sizeof(double))             # <<<<<<<<<<<<<<
 *         if not self._data:
 *             raise MemoryError()
 */
  __pyx_v_self->_data = ((double *)PyMem_Malloc((__pyx_v_size * (sizeof(double)))));

  /* "pyrobuf_list.pyx":245
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <double *>PyMem_Malloc(size * sizeof(double))
 *         if not self._data:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_self->_data != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":246
 *         self._data = <double *>PyMem_Malloc(size * sizeof(double))
 *         if not self._data:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         self._n_items = 0
 */
    PyErr_NoMemory(); __PYX_ERR(0, 246, __pyx_L1_error)

    /* "pyrobuf_list.pyx":245
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <double *>PyMem_Malloc(size * sizeof(double))
 *         if not self._data:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "pyrobuf_list.pyx":248
 *             raise MemoryError()
 * 
 *         self._n_items = 0             # <<<<<<<<<<<<<<
 *         self._size = size
 *         if listener is None:
 */
  __pyx_v_self->_n_items = 0;

  /* "pyrobuf_list.pyx":249
 * 
 *         self._n_items = 0
 *         self._size = size             # <<<<<<<<<<<<<<
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener
 */
  __pyx_v_self->_size = __pyx_v_size;

  /* "pyrobuf_list.pyx":250
 *         self._n_items = 0
 *         self._size = size
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
  __pyx_t_1 = (__pyx_v_listener == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":251
 *         self._size = size
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener             # <<<<<<<<<<<<<<
 *         else:
 *             self._listener = <PyObject *>listener
 */
    __pyx_v_self->_listener = ((PyObject *)null_listener);

    /* "pyrobuf_list.pyx":250
 *         self._n_items = 0
 *         self._size = size
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
    goto __pyx_L4;
  }

  /* "pyrobuf_list.pyx":253
 *             self._listener = <PyObject *>null_listener
 *         else:
 *             self._listener = <PyObject *>listener             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  /*else*/ {
    __pyx_v_self->_listener = ((PyObject *)__pyx_v_listener);
  }
  __pyx_L4:;

  /* "pyrobuf_list.pyx":243
 * cdef class DoubleList:
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):             # <<<<<<<<<<<<<<
 *         self._data = <double *>PyMem_Malloc(size * sizeof(double))
 *         if not self._data:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":255
 *             self._listener = <PyObject *>listener
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self._data)
 * 
 */

/* Python wrapper */
static void __pyx_pw_12pyrobuf_list_10DoubleList_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12pyrobuf_list_10DoubleList_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_12pyrobuf_list_10DoubleList_2__dealloc__(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12pyrobuf_list_10DoubleList_2__dealloc__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "pyrobuf_list.pyx":256
 * 
 *     def __dealloc__(self):
 *         PyMem_Free(self._data)             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, double x):
 */
  PyMem_Free(__pyx_v_self->_data);

  /* "pyrobuf_list.pyx":255
 *             self._listener = <PyObject *>listener
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self._data)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "pyrobuf_list.pyx":258
 *         PyMem_Free(self._data)
 * 
 *     def __contains__(self, double x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self._n_items):
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_10DoubleList_5__contains__(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static int __pyx_pw_12pyrobuf_list_10DoubleList_5__contains__(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  double __pyx_v_x;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __pyx_PyFloat_AsDouble(__pyx_arg_x); if (unlikely((__pyx_v_x == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 258, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList_4__contains__(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self), ((double)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_10DoubleList_4__contains__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, double __pyx_v_x) {
  size_t __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "pyrobuf_list.pyx":260
 *     def __contains__(self, double x):
 *         cdef size_t i
 *         for i in range(self._n_items):             # <<<<<<<<<<<<<<
 *             if self._data[i] == x:
 *                 return True
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "pyrobuf_list.pyx":261
 *         cdef size_t i
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */
    __pyx_t_4 = (((__pyx_v_self->_data[__pyx_v_i]) == __pyx_v_x) != 0);
    if (__pyx_t_4) {

      /* "pyrobuf_list.pyx":262
 *         for i in range(self._n_items):
 *             if self._data[i] == x:
 *                 return True             # <<<<<<<<<<<<<<
 * 
 *         return False
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "pyrobuf_list.pyx":261
 *         cdef size_t i
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */
    }
  }

  /* "pyrobuf_list.pyx":264
 *                 return True
 * 
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     def __delitem__(self, int i):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":258
 *         PyMem_Free(self._data)
 * 
 *     def __contains__(self, double x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self._n_items):
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":266
 *         return False
 * 
 *     def __delitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_10DoubleList_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static int __pyx_pw_12pyrobuf_list_10DoubleList_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 266, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList_6__delitem__(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self), ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_10DoubleList_6__delitem__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, int __pyx_v_i) {
  size_t __pyx_v_j;
  PyObject *__pyx_v_listener = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__delitem__", 0);

  /* "pyrobuf_list.pyx":267
 * 
 *     def __delitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":268
 *     def __delitem__(self, int i):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":267
 * 
 *     def __delitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":270
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":271
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t j
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 271, __pyx_L1_error)

    /* "pyrobuf_list.pyx":270
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":274
 * 
 *         cdef size_t j
 *         for j in range(i, self._n_items):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j + 1]
 * 
 */
  __pyx_t_4 = __pyx_v_self->_n_items;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = __pyx_v_i; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_j = __pyx_t_6;

    /* "pyrobuf_list.pyx":275
 *         cdef size_t j
 *         for j in range(i, self._n_items):
 *             self._data[j] = self._data[j + 1]             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j + 1)]);
  }

  /* "pyrobuf_list.pyx":277
 *             self._data[j] = self._data[j + 1]
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":278
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":279
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":280
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, int i):
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    if (__pyx_t_8) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 280, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 280, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":279
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":266
 *         return False
 * 
 *     def __delitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":282
 *             listener._Modified()
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 282, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList_8__getitem__(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self), ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_8__getitem__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, int __pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pyrobuf_list.pyx":283
 * 
 *     def __getitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":284
 *     def __getitem__(self, int i):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":283
 * 
 *     def __getitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":286
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":287
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         return self._data[i]
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 287, __pyx_L1_error)

    /* "pyrobuf_list.pyx":286
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":289
 *             raise IndexError("list index out of range")
 * 
 *         return self._data[i]             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyFloat_FromDouble((__pyx_v_self->_data[__pyx_v_i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":282
 *             listener._Modified()
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_12pyrobuf_list_10DoubleList_12generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "pyrobuf_list.pyx":291
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_11__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_11__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList_10__iter__(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_10__iter__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self) {
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct____iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct____iter__ *)__pyx_tp_new_12pyrobuf_list___pyx_scope_struct____iter__(__pyx_ptype_12pyrobuf_list___pyx_scope_struct____iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct____iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 291, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_12pyrobuf_list_10DoubleList_12generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_DoubleList___iter, __pyx_n_s_pyrobuf_list); if (unlikely(!gen)) __PYX_ERR(0, 291, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_12pyrobuf_list_10DoubleList_12generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct____iter__ *__pyx_cur_scope = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct____iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 291, __pyx_L1_error)

  /* "pyrobuf_list.pyx":292
 * 
 *     def __iter__(self):
 *         cdef size_t idx = 0             # <<<<<<<<<<<<<<
 *         while idx < self._n_items:
 *             yield self._data[idx]
 */
  __pyx_cur_scope->__pyx_v_idx = 0;

  /* "pyrobuf_list.pyx":293
 *     def __iter__(self):
 *         cdef size_t idx = 0
 *         while idx < self._n_items:             # <<<<<<<<<<<<<<
 *             yield self._data[idx]
 *             idx += 1
 */
  while (1) {
    __pyx_t_1 = ((__pyx_cur_scope->__pyx_v_idx < __pyx_cur_scope->__pyx_v_self->_n_items) != 0);
    if (!__pyx_t_1) break;

    /* "pyrobuf_list.pyx":294
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 *             yield self._data[idx]             # <<<<<<<<<<<<<<
 *             idx += 1
 * 
 */
    __pyx_t_2 = PyFloat_FromDouble((__pyx_cur_scope->__pyx_v_self->_data[__pyx_cur_scope->__pyx_v_idx])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 294, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 294, __pyx_L1_error)

    /* "pyrobuf_list.pyx":295
 *         while idx < self._n_items:
 *             yield self._data[idx]
 *             idx += 1             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
    __pyx_cur_scope->__pyx_v_idx = (__pyx_cur_scope->__pyx_v_idx + 1);
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "pyrobuf_list.pyx":291
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":297
 *             idx += 1
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._n_items
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_12pyrobuf_list_10DoubleList_14__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_12pyrobuf_list_10DoubleList_14__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList_13__len__(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_12pyrobuf_list_10DoubleList_13__len__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pyrobuf_list.pyx":298
 * 
 *     def __len__(self):
 *         return self._n_items             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = __pyx_v_self->_n_items;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":297
 *             idx += 1
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._n_items
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":300
 *         return self._n_items
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(list(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_16__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_16__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList_15__repr__(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_15__repr__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pyrobuf_list.pyx":301
 * 
 *     def __repr__(self):
 *         return repr(list(self))             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, int i, double x):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":300
 *         return self._n_items
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(list(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":303
 *         return repr(list(self))
 * 
 *     def __setitem__(self, int i, double x):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_10DoubleList_18__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i, PyObject *__pyx_arg_x); /*proto*/
static int __pyx_pw_12pyrobuf_list_10DoubleList_18__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i, PyObject *__pyx_arg_x) {
  int __pyx_v_i;
  double __pyx_v_x;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 303, __pyx_L3_error)
  }
  assert(__pyx_arg_x); {
    __pyx_v_x = __pyx_PyFloat_AsDouble(__pyx_arg_x); if (unlikely((__pyx_v_x == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 303, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList_17__setitem__(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self), ((int)__pyx_v_i), ((double)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_10DoubleList_17__setitem__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, int __pyx_v_i, double __pyx_v_x) {
  PyObject *__pyx_v_listener = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "pyrobuf_list.pyx":304
 * 
 *     def __setitem__(self, int i, double x):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":305
 *     def __setitem__(self, int i, double x):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":304
 * 
 *     def __setitem__(self, int i, double x):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":307
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":308
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 308, __pyx_L1_error)

    /* "pyrobuf_list.pyx":307
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":310
 *             raise IndexError("list index out of range")
 * 
 *         self._data[i] = x             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  (__pyx_v_self->_data[__pyx_v_i]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":311
 * 
 *         self._data[i] = x
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":312
 *         self._data[i] = x
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":313
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 313, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 313, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 313, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":312
 *         self._data[i] = x
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":303
 *         return repr(list(self))
 * 
 *     def __setitem__(self, int i, double x):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":315
 *             listener._Modified()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(list(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_20__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_20__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList_19__str__(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_19__str__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pyrobuf_list.pyx":316
 * 
 *     def __str__(self):
 *         return str(list(self))             # <<<<<<<<<<<<<<
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":315
 *             listener._Modified()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(list(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":318
 *         return str(list(self))
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t itemsize = sizeof(double)
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_12pyrobuf_list_10DoubleList_22__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_12pyrobuf_list_10DoubleList_22__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList_21__getbuffer__(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_10DoubleList_21__getbuffer__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags) {
  Py_ssize_t __pyx_v_itemsize;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  if (__pyx_v_buffer == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_buffer->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_buffer->obj);

  /* "pyrobuf_list.pyx":320
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 * 
 *         cdef Py_ssize_t itemsize = sizeof(double)             # <<<<<<<<<<<<<<
 *         self.shape[0] = self._n_items
 *         self.strides[0] = 1
 */
  __pyx_v_itemsize = (sizeof(double));

  /* "pyrobuf_list.pyx":321
 * 
 *         cdef Py_ssize_t itemsize = sizeof(double)
 *         self.shape[0] = self._n_items             # <<<<<<<<<<<<<<
 *         self.strides[0] = 1
 * 
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  (__pyx_v_self->shape[0]) = __pyx_t_1;

  /* "pyrobuf_list.pyx":322
 *         cdef Py_ssize_t itemsize = sizeof(double)
 *         self.shape[0] = self._n_items
 *         self.strides[0] = 1             # <<<<<<<<<<<<<<
 * 
 *         buffer.buf = <char *>(self._data)
 */
  (__pyx_v_self->strides[0]) = 1;

  /* "pyrobuf_list.pyx":324
 *         self.strides[0] = 1
 * 
 *         buffer.buf = <char *>(self._data)             # <<<<<<<<<<<<<<
 *         buffer.format = 'd'
 *         buffer.internal = NULL                  # see References
 */
  __pyx_v_buffer->buf = ((char *)__pyx_v_self->_data);

  /* "pyrobuf_list.pyx":325
 * 
 *         buffer.buf = <char *>(self._data)
 *         buffer.format = 'd'             # <<<<<<<<<<<<<<
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 */
  __pyx_v_buffer->format = ((char *)"d");

  /* "pyrobuf_list.pyx":326
 *         buffer.buf = <char *>(self._data)
 *         buffer.format = 'd'
 *         buffer.internal = NULL                  # see References             # <<<<<<<<<<<<<<
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items
 */
  __pyx_v_buffer->internal = NULL;

  /* "pyrobuf_list.pyx":327
 *         buffer.format = 'd'
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize             # <<<<<<<<<<<<<<
 *         buffer.len = self._n_items
 *         buffer.ndim = 1
 */
  __pyx_v_buffer->itemsize = __pyx_v_itemsize;

  /* "pyrobuf_list.pyx":328
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items             # <<<<<<<<<<<<<<
 *         buffer.ndim = 1
 *         buffer.obj = self
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  __pyx_v_buffer->len = __pyx_t_1;

  /* "pyrobuf_list.pyx":329
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items
 *         buffer.ndim = 1             # <<<<<<<<<<<<<<
 *         buffer.obj = self
 *         buffer.readonly = 1
 */
  __pyx_v_buffer->ndim = 1;

  /* "pyrobuf_list.pyx":330
 *         buffer.len = self._n_items
 *         buffer.ndim = 1
 *         buffer.obj = self             # <<<<<<<<<<<<<<
 *         buffer.readonly = 1
 *         buffer.shape = self.shape
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_buffer->obj);
  __Pyx_DECREF(__pyx_v_buffer->obj);
  __pyx_v_buffer->obj = ((PyObject *)__pyx_v_self);

  /* "pyrobuf_list.pyx":331
 *         buffer.ndim = 1
 *         buffer.obj = self
 *         buffer.readonly = 1             # <<<<<<<<<<<<<<
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides
 */
  __pyx_v_buffer->readonly = 1;

  /* "pyrobuf_list.pyx":332
 *         buffer.obj = self
 *         buffer.readonly = 1
 *         buffer.shape = self.shape             # <<<<<<<<<<<<<<
 *         buffer.strides = self.strides
 *         buffer.suboffsets = NULL                # for pointer arrays only
 */
  __pyx_t_2 = __pyx_v_self->shape;
  __pyx_v_buffer->shape = __pyx_t_2;

  /* "pyrobuf_list.pyx":333
 *         buffer.readonly = 1
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides             # <<<<<<<<<<<<<<
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 */
  __pyx_t_2 = __pyx_v_self->strides;
  __pyx_v_buffer->strides = __pyx_t_2;

  /* "pyrobuf_list.pyx":334
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides
 *         buffer.suboffsets = NULL                # for pointer arrays only             # <<<<<<<<<<<<<<
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):
 */
  __pyx_v_buffer->suboffsets = NULL;

  /* "pyrobuf_list.pyx":318
 *         return str(list(self))
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t itemsize = sizeof(double)
 */

  /* function exit code */
  __pyx_r = 0;
  if (__pyx_v_buffer->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_buffer->obj);
    __Pyx_DECREF(__pyx_v_buffer->obj); __pyx_v_buffer->obj = 0;
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":336
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_12pyrobuf_list_10DoubleList_24__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer); /*proto*/
static CYTHON_UNUSED void __pyx_pw_12pyrobuf_list_10DoubleList_24__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_12pyrobuf_list_10DoubleList_23__releasebuffer__(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12pyrobuf_list_10DoubleList_23__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "pyrobuf_list.pyx":339
 *         pass
 * 
 *     def __richcmp__(self, other, op):             # <<<<<<<<<<<<<<
 *         if op == Py_EQ:
 *             return self._equal(other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_26__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_26__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op) {
  PyObject *__pyx_v_op = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_v_op = __Pyx_PyInt_From_int(__pyx_arg_op); if (unlikely(!__pyx_v_op)) __PYX_ERR(0, 339, __pyx_L3_error)
  __Pyx_GOTREF(__pyx_v_op);
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList_25__richcmp__(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((PyObject *)__pyx_v_op));

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_op);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_25__richcmp__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "pyrobuf_list.pyx":340
 * 
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:             # <<<<<<<<<<<<<<
 *             return self._equal(other)
 *         elif op == Py_NE:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(Py_EQ); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_op, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "pyrobuf_list.pyx":341
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:
 *             return self._equal(other)             # <<<<<<<<<<<<<<
 *         elif op == Py_NE:
 *             return not self._equal(other)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = ((struct __pyx_vtabstruct_12pyrobuf_list_DoubleList *)__pyx_v_self->__pyx_vtab)->_equal(__pyx_v_self, __pyx_v_other, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pyrobuf_list.pyx":340
 * 
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:             # <<<<<<<<<<<<<<
 *             return self._equal(other)
 *         elif op == Py_NE:
 */
  }

  /* "pyrobuf_list.pyx":342
 *         if op == Py_EQ:
 *             return self._equal(other)
 *         elif op == Py_NE:             # <<<<<<<<<<<<<<
 *             return not self._equal(other)
 *         else:
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(Py_NE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_op, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(__pyx_t_3)) {

    /* "pyrobuf_list.pyx":343
 *             return self._equal(other)
 *         elif op == Py_NE:
 *             return not self._equal(other)             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError('Operation not supported for lists')
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((struct __pyx_vtabstruct_12pyrobuf_list_DoubleList *)__pyx_v_self->__pyx_vtab)->_equal(__pyx_v_self, __pyx_v_other, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 343, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_3)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pyrobuf_list.pyx":342
 *         if op == Py_EQ:
 *             return self._equal(other)
 *         elif op == Py_NE:             # <<<<<<<<<<<<<<
 *             return not self._equal(other)
 *         else:
 */
  }

  /* "pyrobuf_list.pyx":345
 *             return not self._equal(other)
 *         else:
 *             raise TypeError('Operation not supported for lists')             # <<<<<<<<<<<<<<
 * 
 *     def __neq__(self, other):
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 345, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 345, __pyx_L1_error)
  }

  /* "pyrobuf_list.pyx":339
 *         pass
 * 
 *     def __richcmp__(self, other, op):             # <<<<<<<<<<<<<<
 *         if op == Py_EQ:
 *             return self._equal(other)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":347
 *             raise TypeError('Operation not supported for lists')
 * 
 *     def __neq__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_28__neq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_28__neq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__neq__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList_27__neq__(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_27__neq__(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__neq__", 0);

  /* "pyrobuf_list.pyx":348
 * 
 *     def __neq__(self, other):
 *         return not self == other             # <<<<<<<<<<<<<<
 * 
 *     cpdef _equal(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_self), __pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 348, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":347
 *             raise TypeError('Operation not supported for lists')
 * 
 *     def __neq__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.__neq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":350
 *         return not self == other
 * 
 *     cpdef _equal(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             if len(self) != len(other):
 */

static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_10DoubleList__equal(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_v_y = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *(*__pyx_t_14)(PyObject *);
  __Pyx_RefNannySetupContext("_equal", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_equal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_10DoubleList_30_equal)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_other); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 350, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_other};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 350, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_other};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 350, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 350, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_other);
          __Pyx_GIVEREF(__pyx_v_other);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_other);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 350, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":351
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    /*try:*/ {

      /* "pyrobuf_list.pyx":352
 *     cpdef _equal(self, other):
 *         try:
 *             if len(self) != len(other):             # <<<<<<<<<<<<<<
 *                 return False
 *         except TypeError:
 */
      __pyx_t_9 = PyObject_Length(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 352, __pyx_L3_error)
      __pyx_t_10 = PyObject_Length(__pyx_v_other); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 352, __pyx_L3_error)
      __pyx_t_11 = ((__pyx_t_9 != __pyx_t_10) != 0);
      if (__pyx_t_11) {

        /* "pyrobuf_list.pyx":353
 *         try:
 *             if len(self) != len(other):
 *                 return False             # <<<<<<<<<<<<<<
 *         except TypeError:
 *             return False
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_False);
        __pyx_r = Py_False;
        goto __pyx_L7_try_return;

        /* "pyrobuf_list.pyx":352
 *     cpdef _equal(self, other):
 *         try:
 *             if len(self) != len(other):             # <<<<<<<<<<<<<<
 *                 return False
 *         except TypeError:
 */
      }

      /* "pyrobuf_list.pyx":351
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
    }
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":354
 *             if len(self) != len(other):
 *                 return False
 *         except TypeError:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
    __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_12) {
      __Pyx_AddTraceback("pyrobuf_list.DoubleList._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 354, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_3);

      /* "pyrobuf_list.pyx":355
 *                 return False
 *         except TypeError:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         for x, y in zip(self, other):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "pyrobuf_list.pyx":351
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "pyrobuf_list.pyx":357
 *             return False
 * 
 *         for x, y in zip(self, other):             # <<<<<<<<<<<<<<
 *             if x != y:
 *                 return False
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));
  __Pyx_INCREF(__pyx_v_other);
  __Pyx_GIVEREF(__pyx_v_other);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_other);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_10 = 0;
    __pyx_t_13 = NULL;
  } else {
    __pyx_t_10 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_13 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 357, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_13)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 357, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 357, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 357, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 357, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_13(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 357, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 357, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 357, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 357, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 357, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_14 = Py_TYPE(__pyx_t_4)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_5 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_5)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_4), 2) < 0) __PYX_ERR(0, 357, __pyx_L1_error)
      __pyx_t_14 = NULL;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L15_unpacking_done;
      __pyx_L14_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_14 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 357, __pyx_L1_error)
      __pyx_L15_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_x, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_y, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "pyrobuf_list.pyx":358
 * 
 *         for x, y in zip(self, other):
 *             if x != y:             # <<<<<<<<<<<<<<
 *                 return False
 * 
 */
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_x, __pyx_v_y, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 358, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_11) {

      /* "pyrobuf_list.pyx":359
 *         for x, y in zip(self, other):
 *             if x != y:
 *                 return False             # <<<<<<<<<<<<<<
 * 
 *         return True
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "pyrobuf_list.pyx":358
 * 
 *         for x, y in zip(self, other):
 *             if x != y:             # <<<<<<<<<<<<<<
 *                 return False
 * 
 */
    }

    /* "pyrobuf_list.pyx":357
 *             return False
 * 
 *         for x, y in zip(self, other):             # <<<<<<<<<<<<<<
 *             if x != y:
 *                 return False
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":361
 *                 return False
 * 
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     cpdef append(self, double x):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":350
 *         return not self == other
 * 
 *     cpdef _equal(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             if len(self) != len(other):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_y);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_equal (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList_29_equal(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_29_equal(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_equal", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_10DoubleList__equal(__pyx_v_self, __pyx_v_other, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":363
 *         return True
 * 
 *     cpdef append(self, double x):             # <<<<<<<<<<<<<<
 *         self._append(x)
 *         listener = <object>self._listener
 */

static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_10DoubleList_append(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, double __pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("append", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_append); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_10DoubleList_32append)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = PyFloat_FromDouble(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 363, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 363, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 363, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 363, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 363, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 363, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":364
 * 
 *     cpdef append(self, double x):
 *         self._append(x)             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  ((struct __pyx_vtabstruct_12pyrobuf_list_DoubleList *)__pyx_v_self->__pyx_vtab)->_append(__pyx_v_self, __pyx_v_x);

  /* "pyrobuf_list.pyx":365
 *     cpdef append(self, double x):
 *         self._append(x)
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":366
 *         self._append(x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = ((!__pyx_t_7) != 0);
  if (__pyx_t_8) {

    /* "pyrobuf_list.pyx":367
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef extend(self, x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 367, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 367, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 367, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":366
 *         self._append(x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":363
 *         return True
 * 
 *     cpdef append(self, double x):             # <<<<<<<<<<<<<<
 *         self._append(x)
 *         listener = <object>self._listener
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  double __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __pyx_PyFloat_AsDouble(__pyx_arg_x); if (unlikely((__pyx_v_x == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 363, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList_31append(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self), ((double)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_31append(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, double __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("append", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_10DoubleList_append(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":369
 *             listener._Modified()
 * 
 *     cpdef extend(self, x):             # <<<<<<<<<<<<<<
 *         for i in x:
 *             self._append(i)
 */

static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_10DoubleList_extend(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, PyObject *__pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  double __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("extend", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extend); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 369, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_10DoubleList_34extend)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 369, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_x};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 369, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_x};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 369, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 369, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_x);
          __Pyx_GIVEREF(__pyx_v_x);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_x);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 369, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":370
 * 
 *     cpdef extend(self, x):
 *         for i in x:             # <<<<<<<<<<<<<<
 *             self._append(i)
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_x)) || PyTuple_CheckExact(__pyx_v_x)) {
    __pyx_t_1 = __pyx_v_x; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 370, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 370, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 370, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 370, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 370, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_7(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 370, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyrobuf_list.pyx":371
 *     cpdef extend(self, x):
 *         for i in x:
 *             self._append(i)             # <<<<<<<<<<<<<<
 * 
 *         listener = <object>self._listener
 */
    __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_v_i); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 371, __pyx_L1_error)
    ((struct __pyx_vtabstruct_12pyrobuf_list_DoubleList *)__pyx_v_self->__pyx_vtab)->_append(__pyx_v_self, __pyx_t_8);

    /* "pyrobuf_list.pyx":370
 * 
 *     cpdef extend(self, x):
 *         for i in x:             # <<<<<<<<<<<<<<
 *             self._append(i)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":373
 *             self._append(i)
 * 
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":374
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = ((!__pyx_t_9) != 0);
  if (__pyx_t_10) {

    /* "pyrobuf_list.pyx":375
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef insert(self, int i, double x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 375, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 375, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":374
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":369
 *             listener._Modified()
 * 
 *     cpdef extend(self, x):             # <<<<<<<<<<<<<<
 *         for i in x:
 *             self._append(i)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extend (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList_33extend(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_33extend(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("extend", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_10DoubleList_extend(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":377
 *             listener._Modified()
 * 
 *     cpdef insert(self, int i, double x):             # <<<<<<<<<<<<<<
 *         cdef double *mem
 *         cdef size_t j
 */

static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_10DoubleList_insert(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, int __pyx_v_i, double __pyx_v_x, int __pyx_skip_dispatch) {
  double *__pyx_v_mem;
  size_t __pyx_v_j;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  size_t __pyx_t_11;
  size_t __pyx_t_12;
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("insert", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_insert); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 377, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_10DoubleList_36insert)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 377, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyFloat_FromDouble(__pyx_v_x); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 377, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 377, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 377, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 377, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
        __pyx_t_3 = 0;
        __pyx_t_4 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 377, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":381
 *         cdef size_t j
 * 
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_9 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_9) {

    /* "pyrobuf_list.pyx":382
 * 
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":381
 *         cdef size_t j
 * 
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":384
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i > self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_10 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_10) {
  } else {
    __pyx_t_9 = __pyx_t_10;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_10 = ((((size_t)__pyx_v_i) > __pyx_v_self->_n_items) != 0);
  __pyx_t_9 = __pyx_t_10;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_9)) {

    /* "pyrobuf_list.pyx":385
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         if self._n_items == self._size:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 385, __pyx_L1_error)

    /* "pyrobuf_list.pyx":384
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i > self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":387
 *             raise IndexError("list index out of range")
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <double *>PyMem_Realloc(self._data, 2 * self._size * sizeof(double))
 *             if not mem:
 */
  __pyx_t_9 = ((__pyx_v_self->_n_items == __pyx_v_self->_size) != 0);
  if (__pyx_t_9) {

    /* "pyrobuf_list.pyx":388
 * 
 *         if self._n_items == self._size:
 *             mem = <double *>PyMem_Realloc(self._data, 2 * self._size * sizeof(double))             # <<<<<<<<<<<<<<
 *             if not mem:
 *                 raise MemoryError()
 */
    __pyx_v_mem = ((double *)PyMem_Realloc(__pyx_v_self->_data, ((2 * __pyx_v_self->_size) * (sizeof(double)))));

    /* "pyrobuf_list.pyx":389
 *         if self._n_items == self._size:
 *             mem = <double *>PyMem_Realloc(self._data, 2 * self._size * sizeof(double))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    __pyx_t_9 = ((!(__pyx_v_mem != 0)) != 0);
    if (unlikely(__pyx_t_9)) {

      /* "pyrobuf_list.pyx":390
 *             mem = <double *>PyMem_Realloc(self._data, 2 * self._size * sizeof(double))
 *             if not mem:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *             self._data = mem
 */
      PyErr_NoMemory(); __PYX_ERR(0, 390, __pyx_L1_error)

      /* "pyrobuf_list.pyx":389
 *         if self._n_items == self._size:
 *             mem = <double *>PyMem_Realloc(self._data, 2 * self._size * sizeof(double))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    }

    /* "pyrobuf_list.pyx":392
 *                 raise MemoryError()
 * 
 *             self._data = mem             # <<<<<<<<<<<<<<
 *             self._size *= 2
 * 
 */
    __pyx_v_self->_data = __pyx_v_mem;

    /* "pyrobuf_list.pyx":393
 * 
 *             self._data = mem
 *             self._size *= 2             # <<<<<<<<<<<<<<
 * 
 *         for j in range(self._n_items, <size_t>i, -1):
 */
    __pyx_v_self->_size = (__pyx_v_self->_size * 2);

    /* "pyrobuf_list.pyx":387
 *             raise IndexError("list index out of range")
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <double *>PyMem_Realloc(self._data, 2 * self._size * sizeof(double))
 *             if not mem:
 */
  }

  /* "pyrobuf_list.pyx":395
 *             self._size *= 2
 * 
 *         for j in range(self._n_items, <size_t>i, -1):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j - 1]
 * 
 */
  __pyx_t_11 = ((size_t)__pyx_v_i);
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = __pyx_v_self->_n_items + 1; __pyx_t_13 > __pyx_t_12 + 1; ) { __pyx_t_13-=1;
    __pyx_v_j = __pyx_t_13;

    /* "pyrobuf_list.pyx":396
 * 
 *         for j in range(self._n_items, <size_t>i, -1):
 *             self._data[j] = self._data[j - 1]             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j - 1)]);
  }

  /* "pyrobuf_list.pyx":398
 *             self._data[j] = self._data[j - 1]
 * 
 *         self._data[i] = x             # <<<<<<<<<<<<<<
 *         self._n_items += 1
 *         listener = <object>self._listener
 */
  (__pyx_v_self->_data[__pyx_v_i]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":399
 * 
 *         self._data[i] = x
 *         self._n_items += 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items + 1);

  /* "pyrobuf_list.pyx":400
 *         self._data[i] = x
 *         self._n_items += 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":401
 *         self._n_items += 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = ((!__pyx_t_9) != 0);
  if (__pyx_t_10) {

    /* "pyrobuf_list.pyx":402
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef pop(self):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 402, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 402, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 402, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":401
 *         self._n_items += 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":377
 *             listener._Modified()
 * 
 *     cpdef insert(self, int i, double x):             # <<<<<<<<<<<<<<
 *         cdef double *mem
 *         cdef size_t j
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_i;
  double __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("insert (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_i,&__pyx_n_s_x,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, 1); __PYX_ERR(0, 377, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "insert") < 0)) __PYX_ERR(0, 377, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_i = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 377, __pyx_L3_error)
    __pyx_v_x = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_x == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 377, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 377, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList_35insert(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self), __pyx_v_i, __pyx_v_x);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_35insert(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, int __pyx_v_i, double __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("insert", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_10DoubleList_insert(__pyx_v_self, __pyx_v_i, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":404
 *             listener._Modified()
 * 
 *     cpdef pop(self):             # <<<<<<<<<<<<<<
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")
 */

static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_10DoubleList_pop(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("pop", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_10DoubleList_38pop)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 404, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 404, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":405
 * 
 *     cpdef pop(self):
 *         if self._n_items == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("pop from empty list")
 * 
 */
  __pyx_t_5 = ((__pyx_v_self->_n_items == 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "pyrobuf_list.pyx":406
 *     cpdef pop(self):
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 406, __pyx_L1_error)

    /* "pyrobuf_list.pyx":405
 * 
 *     cpdef pop(self):
 *         if self._n_items == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("pop from empty list")
 * 
 */
  }

  /* "pyrobuf_list.pyx":408
 *             raise IndexError("pop from empty list")
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":409
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":410
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 *         return self._data[self._n_items]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = ((!__pyx_t_5) != 0);
  if (__pyx_t_6) {

    /* "pyrobuf_list.pyx":411
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 *         return self._data[self._n_items]
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 411, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 411, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":410
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 *         return self._data[self._n_items]
 */
  }

  /* "pyrobuf_list.pyx":412
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 *         return self._data[self._n_items]             # <<<<<<<<<<<<<<
 * 
 *     cpdef remove(self, double x):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->_data[__pyx_v_self->_n_items])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":404
 *             listener._Modified()
 * 
 *     cpdef pop(self):             # <<<<<<<<<<<<<<
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pop (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList_37pop(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_37pop(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("pop", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_10DoubleList_pop(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":414
 *         return self._data[self._n_items]
 * 
 *     cpdef remove(self, double x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef size_t j
 */

static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_10DoubleList_remove(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, double __pyx_v_x, int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  int __pyx_v_found;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("remove", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_remove); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_10DoubleList_40remove)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = PyFloat_FromDouble(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 414, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 414, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 414, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 414, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 414, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 414, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":417
 *         cdef size_t i
 *         cdef size_t j
 *         cdef bint found = 0             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self._n_items):
 */
  __pyx_v_found = 0;

  /* "pyrobuf_list.pyx":419
 *         cdef bint found = 0
 * 
 *         for i in range(self._n_items):             # <<<<<<<<<<<<<<
 *             if self._data[i] == x:
 *                 found = 1
 */
  __pyx_t_7 = __pyx_v_self->_n_items;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "pyrobuf_list.pyx":420
 * 
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 found = 1
 *                 break
 */
    __pyx_t_10 = (((__pyx_v_self->_data[__pyx_v_i]) == __pyx_v_x) != 0);
    if (__pyx_t_10) {

      /* "pyrobuf_list.pyx":421
 *         for i in range(self._n_items):
 *             if self._data[i] == x:
 *                 found = 1             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
      __pyx_v_found = 1;

      /* "pyrobuf_list.pyx":422
 *             if self._data[i] == x:
 *                 found = 1
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         if found == 0:
 */
      goto __pyx_L4_break;

      /* "pyrobuf_list.pyx":420
 * 
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 found = 1
 *                 break
 */
    }
  }
  __pyx_L4_break:;

  /* "pyrobuf_list.pyx":424
 *                 break
 * 
 *         if found == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("x not in list")
 * 
 */
  __pyx_t_10 = ((__pyx_v_found == 0) != 0);
  if (unlikely(__pyx_t_10)) {

    /* "pyrobuf_list.pyx":425
 * 
 *         if found == 0:
 *             raise ValueError("x not in list")             # <<<<<<<<<<<<<<
 * 
 *         for j in range(i, self._n_items):
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 425, __pyx_L1_error)

    /* "pyrobuf_list.pyx":424
 *                 break
 * 
 *         if found == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("x not in list")
 * 
 */
  }

  /* "pyrobuf_list.pyx":427
 *             raise ValueError("x not in list")
 * 
 *         for j in range(i, self._n_items):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j + 1]
 * 
 */
  __pyx_t_7 = __pyx_v_self->_n_items;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = __pyx_v_i; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_j = __pyx_t_9;

    /* "pyrobuf_list.pyx":428
 * 
 *         for j in range(i, self._n_items):
 *             self._data[j] = self._data[j + 1]             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j + 1)]);
  }

  /* "pyrobuf_list.pyx":430
 *             self._data[j] = self._data[j + 1]
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":431
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":432
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = ((!__pyx_t_10) != 0);
  if (__pyx_t_11) {

    /* "pyrobuf_list.pyx":433
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cdef void _append(self, double x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 433, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 433, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":432
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":414
 *         return self._data[self._n_items]
 * 
 *     cpdef remove(self, double x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef size_t j
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  double __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __pyx_PyFloat_AsDouble(__pyx_arg_x); if (unlikely((__pyx_v_x == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 414, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList_39remove(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self), ((double)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_39remove(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, double __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("remove", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_10DoubleList_remove(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":435
 *             listener._Modified()
 * 
 *     cdef void _append(self, double x):             # <<<<<<<<<<<<<<
 *         cdef double *mem
 * 
 */

static void __pyx_f_12pyrobuf_list_10DoubleList__append(struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, double __pyx_v_x) {
  double *__pyx_v_mem;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_append", 0);

  /* "pyrobuf_list.pyx":438
 *         cdef double *mem
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <double *>PyMem_Realloc(self._data, 2 * self._size * sizeof(double))
 *             if not mem:
 */
  __pyx_t_1 = ((__pyx_v_self->_n_items == __pyx_v_self->_size) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":439
 * 
 *         if self._n_items == self._size:
 *             mem = <double *>PyMem_Realloc(self._data, 2 * self._size * sizeof(double))             # <<<<<<<<<<<<<<
 *             if not mem:
 *                 raise MemoryError()
 */
    __pyx_v_mem = ((double *)PyMem_Realloc(__pyx_v_self->_data, ((2 * __pyx_v_self->_size) * (sizeof(double)))));

    /* "pyrobuf_list.pyx":440
 *         if self._n_items == self._size:
 *             mem = <double *>PyMem_Realloc(self._data, 2 * self._size * sizeof(double))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    __pyx_t_1 = ((!(__pyx_v_mem != 0)) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "pyrobuf_list.pyx":441
 *             mem = <double *>PyMem_Realloc(self._data, 2 * self._size * sizeof(double))
 *             if not mem:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *             self._data = mem
 */
      PyErr_NoMemory(); __PYX_ERR(0, 441, __pyx_L1_error)

      /* "pyrobuf_list.pyx":440
 *         if self._n_items == self._size:
 *             mem = <double *>PyMem_Realloc(self._data, 2 * self._size * sizeof(double))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    }

    /* "pyrobuf_list.pyx":443
 *                 raise MemoryError()
 * 
 *             self._data = mem             # <<<<<<<<<<<<<<
 *             self._size *= 2
 * 
 */
    __pyx_v_self->_data = __pyx_v_mem;

    /* "pyrobuf_list.pyx":444
 * 
 *             self._data = mem
 *             self._size *= 2             # <<<<<<<<<<<<<<
 * 
 *         self._data[self._n_items] = x
 */
    __pyx_v_self->_size = (__pyx_v_self->_size * 2);

    /* "pyrobuf_list.pyx":438
 *         cdef double *mem
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <double *>PyMem_Realloc(self._data, 2 * self._size * sizeof(double))
 *             if not mem:
 */
  }

  /* "pyrobuf_list.pyx":446
 *             self._size *= 2
 * 
 *         self._data[self._n_items] = x             # <<<<<<<<<<<<<<
 *         self._n_items += 1
 * 
 */
  (__pyx_v_self->_data[__pyx_v_self->_n_items]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":447
 * 
 *         self._data[self._n_items] = x
 *         self._n_items += 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items + 1);

  /* "pyrobuf_list.pyx":435
 *             listener._Modified()
 * 
 *     cdef void _append(self, double x):             # <<<<<<<<<<<<<<
 *         cdef double *mem
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("pyrobuf_list.DoubleList._append", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_42__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_42__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList_41__reduce_cython__(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_41__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_44__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10DoubleList_44__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10DoubleList_43__setstate_cython__(((struct __pyx_obj_12pyrobuf_list_DoubleList *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10DoubleList_43__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_DoubleList *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.DoubleList.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":453
 * cdef class FloatList:
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):             # <<<<<<<<<<<<<<
 *         self._data = <float *>PyMem_Malloc(size * sizeof(float))
 *         if not self._data:
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_9FloatList_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12pyrobuf_list_9FloatList_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_size;
  PyObject *__pyx_v_listener = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_size,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_listener);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 453, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_size = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_size == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 453, __pyx_L3_error)
    } else {
      __pyx_v_size = ((size_t)16);
    }
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 453, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.FloatList.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList___cinit__(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self), __pyx_v_size, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_9FloatList___cinit__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, size_t __pyx_v_size, PyObject *__pyx_v_listener) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "pyrobuf_list.pyx":454
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <float *>PyMem_Malloc(size * sizeof(float))             # <<<<<<<<<<<<<<
 *         if not self._data:
 *             raise MemoryError()
 */
  __pyx_v_self->_data = ((float *)PyMem_Malloc((__pyx_v_size * (sizeof(float)))));

  /* "pyrobuf_list.pyx":455
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <float *>PyMem_Malloc(size * sizeof(float))
 *         if not self._data:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_self->_data != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":456
 *         self._data = <float *>PyMem_Malloc(size * sizeof(float))
 *         if not self._data:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         self._n_items = 0
 */
    PyErr_NoMemory(); __PYX_ERR(0, 456, __pyx_L1_error)

    /* "pyrobuf_list.pyx":455
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <float *>PyMem_Malloc(size * sizeof(float))
 *         if not self._data:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "pyrobuf_list.pyx":458
 *             raise MemoryError()
 * 
 *         self._n_items = 0             # <<<<<<<<<<<<<<
 *         self._size = size
 *         if listener is None:
 */
  __pyx_v_self->_n_items = 0;

  /* "pyrobuf_list.pyx":459
 * 
 *         self._n_items = 0
 *         self._size = size             # <<<<<<<<<<<<<<
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener
 */
  __pyx_v_self->_size = __pyx_v_size;

  /* "pyrobuf_list.pyx":460
 *         self._n_items = 0
 *         self._size = size
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
  __pyx_t_1 = (__pyx_v_listener == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":461
 *         self._size = size
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener             # <<<<<<<<<<<<<<
 *         else:
 *             self._listener = <PyObject *>listener
 */
    __pyx_v_self->_listener = ((PyObject *)null_listener);

    /* "pyrobuf_list.pyx":460
 *         self._n_items = 0
 *         self._size = size
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
    goto __pyx_L4;
  }

  /* "pyrobuf_list.pyx":463
 *             self._listener = <PyObject *>null_listener
 *         else:
 *             self._listener = <PyObject *>listener             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  /*else*/ {
    __pyx_v_self->_listener = ((PyObject *)__pyx_v_listener);
  }
  __pyx_L4:;

  /* "pyrobuf_list.pyx":453
 * cdef class FloatList:
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):             # <<<<<<<<<<<<<<
 *         self._data = <float *>PyMem_Malloc(size * sizeof(float))
 *         if not self._data:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyrobuf_list.FloatList.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":465
 *             self._listener = <PyObject *>listener
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self._data)
 * 
 */

/* Python wrapper */
static void __pyx_pw_12pyrobuf_list_9FloatList_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12pyrobuf_list_9FloatList_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_12pyrobuf_list_9FloatList_2__dealloc__(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12pyrobuf_list_9FloatList_2__dealloc__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "pyrobuf_list.pyx":466
 * 
 *     def __dealloc__(self):
 *         PyMem_Free(self._data)             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, float x):
 */
  PyMem_Free(__pyx_v_self->_data);

  /* "pyrobuf_list.pyx":465
 *             self._listener = <PyObject *>listener
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self._data)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "pyrobuf_list.pyx":468
 *         PyMem_Free(self._data)
 * 
 *     def __contains__(self, float x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self._n_items):
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_9FloatList_5__contains__(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static int __pyx_pw_12pyrobuf_list_9FloatList_5__contains__(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  float __pyx_v_x;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __pyx_PyFloat_AsFloat(__pyx_arg_x); if (unlikely((__pyx_v_x == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 468, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.FloatList.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList_4__contains__(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self), ((float)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_9FloatList_4__contains__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, float __pyx_v_x) {
  size_t __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "pyrobuf_list.pyx":470
 *     def __contains__(self, float x):
 *         cdef size_t i
 *         for i in range(self._n_items):             # <<<<<<<<<<<<<<
 *             if self._data[i] == x:
 *                 return True
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "pyrobuf_list.pyx":471
 *         cdef size_t i
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */
    __pyx_t_4 = (((__pyx_v_self->_data[__pyx_v_i]) == __pyx_v_x) != 0);
    if (__pyx_t_4) {

      /* "pyrobuf_list.pyx":472
 *         for i in range(self._n_items):
 *             if self._data[i] == x:
 *                 return True             # <<<<<<<<<<<<<<
 * 
 *         return False
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "pyrobuf_list.pyx":471
 *         cdef size_t i
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */
    }
  }

  /* "pyrobuf_list.pyx":474
 *                 return True
 * 
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     def __delitem__(self, int i):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":468
 *         PyMem_Free(self._data)
 * 
 *     def __contains__(self, float x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self._n_items):
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":476
 *         return False
 * 
 *     def __delitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_9FloatList_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static int __pyx_pw_12pyrobuf_list_9FloatList_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 476, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.FloatList.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList_6__delitem__(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self), ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_9FloatList_6__delitem__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, int __pyx_v_i) {
  size_t __pyx_v_j;
  PyObject *__pyx_v_listener = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__delitem__", 0);

  /* "pyrobuf_list.pyx":477
 * 
 *     def __delitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":478
 *     def __delitem__(self, int i):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":477
 * 
 *     def __delitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":480
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":481
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t j
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 481, __pyx_L1_error)

    /* "pyrobuf_list.pyx":480
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":484
 * 
 *         cdef size_t j
 *         for j in range(i, self._n_items):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j + 1]
 * 
 */
  __pyx_t_4 = __pyx_v_self->_n_items;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = __pyx_v_i; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_j = __pyx_t_6;

    /* "pyrobuf_list.pyx":485
 *         cdef size_t j
 *         for j in range(i, self._n_items):
 *             self._data[j] = self._data[j + 1]             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j + 1)]);
  }

  /* "pyrobuf_list.pyx":487
 *             self._data[j] = self._data[j + 1]
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":488
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":489
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":490
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, int i):
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    if (__pyx_t_8) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 490, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 490, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":489
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":476
 *         return False
 * 
 *     def __delitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyrobuf_list.FloatList.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":492
 *             listener._Modified()
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 492, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.FloatList.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList_8__getitem__(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self), ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_8__getitem__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, int __pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pyrobuf_list.pyx":493
 * 
 *     def __getitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":494
 *     def __getitem__(self, int i):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":493
 * 
 *     def __getitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":496
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":497
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         return self._data[i]
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 497, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 497, __pyx_L1_error)

    /* "pyrobuf_list.pyx":496
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":499
 *             raise IndexError("list index out of range")
 * 
 *         return self._data[i]             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyFloat_FromDouble((__pyx_v_self->_data[__pyx_v_i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":492
 *             listener._Modified()
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyrobuf_list.FloatList.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_12pyrobuf_list_9FloatList_12generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "pyrobuf_list.pyx":501
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_11__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_11__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList_10__iter__(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_10__iter__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self) {
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_1___iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_1___iter__ *)__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_1___iter__(__pyx_ptype_12pyrobuf_list___pyx_scope_struct_1___iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_1___iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 501, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_12pyrobuf_list_9FloatList_12generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_FloatList___iter, __pyx_n_s_pyrobuf_list); if (unlikely(!gen)) __PYX_ERR(0, 501, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyrobuf_list.FloatList.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_12pyrobuf_list_9FloatList_12generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_1___iter__ *__pyx_cur_scope = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_1___iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 501, __pyx_L1_error)

  /* "pyrobuf_list.pyx":502
 * 
 *     def __iter__(self):
 *         cdef size_t idx = 0             # <<<<<<<<<<<<<<
 *         while idx < self._n_items:
 *             yield self._data[idx]
 */
  __pyx_cur_scope->__pyx_v_idx = 0;

  /* "pyrobuf_list.pyx":503
 *     def __iter__(self):
 *         cdef size_t idx = 0
 *         while idx < self._n_items:             # <<<<<<<<<<<<<<
 *             yield self._data[idx]
 *             idx += 1
 */
  while (1) {
    __pyx_t_1 = ((__pyx_cur_scope->__pyx_v_idx < __pyx_cur_scope->__pyx_v_self->_n_items) != 0);
    if (!__pyx_t_1) break;

    /* "pyrobuf_list.pyx":504
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 *             yield self._data[idx]             # <<<<<<<<<<<<<<
 *             idx += 1
 * 
 */
    __pyx_t_2 = PyFloat_FromDouble((__pyx_cur_scope->__pyx_v_self->_data[__pyx_cur_scope->__pyx_v_idx])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 504, __pyx_L1_error)

    /* "pyrobuf_list.pyx":505
 *         while idx < self._n_items:
 *             yield self._data[idx]
 *             idx += 1             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
    __pyx_cur_scope->__pyx_v_idx = (__pyx_cur_scope->__pyx_v_idx + 1);
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "pyrobuf_list.pyx":501
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":507
 *             idx += 1
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._n_items
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_12pyrobuf_list_9FloatList_14__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_12pyrobuf_list_9FloatList_14__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList_13__len__(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_12pyrobuf_list_9FloatList_13__len__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pyrobuf_list.pyx":508
 * 
 *     def __len__(self):
 *         return self._n_items             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = __pyx_v_self->_n_items;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":507
 *             idx += 1
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._n_items
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":510
 *         return self._n_items
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(list(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_16__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_16__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList_15__repr__(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_15__repr__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pyrobuf_list.pyx":511
 * 
 *     def __repr__(self):
 *         return repr(list(self))             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, int i, float x):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":510
 *         return self._n_items
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(list(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.FloatList.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":513
 *         return repr(list(self))
 * 
 *     def __setitem__(self, int i, float x):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_9FloatList_18__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i, PyObject *__pyx_arg_x); /*proto*/
static int __pyx_pw_12pyrobuf_list_9FloatList_18__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i, PyObject *__pyx_arg_x) {
  int __pyx_v_i;
  float __pyx_v_x;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 513, __pyx_L3_error)
  }
  assert(__pyx_arg_x); {
    __pyx_v_x = __pyx_PyFloat_AsFloat(__pyx_arg_x); if (unlikely((__pyx_v_x == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 513, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.FloatList.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList_17__setitem__(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self), ((int)__pyx_v_i), ((float)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_9FloatList_17__setitem__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, int __pyx_v_i, float __pyx_v_x) {
  PyObject *__pyx_v_listener = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "pyrobuf_list.pyx":514
 * 
 *     def __setitem__(self, int i, float x):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":515
 *     def __setitem__(self, int i, float x):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":514
 * 
 *     def __setitem__(self, int i, float x):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":517
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":518
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 518, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 518, __pyx_L1_error)

    /* "pyrobuf_list.pyx":517
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":520
 *             raise IndexError("list index out of range")
 * 
 *         self._data[i] = x             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  (__pyx_v_self->_data[__pyx_v_i]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":521
 * 
 *         self._data[i] = x
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":522
 *         self._data[i] = x
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":523
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 523, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 523, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":522
 *         self._data[i] = x
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":513
 *         return repr(list(self))
 * 
 *     def __setitem__(self, int i, float x):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.FloatList.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":525
 *             listener._Modified()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(list(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_20__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_20__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList_19__str__(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_19__str__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pyrobuf_list.pyx":526
 * 
 *     def __str__(self):
 *         return str(list(self))             # <<<<<<<<<<<<<<
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":525
 *             listener._Modified()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(list(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.FloatList.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":528
 *         return str(list(self))
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t itemsize = sizeof(float)
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_12pyrobuf_list_9FloatList_22__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_12pyrobuf_list_9FloatList_22__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList_21__getbuffer__(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_9FloatList_21__getbuffer__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags) {
  Py_ssize_t __pyx_v_itemsize;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  if (__pyx_v_buffer == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_buffer->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_buffer->obj);

  /* "pyrobuf_list.pyx":530
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 * 
 *         cdef Py_ssize_t itemsize = sizeof(float)             # <<<<<<<<<<<<<<
 *         self.shape[0] = self._n_items
 *         self.strides[0] = 1
 */
  __pyx_v_itemsize = (sizeof(float));

  /* "pyrobuf_list.pyx":531
 * 
 *         cdef Py_ssize_t itemsize = sizeof(float)
 *         self.shape[0] = self._n_items             # <<<<<<<<<<<<<<
 *         self.strides[0] = 1
 * 
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  (__pyx_v_self->shape[0]) = __pyx_t_1;

  /* "pyrobuf_list.pyx":532
 *         cdef Py_ssize_t itemsize = sizeof(float)
 *         self.shape[0] = self._n_items
 *         self.strides[0] = 1             # <<<<<<<<<<<<<<
 * 
 *         buffer.buf = <char *>(self._data)
 */
  (__pyx_v_self->strides[0]) = 1;

  /* "pyrobuf_list.pyx":534
 *         self.strides[0] = 1
 * 
 *         buffer.buf = <char *>(self._data)             # <<<<<<<<<<<<<<
 *         buffer.format = 'f'
 *         buffer.internal = NULL                  # see References
 */
  __pyx_v_buffer->buf = ((char *)__pyx_v_self->_data);

  /* "pyrobuf_list.pyx":535
 * 
 *         buffer.buf = <char *>(self._data)
 *         buffer.format = 'f'             # <<<<<<<<<<<<<<
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 */
  __pyx_v_buffer->format = ((char *)"f");

  /* "pyrobuf_list.pyx":536
 *         buffer.buf = <char *>(self._data)
 *         buffer.format = 'f'
 *         buffer.internal = NULL                  # see References             # <<<<<<<<<<<<<<
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items
 */
  __pyx_v_buffer->internal = NULL;

  /* "pyrobuf_list.pyx":537
 *         buffer.format = 'f'
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize             # <<<<<<<<<<<<<<
 *         buffer.len = self._n_items
 *         buffer.ndim = 1
 */
  __pyx_v_buffer->itemsize = __pyx_v_itemsize;

  /* "pyrobuf_list.pyx":538
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items             # <<<<<<<<<<<<<<
 *         buffer.ndim = 1
 *         buffer.obj = self
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  __pyx_v_buffer->len = __pyx_t_1;

  /* "pyrobuf_list.pyx":539
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items
 *         buffer.ndim = 1             # <<<<<<<<<<<<<<
 *         buffer.obj = self
 *         buffer.readonly = 1
 */
  __pyx_v_buffer->ndim = 1;

  /* "pyrobuf_list.pyx":540
 *         buffer.len = self._n_items
 *         buffer.ndim = 1
 *         buffer.obj = self             # <<<<<<<<<<<<<<
 *         buffer.readonly = 1
 *         buffer.shape = self.shape
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_buffer->obj);
  __Pyx_DECREF(__pyx_v_buffer->obj);
  __pyx_v_buffer->obj = ((PyObject *)__pyx_v_self);

  /* "pyrobuf_list.pyx":541
 *         buffer.ndim = 1
 *         buffer.obj = self
 *         buffer.readonly = 1             # <<<<<<<<<<<<<<
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides
 */
  __pyx_v_buffer->readonly = 1;

  /* "pyrobuf_list.pyx":542
 *         buffer.obj = self
 *         buffer.readonly = 1
 *         buffer.shape = self.shape             # <<<<<<<<<<<<<<
 *         buffer.strides = self.strides
 *         buffer.suboffsets = NULL                # for pointer arrays only
 */
  __pyx_t_2 = __pyx_v_self->shape;
  __pyx_v_buffer->shape = __pyx_t_2;

  /* "pyrobuf_list.pyx":543
 *         buffer.readonly = 1
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides             # <<<<<<<<<<<<<<
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 */
  __pyx_t_2 = __pyx_v_self->strides;
  __pyx_v_buffer->strides = __pyx_t_2;

  /* "pyrobuf_list.pyx":544
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides
 *         buffer.suboffsets = NULL                # for pointer arrays only             # <<<<<<<<<<<<<<
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):
 */
  __pyx_v_buffer->suboffsets = NULL;

  /* "pyrobuf_list.pyx":528
 *         return str(list(self))
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t itemsize = sizeof(float)
 */

  /* function exit code */
  __pyx_r = 0;
  if (__pyx_v_buffer->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_buffer->obj);
    __Pyx_DECREF(__pyx_v_buffer->obj); __pyx_v_buffer->obj = 0;
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":546
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_12pyrobuf_list_9FloatList_24__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer); /*proto*/
static CYTHON_UNUSED void __pyx_pw_12pyrobuf_list_9FloatList_24__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_12pyrobuf_list_9FloatList_23__releasebuffer__(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12pyrobuf_list_9FloatList_23__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "pyrobuf_list.pyx":549
 *         pass
 * 
 *     def __richcmp__(self, other, op):             # <<<<<<<<<<<<<<
 *         if op == Py_EQ:
 *             return self._equal(other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_26__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_26__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op) {
  PyObject *__pyx_v_op = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_v_op = __Pyx_PyInt_From_int(__pyx_arg_op); if (unlikely(!__pyx_v_op)) __PYX_ERR(0, 549, __pyx_L3_error)
  __Pyx_GOTREF(__pyx_v_op);
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.FloatList.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList_25__richcmp__(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((PyObject *)__pyx_v_op));

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_op);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_25__richcmp__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "pyrobuf_list.pyx":550
 * 
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:             # <<<<<<<<<<<<<<
 *             return self._equal(other)
 *         elif op == Py_NE:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(Py_EQ); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_op, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 550, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "pyrobuf_list.pyx":551
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:
 *             return self._equal(other)             # <<<<<<<<<<<<<<
 *         elif op == Py_NE:
 *             return not self._equal(other)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = ((struct __pyx_vtabstruct_12pyrobuf_list_FloatList *)__pyx_v_self->__pyx_vtab)->_equal(__pyx_v_self, __pyx_v_other, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pyrobuf_list.pyx":550
 * 
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:             # <<<<<<<<<<<<<<
 *             return self._equal(other)
 *         elif op == Py_NE:
 */
  }

  /* "pyrobuf_list.pyx":552
 *         if op == Py_EQ:
 *             return self._equal(other)
 *         elif op == Py_NE:             # <<<<<<<<<<<<<<
 *             return not self._equal(other)
 *         else:
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(Py_NE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_op, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(__pyx_t_3)) {

    /* "pyrobuf_list.pyx":553
 *             return self._equal(other)
 *         elif op == Py_NE:
 *             return not self._equal(other)             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError('Operation not supported for lists')
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((struct __pyx_vtabstruct_12pyrobuf_list_FloatList *)__pyx_v_self->__pyx_vtab)->_equal(__pyx_v_self, __pyx_v_other, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 553, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 553, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_3)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 553, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pyrobuf_list.pyx":552
 *         if op == Py_EQ:
 *             return self._equal(other)
 *         elif op == Py_NE:             # <<<<<<<<<<<<<<
 *             return not self._equal(other)
 *         else:
 */
  }

  /* "pyrobuf_list.pyx":555
 *             return not self._equal(other)
 *         else:
 *             raise TypeError('Operation not supported for lists')             # <<<<<<<<<<<<<<
 * 
 *     def __neq__(self, other):
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 555, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 555, __pyx_L1_error)
  }

  /* "pyrobuf_list.pyx":549
 *         pass
 * 
 *     def __richcmp__(self, other, op):             # <<<<<<<<<<<<<<
 *         if op == Py_EQ:
 *             return self._equal(other)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.FloatList.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":557
 *             raise TypeError('Operation not supported for lists')
 * 
 *     def __neq__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_28__neq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_28__neq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__neq__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList_27__neq__(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_27__neq__(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__neq__", 0);

  /* "pyrobuf_list.pyx":558
 * 
 *     def __neq__(self, other):
 *         return not self == other             # <<<<<<<<<<<<<<
 * 
 *     cpdef _equal(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_self), __pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 558, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":557
 *             raise TypeError('Operation not supported for lists')
 * 
 *     def __neq__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.FloatList.__neq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":560
 *         return not self == other
 * 
 *     cpdef _equal(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             if len(self) != len(other):
 */

static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_9FloatList__equal(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_v_y = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *(*__pyx_t_14)(PyObject *);
  __Pyx_RefNannySetupContext("_equal", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_equal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_9FloatList_30_equal)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_other); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 560, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_other};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 560, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_other};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 560, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 560, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_other);
          __Pyx_GIVEREF(__pyx_v_other);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_other);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 560, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":561
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    /*try:*/ {

      /* "pyrobuf_list.pyx":562
 *     cpdef _equal(self, other):
 *         try:
 *             if len(self) != len(other):             # <<<<<<<<<<<<<<
 *                 return False
 *         except TypeError:
 */
      __pyx_t_9 = PyObject_Length(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 562, __pyx_L3_error)
      __pyx_t_10 = PyObject_Length(__pyx_v_other); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 562, __pyx_L3_error)
      __pyx_t_11 = ((__pyx_t_9 != __pyx_t_10) != 0);
      if (__pyx_t_11) {

        /* "pyrobuf_list.pyx":563
 *         try:
 *             if len(self) != len(other):
 *                 return False             # <<<<<<<<<<<<<<
 *         except TypeError:
 *             return False
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_False);
        __pyx_r = Py_False;
        goto __pyx_L7_try_return;

        /* "pyrobuf_list.pyx":562
 *     cpdef _equal(self, other):
 *         try:
 *             if len(self) != len(other):             # <<<<<<<<<<<<<<
 *                 return False
 *         except TypeError:
 */
      }

      /* "pyrobuf_list.pyx":561
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
    }
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":564
 *             if len(self) != len(other):
 *                 return False
 *         except TypeError:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
    __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_12) {
      __Pyx_AddTraceback("pyrobuf_list.FloatList._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 564, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_3);

      /* "pyrobuf_list.pyx":565
 *                 return False
 *         except TypeError:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         for x, y in zip(self, other):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "pyrobuf_list.pyx":561
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "pyrobuf_list.pyx":567
 *             return False
 * 
 *         for x, y in zip(self, other):             # <<<<<<<<<<<<<<
 *             if x != y:
 *                 return False
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));
  __Pyx_INCREF(__pyx_v_other);
  __Pyx_GIVEREF(__pyx_v_other);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_other);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_10 = 0;
    __pyx_t_13 = NULL;
  } else {
    __pyx_t_10 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_13 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 567, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_13)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 567, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 567, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 567, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 567, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_13(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 567, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 567, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 567, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 567, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 567, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_14 = Py_TYPE(__pyx_t_4)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_5 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_5)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_4), 2) < 0) __PYX_ERR(0, 567, __pyx_L1_error)
      __pyx_t_14 = NULL;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L15_unpacking_done;
      __pyx_L14_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_14 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 567, __pyx_L1_error)
      __pyx_L15_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_x, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_y, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "pyrobuf_list.pyx":568
 * 
 *         for x, y in zip(self, other):
 *             if x != y:             # <<<<<<<<<<<<<<
 *                 return False
 * 
 */
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_x, __pyx_v_y, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 568, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 568, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_11) {

      /* "pyrobuf_list.pyx":569
 *         for x, y in zip(self, other):
 *             if x != y:
 *                 return False             # <<<<<<<<<<<<<<
 * 
 *         return True
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "pyrobuf_list.pyx":568
 * 
 *         for x, y in zip(self, other):
 *             if x != y:             # <<<<<<<<<<<<<<
 *                 return False
 * 
 */
    }

    /* "pyrobuf_list.pyx":567
 *             return False
 * 
 *         for x, y in zip(self, other):             # <<<<<<<<<<<<<<
 *             if x != y:
 *                 return False
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":571
 *                 return False
 * 
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     cpdef append(self, float x):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":560
 *         return not self == other
 * 
 *     cpdef _equal(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             if len(self) != len(other):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.FloatList._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_y);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_equal (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList_29_equal(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_29_equal(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_equal", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_9FloatList__equal(__pyx_v_self, __pyx_v_other, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 560, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.FloatList._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":573
 *         return True
 * 
 *     cpdef append(self, float x):             # <<<<<<<<<<<<<<
 *         self._append(x)
 *         listener = <object>self._listener
 */

static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_9FloatList_append(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, float __pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("append", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_append); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_9FloatList_32append)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = PyFloat_FromDouble(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 573, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 573, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 573, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 573, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 573, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 573, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":574
 * 
 *     cpdef append(self, float x):
 *         self._append(x)             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  ((struct __pyx_vtabstruct_12pyrobuf_list_FloatList *)__pyx_v_self->__pyx_vtab)->_append(__pyx_v_self, __pyx_v_x);

  /* "pyrobuf_list.pyx":575
 *     cpdef append(self, float x):
 *         self._append(x)
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":576
 *         self._append(x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = ((!__pyx_t_7) != 0);
  if (__pyx_t_8) {

    /* "pyrobuf_list.pyx":577
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef extend(self, x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 577, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 577, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":576
 *         self._append(x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":573
 *         return True
 * 
 *     cpdef append(self, float x):             # <<<<<<<<<<<<<<
 *         self._append(x)
 *         listener = <object>self._listener
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.FloatList.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  float __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __pyx_PyFloat_AsFloat(__pyx_arg_x); if (unlikely((__pyx_v_x == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 573, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.FloatList.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList_31append(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self), ((float)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_31append(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, float __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("append", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_9FloatList_append(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.FloatList.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":579
 *             listener._Modified()
 * 
 *     cpdef extend(self, x):             # <<<<<<<<<<<<<<
 *         for i in x:
 *             self._append(i)
 */

static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_9FloatList_extend(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, PyObject *__pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  float __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("extend", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extend); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_9FloatList_34extend)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_x};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_x};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 579, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_x);
          __Pyx_GIVEREF(__pyx_v_x);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_x);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":580
 * 
 *     cpdef extend(self, x):
 *         for i in x:             # <<<<<<<<<<<<<<
 *             self._append(i)
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_x)) || PyTuple_CheckExact(__pyx_v_x)) {
    __pyx_t_1 = __pyx_v_x; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 580, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 580, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 580, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 580, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 580, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_7(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 580, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyrobuf_list.pyx":581
 *     cpdef extend(self, x):
 *         for i in x:
 *             self._append(i)             # <<<<<<<<<<<<<<
 * 
 *         listener = <object>self._listener
 */
    __pyx_t_8 = __pyx_PyFloat_AsFloat(__pyx_v_i); if (unlikely((__pyx_t_8 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 581, __pyx_L1_error)
    ((struct __pyx_vtabstruct_12pyrobuf_list_FloatList *)__pyx_v_self->__pyx_vtab)->_append(__pyx_v_self, __pyx_t_8);

    /* "pyrobuf_list.pyx":580
 * 
 *     cpdef extend(self, x):
 *         for i in x:             # <<<<<<<<<<<<<<
 *             self._append(i)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":583
 *             self._append(i)
 * 
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":584
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = ((!__pyx_t_9) != 0);
  if (__pyx_t_10) {

    /* "pyrobuf_list.pyx":585
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef insert(self, int i, float x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 585, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 585, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 585, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":584
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":579
 *             listener._Modified()
 * 
 *     cpdef extend(self, x):             # <<<<<<<<<<<<<<
 *         for i in x:
 *             self._append(i)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.FloatList.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extend (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList_33extend(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_33extend(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("extend", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_9FloatList_extend(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.FloatList.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":587
 *             listener._Modified()
 * 
 *     cpdef insert(self, int i, float x):             # <<<<<<<<<<<<<<
 *         cdef float *mem
 *         cdef size_t j
 */

static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_9FloatList_insert(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, int __pyx_v_i, float __pyx_v_x, int __pyx_skip_dispatch) {
  float *__pyx_v_mem;
  size_t __pyx_v_j;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  size_t __pyx_t_11;
  size_t __pyx_t_12;
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("insert", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_insert); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 587, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_9FloatList_36insert)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 587, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyFloat_FromDouble(__pyx_v_x); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 587, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 587, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 587, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 587, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
        __pyx_t_3 = 0;
        __pyx_t_4 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 587, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":591
 *         cdef size_t j
 * 
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_9 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_9) {

    /* "pyrobuf_list.pyx":592
 * 
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":591
 *         cdef size_t j
 * 
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":594
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i > self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_10 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_10) {
  } else {
    __pyx_t_9 = __pyx_t_10;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_10 = ((((size_t)__pyx_v_i) > __pyx_v_self->_n_items) != 0);
  __pyx_t_9 = __pyx_t_10;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_9)) {

    /* "pyrobuf_list.pyx":595
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         if self._n_items == self._size:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 595, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 595, __pyx_L1_error)

    /* "pyrobuf_list.pyx":594
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i > self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":597
 *             raise IndexError("list index out of range")
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <float *>PyMem_Realloc(self._data, 2 * self._size * sizeof(float))
 *             if not mem:
 */
  __pyx_t_9 = ((__pyx_v_self->_n_items == __pyx_v_self->_size) != 0);
  if (__pyx_t_9) {

    /* "pyrobuf_list.pyx":598
 * 
 *         if self._n_items == self._size:
 *             mem = <float *>PyMem_Realloc(self._data, 2 * self._size * sizeof(float))             # <<<<<<<<<<<<<<
 *             if not mem:
 *                 raise MemoryError()
 */
    __pyx_v_mem = ((float *)PyMem_Realloc(__pyx_v_self->_data, ((2 * __pyx_v_self->_size) * (sizeof(float)))));

    /* "pyrobuf_list.pyx":599
 *         if self._n_items == self._size:
 *             mem = <float *>PyMem_Realloc(self._data, 2 * self._size * sizeof(float))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    __pyx_t_9 = ((!(__pyx_v_mem != 0)) != 0);
    if (unlikely(__pyx_t_9)) {

      /* "pyrobuf_list.pyx":600
 *             mem = <float *>PyMem_Realloc(self._data, 2 * self._size * sizeof(float))
 *             if not mem:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *             self._data = mem
 */
      PyErr_NoMemory(); __PYX_ERR(0, 600, __pyx_L1_error)

      /* "pyrobuf_list.pyx":599
 *         if self._n_items == self._size:
 *             mem = <float *>PyMem_Realloc(self._data, 2 * self._size * sizeof(float))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    }

    /* "pyrobuf_list.pyx":602
 *                 raise MemoryError()
 * 
 *             self._data = mem             # <<<<<<<<<<<<<<
 *             self._size *= 2
 * 
 */
    __pyx_v_self->_data = __pyx_v_mem;

    /* "pyrobuf_list.pyx":603
 * 
 *             self._data = mem
 *             self._size *= 2             # <<<<<<<<<<<<<<
 * 
 *         for j in range(self._n_items, <size_t>i, -1):
 */
    __pyx_v_self->_size = (__pyx_v_self->_size * 2);

    /* "pyrobuf_list.pyx":597
 *             raise IndexError("list index out of range")
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <float *>PyMem_Realloc(self._data, 2 * self._size * sizeof(float))
 *             if not mem:
 */
  }

  /* "pyrobuf_list.pyx":605
 *             self._size *= 2
 * 
 *         for j in range(self._n_items, <size_t>i, -1):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j - 1]
 * 
 */
  __pyx_t_11 = ((size_t)__pyx_v_i);
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = __pyx_v_self->_n_items + 1; __pyx_t_13 > __pyx_t_12 + 1; ) { __pyx_t_13-=1;
    __pyx_v_j = __pyx_t_13;

    /* "pyrobuf_list.pyx":606
 * 
 *         for j in range(self._n_items, <size_t>i, -1):
 *             self._data[j] = self._data[j - 1]             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j - 1)]);
  }

  /* "pyrobuf_list.pyx":608
 *             self._data[j] = self._data[j - 1]
 * 
 *         self._data[i] = x             # <<<<<<<<<<<<<<
 *         self._n_items += 1
 *         listener = <object>self._listener
 */
  (__pyx_v_self->_data[__pyx_v_i]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":609
 * 
 *         self._data[i] = x
 *         self._n_items += 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items + 1);

  /* "pyrobuf_list.pyx":610
 *         self._data[i] = x
 *         self._n_items += 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":611
 *         self._n_items += 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 611, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = ((!__pyx_t_9) != 0);
  if (__pyx_t_10) {

    /* "pyrobuf_list.pyx":612
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef pop(self):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 612, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 612, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 612, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":611
 *         self._n_items += 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":587
 *             listener._Modified()
 * 
 *     cpdef insert(self, int i, float x):             # <<<<<<<<<<<<<<
 *         cdef float *mem
 *         cdef size_t j
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyrobuf_list.FloatList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_i;
  float __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("insert (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_i,&__pyx_n_s_x,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, 1); __PYX_ERR(0, 587, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "insert") < 0)) __PYX_ERR(0, 587, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_i = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 587, __pyx_L3_error)
    __pyx_v_x = __pyx_PyFloat_AsFloat(values[1]); if (unlikely((__pyx_v_x == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 587, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 587, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.FloatList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList_35insert(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self), __pyx_v_i, __pyx_v_x);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_35insert(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, int __pyx_v_i, float __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("insert", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_9FloatList_insert(__pyx_v_self, __pyx_v_i, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.FloatList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":614
 *             listener._Modified()
 * 
 *     cpdef pop(self):             # <<<<<<<<<<<<<<
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")
 */

static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_9FloatList_pop(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("pop", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_9FloatList_38pop)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 614, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 614, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":615
 * 
 *     cpdef pop(self):
 *         if self._n_items == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("pop from empty list")
 * 
 */
  __pyx_t_5 = ((__pyx_v_self->_n_items == 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "pyrobuf_list.pyx":616
 *     cpdef pop(self):
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 616, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 616, __pyx_L1_error)

    /* "pyrobuf_list.pyx":615
 * 
 *     cpdef pop(self):
 *         if self._n_items == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("pop from empty list")
 * 
 */
  }

  /* "pyrobuf_list.pyx":618
 *             raise IndexError("pop from empty list")
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":619
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":620
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 *         return self._data[self._n_items]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = ((!__pyx_t_5) != 0);
  if (__pyx_t_6) {

    /* "pyrobuf_list.pyx":621
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 *         return self._data[self._n_items]
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 621, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 621, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":620
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 *         return self._data[self._n_items]
 */
  }

  /* "pyrobuf_list.pyx":622
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 *         return self._data[self._n_items]             # <<<<<<<<<<<<<<
 * 
 *     cpdef remove(self, float x):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble((__pyx_v_self->_data[__pyx_v_self->_n_items])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":614
 *             listener._Modified()
 * 
 *     cpdef pop(self):             # <<<<<<<<<<<<<<
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyrobuf_list.FloatList.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pop (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList_37pop(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_37pop(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("pop", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_9FloatList_pop(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.FloatList.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":624
 *         return self._data[self._n_items]
 * 
 *     cpdef remove(self, float x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef size_t j
 */

static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_9FloatList_remove(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, float __pyx_v_x, int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  int __pyx_v_found;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("remove", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_remove); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 624, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_9FloatList_40remove)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = PyFloat_FromDouble(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 624, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 624, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 624, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 624, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 624, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 624, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":627
 *         cdef size_t i
 *         cdef size_t j
 *         cdef bint found = 0             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self._n_items):
 */
  __pyx_v_found = 0;

  /* "pyrobuf_list.pyx":629
 *         cdef bint found = 0
 * 
 *         for i in range(self._n_items):             # <<<<<<<<<<<<<<
 *             if self._data[i] == x:
 *                 found = 1
 */
  __pyx_t_7 = __pyx_v_self->_n_items;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "pyrobuf_list.pyx":630
 * 
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 found = 1
 *                 break
 */
    __pyx_t_10 = (((__pyx_v_self->_data[__pyx_v_i]) == __pyx_v_x) != 0);
    if (__pyx_t_10) {

      /* "pyrobuf_list.pyx":631
 *         for i in range(self._n_items):
 *             if self._data[i] == x:
 *                 found = 1             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
      __pyx_v_found = 1;

      /* "pyrobuf_list.pyx":632
 *             if self._data[i] == x:
 *                 found = 1
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         if found == 0:
 */
      goto __pyx_L4_break;

      /* "pyrobuf_list.pyx":630
 * 
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 found = 1
 *                 break
 */
    }
  }
  __pyx_L4_break:;

  /* "pyrobuf_list.pyx":634
 *                 break
 * 
 *         if found == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("x not in list")
 * 
 */
  __pyx_t_10 = ((__pyx_v_found == 0) != 0);
  if (unlikely(__pyx_t_10)) {

    /* "pyrobuf_list.pyx":635
 * 
 *         if found == 0:
 *             raise ValueError("x not in list")             # <<<<<<<<<<<<<<
 * 
 *         for j in range(i, self._n_items):
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 635, __pyx_L1_error)

    /* "pyrobuf_list.pyx":634
 *                 break
 * 
 *         if found == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("x not in list")
 * 
 */
  }

  /* "pyrobuf_list.pyx":637
 *             raise ValueError("x not in list")
 * 
 *         for j in range(i, self._n_items):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j + 1]
 * 
 */
  __pyx_t_7 = __pyx_v_self->_n_items;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = __pyx_v_i; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_j = __pyx_t_9;

    /* "pyrobuf_list.pyx":638
 * 
 *         for j in range(i, self._n_items):
 *             self._data[j] = self._data[j + 1]             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j + 1)]);
  }

  /* "pyrobuf_list.pyx":640
 *             self._data[j] = self._data[j + 1]
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":641
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":642
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = ((!__pyx_t_10) != 0);
  if (__pyx_t_11) {

    /* "pyrobuf_list.pyx":643
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cdef void _append(self, float x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 643, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 643, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 643, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":642
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":624
 *         return self._data[self._n_items]
 * 
 *     cpdef remove(self, float x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef size_t j
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.FloatList.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  float __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __pyx_PyFloat_AsFloat(__pyx_arg_x); if (unlikely((__pyx_v_x == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 624, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.FloatList.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList_39remove(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self), ((float)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_39remove(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, float __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("remove", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_9FloatList_remove(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.FloatList.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":645
 *             listener._Modified()
 * 
 *     cdef void _append(self, float x):             # <<<<<<<<<<<<<<
 *         cdef float *mem
 * 
 */

static void __pyx_f_12pyrobuf_list_9FloatList__append(struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, float __pyx_v_x) {
  float *__pyx_v_mem;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_append", 0);

  /* "pyrobuf_list.pyx":648
 *         cdef float *mem
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <float *>PyMem_Realloc(self._data, 2 * self._size * sizeof(float))
 *             if not mem:
 */
  __pyx_t_1 = ((__pyx_v_self->_n_items == __pyx_v_self->_size) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":649
 * 
 *         if self._n_items == self._size:
 *             mem = <float *>PyMem_Realloc(self._data, 2 * self._size * sizeof(float))             # <<<<<<<<<<<<<<
 *             if not mem:
 *                 raise MemoryError()
 */
    __pyx_v_mem = ((float *)PyMem_Realloc(__pyx_v_self->_data, ((2 * __pyx_v_self->_size) * (sizeof(float)))));

    /* "pyrobuf_list.pyx":650
 *         if self._n_items == self._size:
 *             mem = <float *>PyMem_Realloc(self._data, 2 * self._size * sizeof(float))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    __pyx_t_1 = ((!(__pyx_v_mem != 0)) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "pyrobuf_list.pyx":651
 *             mem = <float *>PyMem_Realloc(self._data, 2 * self._size * sizeof(float))
 *             if not mem:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *             self._data = mem
 */
      PyErr_NoMemory(); __PYX_ERR(0, 651, __pyx_L1_error)

      /* "pyrobuf_list.pyx":650
 *         if self._n_items == self._size:
 *             mem = <float *>PyMem_Realloc(self._data, 2 * self._size * sizeof(float))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    }

    /* "pyrobuf_list.pyx":653
 *                 raise MemoryError()
 * 
 *             self._data = mem             # <<<<<<<<<<<<<<
 *             self._size *= 2
 * 
 */
    __pyx_v_self->_data = __pyx_v_mem;

    /* "pyrobuf_list.pyx":654
 * 
 *             self._data = mem
 *             self._size *= 2             # <<<<<<<<<<<<<<
 * 
 *         self._data[self._n_items] = x
 */
    __pyx_v_self->_size = (__pyx_v_self->_size * 2);

    /* "pyrobuf_list.pyx":648
 *         cdef float *mem
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <float *>PyMem_Realloc(self._data, 2 * self._size * sizeof(float))
 *             if not mem:
 */
  }

  /* "pyrobuf_list.pyx":656
 *             self._size *= 2
 * 
 *         self._data[self._n_items] = x             # <<<<<<<<<<<<<<
 *         self._n_items += 1
 * 
 */
  (__pyx_v_self->_data[__pyx_v_self->_n_items]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":657
 * 
 *         self._data[self._n_items] = x
 *         self._n_items += 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items + 1);

  /* "pyrobuf_list.pyx":645
 *             listener._Modified()
 * 
 *     cdef void _append(self, float x):             # <<<<<<<<<<<<<<
 *         cdef float *mem
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("pyrobuf_list.FloatList._append", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_42__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_42__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList_41__reduce_cython__(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_41__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.FloatList.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_44__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9FloatList_44__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9FloatList_43__setstate_cython__(((struct __pyx_obj_12pyrobuf_list_FloatList *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9FloatList_43__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_FloatList *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.FloatList.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":663
 * cdef class IntList:
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):             # <<<<<<<<<<<<<<
 *         self._data = <int *>PyMem_Malloc(size * sizeof(int))
 *         if not self._data:
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_7IntList_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12pyrobuf_list_7IntList_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_size;
  PyObject *__pyx_v_listener = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_size,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_listener);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 663, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_size = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_size == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 663, __pyx_L3_error)
    } else {
      __pyx_v_size = ((size_t)16);
    }
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 663, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.IntList.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList___cinit__(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self), __pyx_v_size, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_7IntList___cinit__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, size_t __pyx_v_size, PyObject *__pyx_v_listener) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "pyrobuf_list.pyx":664
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <int *>PyMem_Malloc(size * sizeof(int))             # <<<<<<<<<<<<<<
 *         if not self._data:
 *             raise MemoryError()
 */
  __pyx_v_self->_data = ((int *)PyMem_Malloc((__pyx_v_size * (sizeof(int)))));

  /* "pyrobuf_list.pyx":665
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <int *>PyMem_Malloc(size * sizeof(int))
 *         if not self._data:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_self->_data != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":666
 *         self._data = <int *>PyMem_Malloc(size * sizeof(int))
 *         if not self._data:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         self._n_items = 0
 */
    PyErr_NoMemory(); __PYX_ERR(0, 666, __pyx_L1_error)

    /* "pyrobuf_list.pyx":665
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <int *>PyMem_Malloc(size * sizeof(int))
 *         if not self._data:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "pyrobuf_list.pyx":668
 *             raise MemoryError()
 * 
 *         self._n_items = 0             # <<<<<<<<<<<<<<
 *         self._size = size
 *         if listener is None:
 */
  __pyx_v_self->_n_items = 0;

  /* "pyrobuf_list.pyx":669
 * 
 *         self._n_items = 0
 *         self._size = size             # <<<<<<<<<<<<<<
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener
 */
  __pyx_v_self->_size = __pyx_v_size;

  /* "pyrobuf_list.pyx":670
 *         self._n_items = 0
 *         self._size = size
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
  __pyx_t_1 = (__pyx_v_listener == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":671
 *         self._size = size
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener             # <<<<<<<<<<<<<<
 *         else:
 *             self._listener = <PyObject *>listener
 */
    __pyx_v_self->_listener = ((PyObject *)null_listener);

    /* "pyrobuf_list.pyx":670
 *         self._n_items = 0
 *         self._size = size
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
    goto __pyx_L4;
  }

  /* "pyrobuf_list.pyx":673
 *             self._listener = <PyObject *>null_listener
 *         else:
 *             self._listener = <PyObject *>listener             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  /*else*/ {
    __pyx_v_self->_listener = ((PyObject *)__pyx_v_listener);
  }
  __pyx_L4:;

  /* "pyrobuf_list.pyx":663
 * cdef class IntList:
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):             # <<<<<<<<<<<<<<
 *         self._data = <int *>PyMem_Malloc(size * sizeof(int))
 *         if not self._data:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyrobuf_list.IntList.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":675
 *             self._listener = <PyObject *>listener
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self._data)
 * 
 */

/* Python wrapper */
static void __pyx_pw_12pyrobuf_list_7IntList_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12pyrobuf_list_7IntList_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_12pyrobuf_list_7IntList_2__dealloc__(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12pyrobuf_list_7IntList_2__dealloc__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "pyrobuf_list.pyx":676
 * 
 *     def __dealloc__(self):
 *         PyMem_Free(self._data)             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, int x):
 */
  PyMem_Free(__pyx_v_self->_data);

  /* "pyrobuf_list.pyx":675
 *             self._listener = <PyObject *>listener
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self._data)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "pyrobuf_list.pyx":678
 *         PyMem_Free(self._data)
 * 
 *     def __contains__(self, int x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self._n_items):
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_7IntList_5__contains__(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static int __pyx_pw_12pyrobuf_list_7IntList_5__contains__(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  int __pyx_v_x;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_int(__pyx_arg_x); if (unlikely((__pyx_v_x == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 678, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.IntList.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList_4__contains__(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self), ((int)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_7IntList_4__contains__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_x) {
  size_t __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "pyrobuf_list.pyx":680
 *     def __contains__(self, int x):
 *         cdef size_t i
 *         for i in range(self._n_items):             # <<<<<<<<<<<<<<
 *             if self._data[i] == x:
 *                 return True
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "pyrobuf_list.pyx":681
 *         cdef size_t i
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */
    __pyx_t_4 = (((__pyx_v_self->_data[__pyx_v_i]) == __pyx_v_x) != 0);
    if (__pyx_t_4) {

      /* "pyrobuf_list.pyx":682
 *         for i in range(self._n_items):
 *             if self._data[i] == x:
 *                 return True             # <<<<<<<<<<<<<<
 * 
 *         return False
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "pyrobuf_list.pyx":681
 *         cdef size_t i
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */
    }
  }

  /* "pyrobuf_list.pyx":684
 *                 return True
 * 
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     def __delitem__(self, int i):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":678
 *         PyMem_Free(self._data)
 * 
 *     def __contains__(self, int x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self._n_items):
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":686
 *         return False
 * 
 *     def __delitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_7IntList_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static int __pyx_pw_12pyrobuf_list_7IntList_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 686, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.IntList.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList_6__delitem__(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self), ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_7IntList_6__delitem__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_i) {
  size_t __pyx_v_j;
  PyObject *__pyx_v_listener = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__delitem__", 0);

  /* "pyrobuf_list.pyx":687
 * 
 *     def __delitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":688
 *     def __delitem__(self, int i):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":687
 * 
 *     def __delitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":690
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":691
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t j
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 691, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 691, __pyx_L1_error)

    /* "pyrobuf_list.pyx":690
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":694
 * 
 *         cdef size_t j
 *         for j in range(i, self._n_items):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j + 1]
 * 
 */
  __pyx_t_4 = __pyx_v_self->_n_items;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = __pyx_v_i; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_j = __pyx_t_6;

    /* "pyrobuf_list.pyx":695
 *         cdef size_t j
 *         for j in range(i, self._n_items):
 *             self._data[j] = self._data[j + 1]             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j + 1)]);
  }

  /* "pyrobuf_list.pyx":697
 *             self._data[j] = self._data[j + 1]
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":698
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":699
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":700
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, int i):
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    if (__pyx_t_8) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 700, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 700, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":699
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":686
 *         return False
 * 
 *     def __delitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyrobuf_list.IntList.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":702
 *             listener._Modified()
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 702, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.IntList.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList_8__getitem__(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self), ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_7IntList_8__getitem__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pyrobuf_list.pyx":703
 * 
 *     def __getitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":704
 *     def __getitem__(self, int i):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":703
 * 
 *     def __getitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":706
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":707
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         return self._data[i]
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 707, __pyx_L1_error)

    /* "pyrobuf_list.pyx":706
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":709
 *             raise IndexError("list index out of range")
 * 
 *         return self._data[i]             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_self->_data[__pyx_v_i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":702
 *             listener._Modified()
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyrobuf_list.IntList.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_12pyrobuf_list_7IntList_12generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "pyrobuf_list.pyx":711
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_11__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_11__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList_10__iter__(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_7IntList_10__iter__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self) {
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_2___iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_2___iter__ *)__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_2___iter__(__pyx_ptype_12pyrobuf_list___pyx_scope_struct_2___iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_2___iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 711, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_12pyrobuf_list_7IntList_12generator2, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_IntList___iter, __pyx_n_s_pyrobuf_list); if (unlikely(!gen)) __PYX_ERR(0, 711, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyrobuf_list.IntList.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_12pyrobuf_list_7IntList_12generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_2___iter__ *__pyx_cur_scope = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_2___iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 711, __pyx_L1_error)

  /* "pyrobuf_list.pyx":712
 * 
 *     def __iter__(self):
 *         cdef size_t idx = 0             # <<<<<<<<<<<<<<
 *         while idx < self._n_items:
 *             yield self._data[idx]
 */
  __pyx_cur_scope->__pyx_v_idx = 0;

  /* "pyrobuf_list.pyx":713
 *     def __iter__(self):
 *         cdef size_t idx = 0
 *         while idx < self._n_items:             # <<<<<<<<<<<<<<
 *             yield self._data[idx]
 *             idx += 1
 */
  while (1) {
    __pyx_t_1 = ((__pyx_cur_scope->__pyx_v_idx < __pyx_cur_scope->__pyx_v_self->_n_items) != 0);
    if (!__pyx_t_1) break;

    /* "pyrobuf_list.pyx":714
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 *             yield self._data[idx]             # <<<<<<<<<<<<<<
 *             idx += 1
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_int((__pyx_cur_scope->__pyx_v_self->_data[__pyx_cur_scope->__pyx_v_idx])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 714, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 714, __pyx_L1_error)

    /* "pyrobuf_list.pyx":715
 *         while idx < self._n_items:
 *             yield self._data[idx]
 *             idx += 1             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
    __pyx_cur_scope->__pyx_v_idx = (__pyx_cur_scope->__pyx_v_idx + 1);
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "pyrobuf_list.pyx":711
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":717
 *             idx += 1
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._n_items
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_12pyrobuf_list_7IntList_14__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_12pyrobuf_list_7IntList_14__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList_13__len__(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_12pyrobuf_list_7IntList_13__len__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pyrobuf_list.pyx":718
 * 
 *     def __len__(self):
 *         return self._n_items             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = __pyx_v_self->_n_items;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":717
 *             idx += 1
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._n_items
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":720
 *         return self._n_items
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(list(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_16__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_16__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList_15__repr__(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_7IntList_15__repr__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pyrobuf_list.pyx":721
 * 
 *     def __repr__(self):
 *         return repr(list(self))             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, int i, int x):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":720
 *         return self._n_items
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(list(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.IntList.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":723
 *         return repr(list(self))
 * 
 *     def __setitem__(self, int i, int x):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_7IntList_18__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i, PyObject *__pyx_arg_x); /*proto*/
static int __pyx_pw_12pyrobuf_list_7IntList_18__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i, PyObject *__pyx_arg_x) {
  int __pyx_v_i;
  int __pyx_v_x;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 723, __pyx_L3_error)
  }
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_int(__pyx_arg_x); if (unlikely((__pyx_v_x == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 723, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.IntList.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList_17__setitem__(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self), ((int)__pyx_v_i), ((int)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_7IntList_17__setitem__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_i, int __pyx_v_x) {
  PyObject *__pyx_v_listener = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "pyrobuf_list.pyx":724
 * 
 *     def __setitem__(self, int i, int x):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":725
 *     def __setitem__(self, int i, int x):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":724
 * 
 *     def __setitem__(self, int i, int x):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":727
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":728
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__35, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 728, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 728, __pyx_L1_error)

    /* "pyrobuf_list.pyx":727
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":730
 *             raise IndexError("list index out of range")
 * 
 *         self._data[i] = x             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  (__pyx_v_self->_data[__pyx_v_i]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":731
 * 
 *         self._data[i] = x
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":732
 *         self._data[i] = x
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":733
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 733, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 733, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 733, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":732
 *         self._data[i] = x
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":723
 *         return repr(list(self))
 * 
 *     def __setitem__(self, int i, int x):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.IntList.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":735
 *             listener._Modified()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(list(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_20__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_20__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList_19__str__(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_7IntList_19__str__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pyrobuf_list.pyx":736
 * 
 *     def __str__(self):
 *         return str(list(self))             # <<<<<<<<<<<<<<
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":735
 *             listener._Modified()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(list(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.IntList.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":738
 *         return str(list(self))
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t itemsize = sizeof(int)
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_12pyrobuf_list_7IntList_22__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_12pyrobuf_list_7IntList_22__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList_21__getbuffer__(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_7IntList_21__getbuffer__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags) {
  Py_ssize_t __pyx_v_itemsize;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  if (__pyx_v_buffer == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_buffer->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_buffer->obj);

  /* "pyrobuf_list.pyx":740
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 * 
 *         cdef Py_ssize_t itemsize = sizeof(int)             # <<<<<<<<<<<<<<
 *         self.shape[0] = self._n_items
 *         self.strides[0] = 1
 */
  __pyx_v_itemsize = (sizeof(int));

  /* "pyrobuf_list.pyx":741
 * 
 *         cdef Py_ssize_t itemsize = sizeof(int)
 *         self.shape[0] = self._n_items             # <<<<<<<<<<<<<<
 *         self.strides[0] = 1
 * 
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  (__pyx_v_self->shape[0]) = __pyx_t_1;

  /* "pyrobuf_list.pyx":742
 *         cdef Py_ssize_t itemsize = sizeof(int)
 *         self.shape[0] = self._n_items
 *         self.strides[0] = 1             # <<<<<<<<<<<<<<
 * 
 *         buffer.buf = <char *>(self._data)
 */
  (__pyx_v_self->strides[0]) = 1;

  /* "pyrobuf_list.pyx":744
 *         self.strides[0] = 1
 * 
 *         buffer.buf = <char *>(self._data)             # <<<<<<<<<<<<<<
 *         buffer.format = 'i'
 *         buffer.internal = NULL                  # see References
 */
  __pyx_v_buffer->buf = ((char *)__pyx_v_self->_data);

  /* "pyrobuf_list.pyx":745
 * 
 *         buffer.buf = <char *>(self._data)
 *         buffer.format = 'i'             # <<<<<<<<<<<<<<
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 */
  __pyx_v_buffer->format = ((char *)"i");

  /* "pyrobuf_list.pyx":746
 *         buffer.buf = <char *>(self._data)
 *         buffer.format = 'i'
 *         buffer.internal = NULL                  # see References             # <<<<<<<<<<<<<<
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items
 */
  __pyx_v_buffer->internal = NULL;

  /* "pyrobuf_list.pyx":747
 *         buffer.format = 'i'
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize             # <<<<<<<<<<<<<<
 *         buffer.len = self._n_items
 *         buffer.ndim = 1
 */
  __pyx_v_buffer->itemsize = __pyx_v_itemsize;

  /* "pyrobuf_list.pyx":748
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items             # <<<<<<<<<<<<<<
 *         buffer.ndim = 1
 *         buffer.obj = self
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  __pyx_v_buffer->len = __pyx_t_1;

  /* "pyrobuf_list.pyx":749
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items
 *         buffer.ndim = 1             # <<<<<<<<<<<<<<
 *         buffer.obj = self
 *         buffer.readonly = 1
 */
  __pyx_v_buffer->ndim = 1;

  /* "pyrobuf_list.pyx":750
 *         buffer.len = self._n_items
 *         buffer.ndim = 1
 *         buffer.obj = self             # <<<<<<<<<<<<<<
 *         buffer.readonly = 1
 *         buffer.shape = self.shape
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_buffer->obj);
  __Pyx_DECREF(__pyx_v_buffer->obj);
  __pyx_v_buffer->obj = ((PyObject *)__pyx_v_self);

  /* "pyrobuf_list.pyx":751
 *         buffer.ndim = 1
 *         buffer.obj = self
 *         buffer.readonly = 1             # <<<<<<<<<<<<<<
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides
 */
  __pyx_v_buffer->readonly = 1;

  /* "pyrobuf_list.pyx":752
 *         buffer.obj = self
 *         buffer.readonly = 1
 *         buffer.shape = self.shape             # <<<<<<<<<<<<<<
 *         buffer.strides = self.strides
 *         buffer.suboffsets = NULL                # for pointer arrays only
 */
  __pyx_t_2 = __pyx_v_self->shape;
  __pyx_v_buffer->shape = __pyx_t_2;

  /* "pyrobuf_list.pyx":753
 *         buffer.readonly = 1
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides             # <<<<<<<<<<<<<<
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 */
  __pyx_t_2 = __pyx_v_self->strides;
  __pyx_v_buffer->strides = __pyx_t_2;

  /* "pyrobuf_list.pyx":754
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides
 *         buffer.suboffsets = NULL                # for pointer arrays only             # <<<<<<<<<<<<<<
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):
 */
  __pyx_v_buffer->suboffsets = NULL;

  /* "pyrobuf_list.pyx":738
 *         return str(list(self))
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t itemsize = sizeof(int)
 */

  /* function exit code */
  __pyx_r = 0;
  if (__pyx_v_buffer->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_buffer->obj);
    __Pyx_DECREF(__pyx_v_buffer->obj); __pyx_v_buffer->obj = 0;
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":756
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_12pyrobuf_list_7IntList_24__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer); /*proto*/
static CYTHON_UNUSED void __pyx_pw_12pyrobuf_list_7IntList_24__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_12pyrobuf_list_7IntList_23__releasebuffer__(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12pyrobuf_list_7IntList_23__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "pyrobuf_list.pyx":759
 *         pass
 * 
 *     def __richcmp__(self, other, op):             # <<<<<<<<<<<<<<
 *         if op == Py_EQ:
 *             return self._equal(other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_26__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_26__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op) {
  PyObject *__pyx_v_op = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_v_op = __Pyx_PyInt_From_int(__pyx_arg_op); if (unlikely(!__pyx_v_op)) __PYX_ERR(0, 759, __pyx_L3_error)
  __Pyx_GOTREF(__pyx_v_op);
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.IntList.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList_25__richcmp__(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((PyObject *)__pyx_v_op));

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_op);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_7IntList_25__richcmp__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "pyrobuf_list.pyx":760
 * 
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:             # <<<<<<<<<<<<<<
 *             return self._equal(other)
 *         elif op == Py_NE:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(Py_EQ); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 760, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_op, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 760, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 760, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "pyrobuf_list.pyx":761
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:
 *             return self._equal(other)             # <<<<<<<<<<<<<<
 *         elif op == Py_NE:
 *             return not self._equal(other)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = ((struct __pyx_vtabstruct_12pyrobuf_list_IntList *)__pyx_v_self->__pyx_vtab)->_equal(__pyx_v_self, __pyx_v_other, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 761, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pyrobuf_list.pyx":760
 * 
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:             # <<<<<<<<<<<<<<
 *             return self._equal(other)
 *         elif op == Py_NE:
 */
  }

  /* "pyrobuf_list.pyx":762
 *         if op == Py_EQ:
 *             return self._equal(other)
 *         elif op == Py_NE:             # <<<<<<<<<<<<<<
 *             return not self._equal(other)
 *         else:
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(Py_NE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_op, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(__pyx_t_3)) {

    /* "pyrobuf_list.pyx":763
 *             return self._equal(other)
 *         elif op == Py_NE:
 *             return not self._equal(other)             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError('Operation not supported for lists')
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((struct __pyx_vtabstruct_12pyrobuf_list_IntList *)__pyx_v_self->__pyx_vtab)->_equal(__pyx_v_self, __pyx_v_other, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 763, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 763, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_3)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 763, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pyrobuf_list.pyx":762
 *         if op == Py_EQ:
 *             return self._equal(other)
 *         elif op == Py_NE:             # <<<<<<<<<<<<<<
 *             return not self._equal(other)
 *         else:
 */
  }

  /* "pyrobuf_list.pyx":765
 *             return not self._equal(other)
 *         else:
 *             raise TypeError('Operation not supported for lists')             # <<<<<<<<<<<<<<
 * 
 *     def __neq__(self, other):
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__36, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 765, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 765, __pyx_L1_error)
  }

  /* "pyrobuf_list.pyx":759
 *         pass
 * 
 *     def __richcmp__(self, other, op):             # <<<<<<<<<<<<<<
 *         if op == Py_EQ:
 *             return self._equal(other)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.IntList.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":767
 *             raise TypeError('Operation not supported for lists')
 * 
 *     def __neq__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_28__neq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_28__neq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__neq__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList_27__neq__(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_7IntList_27__neq__(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__neq__", 0);

  /* "pyrobuf_list.pyx":768
 * 
 *     def __neq__(self, other):
 *         return not self == other             # <<<<<<<<<<<<<<
 * 
 *     cpdef _equal(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_self), __pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 768, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":767
 *             raise TypeError('Operation not supported for lists')
 * 
 *     def __neq__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.IntList.__neq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":770
 *         return not self == other
 * 
 *     cpdef _equal(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             if len(self) != len(other):
 */

static PyObject *__pyx_pw_12pyrobuf_list_7IntList_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_7IntList__equal(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_v_y = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *(*__pyx_t_14)(PyObject *);
  __Pyx_RefNannySetupContext("_equal", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_equal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 770, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_7IntList_30_equal)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_other); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 770, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_other};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 770, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_other};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 770, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 770, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_other);
          __Pyx_GIVEREF(__pyx_v_other);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_other);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 770, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":771
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    /*try:*/ {

      /* "pyrobuf_list.pyx":772
 *     cpdef _equal(self, other):
 *         try:
 *             if len(self) != len(other):             # <<<<<<<<<<<<<<
 *                 return False
 *         except TypeError:
 */
      __pyx_t_9 = PyObject_Length(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 772, __pyx_L3_error)
      __pyx_t_10 = PyObject_Length(__pyx_v_other); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 772, __pyx_L3_error)
      __pyx_t_11 = ((__pyx_t_9 != __pyx_t_10) != 0);
      if (__pyx_t_11) {

        /* "pyrobuf_list.pyx":773
 *         try:
 *             if len(self) != len(other):
 *                 return False             # <<<<<<<<<<<<<<
 *         except TypeError:
 *             return False
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_False);
        __pyx_r = Py_False;
        goto __pyx_L7_try_return;

        /* "pyrobuf_list.pyx":772
 *     cpdef _equal(self, other):
 *         try:
 *             if len(self) != len(other):             # <<<<<<<<<<<<<<
 *                 return False
 *         except TypeError:
 */
      }

      /* "pyrobuf_list.pyx":771
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
    }
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":774
 *             if len(self) != len(other):
 *                 return False
 *         except TypeError:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
    __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_12) {
      __Pyx_AddTraceback("pyrobuf_list.IntList._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 774, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_3);

      /* "pyrobuf_list.pyx":775
 *                 return False
 *         except TypeError:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         for x, y in zip(self, other):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "pyrobuf_list.pyx":771
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "pyrobuf_list.pyx":777
 *             return False
 * 
 *         for x, y in zip(self, other):             # <<<<<<<<<<<<<<
 *             if x != y:
 *                 return False
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));
  __Pyx_INCREF(__pyx_v_other);
  __Pyx_GIVEREF(__pyx_v_other);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_other);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_10 = 0;
    __pyx_t_13 = NULL;
  } else {
    __pyx_t_10 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_13 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 777, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_13)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 777, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 777, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 777, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 777, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_13(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 777, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 777, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 777, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 777, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 777, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_14 = Py_TYPE(__pyx_t_4)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_5 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_5)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_4), 2) < 0) __PYX_ERR(0, 777, __pyx_L1_error)
      __pyx_t_14 = NULL;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L15_unpacking_done;
      __pyx_L14_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_14 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 777, __pyx_L1_error)
      __pyx_L15_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_x, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_y, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "pyrobuf_list.pyx":778
 * 
 *         for x, y in zip(self, other):
 *             if x != y:             # <<<<<<<<<<<<<<
 *                 return False
 * 
 */
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_x, __pyx_v_y, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 778, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 778, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_11) {

      /* "pyrobuf_list.pyx":779
 *         for x, y in zip(self, other):
 *             if x != y:
 *                 return False             # <<<<<<<<<<<<<<
 * 
 *         return True
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "pyrobuf_list.pyx":778
 * 
 *         for x, y in zip(self, other):
 *             if x != y:             # <<<<<<<<<<<<<<
 *                 return False
 * 
 */
    }

    /* "pyrobuf_list.pyx":777
 *             return False
 * 
 *         for x, y in zip(self, other):             # <<<<<<<<<<<<<<
 *             if x != y:
 *                 return False
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":781
 *                 return False
 * 
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     cpdef append(self, int x):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":770
 *         return not self == other
 * 
 *     cpdef _equal(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             if len(self) != len(other):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.IntList._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_y);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_equal (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList_29_equal(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_7IntList_29_equal(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_equal", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_7IntList__equal(__pyx_v_self, __pyx_v_other, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 770, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.IntList._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":783
 *         return True
 * 
 *     cpdef append(self, int x):             # <<<<<<<<<<<<<<
 *         self._append(x)
 *         listener = <object>self._listener
 */

static PyObject *__pyx_pw_12pyrobuf_list_7IntList_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_7IntList_append(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("append", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_append); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 783, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_7IntList_32append)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 783, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 783, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 783, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 783, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 783, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 783, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":784
 * 
 *     cpdef append(self, int x):
 *         self._append(x)             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  ((struct __pyx_vtabstruct_12pyrobuf_list_IntList *)__pyx_v_self->__pyx_vtab)->_append(__pyx_v_self, __pyx_v_x);

  /* "pyrobuf_list.pyx":785
 *     cpdef append(self, int x):
 *         self._append(x)
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":786
 *         self._append(x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = ((!__pyx_t_7) != 0);
  if (__pyx_t_8) {

    /* "pyrobuf_list.pyx":787
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef extend(self, x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 787, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 787, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 787, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":786
 *         self._append(x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":783
 *         return True
 * 
 *     cpdef append(self, int x):             # <<<<<<<<<<<<<<
 *         self._append(x)
 *         listener = <object>self._listener
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.IntList.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  int __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_int(__pyx_arg_x); if (unlikely((__pyx_v_x == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 783, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.IntList.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList_31append(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self), ((int)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_7IntList_31append(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("append", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_7IntList_append(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.IntList.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":789
 *             listener._Modified()
 * 
 *     cpdef extend(self, x):             # <<<<<<<<<<<<<<
 *         for i in x:
 *             self._append(i)
 */

static PyObject *__pyx_pw_12pyrobuf_list_7IntList_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_7IntList_extend(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, PyObject *__pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("extend", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extend); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 789, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_7IntList_34extend)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 789, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_x};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 789, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_x};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 789, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 789, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_x);
          __Pyx_GIVEREF(__pyx_v_x);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_x);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 789, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":790
 * 
 *     cpdef extend(self, x):
 *         for i in x:             # <<<<<<<<<<<<<<
 *             self._append(i)
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_x)) || PyTuple_CheckExact(__pyx_v_x)) {
    __pyx_t_1 = __pyx_v_x; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 790, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 790, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 790, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_7(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 790, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyrobuf_list.pyx":791
 *     cpdef extend(self, x):
 *         for i in x:
 *             self._append(i)             # <<<<<<<<<<<<<<
 * 
 *         listener = <object>self._listener
 */
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_v_i); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 791, __pyx_L1_error)
    ((struct __pyx_vtabstruct_12pyrobuf_list_IntList *)__pyx_v_self->__pyx_vtab)->_append(__pyx_v_self, __pyx_t_8);

    /* "pyrobuf_list.pyx":790
 * 
 *     cpdef extend(self, x):
 *         for i in x:             # <<<<<<<<<<<<<<
 *             self._append(i)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":793
 *             self._append(i)
 * 
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":794
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = ((!__pyx_t_9) != 0);
  if (__pyx_t_10) {

    /* "pyrobuf_list.pyx":795
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef insert(self, int i, int x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 795, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 795, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 795, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":794
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":789
 *             listener._Modified()
 * 
 *     cpdef extend(self, x):             # <<<<<<<<<<<<<<
 *         for i in x:
 *             self._append(i)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.IntList.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extend (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList_33extend(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_7IntList_33extend(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("extend", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_7IntList_extend(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.IntList.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":797
 *             listener._Modified()
 * 
 *     cpdef insert(self, int i, int x):             # <<<<<<<<<<<<<<
 *         cdef int *mem
 *         cdef size_t j
 */

static PyObject *__pyx_pw_12pyrobuf_list_7IntList_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_7IntList_insert(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_i, int __pyx_v_x, int __pyx_skip_dispatch) {
  int *__pyx_v_mem;
  size_t __pyx_v_j;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  size_t __pyx_t_11;
  size_t __pyx_t_12;
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("insert", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_insert); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 797, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_7IntList_36insert)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 797, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_x); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 797, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 797, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 797, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 797, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
        __pyx_t_3 = 0;
        __pyx_t_4 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 797, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":801
 *         cdef size_t j
 * 
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_9 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_9) {

    /* "pyrobuf_list.pyx":802
 * 
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":801
 *         cdef size_t j
 * 
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":804
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i > self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_10 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_10) {
  } else {
    __pyx_t_9 = __pyx_t_10;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_10 = ((((size_t)__pyx_v_i) > __pyx_v_self->_n_items) != 0);
  __pyx_t_9 = __pyx_t_10;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_9)) {

    /* "pyrobuf_list.pyx":805
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         if self._n_items == self._size:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__37, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 805, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 805, __pyx_L1_error)

    /* "pyrobuf_list.pyx":804
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i > self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":807
 *             raise IndexError("list index out of range")
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <int *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int))
 *             if not mem:
 */
  __pyx_t_9 = ((__pyx_v_self->_n_items == __pyx_v_self->_size) != 0);
  if (__pyx_t_9) {

    /* "pyrobuf_list.pyx":808
 * 
 *         if self._n_items == self._size:
 *             mem = <int *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int))             # <<<<<<<<<<<<<<
 *             if not mem:
 *                 raise MemoryError()
 */
    __pyx_v_mem = ((int *)PyMem_Realloc(__pyx_v_self->_data, ((2 * __pyx_v_self->_size) * (sizeof(int)))));

    /* "pyrobuf_list.pyx":809
 *         if self._n_items == self._size:
 *             mem = <int *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    __pyx_t_9 = ((!(__pyx_v_mem != 0)) != 0);
    if (unlikely(__pyx_t_9)) {

      /* "pyrobuf_list.pyx":810
 *             mem = <int *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int))
 *             if not mem:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *             self._data = mem
 */
      PyErr_NoMemory(); __PYX_ERR(0, 810, __pyx_L1_error)

      /* "pyrobuf_list.pyx":809
 *         if self._n_items == self._size:
 *             mem = <int *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    }

    /* "pyrobuf_list.pyx":812
 *                 raise MemoryError()
 * 
 *             self._data = mem             # <<<<<<<<<<<<<<
 *             self._size *= 2
 * 
 */
    __pyx_v_self->_data = __pyx_v_mem;

    /* "pyrobuf_list.pyx":813
 * 
 *             self._data = mem
 *             self._size *= 2             # <<<<<<<<<<<<<<
 * 
 *         for j in range(self._n_items, <size_t>i, -1):
 */
    __pyx_v_self->_size = (__pyx_v_self->_size * 2);

    /* "pyrobuf_list.pyx":807
 *             raise IndexError("list index out of range")
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <int *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int))
 *             if not mem:
 */
  }

  /* "pyrobuf_list.pyx":815
 *             self._size *= 2
 * 
 *         for j in range(self._n_items, <size_t>i, -1):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j - 1]
 * 
 */
  __pyx_t_11 = ((size_t)__pyx_v_i);
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = __pyx_v_self->_n_items + 1; __pyx_t_13 > __pyx_t_12 + 1; ) { __pyx_t_13-=1;
    __pyx_v_j = __pyx_t_13;

    /* "pyrobuf_list.pyx":816
 * 
 *         for j in range(self._n_items, <size_t>i, -1):
 *             self._data[j] = self._data[j - 1]             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j - 1)]);
  }

  /* "pyrobuf_list.pyx":818
 *             self._data[j] = self._data[j - 1]
 * 
 *         self._data[i] = x             # <<<<<<<<<<<<<<
 *         self._n_items += 1
 *         listener = <object>self._listener
 */
  (__pyx_v_self->_data[__pyx_v_i]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":819
 * 
 *         self._data[i] = x
 *         self._n_items += 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items + 1);

  /* "pyrobuf_list.pyx":820
 *         self._data[i] = x
 *         self._n_items += 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":821
 *         self._n_items += 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = ((!__pyx_t_9) != 0);
  if (__pyx_t_10) {

    /* "pyrobuf_list.pyx":822
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef pop(self):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 822, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 822, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 822, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":821
 *         self._n_items += 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":797
 *             listener._Modified()
 * 
 *     cpdef insert(self, int i, int x):             # <<<<<<<<<<<<<<
 *         cdef int *mem
 *         cdef size_t j
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyrobuf_list.IntList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_i;
  int __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("insert (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_i,&__pyx_n_s_x,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, 1); __PYX_ERR(0, 797, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "insert") < 0)) __PYX_ERR(0, 797, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_i = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 797, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_x == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 797, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 797, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.IntList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList_35insert(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self), __pyx_v_i, __pyx_v_x);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_7IntList_35insert(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_i, int __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("insert", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_7IntList_insert(__pyx_v_self, __pyx_v_i, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.IntList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":824
 *             listener._Modified()
 * 
 *     cpdef pop(self):             # <<<<<<<<<<<<<<
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")
 */

static PyObject *__pyx_pw_12pyrobuf_list_7IntList_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_7IntList_pop(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("pop", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 824, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_7IntList_38pop)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 824, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 824, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":825
 * 
 *     cpdef pop(self):
 *         if self._n_items == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("pop from empty list")
 * 
 */
  __pyx_t_5 = ((__pyx_v_self->_n_items == 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "pyrobuf_list.pyx":826
 *     cpdef pop(self):
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__38, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 826, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 826, __pyx_L1_error)

    /* "pyrobuf_list.pyx":825
 * 
 *     cpdef pop(self):
 *         if self._n_items == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("pop from empty list")
 * 
 */
  }

  /* "pyrobuf_list.pyx":828
 *             raise IndexError("pop from empty list")
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":829
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":830
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 *         return self._data[self._n_items]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = ((!__pyx_t_5) != 0);
  if (__pyx_t_6) {

    /* "pyrobuf_list.pyx":831
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 *         return self._data[self._n_items]
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 831, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 831, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 831, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":830
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 *         return self._data[self._n_items]
 */
  }

  /* "pyrobuf_list.pyx":832
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 *         return self._data[self._n_items]             # <<<<<<<<<<<<<<
 * 
 *     cpdef remove(self, int x):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int((__pyx_v_self->_data[__pyx_v_self->_n_items])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":824
 *             listener._Modified()
 * 
 *     cpdef pop(self):             # <<<<<<<<<<<<<<
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyrobuf_list.IntList.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pop (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList_37pop(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_7IntList_37pop(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("pop", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_7IntList_pop(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.IntList.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":834
 *         return self._data[self._n_items]
 * 
 *     cpdef remove(self, int x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef size_t j
 */

static PyObject *__pyx_pw_12pyrobuf_list_7IntList_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_7IntList_remove(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_x, int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  int __pyx_v_found;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("remove", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_remove); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 834, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_7IntList_40remove)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 834, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 834, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 834, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 834, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 834, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 834, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":837
 *         cdef size_t i
 *         cdef size_t j
 *         cdef bint found = 0             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self._n_items):
 */
  __pyx_v_found = 0;

  /* "pyrobuf_list.pyx":839
 *         cdef bint found = 0
 * 
 *         for i in range(self._n_items):             # <<<<<<<<<<<<<<
 *             if self._data[i] == x:
 *                 found = 1
 */
  __pyx_t_7 = __pyx_v_self->_n_items;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "pyrobuf_list.pyx":840
 * 
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 found = 1
 *                 break
 */
    __pyx_t_10 = (((__pyx_v_self->_data[__pyx_v_i]) == __pyx_v_x) != 0);
    if (__pyx_t_10) {

      /* "pyrobuf_list.pyx":841
 *         for i in range(self._n_items):
 *             if self._data[i] == x:
 *                 found = 1             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
      __pyx_v_found = 1;

      /* "pyrobuf_list.pyx":842
 *             if self._data[i] == x:
 *                 found = 1
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         if found == 0:
 */
      goto __pyx_L4_break;

      /* "pyrobuf_list.pyx":840
 * 
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 found = 1
 *                 break
 */
    }
  }
  __pyx_L4_break:;

  /* "pyrobuf_list.pyx":844
 *                 break
 * 
 *         if found == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("x not in list")
 * 
 */
  __pyx_t_10 = ((__pyx_v_found == 0) != 0);
  if (unlikely(__pyx_t_10)) {

    /* "pyrobuf_list.pyx":845
 * 
 *         if found == 0:
 *             raise ValueError("x not in list")             # <<<<<<<<<<<<<<
 * 
 *         for j in range(i, self._n_items):
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__39, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 845, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 845, __pyx_L1_error)

    /* "pyrobuf_list.pyx":844
 *                 break
 * 
 *         if found == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("x not in list")
 * 
 */
  }

  /* "pyrobuf_list.pyx":847
 *             raise ValueError("x not in list")
 * 
 *         for j in range(i, self._n_items):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j + 1]
 * 
 */
  __pyx_t_7 = __pyx_v_self->_n_items;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = __pyx_v_i; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_j = __pyx_t_9;

    /* "pyrobuf_list.pyx":848
 * 
 *         for j in range(i, self._n_items):
 *             self._data[j] = self._data[j + 1]             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j + 1)]);
  }

  /* "pyrobuf_list.pyx":850
 *             self._data[j] = self._data[j + 1]
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":851
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":852
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = ((!__pyx_t_10) != 0);
  if (__pyx_t_11) {

    /* "pyrobuf_list.pyx":853
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cdef void _append(self, int x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 853, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 853, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 853, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":852
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":834
 *         return self._data[self._n_items]
 * 
 *     cpdef remove(self, int x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef size_t j
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.IntList.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  int __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_int(__pyx_arg_x); if (unlikely((__pyx_v_x == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 834, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.IntList.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList_39remove(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self), ((int)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_7IntList_39remove(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("remove", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_7IntList_remove(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.IntList.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":855
 *             listener._Modified()
 * 
 *     cdef void _append(self, int x):             # <<<<<<<<<<<<<<
 *         cdef int *mem
 * 
 */

static void __pyx_f_12pyrobuf_list_7IntList__append(struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, int __pyx_v_x) {
  int *__pyx_v_mem;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_append", 0);

  /* "pyrobuf_list.pyx":858
 *         cdef int *mem
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <int *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int))
 *             if not mem:
 */
  __pyx_t_1 = ((__pyx_v_self->_n_items == __pyx_v_self->_size) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":859
 * 
 *         if self._n_items == self._size:
 *             mem = <int *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int))             # <<<<<<<<<<<<<<
 *             if not mem:
 *                 raise MemoryError()
 */
    __pyx_v_mem = ((int *)PyMem_Realloc(__pyx_v_self->_data, ((2 * __pyx_v_self->_size) * (sizeof(int)))));

    /* "pyrobuf_list.pyx":860
 *         if self._n_items == self._size:
 *             mem = <int *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    __pyx_t_1 = ((!(__pyx_v_mem != 0)) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "pyrobuf_list.pyx":861
 *             mem = <int *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int))
 *             if not mem:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *             self._data = mem
 */
      PyErr_NoMemory(); __PYX_ERR(0, 861, __pyx_L1_error)

      /* "pyrobuf_list.pyx":860
 *         if self._n_items == self._size:
 *             mem = <int *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    }

    /* "pyrobuf_list.pyx":863
 *                 raise MemoryError()
 * 
 *             self._data = mem             # <<<<<<<<<<<<<<
 *             self._size *= 2
 * 
 */
    __pyx_v_self->_data = __pyx_v_mem;

    /* "pyrobuf_list.pyx":864
 * 
 *             self._data = mem
 *             self._size *= 2             # <<<<<<<<<<<<<<
 * 
 *         self._data[self._n_items] = x
 */
    __pyx_v_self->_size = (__pyx_v_self->_size * 2);

    /* "pyrobuf_list.pyx":858
 *         cdef int *mem
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <int *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int))
 *             if not mem:
 */
  }

  /* "pyrobuf_list.pyx":866
 *             self._size *= 2
 * 
 *         self._data[self._n_items] = x             # <<<<<<<<<<<<<<
 *         self._n_items += 1
 * 
 */
  (__pyx_v_self->_data[__pyx_v_self->_n_items]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":867
 * 
 *         self._data[self._n_items] = x
 *         self._n_items += 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items + 1);

  /* "pyrobuf_list.pyx":855
 *             listener._Modified()
 * 
 *     cdef void _append(self, int x):             # <<<<<<<<<<<<<<
 *         cdef int *mem
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("pyrobuf_list.IntList._append", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_42__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_42__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList_41__reduce_cython__(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_7IntList_41__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__40, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.IntList.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_44__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_7IntList_44__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_7IntList_43__setstate_cython__(((struct __pyx_obj_12pyrobuf_list_IntList *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_7IntList_43__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_IntList *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__41, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.IntList.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":873
 * cdef class Int32List:
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):             # <<<<<<<<<<<<<<
 *         self._data = <int32_t *>PyMem_Malloc(size * sizeof(int32_t))
 *         if not self._data:
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_9Int32List_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12pyrobuf_list_9Int32List_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_size;
  PyObject *__pyx_v_listener = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_size,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_listener);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 873, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_size = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_size == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 873, __pyx_L3_error)
    } else {
      __pyx_v_size = ((size_t)16);
    }
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 873, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int32List.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List___cinit__(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self), __pyx_v_size, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_9Int32List___cinit__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, size_t __pyx_v_size, PyObject *__pyx_v_listener) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "pyrobuf_list.pyx":874
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <int32_t *>PyMem_Malloc(size * sizeof(int32_t))             # <<<<<<<<<<<<<<
 *         if not self._data:
 *             raise MemoryError()
 */
  __pyx_v_self->_data = ((int32_t *)PyMem_Malloc((__pyx_v_size * (sizeof(int32_t)))));

  /* "pyrobuf_list.pyx":875
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <int32_t *>PyMem_Malloc(size * sizeof(int32_t))
 *         if not self._data:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_self->_data != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":876
 *         self._data = <int32_t *>PyMem_Malloc(size * sizeof(int32_t))
 *         if not self._data:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         self._n_items = 0
 */
    PyErr_NoMemory(); __PYX_ERR(0, 876, __pyx_L1_error)

    /* "pyrobuf_list.pyx":875
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <int32_t *>PyMem_Malloc(size * sizeof(int32_t))
 *         if not self._data:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "pyrobuf_list.pyx":878
 *             raise MemoryError()
 * 
 *         self._n_items = 0             # <<<<<<<<<<<<<<
 *         self._size = size
 *         if listener is None:
 */
  __pyx_v_self->_n_items = 0;

  /* "pyrobuf_list.pyx":879
 * 
 *         self._n_items = 0
 *         self._size = size             # <<<<<<<<<<<<<<
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener
 */
  __pyx_v_self->_size = __pyx_v_size;

  /* "pyrobuf_list.pyx":880
 *         self._n_items = 0
 *         self._size = size
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
  __pyx_t_1 = (__pyx_v_listener == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":881
 *         self._size = size
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener             # <<<<<<<<<<<<<<
 *         else:
 *             self._listener = <PyObject *>listener
 */
    __pyx_v_self->_listener = ((PyObject *)null_listener);

    /* "pyrobuf_list.pyx":880
 *         self._n_items = 0
 *         self._size = size
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
    goto __pyx_L4;
  }

  /* "pyrobuf_list.pyx":883
 *             self._listener = <PyObject *>null_listener
 *         else:
 *             self._listener = <PyObject *>listener             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  /*else*/ {
    __pyx_v_self->_listener = ((PyObject *)__pyx_v_listener);
  }
  __pyx_L4:;

  /* "pyrobuf_list.pyx":873
 * cdef class Int32List:
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):             # <<<<<<<<<<<<<<
 *         self._data = <int32_t *>PyMem_Malloc(size * sizeof(int32_t))
 *         if not self._data:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int32List.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":885
 *             self._listener = <PyObject *>listener
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self._data)
 * 
 */

/* Python wrapper */
static void __pyx_pw_12pyrobuf_list_9Int32List_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12pyrobuf_list_9Int32List_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_12pyrobuf_list_9Int32List_2__dealloc__(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12pyrobuf_list_9Int32List_2__dealloc__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "pyrobuf_list.pyx":886
 * 
 *     def __dealloc__(self):
 *         PyMem_Free(self._data)             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, int32_t x):
 */
  PyMem_Free(__pyx_v_self->_data);

  /* "pyrobuf_list.pyx":885
 *             self._listener = <PyObject *>listener
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self._data)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "pyrobuf_list.pyx":888
 *         PyMem_Free(self._data)
 * 
 *     def __contains__(self, int32_t x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self._n_items):
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_9Int32List_5__contains__(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static int __pyx_pw_12pyrobuf_list_9Int32List_5__contains__(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  int32_t __pyx_v_x;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_int32_t(__pyx_arg_x); if (unlikely((__pyx_v_x == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 888, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int32List.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List_4__contains__(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self), ((int32_t)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_9Int32List_4__contains__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int32_t __pyx_v_x) {
  size_t __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "pyrobuf_list.pyx":890
 *     def __contains__(self, int32_t x):
 *         cdef size_t i
 *         for i in range(self._n_items):             # <<<<<<<<<<<<<<
 *             if self._data[i] == x:
 *                 return True
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "pyrobuf_list.pyx":891
 *         cdef size_t i
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */
    __pyx_t_4 = (((__pyx_v_self->_data[__pyx_v_i]) == __pyx_v_x) != 0);
    if (__pyx_t_4) {

      /* "pyrobuf_list.pyx":892
 *         for i in range(self._n_items):
 *             if self._data[i] == x:
 *                 return True             # <<<<<<<<<<<<<<
 * 
 *         return False
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "pyrobuf_list.pyx":891
 *         cdef size_t i
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */
    }
  }

  /* "pyrobuf_list.pyx":894
 *                 return True
 * 
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     def __delitem__(self, int i):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":888
 *         PyMem_Free(self._data)
 * 
 *     def __contains__(self, int32_t x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self._n_items):
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":896
 *         return False
 * 
 *     def __delitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_9Int32List_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static int __pyx_pw_12pyrobuf_list_9Int32List_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 896, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int32List.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List_6__delitem__(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self), ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_9Int32List_6__delitem__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int __pyx_v_i) {
  size_t __pyx_v_j;
  PyObject *__pyx_v_listener = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__delitem__", 0);

  /* "pyrobuf_list.pyx":897
 * 
 *     def __delitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":898
 *     def __delitem__(self, int i):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":897
 * 
 *     def __delitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":900
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":901
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t j
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__42, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 901, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 901, __pyx_L1_error)

    /* "pyrobuf_list.pyx":900
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":904
 * 
 *         cdef size_t j
 *         for j in range(i, self._n_items):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j + 1]
 * 
 */
  __pyx_t_4 = __pyx_v_self->_n_items;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = __pyx_v_i; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_j = __pyx_t_6;

    /* "pyrobuf_list.pyx":905
 *         cdef size_t j
 *         for j in range(i, self._n_items):
 *             self._data[j] = self._data[j + 1]             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j + 1)]);
  }

  /* "pyrobuf_list.pyx":907
 *             self._data[j] = self._data[j + 1]
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":908
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":909
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 909, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":910
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, int i):
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 910, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    if (__pyx_t_8) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 910, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 910, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":909
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":896
 *         return False
 * 
 *     def __delitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyrobuf_list.Int32List.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":912
 *             listener._Modified()
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 912, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int32List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List_8__getitem__(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self), ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_8__getitem__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int __pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pyrobuf_list.pyx":913
 * 
 *     def __getitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":914
 *     def __getitem__(self, int i):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":913
 * 
 *     def __getitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":916
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":917
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         return self._data[i]
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__43, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 917, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 917, __pyx_L1_error)

    /* "pyrobuf_list.pyx":916
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":919
 *             raise IndexError("list index out of range")
 * 
 *         return self._data[i]             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_int32_t((__pyx_v_self->_data[__pyx_v_i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":912
 *             listener._Modified()
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyrobuf_list.Int32List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_12pyrobuf_list_9Int32List_12generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "pyrobuf_list.pyx":921
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_11__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_11__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List_10__iter__(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_10__iter__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self) {
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_3___iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_3___iter__ *)__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_3___iter__(__pyx_ptype_12pyrobuf_list___pyx_scope_struct_3___iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_3___iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 921, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_12pyrobuf_list_9Int32List_12generator3, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_Int32List___iter, __pyx_n_s_pyrobuf_list); if (unlikely(!gen)) __PYX_ERR(0, 921, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int32List.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_12pyrobuf_list_9Int32List_12generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_3___iter__ *__pyx_cur_scope = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_3___iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 921, __pyx_L1_error)

  /* "pyrobuf_list.pyx":922
 * 
 *     def __iter__(self):
 *         cdef size_t idx = 0             # <<<<<<<<<<<<<<
 *         while idx < self._n_items:
 *             yield self._data[idx]
 */
  __pyx_cur_scope->__pyx_v_idx = 0;

  /* "pyrobuf_list.pyx":923
 *     def __iter__(self):
 *         cdef size_t idx = 0
 *         while idx < self._n_items:             # <<<<<<<<<<<<<<
 *             yield self._data[idx]
 *             idx += 1
 */
  while (1) {
    __pyx_t_1 = ((__pyx_cur_scope->__pyx_v_idx < __pyx_cur_scope->__pyx_v_self->_n_items) != 0);
    if (!__pyx_t_1) break;

    /* "pyrobuf_list.pyx":924
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 *             yield self._data[idx]             # <<<<<<<<<<<<<<
 *             idx += 1
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_int32_t((__pyx_cur_scope->__pyx_v_self->_data[__pyx_cur_scope->__pyx_v_idx])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 924, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 924, __pyx_L1_error)

    /* "pyrobuf_list.pyx":925
 *         while idx < self._n_items:
 *             yield self._data[idx]
 *             idx += 1             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
    __pyx_cur_scope->__pyx_v_idx = (__pyx_cur_scope->__pyx_v_idx + 1);
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "pyrobuf_list.pyx":921
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":927
 *             idx += 1
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._n_items
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_12pyrobuf_list_9Int32List_14__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_12pyrobuf_list_9Int32List_14__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List_13__len__(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_12pyrobuf_list_9Int32List_13__len__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pyrobuf_list.pyx":928
 * 
 *     def __len__(self):
 *         return self._n_items             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = __pyx_v_self->_n_items;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":927
 *             idx += 1
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._n_items
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":930
 *         return self._n_items
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(list(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_16__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_16__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List_15__repr__(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_15__repr__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pyrobuf_list.pyx":931
 * 
 *     def __repr__(self):
 *         return repr(list(self))             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, int i, int32_t x):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 931, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":930
 *         return self._n_items
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(list(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.Int32List.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":933
 *         return repr(list(self))
 * 
 *     def __setitem__(self, int i, int32_t x):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_9Int32List_18__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i, PyObject *__pyx_arg_x); /*proto*/
static int __pyx_pw_12pyrobuf_list_9Int32List_18__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i, PyObject *__pyx_arg_x) {
  int __pyx_v_i;
  int32_t __pyx_v_x;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 933, __pyx_L3_error)
  }
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_int32_t(__pyx_arg_x); if (unlikely((__pyx_v_x == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 933, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int32List.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List_17__setitem__(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self), ((int)__pyx_v_i), ((int32_t)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_9Int32List_17__setitem__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int __pyx_v_i, int32_t __pyx_v_x) {
  PyObject *__pyx_v_listener = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "pyrobuf_list.pyx":934
 * 
 *     def __setitem__(self, int i, int32_t x):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":935
 *     def __setitem__(self, int i, int32_t x):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":934
 * 
 *     def __setitem__(self, int i, int32_t x):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":937
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":938
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__44, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 938, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 938, __pyx_L1_error)

    /* "pyrobuf_list.pyx":937
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":940
 *             raise IndexError("list index out of range")
 * 
 *         self._data[i] = x             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  (__pyx_v_self->_data[__pyx_v_i]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":941
 * 
 *         self._data[i] = x
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":942
 *         self._data[i] = x
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":943
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 943, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 943, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 943, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":942
 *         self._data[i] = x
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":933
 *         return repr(list(self))
 * 
 *     def __setitem__(self, int i, int32_t x):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.Int32List.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":945
 *             listener._Modified()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(list(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_20__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_20__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List_19__str__(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_19__str__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pyrobuf_list.pyx":946
 * 
 *     def __str__(self):
 *         return str(list(self))             # <<<<<<<<<<<<<<
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":945
 *             listener._Modified()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(list(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.Int32List.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":948
 *         return str(list(self))
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t itemsize = sizeof(int32_t)
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_12pyrobuf_list_9Int32List_22__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_12pyrobuf_list_9Int32List_22__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List_21__getbuffer__(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_9Int32List_21__getbuffer__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags) {
  Py_ssize_t __pyx_v_itemsize;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  if (__pyx_v_buffer == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_buffer->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_buffer->obj);

  /* "pyrobuf_list.pyx":950
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 * 
 *         cdef Py_ssize_t itemsize = sizeof(int32_t)             # <<<<<<<<<<<<<<
 *         self.shape[0] = self._n_items
 *         self.strides[0] = 1
 */
  __pyx_v_itemsize = (sizeof(int32_t));

  /* "pyrobuf_list.pyx":951
 * 
 *         cdef Py_ssize_t itemsize = sizeof(int32_t)
 *         self.shape[0] = self._n_items             # <<<<<<<<<<<<<<
 *         self.strides[0] = 1
 * 
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  (__pyx_v_self->shape[0]) = __pyx_t_1;

  /* "pyrobuf_list.pyx":952
 *         cdef Py_ssize_t itemsize = sizeof(int32_t)
 *         self.shape[0] = self._n_items
 *         self.strides[0] = 1             # <<<<<<<<<<<<<<
 * 
 *         buffer.buf = <char *>(self._data)
 */
  (__pyx_v_self->strides[0]) = 1;

  /* "pyrobuf_list.pyx":954
 *         self.strides[0] = 1
 * 
 *         buffer.buf = <char *>(self._data)             # <<<<<<<<<<<<<<
 *         buffer.format = 'i'
 *         buffer.internal = NULL                  # see References
 */
  __pyx_v_buffer->buf = ((char *)__pyx_v_self->_data);

  /* "pyrobuf_list.pyx":955
 * 
 *         buffer.buf = <char *>(self._data)
 *         buffer.format = 'i'             # <<<<<<<<<<<<<<
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 */
  __pyx_v_buffer->format = ((char *)"i");

  /* "pyrobuf_list.pyx":956
 *         buffer.buf = <char *>(self._data)
 *         buffer.format = 'i'
 *         buffer.internal = NULL                  # see References             # <<<<<<<<<<<<<<
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items
 */
  __pyx_v_buffer->internal = NULL;

  /* "pyrobuf_list.pyx":957
 *         buffer.format = 'i'
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize             # <<<<<<<<<<<<<<
 *         buffer.len = self._n_items
 *         buffer.ndim = 1
 */
  __pyx_v_buffer->itemsize = __pyx_v_itemsize;

  /* "pyrobuf_list.pyx":958
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items             # <<<<<<<<<<<<<<
 *         buffer.ndim = 1
 *         buffer.obj = self
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  __pyx_v_buffer->len = __pyx_t_1;

  /* "pyrobuf_list.pyx":959
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items
 *         buffer.ndim = 1             # <<<<<<<<<<<<<<
 *         buffer.obj = self
 *         buffer.readonly = 1
 */
  __pyx_v_buffer->ndim = 1;

  /* "pyrobuf_list.pyx":960
 *         buffer.len = self._n_items
 *         buffer.ndim = 1
 *         buffer.obj = self             # <<<<<<<<<<<<<<
 *         buffer.readonly = 1
 *         buffer.shape = self.shape
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_buffer->obj);
  __Pyx_DECREF(__pyx_v_buffer->obj);
  __pyx_v_buffer->obj = ((PyObject *)__pyx_v_self);

  /* "pyrobuf_list.pyx":961
 *         buffer.ndim = 1
 *         buffer.obj = self
 *         buffer.readonly = 1             # <<<<<<<<<<<<<<
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides
 */
  __pyx_v_buffer->readonly = 1;

  /* "pyrobuf_list.pyx":962
 *         buffer.obj = self
 *         buffer.readonly = 1
 *         buffer.shape = self.shape             # <<<<<<<<<<<<<<
 *         buffer.strides = self.strides
 *         buffer.suboffsets = NULL                # for pointer arrays only
 */
  __pyx_t_2 = __pyx_v_self->shape;
  __pyx_v_buffer->shape = __pyx_t_2;

  /* "pyrobuf_list.pyx":963
 *         buffer.readonly = 1
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides             # <<<<<<<<<<<<<<
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 */
  __pyx_t_2 = __pyx_v_self->strides;
  __pyx_v_buffer->strides = __pyx_t_2;

  /* "pyrobuf_list.pyx":964
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides
 *         buffer.suboffsets = NULL                # for pointer arrays only             # <<<<<<<<<<<<<<
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):
 */
  __pyx_v_buffer->suboffsets = NULL;

  /* "pyrobuf_list.pyx":948
 *         return str(list(self))
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t itemsize = sizeof(int32_t)
 */

  /* function exit code */
  __pyx_r = 0;
  if (__pyx_v_buffer->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_buffer->obj);
    __Pyx_DECREF(__pyx_v_buffer->obj); __pyx_v_buffer->obj = 0;
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":966
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_12pyrobuf_list_9Int32List_24__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer); /*proto*/
static CYTHON_UNUSED void __pyx_pw_12pyrobuf_list_9Int32List_24__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_12pyrobuf_list_9Int32List_23__releasebuffer__(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12pyrobuf_list_9Int32List_23__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "pyrobuf_list.pyx":969
 *         pass
 * 
 *     def __richcmp__(self, other, op):             # <<<<<<<<<<<<<<
 *         if op == Py_EQ:
 *             return self._equal(other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_26__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_26__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op) {
  PyObject *__pyx_v_op = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_v_op = __Pyx_PyInt_From_int(__pyx_arg_op); if (unlikely(!__pyx_v_op)) __PYX_ERR(0, 969, __pyx_L3_error)
  __Pyx_GOTREF(__pyx_v_op);
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int32List.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List_25__richcmp__(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((PyObject *)__pyx_v_op));

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_op);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_25__richcmp__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "pyrobuf_list.pyx":970
 * 
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:             # <<<<<<<<<<<<<<
 *             return self._equal(other)
 *         elif op == Py_NE:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(Py_EQ); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_op, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "pyrobuf_list.pyx":971
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:
 *             return self._equal(other)             # <<<<<<<<<<<<<<
 *         elif op == Py_NE:
 *             return not self._equal(other)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = ((struct __pyx_vtabstruct_12pyrobuf_list_Int32List *)__pyx_v_self->__pyx_vtab)->_equal(__pyx_v_self, __pyx_v_other, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 971, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pyrobuf_list.pyx":970
 * 
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:             # <<<<<<<<<<<<<<
 *             return self._equal(other)
 *         elif op == Py_NE:
 */
  }

  /* "pyrobuf_list.pyx":972
 *         if op == Py_EQ:
 *             return self._equal(other)
 *         elif op == Py_NE:             # <<<<<<<<<<<<<<
 *             return not self._equal(other)
 *         else:
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(Py_NE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_op, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 972, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 972, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(__pyx_t_3)) {

    /* "pyrobuf_list.pyx":973
 *             return self._equal(other)
 *         elif op == Py_NE:
 *             return not self._equal(other)             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError('Operation not supported for lists')
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((struct __pyx_vtabstruct_12pyrobuf_list_Int32List *)__pyx_v_self->__pyx_vtab)->_equal(__pyx_v_self, __pyx_v_other, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 973, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 973, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_3)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 973, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pyrobuf_list.pyx":972
 *         if op == Py_EQ:
 *             return self._equal(other)
 *         elif op == Py_NE:             # <<<<<<<<<<<<<<
 *             return not self._equal(other)
 *         else:
 */
  }

  /* "pyrobuf_list.pyx":975
 *             return not self._equal(other)
 *         else:
 *             raise TypeError('Operation not supported for lists')             # <<<<<<<<<<<<<<
 * 
 *     def __neq__(self, other):
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__45, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 975, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 975, __pyx_L1_error)
  }

  /* "pyrobuf_list.pyx":969
 *         pass
 * 
 *     def __richcmp__(self, other, op):             # <<<<<<<<<<<<<<
 *         if op == Py_EQ:
 *             return self._equal(other)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.Int32List.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":977
 *             raise TypeError('Operation not supported for lists')
 * 
 *     def __neq__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_28__neq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_28__neq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__neq__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List_27__neq__(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_27__neq__(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__neq__", 0);

  /* "pyrobuf_list.pyx":978
 * 
 *     def __neq__(self, other):
 *         return not self == other             # <<<<<<<<<<<<<<
 * 
 *     cpdef _equal(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_self), __pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 978, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 978, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 978, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":977
 *             raise TypeError('Operation not supported for lists')
 * 
 *     def __neq__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Int32List.__neq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":980
 *         return not self == other
 * 
 *     cpdef _equal(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             if len(self) != len(other):
 */

static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int32List__equal(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_v_y = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *(*__pyx_t_14)(PyObject *);
  __Pyx_RefNannySetupContext("_equal", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_equal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 980, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_9Int32List_30_equal)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_other); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 980, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_other};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 980, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_other};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 980, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 980, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_other);
          __Pyx_GIVEREF(__pyx_v_other);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_other);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 980, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":981
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    /*try:*/ {

      /* "pyrobuf_list.pyx":982
 *     cpdef _equal(self, other):
 *         try:
 *             if len(self) != len(other):             # <<<<<<<<<<<<<<
 *                 return False
 *         except TypeError:
 */
      __pyx_t_9 = PyObject_Length(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 982, __pyx_L3_error)
      __pyx_t_10 = PyObject_Length(__pyx_v_other); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 982, __pyx_L3_error)
      __pyx_t_11 = ((__pyx_t_9 != __pyx_t_10) != 0);
      if (__pyx_t_11) {

        /* "pyrobuf_list.pyx":983
 *         try:
 *             if len(self) != len(other):
 *                 return False             # <<<<<<<<<<<<<<
 *         except TypeError:
 *             return False
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_False);
        __pyx_r = Py_False;
        goto __pyx_L7_try_return;

        /* "pyrobuf_list.pyx":982
 *     cpdef _equal(self, other):
 *         try:
 *             if len(self) != len(other):             # <<<<<<<<<<<<<<
 *                 return False
 *         except TypeError:
 */
      }

      /* "pyrobuf_list.pyx":981
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
    }
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":984
 *             if len(self) != len(other):
 *                 return False
 *         except TypeError:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
    __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_12) {
      __Pyx_AddTraceback("pyrobuf_list.Int32List._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 984, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_3);

      /* "pyrobuf_list.pyx":985
 *                 return False
 *         except TypeError:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         for x, y in zip(self, other):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "pyrobuf_list.pyx":981
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "pyrobuf_list.pyx":987
 *             return False
 * 
 *         for x, y in zip(self, other):             # <<<<<<<<<<<<<<
 *             if x != y:
 *                 return False
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));
  __Pyx_INCREF(__pyx_v_other);
  __Pyx_GIVEREF(__pyx_v_other);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_other);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_10 = 0;
    __pyx_t_13 = NULL;
  } else {
    __pyx_t_10 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 987, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_13 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 987, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_13)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 987, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 987, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 987, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 987, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_13(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 987, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 987, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 987, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 987, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 987, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_14 = Py_TYPE(__pyx_t_4)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_5 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_5)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_4), 2) < 0) __PYX_ERR(0, 987, __pyx_L1_error)
      __pyx_t_14 = NULL;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L15_unpacking_done;
      __pyx_L14_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_14 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 987, __pyx_L1_error)
      __pyx_L15_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_x, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_y, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "pyrobuf_list.pyx":988
 * 
 *         for x, y in zip(self, other):
 *             if x != y:             # <<<<<<<<<<<<<<
 *                 return False
 * 
 */
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_x, __pyx_v_y, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 988, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 988, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_11) {

      /* "pyrobuf_list.pyx":989
 *         for x, y in zip(self, other):
 *             if x != y:
 *                 return False             # <<<<<<<<<<<<<<
 * 
 *         return True
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "pyrobuf_list.pyx":988
 * 
 *         for x, y in zip(self, other):
 *             if x != y:             # <<<<<<<<<<<<<<
 *                 return False
 * 
 */
    }

    /* "pyrobuf_list.pyx":987
 *             return False
 * 
 *         for x, y in zip(self, other):             # <<<<<<<<<<<<<<
 *             if x != y:
 *                 return False
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":991
 *                 return False
 * 
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     cpdef append(self, int32_t x):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":980
 *         return not self == other
 * 
 *     cpdef _equal(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             if len(self) != len(other):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.Int32List._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_y);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_equal (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List_29_equal(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_29_equal(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_equal", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_9Int32List__equal(__pyx_v_self, __pyx_v_other, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Int32List._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":993
 *         return True
 * 
 *     cpdef append(self, int32_t x):             # <<<<<<<<<<<<<<
 *         self._append(x)
 *         listener = <object>self._listener
 */

static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int32List_append(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int32_t __pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("append", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_append); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_9Int32List_32append)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_int32_t(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 993, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 993, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 993, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 993, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 993, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 993, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":994
 * 
 *     cpdef append(self, int32_t x):
 *         self._append(x)             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  ((struct __pyx_vtabstruct_12pyrobuf_list_Int32List *)__pyx_v_self->__pyx_vtab)->_append(__pyx_v_self, __pyx_v_x);

  /* "pyrobuf_list.pyx":995
 *     cpdef append(self, int32_t x):
 *         self._append(x)
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":996
 *         self._append(x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 996, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = ((!__pyx_t_7) != 0);
  if (__pyx_t_8) {

    /* "pyrobuf_list.pyx":997
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef extend(self, x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 997, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 997, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":996
 *         self._append(x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":993
 *         return True
 * 
 *     cpdef append(self, int32_t x):             # <<<<<<<<<<<<<<
 *         self._append(x)
 *         listener = <object>self._listener
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.Int32List.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  int32_t __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_int32_t(__pyx_arg_x); if (unlikely((__pyx_v_x == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 993, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int32List.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List_31append(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self), ((int32_t)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_31append(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int32_t __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("append", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_9Int32List_append(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Int32List.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":999
 *             listener._Modified()
 * 
 *     cpdef extend(self, x):             # <<<<<<<<<<<<<<
 *         for i in x:
 *             self._append(i)
 */

static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int32List_extend(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, PyObject *__pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int32_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("extend", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extend); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 999, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_9Int32List_34extend)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 999, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_x};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 999, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_x};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 999, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 999, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_x);
          __Pyx_GIVEREF(__pyx_v_x);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_x);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 999, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1000
 * 
 *     cpdef extend(self, x):
 *         for i in x:             # <<<<<<<<<<<<<<
 *             self._append(i)
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_x)) || PyTuple_CheckExact(__pyx_v_x)) {
    __pyx_t_1 = __pyx_v_x; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1000, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1000, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1000, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1000, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1000, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1000, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_7(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1000, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyrobuf_list.pyx":1001
 *     cpdef extend(self, x):
 *         for i in x:
 *             self._append(i)             # <<<<<<<<<<<<<<
 * 
 *         listener = <object>self._listener
 */
    __pyx_t_8 = __Pyx_PyInt_As_int32_t(__pyx_v_i); if (unlikely((__pyx_t_8 == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1001, __pyx_L1_error)
    ((struct __pyx_vtabstruct_12pyrobuf_list_Int32List *)__pyx_v_self->__pyx_vtab)->_append(__pyx_v_self, __pyx_t_8);

    /* "pyrobuf_list.pyx":1000
 * 
 *     cpdef extend(self, x):
 *         for i in x:             # <<<<<<<<<<<<<<
 *             self._append(i)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1003
 *             self._append(i)
 * 
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1004
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1004, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = ((!__pyx_t_9) != 0);
  if (__pyx_t_10) {

    /* "pyrobuf_list.pyx":1005
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef insert(self, int i, int32_t x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1005, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1005, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1005, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1004
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":999
 *             listener._Modified()
 * 
 *     cpdef extend(self, x):             # <<<<<<<<<<<<<<
 *         for i in x:
 *             self._append(i)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.Int32List.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extend (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List_33extend(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_33extend(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("extend", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_9Int32List_extend(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Int32List.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1007
 *             listener._Modified()
 * 
 *     cpdef insert(self, int i, int32_t x):             # <<<<<<<<<<<<<<
 *         cdef int32_t *mem
 *         cdef size_t j
 */

static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int32List_insert(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int __pyx_v_i, int32_t __pyx_v_x, int __pyx_skip_dispatch) {
  int32_t *__pyx_v_mem;
  size_t __pyx_v_j;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  size_t __pyx_t_11;
  size_t __pyx_t_12;
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("insert", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_insert); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1007, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_9Int32List_36insert)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1007, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyInt_From_int32_t(__pyx_v_x); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1007, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1007, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1007, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1007, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
        __pyx_t_3 = 0;
        __pyx_t_4 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1007, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1011
 *         cdef size_t j
 * 
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_9 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_9) {

    /* "pyrobuf_list.pyx":1012
 * 
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":1011
 *         cdef size_t j
 * 
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":1014
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i > self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_10 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_10) {
  } else {
    __pyx_t_9 = __pyx_t_10;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_10 = ((((size_t)__pyx_v_i) > __pyx_v_self->_n_items) != 0);
  __pyx_t_9 = __pyx_t_10;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_9)) {

    /* "pyrobuf_list.pyx":1015
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         if self._n_items == self._size:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__46, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1015, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1015, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1014
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i > self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1017
 *             raise IndexError("list index out of range")
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <int32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int32_t))
 *             if not mem:
 */
  __pyx_t_9 = ((__pyx_v_self->_n_items == __pyx_v_self->_size) != 0);
  if (__pyx_t_9) {

    /* "pyrobuf_list.pyx":1018
 * 
 *         if self._n_items == self._size:
 *             mem = <int32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int32_t))             # <<<<<<<<<<<<<<
 *             if not mem:
 *                 raise MemoryError()
 */
    __pyx_v_mem = ((int32_t *)PyMem_Realloc(__pyx_v_self->_data, ((2 * __pyx_v_self->_size) * (sizeof(int32_t)))));

    /* "pyrobuf_list.pyx":1019
 *         if self._n_items == self._size:
 *             mem = <int32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int32_t))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    __pyx_t_9 = ((!(__pyx_v_mem != 0)) != 0);
    if (unlikely(__pyx_t_9)) {

      /* "pyrobuf_list.pyx":1020
 *             mem = <int32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int32_t))
 *             if not mem:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *             self._data = mem
 */
      PyErr_NoMemory(); __PYX_ERR(0, 1020, __pyx_L1_error)

      /* "pyrobuf_list.pyx":1019
 *         if self._n_items == self._size:
 *             mem = <int32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int32_t))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    }

    /* "pyrobuf_list.pyx":1022
 *                 raise MemoryError()
 * 
 *             self._data = mem             # <<<<<<<<<<<<<<
 *             self._size *= 2
 * 
 */
    __pyx_v_self->_data = __pyx_v_mem;

    /* "pyrobuf_list.pyx":1023
 * 
 *             self._data = mem
 *             self._size *= 2             # <<<<<<<<<<<<<<
 * 
 *         for j in range(self._n_items, <size_t>i, -1):
 */
    __pyx_v_self->_size = (__pyx_v_self->_size * 2);

    /* "pyrobuf_list.pyx":1017
 *             raise IndexError("list index out of range")
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <int32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int32_t))
 *             if not mem:
 */
  }

  /* "pyrobuf_list.pyx":1025
 *             self._size *= 2
 * 
 *         for j in range(self._n_items, <size_t>i, -1):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j - 1]
 * 
 */
  __pyx_t_11 = ((size_t)__pyx_v_i);
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = __pyx_v_self->_n_items + 1; __pyx_t_13 > __pyx_t_12 + 1; ) { __pyx_t_13-=1;
    __pyx_v_j = __pyx_t_13;

    /* "pyrobuf_list.pyx":1026
 * 
 *         for j in range(self._n_items, <size_t>i, -1):
 *             self._data[j] = self._data[j - 1]             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j - 1)]);
  }

  /* "pyrobuf_list.pyx":1028
 *             self._data[j] = self._data[j - 1]
 * 
 *         self._data[i] = x             # <<<<<<<<<<<<<<
 *         self._n_items += 1
 *         listener = <object>self._listener
 */
  (__pyx_v_self->_data[__pyx_v_i]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":1029
 * 
 *         self._data[i] = x
 *         self._n_items += 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items + 1);

  /* "pyrobuf_list.pyx":1030
 *         self._data[i] = x
 *         self._n_items += 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1031
 *         self._n_items += 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1031, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = ((!__pyx_t_9) != 0);
  if (__pyx_t_10) {

    /* "pyrobuf_list.pyx":1032
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef pop(self):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1032, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1032, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1032, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1031
 *         self._n_items += 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1007
 *             listener._Modified()
 * 
 *     cpdef insert(self, int i, int32_t x):             # <<<<<<<<<<<<<<
 *         cdef int32_t *mem
 *         cdef size_t j
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyrobuf_list.Int32List.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_i;
  int32_t __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("insert (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_i,&__pyx_n_s_x,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, 1); __PYX_ERR(0, 1007, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "insert") < 0)) __PYX_ERR(0, 1007, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_i = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1007, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_int32_t(values[1]); if (unlikely((__pyx_v_x == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1007, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1007, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int32List.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List_35insert(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self), __pyx_v_i, __pyx_v_x);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_35insert(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int __pyx_v_i, int32_t __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("insert", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_9Int32List_insert(__pyx_v_self, __pyx_v_i, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Int32List.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1034
 *             listener._Modified()
 * 
 *     cpdef pop(self):             # <<<<<<<<<<<<<<
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")
 */

static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int32List_pop(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("pop", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_9Int32List_38pop)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1034, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1034, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1035
 * 
 *     cpdef pop(self):
 *         if self._n_items == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("pop from empty list")
 * 
 */
  __pyx_t_5 = ((__pyx_v_self->_n_items == 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "pyrobuf_list.pyx":1036
 *     cpdef pop(self):
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__47, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1036, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1036, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1035
 * 
 *     cpdef pop(self):
 *         if self._n_items == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("pop from empty list")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1038
 *             raise IndexError("pop from empty list")
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":1039
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1040
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 *         return self._data[self._n_items]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1040, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1040, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = ((!__pyx_t_5) != 0);
  if (__pyx_t_6) {

    /* "pyrobuf_list.pyx":1041
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 *         return self._data[self._n_items]
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1041, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1041, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1040
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 *         return self._data[self._n_items]
 */
  }

  /* "pyrobuf_list.pyx":1042
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 *         return self._data[self._n_items]             # <<<<<<<<<<<<<<
 * 
 *     cpdef remove(self, int32_t x):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int32_t((__pyx_v_self->_data[__pyx_v_self->_n_items])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1042, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1034
 *             listener._Modified()
 * 
 *     cpdef pop(self):             # <<<<<<<<<<<<<<
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyrobuf_list.Int32List.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pop (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List_37pop(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_37pop(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("pop", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_9Int32List_pop(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Int32List.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1044
 *         return self._data[self._n_items]
 * 
 *     cpdef remove(self, int32_t x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef size_t j
 */

static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int32List_remove(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int32_t __pyx_v_x, int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  int __pyx_v_found;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("remove", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_remove); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_9Int32List_40remove)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_int32_t(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1044, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1044, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1044, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1044, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1044, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1044, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1047
 *         cdef size_t i
 *         cdef size_t j
 *         cdef bint found = 0             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self._n_items):
 */
  __pyx_v_found = 0;

  /* "pyrobuf_list.pyx":1049
 *         cdef bint found = 0
 * 
 *         for i in range(self._n_items):             # <<<<<<<<<<<<<<
 *             if self._data[i] == x:
 *                 found = 1
 */
  __pyx_t_7 = __pyx_v_self->_n_items;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "pyrobuf_list.pyx":1050
 * 
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 found = 1
 *                 break
 */
    __pyx_t_10 = (((__pyx_v_self->_data[__pyx_v_i]) == __pyx_v_x) != 0);
    if (__pyx_t_10) {

      /* "pyrobuf_list.pyx":1051
 *         for i in range(self._n_items):
 *             if self._data[i] == x:
 *                 found = 1             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
      __pyx_v_found = 1;

      /* "pyrobuf_list.pyx":1052
 *             if self._data[i] == x:
 *                 found = 1
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         if found == 0:
 */
      goto __pyx_L4_break;

      /* "pyrobuf_list.pyx":1050
 * 
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 found = 1
 *                 break
 */
    }
  }
  __pyx_L4_break:;

  /* "pyrobuf_list.pyx":1054
 *                 break
 * 
 *         if found == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("x not in list")
 * 
 */
  __pyx_t_10 = ((__pyx_v_found == 0) != 0);
  if (unlikely(__pyx_t_10)) {

    /* "pyrobuf_list.pyx":1055
 * 
 *         if found == 0:
 *             raise ValueError("x not in list")             # <<<<<<<<<<<<<<
 * 
 *         for j in range(i, self._n_items):
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__48, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1055, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1055, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1054
 *                 break
 * 
 *         if found == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("x not in list")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1057
 *             raise ValueError("x not in list")
 * 
 *         for j in range(i, self._n_items):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j + 1]
 * 
 */
  __pyx_t_7 = __pyx_v_self->_n_items;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = __pyx_v_i; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_j = __pyx_t_9;

    /* "pyrobuf_list.pyx":1058
 * 
 *         for j in range(i, self._n_items):
 *             self._data[j] = self._data[j + 1]             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j + 1)]);
  }

  /* "pyrobuf_list.pyx":1060
 *             self._data[j] = self._data[j + 1]
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":1061
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1062
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1062, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = ((!__pyx_t_10) != 0);
  if (__pyx_t_11) {

    /* "pyrobuf_list.pyx":1063
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cdef void _append(self, int32_t x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1063, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1063, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1063, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1062
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1044
 *         return self._data[self._n_items]
 * 
 *     cpdef remove(self, int32_t x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef size_t j
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.Int32List.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  int32_t __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_int32_t(__pyx_arg_x); if (unlikely((__pyx_v_x == ((int32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1044, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int32List.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List_39remove(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self), ((int32_t)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_39remove(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int32_t __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("remove", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_9Int32List_remove(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Int32List.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1065
 *             listener._Modified()
 * 
 *     cdef void _append(self, int32_t x):             # <<<<<<<<<<<<<<
 *         cdef int32_t *mem
 * 
 */

static void __pyx_f_12pyrobuf_list_9Int32List__append(struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, int32_t __pyx_v_x) {
  int32_t *__pyx_v_mem;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_append", 0);

  /* "pyrobuf_list.pyx":1068
 *         cdef int32_t *mem
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <int32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int32_t))
 *             if not mem:
 */
  __pyx_t_1 = ((__pyx_v_self->_n_items == __pyx_v_self->_size) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":1069
 * 
 *         if self._n_items == self._size:
 *             mem = <int32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int32_t))             # <<<<<<<<<<<<<<
 *             if not mem:
 *                 raise MemoryError()
 */
    __pyx_v_mem = ((int32_t *)PyMem_Realloc(__pyx_v_self->_data, ((2 * __pyx_v_self->_size) * (sizeof(int32_t)))));

    /* "pyrobuf_list.pyx":1070
 *         if self._n_items == self._size:
 *             mem = <int32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int32_t))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    __pyx_t_1 = ((!(__pyx_v_mem != 0)) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "pyrobuf_list.pyx":1071
 *             mem = <int32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int32_t))
 *             if not mem:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *             self._data = mem
 */
      PyErr_NoMemory(); __PYX_ERR(0, 1071, __pyx_L1_error)

      /* "pyrobuf_list.pyx":1070
 *         if self._n_items == self._size:
 *             mem = <int32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int32_t))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    }

    /* "pyrobuf_list.pyx":1073
 *                 raise MemoryError()
 * 
 *             self._data = mem             # <<<<<<<<<<<<<<
 *             self._size *= 2
 * 
 */
    __pyx_v_self->_data = __pyx_v_mem;

    /* "pyrobuf_list.pyx":1074
 * 
 *             self._data = mem
 *             self._size *= 2             # <<<<<<<<<<<<<<
 * 
 *         self._data[self._n_items] = x
 */
    __pyx_v_self->_size = (__pyx_v_self->_size * 2);

    /* "pyrobuf_list.pyx":1068
 *         cdef int32_t *mem
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <int32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int32_t))
 *             if not mem:
 */
  }

  /* "pyrobuf_list.pyx":1076
 *             self._size *= 2
 * 
 *         self._data[self._n_items] = x             # <<<<<<<<<<<<<<
 *         self._n_items += 1
 * 
 */
  (__pyx_v_self->_data[__pyx_v_self->_n_items]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":1077
 * 
 *         self._data[self._n_items] = x
 *         self._n_items += 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items + 1);

  /* "pyrobuf_list.pyx":1065
 *             listener._Modified()
 * 
 *     cdef void _append(self, int32_t x):             # <<<<<<<<<<<<<<
 *         cdef int32_t *mem
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("pyrobuf_list.Int32List._append", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_42__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_42__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List_41__reduce_cython__(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_41__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__49, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Int32List.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_44__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int32List_44__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int32List_43__setstate_cython__(((struct __pyx_obj_12pyrobuf_list_Int32List *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int32List_43__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Int32List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__50, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Int32List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1083
 * cdef class Uint32List:
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):             # <<<<<<<<<<<<<<
 *         self._data = <uint32_t *>PyMem_Malloc(size * sizeof(uint32_t))
 *         if not self._data:
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_10Uint32List_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12pyrobuf_list_10Uint32List_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_size;
  PyObject *__pyx_v_listener = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_size,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_listener);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 1083, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_size = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_size == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1083, __pyx_L3_error)
    } else {
      __pyx_v_size = ((size_t)16);
    }
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1083, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List___cinit__(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self), __pyx_v_size, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_10Uint32List___cinit__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, size_t __pyx_v_size, PyObject *__pyx_v_listener) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "pyrobuf_list.pyx":1084
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <uint32_t *>PyMem_Malloc(size * sizeof(uint32_t))             # <<<<<<<<<<<<<<
 *         if not self._data:
 *             raise MemoryError()
 */
  __pyx_v_self->_data = ((uint32_t *)PyMem_Malloc((__pyx_v_size * (sizeof(uint32_t)))));

  /* "pyrobuf_list.pyx":1085
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <uint32_t *>PyMem_Malloc(size * sizeof(uint32_t))
 *         if not self._data:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_self->_data != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":1086
 *         self._data = <uint32_t *>PyMem_Malloc(size * sizeof(uint32_t))
 *         if not self._data:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         self._n_items = 0
 */
    PyErr_NoMemory(); __PYX_ERR(0, 1086, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1085
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <uint32_t *>PyMem_Malloc(size * sizeof(uint32_t))
 *         if not self._data:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1088
 *             raise MemoryError()
 * 
 *         self._n_items = 0             # <<<<<<<<<<<<<<
 *         self._size = size
 *         if listener is None:
 */
  __pyx_v_self->_n_items = 0;

  /* "pyrobuf_list.pyx":1089
 * 
 *         self._n_items = 0
 *         self._size = size             # <<<<<<<<<<<<<<
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener
 */
  __pyx_v_self->_size = __pyx_v_size;

  /* "pyrobuf_list.pyx":1090
 *         self._n_items = 0
 *         self._size = size
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
  __pyx_t_1 = (__pyx_v_listener == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":1091
 *         self._size = size
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener             # <<<<<<<<<<<<<<
 *         else:
 *             self._listener = <PyObject *>listener
 */
    __pyx_v_self->_listener = ((PyObject *)null_listener);

    /* "pyrobuf_list.pyx":1090
 *         self._n_items = 0
 *         self._size = size
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
    goto __pyx_L4;
  }

  /* "pyrobuf_list.pyx":1093
 *             self._listener = <PyObject *>null_listener
 *         else:
 *             self._listener = <PyObject *>listener             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  /*else*/ {
    __pyx_v_self->_listener = ((PyObject *)__pyx_v_listener);
  }
  __pyx_L4:;

  /* "pyrobuf_list.pyx":1083
 * cdef class Uint32List:
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):             # <<<<<<<<<<<<<<
 *         self._data = <uint32_t *>PyMem_Malloc(size * sizeof(uint32_t))
 *         if not self._data:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1095
 *             self._listener = <PyObject *>listener
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self._data)
 * 
 */

/* Python wrapper */
static void __pyx_pw_12pyrobuf_list_10Uint32List_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12pyrobuf_list_10Uint32List_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_12pyrobuf_list_10Uint32List_2__dealloc__(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12pyrobuf_list_10Uint32List_2__dealloc__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "pyrobuf_list.pyx":1096
 * 
 *     def __dealloc__(self):
 *         PyMem_Free(self._data)             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, uint32_t x):
 */
  PyMem_Free(__pyx_v_self->_data);

  /* "pyrobuf_list.pyx":1095
 *             self._listener = <PyObject *>listener
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self._data)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "pyrobuf_list.pyx":1098
 *         PyMem_Free(self._data)
 * 
 *     def __contains__(self, uint32_t x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self._n_items):
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_10Uint32List_5__contains__(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static int __pyx_pw_12pyrobuf_list_10Uint32List_5__contains__(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  uint32_t __pyx_v_x;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_uint32_t(__pyx_arg_x); if (unlikely((__pyx_v_x == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1098, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List_4__contains__(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self), ((uint32_t)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_10Uint32List_4__contains__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, uint32_t __pyx_v_x) {
  size_t __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "pyrobuf_list.pyx":1100
 *     def __contains__(self, uint32_t x):
 *         cdef size_t i
 *         for i in range(self._n_items):             # <<<<<<<<<<<<<<
 *             if self._data[i] == x:
 *                 return True
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "pyrobuf_list.pyx":1101
 *         cdef size_t i
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */
    __pyx_t_4 = (((__pyx_v_self->_data[__pyx_v_i]) == __pyx_v_x) != 0);
    if (__pyx_t_4) {

      /* "pyrobuf_list.pyx":1102
 *         for i in range(self._n_items):
 *             if self._data[i] == x:
 *                 return True             # <<<<<<<<<<<<<<
 * 
 *         return False
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "pyrobuf_list.pyx":1101
 *         cdef size_t i
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */
    }
  }

  /* "pyrobuf_list.pyx":1104
 *                 return True
 * 
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     def __delitem__(self, int i):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1098
 *         PyMem_Free(self._data)
 * 
 *     def __contains__(self, uint32_t x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self._n_items):
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1106
 *         return False
 * 
 *     def __delitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_10Uint32List_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static int __pyx_pw_12pyrobuf_list_10Uint32List_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1106, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List_6__delitem__(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self), ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_10Uint32List_6__delitem__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, int __pyx_v_i) {
  size_t __pyx_v_j;
  PyObject *__pyx_v_listener = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__delitem__", 0);

  /* "pyrobuf_list.pyx":1107
 * 
 *     def __delitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":1108
 *     def __delitem__(self, int i):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":1107
 * 
 *     def __delitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":1110
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":1111
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t j
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__51, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1111, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1110
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1114
 * 
 *         cdef size_t j
 *         for j in range(i, self._n_items):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j + 1]
 * 
 */
  __pyx_t_4 = __pyx_v_self->_n_items;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = __pyx_v_i; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_j = __pyx_t_6;

    /* "pyrobuf_list.pyx":1115
 *         cdef size_t j
 *         for j in range(i, self._n_items):
 *             self._data[j] = self._data[j + 1]             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j + 1)]);
  }

  /* "pyrobuf_list.pyx":1117
 *             self._data[j] = self._data[j + 1]
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":1118
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":1119
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":1120
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, int i):
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    if (__pyx_t_8) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1120, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1120, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":1119
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1106
 *         return False
 * 
 *     def __delitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1122
 *             listener._Modified()
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1122, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List_8__getitem__(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self), ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_8__getitem__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, int __pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pyrobuf_list.pyx":1123
 * 
 *     def __getitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":1124
 *     def __getitem__(self, int i):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":1123
 * 
 *     def __getitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":1126
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":1127
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         return self._data[i]
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__52, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1127, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1126
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1129
 *             raise IndexError("list index out of range")
 * 
 *         return self._data[i]             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_uint32_t((__pyx_v_self->_data[__pyx_v_i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1122
 *             listener._Modified()
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_12pyrobuf_list_10Uint32List_12generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "pyrobuf_list.pyx":1131
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_11__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_11__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List_10__iter__(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_10__iter__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self) {
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_4___iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_4___iter__ *)__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_4___iter__(__pyx_ptype_12pyrobuf_list___pyx_scope_struct_4___iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_4___iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1131, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_12pyrobuf_list_10Uint32List_12generator4, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_Uint32List___iter, __pyx_n_s_pyrobuf_list); if (unlikely(!gen)) __PYX_ERR(0, 1131, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_12pyrobuf_list_10Uint32List_12generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_4___iter__ *__pyx_cur_scope = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_4___iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1131, __pyx_L1_error)

  /* "pyrobuf_list.pyx":1132
 * 
 *     def __iter__(self):
 *         cdef size_t idx = 0             # <<<<<<<<<<<<<<
 *         while idx < self._n_items:
 *             yield self._data[idx]
 */
  __pyx_cur_scope->__pyx_v_idx = 0;

  /* "pyrobuf_list.pyx":1133
 *     def __iter__(self):
 *         cdef size_t idx = 0
 *         while idx < self._n_items:             # <<<<<<<<<<<<<<
 *             yield self._data[idx]
 *             idx += 1
 */
  while (1) {
    __pyx_t_1 = ((__pyx_cur_scope->__pyx_v_idx < __pyx_cur_scope->__pyx_v_self->_n_items) != 0);
    if (!__pyx_t_1) break;

    /* "pyrobuf_list.pyx":1134
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 *             yield self._data[idx]             # <<<<<<<<<<<<<<
 *             idx += 1
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_uint32_t((__pyx_cur_scope->__pyx_v_self->_data[__pyx_cur_scope->__pyx_v_idx])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1134, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1135
 *         while idx < self._n_items:
 *             yield self._data[idx]
 *             idx += 1             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
    __pyx_cur_scope->__pyx_v_idx = (__pyx_cur_scope->__pyx_v_idx + 1);
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "pyrobuf_list.pyx":1131
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1137
 *             idx += 1
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._n_items
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_12pyrobuf_list_10Uint32List_14__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_12pyrobuf_list_10Uint32List_14__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List_13__len__(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_12pyrobuf_list_10Uint32List_13__len__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pyrobuf_list.pyx":1138
 * 
 *     def __len__(self):
 *         return self._n_items             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = __pyx_v_self->_n_items;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1137
 *             idx += 1
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._n_items
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1140
 *         return self._n_items
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(list(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_16__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_16__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List_15__repr__(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_15__repr__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pyrobuf_list.pyx":1141
 * 
 *     def __repr__(self):
 *         return repr(list(self))             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, int i, uint32_t x):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1140
 *         return self._n_items
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(list(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1143
 *         return repr(list(self))
 * 
 *     def __setitem__(self, int i, uint32_t x):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_10Uint32List_18__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i, PyObject *__pyx_arg_x); /*proto*/
static int __pyx_pw_12pyrobuf_list_10Uint32List_18__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i, PyObject *__pyx_arg_x) {
  int __pyx_v_i;
  uint32_t __pyx_v_x;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1143, __pyx_L3_error)
  }
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_uint32_t(__pyx_arg_x); if (unlikely((__pyx_v_x == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1143, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List_17__setitem__(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self), ((int)__pyx_v_i), ((uint32_t)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_10Uint32List_17__setitem__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, int __pyx_v_i, uint32_t __pyx_v_x) {
  PyObject *__pyx_v_listener = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "pyrobuf_list.pyx":1144
 * 
 *     def __setitem__(self, int i, uint32_t x):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":1145
 *     def __setitem__(self, int i, uint32_t x):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":1144
 * 
 *     def __setitem__(self, int i, uint32_t x):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":1147
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":1148
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__53, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1148, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1147
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1150
 *             raise IndexError("list index out of range")
 * 
 *         self._data[i] = x             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  (__pyx_v_self->_data[__pyx_v_i]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":1151
 * 
 *         self._data[i] = x
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":1152
 *         self._data[i] = x
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":1153
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1153, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1153, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":1152
 *         self._data[i] = x
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1143
 *         return repr(list(self))
 * 
 *     def __setitem__(self, int i, uint32_t x):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1155
 *             listener._Modified()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(list(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_20__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_20__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List_19__str__(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_19__str__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pyrobuf_list.pyx":1156
 * 
 *     def __str__(self):
 *         return str(list(self))             # <<<<<<<<<<<<<<
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1155
 *             listener._Modified()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(list(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1158
 *         return str(list(self))
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t itemsize = sizeof(uint32_t)
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_12pyrobuf_list_10Uint32List_22__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_12pyrobuf_list_10Uint32List_22__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List_21__getbuffer__(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_10Uint32List_21__getbuffer__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags) {
  Py_ssize_t __pyx_v_itemsize;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  if (__pyx_v_buffer == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_buffer->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_buffer->obj);

  /* "pyrobuf_list.pyx":1160
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 * 
 *         cdef Py_ssize_t itemsize = sizeof(uint32_t)             # <<<<<<<<<<<<<<
 *         self.shape[0] = self._n_items
 *         self.strides[0] = 1
 */
  __pyx_v_itemsize = (sizeof(uint32_t));

  /* "pyrobuf_list.pyx":1161
 * 
 *         cdef Py_ssize_t itemsize = sizeof(uint32_t)
 *         self.shape[0] = self._n_items             # <<<<<<<<<<<<<<
 *         self.strides[0] = 1
 * 
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  (__pyx_v_self->shape[0]) = __pyx_t_1;

  /* "pyrobuf_list.pyx":1162
 *         cdef Py_ssize_t itemsize = sizeof(uint32_t)
 *         self.shape[0] = self._n_items
 *         self.strides[0] = 1             # <<<<<<<<<<<<<<
 * 
 *         buffer.buf = <char *>(self._data)
 */
  (__pyx_v_self->strides[0]) = 1;

  /* "pyrobuf_list.pyx":1164
 *         self.strides[0] = 1
 * 
 *         buffer.buf = <char *>(self._data)             # <<<<<<<<<<<<<<
 *         buffer.format = 'I'
 *         buffer.internal = NULL                  # see References
 */
  __pyx_v_buffer->buf = ((char *)__pyx_v_self->_data);

  /* "pyrobuf_list.pyx":1165
 * 
 *         buffer.buf = <char *>(self._data)
 *         buffer.format = 'I'             # <<<<<<<<<<<<<<
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 */
  __pyx_v_buffer->format = ((char *)"I");

  /* "pyrobuf_list.pyx":1166
 *         buffer.buf = <char *>(self._data)
 *         buffer.format = 'I'
 *         buffer.internal = NULL                  # see References             # <<<<<<<<<<<<<<
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items
 */
  __pyx_v_buffer->internal = NULL;

  /* "pyrobuf_list.pyx":1167
 *         buffer.format = 'I'
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize             # <<<<<<<<<<<<<<
 *         buffer.len = self._n_items
 *         buffer.ndim = 1
 */
  __pyx_v_buffer->itemsize = __pyx_v_itemsize;

  /* "pyrobuf_list.pyx":1168
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items             # <<<<<<<<<<<<<<
 *         buffer.ndim = 1
 *         buffer.obj = self
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  __pyx_v_buffer->len = __pyx_t_1;

  /* "pyrobuf_list.pyx":1169
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items
 *         buffer.ndim = 1             # <<<<<<<<<<<<<<
 *         buffer.obj = self
 *         buffer.readonly = 1
 */
  __pyx_v_buffer->ndim = 1;

  /* "pyrobuf_list.pyx":1170
 *         buffer.len = self._n_items
 *         buffer.ndim = 1
 *         buffer.obj = self             # <<<<<<<<<<<<<<
 *         buffer.readonly = 1
 *         buffer.shape = self.shape
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_buffer->obj);
  __Pyx_DECREF(__pyx_v_buffer->obj);
  __pyx_v_buffer->obj = ((PyObject *)__pyx_v_self);

  /* "pyrobuf_list.pyx":1171
 *         buffer.ndim = 1
 *         buffer.obj = self
 *         buffer.readonly = 1             # <<<<<<<<<<<<<<
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides
 */
  __pyx_v_buffer->readonly = 1;

  /* "pyrobuf_list.pyx":1172
 *         buffer.obj = self
 *         buffer.readonly = 1
 *         buffer.shape = self.shape             # <<<<<<<<<<<<<<
 *         buffer.strides = self.strides
 *         buffer.suboffsets = NULL                # for pointer arrays only
 */
  __pyx_t_2 = __pyx_v_self->shape;
  __pyx_v_buffer->shape = __pyx_t_2;

  /* "pyrobuf_list.pyx":1173
 *         buffer.readonly = 1
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides             # <<<<<<<<<<<<<<
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 */
  __pyx_t_2 = __pyx_v_self->strides;
  __pyx_v_buffer->strides = __pyx_t_2;

  /* "pyrobuf_list.pyx":1174
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides
 *         buffer.suboffsets = NULL                # for pointer arrays only             # <<<<<<<<<<<<<<
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):
 */
  __pyx_v_buffer->suboffsets = NULL;

  /* "pyrobuf_list.pyx":1158
 *         return str(list(self))
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t itemsize = sizeof(uint32_t)
 */

  /* function exit code */
  __pyx_r = 0;
  if (__pyx_v_buffer->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_buffer->obj);
    __Pyx_DECREF(__pyx_v_buffer->obj); __pyx_v_buffer->obj = 0;
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1176
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_12pyrobuf_list_10Uint32List_24__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer); /*proto*/
static CYTHON_UNUSED void __pyx_pw_12pyrobuf_list_10Uint32List_24__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_12pyrobuf_list_10Uint32List_23__releasebuffer__(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12pyrobuf_list_10Uint32List_23__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "pyrobuf_list.pyx":1179
 *         pass
 * 
 *     def __richcmp__(self, other, op):             # <<<<<<<<<<<<<<
 *         if op == Py_EQ:
 *             return self._equal(other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_26__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_26__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op) {
  PyObject *__pyx_v_op = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_v_op = __Pyx_PyInt_From_int(__pyx_arg_op); if (unlikely(!__pyx_v_op)) __PYX_ERR(0, 1179, __pyx_L3_error)
  __Pyx_GOTREF(__pyx_v_op);
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List_25__richcmp__(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((PyObject *)__pyx_v_op));

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_op);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_25__richcmp__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "pyrobuf_list.pyx":1180
 * 
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:             # <<<<<<<<<<<<<<
 *             return self._equal(other)
 *         elif op == Py_NE:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(Py_EQ); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_op, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1180, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1180, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "pyrobuf_list.pyx":1181
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:
 *             return self._equal(other)             # <<<<<<<<<<<<<<
 *         elif op == Py_NE:
 *             return not self._equal(other)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = ((struct __pyx_vtabstruct_12pyrobuf_list_Uint32List *)__pyx_v_self->__pyx_vtab)->_equal(__pyx_v_self, __pyx_v_other, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pyrobuf_list.pyx":1180
 * 
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:             # <<<<<<<<<<<<<<
 *             return self._equal(other)
 *         elif op == Py_NE:
 */
  }

  /* "pyrobuf_list.pyx":1182
 *         if op == Py_EQ:
 *             return self._equal(other)
 *         elif op == Py_NE:             # <<<<<<<<<<<<<<
 *             return not self._equal(other)
 *         else:
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(Py_NE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_op, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(__pyx_t_3)) {

    /* "pyrobuf_list.pyx":1183
 *             return self._equal(other)
 *         elif op == Py_NE:
 *             return not self._equal(other)             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError('Operation not supported for lists')
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((struct __pyx_vtabstruct_12pyrobuf_list_Uint32List *)__pyx_v_self->__pyx_vtab)->_equal(__pyx_v_self, __pyx_v_other, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1183, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_3)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pyrobuf_list.pyx":1182
 *         if op == Py_EQ:
 *             return self._equal(other)
 *         elif op == Py_NE:             # <<<<<<<<<<<<<<
 *             return not self._equal(other)
 *         else:
 */
  }

  /* "pyrobuf_list.pyx":1185
 *             return not self._equal(other)
 *         else:
 *             raise TypeError('Operation not supported for lists')             # <<<<<<<<<<<<<<
 * 
 *     def __neq__(self, other):
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__54, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1185, __pyx_L1_error)
  }

  /* "pyrobuf_list.pyx":1179
 *         pass
 * 
 *     def __richcmp__(self, other, op):             # <<<<<<<<<<<<<<
 *         if op == Py_EQ:
 *             return self._equal(other)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1187
 *             raise TypeError('Operation not supported for lists')
 * 
 *     def __neq__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_28__neq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_28__neq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__neq__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List_27__neq__(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_27__neq__(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__neq__", 0);

  /* "pyrobuf_list.pyx":1188
 * 
 *     def __neq__(self, other):
 *         return not self == other             # <<<<<<<<<<<<<<
 * 
 *     cpdef _equal(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_self), __pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1188, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1188, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1187
 *             raise TypeError('Operation not supported for lists')
 * 
 *     def __neq__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.__neq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1190
 *         return not self == other
 * 
 *     cpdef _equal(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             if len(self) != len(other):
 */

static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint32List__equal(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_v_y = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *(*__pyx_t_14)(PyObject *);
  __Pyx_RefNannySetupContext("_equal", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_equal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint32List_30_equal)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_other); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1190, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_other};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1190, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_other};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1190, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1190, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_other);
          __Pyx_GIVEREF(__pyx_v_other);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_other);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1190, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1191
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    /*try:*/ {

      /* "pyrobuf_list.pyx":1192
 *     cpdef _equal(self, other):
 *         try:
 *             if len(self) != len(other):             # <<<<<<<<<<<<<<
 *                 return False
 *         except TypeError:
 */
      __pyx_t_9 = PyObject_Length(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1192, __pyx_L3_error)
      __pyx_t_10 = PyObject_Length(__pyx_v_other); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1192, __pyx_L3_error)
      __pyx_t_11 = ((__pyx_t_9 != __pyx_t_10) != 0);
      if (__pyx_t_11) {

        /* "pyrobuf_list.pyx":1193
 *         try:
 *             if len(self) != len(other):
 *                 return False             # <<<<<<<<<<<<<<
 *         except TypeError:
 *             return False
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_False);
        __pyx_r = Py_False;
        goto __pyx_L7_try_return;

        /* "pyrobuf_list.pyx":1192
 *     cpdef _equal(self, other):
 *         try:
 *             if len(self) != len(other):             # <<<<<<<<<<<<<<
 *                 return False
 *         except TypeError:
 */
      }

      /* "pyrobuf_list.pyx":1191
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
    }
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1194
 *             if len(self) != len(other):
 *                 return False
 *         except TypeError:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
    __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_12) {
      __Pyx_AddTraceback("pyrobuf_list.Uint32List._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 1194, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_3);

      /* "pyrobuf_list.pyx":1195
 *                 return False
 *         except TypeError:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         for x, y in zip(self, other):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "pyrobuf_list.pyx":1191
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "pyrobuf_list.pyx":1197
 *             return False
 * 
 *         for x, y in zip(self, other):             # <<<<<<<<<<<<<<
 *             if x != y:
 *                 return False
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));
  __Pyx_INCREF(__pyx_v_other);
  __Pyx_GIVEREF(__pyx_v_other);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_other);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_10 = 0;
    __pyx_t_13 = NULL;
  } else {
    __pyx_t_10 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_13 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1197, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_13)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 1197, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1197, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 1197, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1197, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_13(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1197, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1197, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1197, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1197, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1197, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_14 = Py_TYPE(__pyx_t_4)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_5 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_5)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_4), 2) < 0) __PYX_ERR(0, 1197, __pyx_L1_error)
      __pyx_t_14 = NULL;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L15_unpacking_done;
      __pyx_L14_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_14 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1197, __pyx_L1_error)
      __pyx_L15_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_x, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_y, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "pyrobuf_list.pyx":1198
 * 
 *         for x, y in zip(self, other):
 *             if x != y:             # <<<<<<<<<<<<<<
 *                 return False
 * 
 */
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_x, __pyx_v_y, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1198, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1198, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_11) {

      /* "pyrobuf_list.pyx":1199
 *         for x, y in zip(self, other):
 *             if x != y:
 *                 return False             # <<<<<<<<<<<<<<
 * 
 *         return True
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "pyrobuf_list.pyx":1198
 * 
 *         for x, y in zip(self, other):
 *             if x != y:             # <<<<<<<<<<<<<<
 *                 return False
 * 
 */
    }

    /* "pyrobuf_list.pyx":1197
 *             return False
 * 
 *         for x, y in zip(self, other):             # <<<<<<<<<<<<<<
 *             if x != y:
 *                 return False
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":1201
 *                 return False
 * 
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     cpdef append(self, uint32_t x):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1190
 *         return not self == other
 * 
 *     cpdef _equal(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             if len(self) != len(other):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_y);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_equal (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List_29_equal(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_29_equal(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_equal", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_10Uint32List__equal(__pyx_v_self, __pyx_v_other, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1203
 *         return True
 * 
 *     cpdef append(self, uint32_t x):             # <<<<<<<<<<<<<<
 *         self._append(x)
 *         listener = <object>self._listener
 */

static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint32List_append(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, uint32_t __pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("append", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_append); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint32List_32append)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_uint32_t(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1203, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1203, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1203, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1203, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1203, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1203, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1204
 * 
 *     cpdef append(self, uint32_t x):
 *         self._append(x)             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  ((struct __pyx_vtabstruct_12pyrobuf_list_Uint32List *)__pyx_v_self->__pyx_vtab)->_append(__pyx_v_self, __pyx_v_x);

  /* "pyrobuf_list.pyx":1205
 *     cpdef append(self, uint32_t x):
 *         self._append(x)
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1206
 *         self._append(x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = ((!__pyx_t_7) != 0);
  if (__pyx_t_8) {

    /* "pyrobuf_list.pyx":1207
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef extend(self, x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1207, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1207, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1206
 *         self._append(x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1203
 *         return True
 * 
 *     cpdef append(self, uint32_t x):             # <<<<<<<<<<<<<<
 *         self._append(x)
 *         listener = <object>self._listener
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  uint32_t __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_uint32_t(__pyx_arg_x); if (unlikely((__pyx_v_x == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1203, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List_31append(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self), ((uint32_t)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_31append(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, uint32_t __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("append", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_10Uint32List_append(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1209
 *             listener._Modified()
 * 
 *     cpdef extend(self, x):             # <<<<<<<<<<<<<<
 *         for i in x:
 *             self._append(i)
 */

static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint32List_extend(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, PyObject *__pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  uint32_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("extend", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extend); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint32List_34extend)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1209, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_x};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1209, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_x};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1209, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1209, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_x);
          __Pyx_GIVEREF(__pyx_v_x);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_x);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1209, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1210
 * 
 *     cpdef extend(self, x):
 *         for i in x:             # <<<<<<<<<<<<<<
 *             self._append(i)
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_x)) || PyTuple_CheckExact(__pyx_v_x)) {
    __pyx_t_1 = __pyx_v_x; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1210, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1210, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1210, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1210, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1210, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_7(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1210, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyrobuf_list.pyx":1211
 *     cpdef extend(self, x):
 *         for i in x:
 *             self._append(i)             # <<<<<<<<<<<<<<
 * 
 *         listener = <object>self._listener
 */
    __pyx_t_8 = __Pyx_PyInt_As_uint32_t(__pyx_v_i); if (unlikely((__pyx_t_8 == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1211, __pyx_L1_error)
    ((struct __pyx_vtabstruct_12pyrobuf_list_Uint32List *)__pyx_v_self->__pyx_vtab)->_append(__pyx_v_self, __pyx_t_8);

    /* "pyrobuf_list.pyx":1210
 * 
 *     cpdef extend(self, x):
 *         for i in x:             # <<<<<<<<<<<<<<
 *             self._append(i)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1213
 *             self._append(i)
 * 
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1214
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1214, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = ((!__pyx_t_9) != 0);
  if (__pyx_t_10) {

    /* "pyrobuf_list.pyx":1215
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef insert(self, int i, uint32_t x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1215, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1215, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1215, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1214
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1209
 *             listener._Modified()
 * 
 *     cpdef extend(self, x):             # <<<<<<<<<<<<<<
 *         for i in x:
 *             self._append(i)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extend (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List_33extend(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_33extend(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("extend", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_10Uint32List_extend(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1217
 *             listener._Modified()
 * 
 *     cpdef insert(self, int i, uint32_t x):             # <<<<<<<<<<<<<<
 *         cdef uint32_t *mem
 *         cdef size_t j
 */

static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint32List_insert(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, int __pyx_v_i, uint32_t __pyx_v_x, int __pyx_skip_dispatch) {
  uint32_t *__pyx_v_mem;
  size_t __pyx_v_j;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  size_t __pyx_t_11;
  size_t __pyx_t_12;
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("insert", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_insert); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint32List_36insert)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1217, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyInt_From_uint32_t(__pyx_v_x); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1217, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1217, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1217, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1217, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
        __pyx_t_3 = 0;
        __pyx_t_4 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1217, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1221
 *         cdef size_t j
 * 
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_9 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_9) {

    /* "pyrobuf_list.pyx":1222
 * 
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":1221
 *         cdef size_t j
 * 
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":1224
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i > self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_10 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_10) {
  } else {
    __pyx_t_9 = __pyx_t_10;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_10 = ((((size_t)__pyx_v_i) > __pyx_v_self->_n_items) != 0);
  __pyx_t_9 = __pyx_t_10;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_9)) {

    /* "pyrobuf_list.pyx":1225
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         if self._n_items == self._size:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__55, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1225, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1225, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1224
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i > self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1227
 *             raise IndexError("list index out of range")
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <uint32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint32_t))
 *             if not mem:
 */
  __pyx_t_9 = ((__pyx_v_self->_n_items == __pyx_v_self->_size) != 0);
  if (__pyx_t_9) {

    /* "pyrobuf_list.pyx":1228
 * 
 *         if self._n_items == self._size:
 *             mem = <uint32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint32_t))             # <<<<<<<<<<<<<<
 *             if not mem:
 *                 raise MemoryError()
 */
    __pyx_v_mem = ((uint32_t *)PyMem_Realloc(__pyx_v_self->_data, ((2 * __pyx_v_self->_size) * (sizeof(uint32_t)))));

    /* "pyrobuf_list.pyx":1229
 *         if self._n_items == self._size:
 *             mem = <uint32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint32_t))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    __pyx_t_9 = ((!(__pyx_v_mem != 0)) != 0);
    if (unlikely(__pyx_t_9)) {

      /* "pyrobuf_list.pyx":1230
 *             mem = <uint32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint32_t))
 *             if not mem:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *             self._data = mem
 */
      PyErr_NoMemory(); __PYX_ERR(0, 1230, __pyx_L1_error)

      /* "pyrobuf_list.pyx":1229
 *         if self._n_items == self._size:
 *             mem = <uint32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint32_t))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    }

    /* "pyrobuf_list.pyx":1232
 *                 raise MemoryError()
 * 
 *             self._data = mem             # <<<<<<<<<<<<<<
 *             self._size *= 2
 * 
 */
    __pyx_v_self->_data = __pyx_v_mem;

    /* "pyrobuf_list.pyx":1233
 * 
 *             self._data = mem
 *             self._size *= 2             # <<<<<<<<<<<<<<
 * 
 *         for j in range(self._n_items, <size_t>i, -1):
 */
    __pyx_v_self->_size = (__pyx_v_self->_size * 2);

    /* "pyrobuf_list.pyx":1227
 *             raise IndexError("list index out of range")
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <uint32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint32_t))
 *             if not mem:
 */
  }

  /* "pyrobuf_list.pyx":1235
 *             self._size *= 2
 * 
 *         for j in range(self._n_items, <size_t>i, -1):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j - 1]
 * 
 */
  __pyx_t_11 = ((size_t)__pyx_v_i);
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = __pyx_v_self->_n_items + 1; __pyx_t_13 > __pyx_t_12 + 1; ) { __pyx_t_13-=1;
    __pyx_v_j = __pyx_t_13;

    /* "pyrobuf_list.pyx":1236
 * 
 *         for j in range(self._n_items, <size_t>i, -1):
 *             self._data[j] = self._data[j - 1]             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j - 1)]);
  }

  /* "pyrobuf_list.pyx":1238
 *             self._data[j] = self._data[j - 1]
 * 
 *         self._data[i] = x             # <<<<<<<<<<<<<<
 *         self._n_items += 1
 *         listener = <object>self._listener
 */
  (__pyx_v_self->_data[__pyx_v_i]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":1239
 * 
 *         self._data[i] = x
 *         self._n_items += 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items + 1);

  /* "pyrobuf_list.pyx":1240
 *         self._data[i] = x
 *         self._n_items += 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1241
 *         self._n_items += 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1241, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = ((!__pyx_t_9) != 0);
  if (__pyx_t_10) {

    /* "pyrobuf_list.pyx":1242
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef pop(self):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1242, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1242, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1242, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1241
 *         self._n_items += 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1217
 *             listener._Modified()
 * 
 *     cpdef insert(self, int i, uint32_t x):             # <<<<<<<<<<<<<<
 *         cdef uint32_t *mem
 *         cdef size_t j
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_i;
  uint32_t __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("insert (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_i,&__pyx_n_s_x,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, 1); __PYX_ERR(0, 1217, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "insert") < 0)) __PYX_ERR(0, 1217, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_i = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1217, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_uint32_t(values[1]); if (unlikely((__pyx_v_x == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1217, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1217, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List_35insert(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self), __pyx_v_i, __pyx_v_x);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_35insert(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, int __pyx_v_i, uint32_t __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("insert", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_10Uint32List_insert(__pyx_v_self, __pyx_v_i, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1244
 *             listener._Modified()
 * 
 *     cpdef pop(self):             # <<<<<<<<<<<<<<
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")
 */

static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint32List_pop(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("pop", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint32List_38pop)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1244, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1244, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1245
 * 
 *     cpdef pop(self):
 *         if self._n_items == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("pop from empty list")
 * 
 */
  __pyx_t_5 = ((__pyx_v_self->_n_items == 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "pyrobuf_list.pyx":1246
 *     cpdef pop(self):
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__56, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1246, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1246, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1245
 * 
 *     cpdef pop(self):
 *         if self._n_items == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("pop from empty list")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1248
 *             raise IndexError("pop from empty list")
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":1249
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1250
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 *         return self._data[self._n_items]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1250, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = ((!__pyx_t_5) != 0);
  if (__pyx_t_6) {

    /* "pyrobuf_list.pyx":1251
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 *         return self._data[self._n_items]
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1251, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1251, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1250
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 *         return self._data[self._n_items]
 */
  }

  /* "pyrobuf_list.pyx":1252
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 *         return self._data[self._n_items]             # <<<<<<<<<<<<<<
 * 
 *     cpdef remove(self, uint32_t x):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t((__pyx_v_self->_data[__pyx_v_self->_n_items])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1244
 *             listener._Modified()
 * 
 *     cpdef pop(self):             # <<<<<<<<<<<<<<
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pop (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List_37pop(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_37pop(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("pop", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_10Uint32List_pop(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1254
 *         return self._data[self._n_items]
 * 
 *     cpdef remove(self, uint32_t x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef size_t j
 */

static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint32List_remove(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, uint32_t __pyx_v_x, int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  int __pyx_v_found;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("remove", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_remove); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint32List_40remove)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_uint32_t(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1254, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1254, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1254, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1254, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1254, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1254, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1257
 *         cdef size_t i
 *         cdef size_t j
 *         cdef bint found = 0             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self._n_items):
 */
  __pyx_v_found = 0;

  /* "pyrobuf_list.pyx":1259
 *         cdef bint found = 0
 * 
 *         for i in range(self._n_items):             # <<<<<<<<<<<<<<
 *             if self._data[i] == x:
 *                 found = 1
 */
  __pyx_t_7 = __pyx_v_self->_n_items;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "pyrobuf_list.pyx":1260
 * 
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 found = 1
 *                 break
 */
    __pyx_t_10 = (((__pyx_v_self->_data[__pyx_v_i]) == __pyx_v_x) != 0);
    if (__pyx_t_10) {

      /* "pyrobuf_list.pyx":1261
 *         for i in range(self._n_items):
 *             if self._data[i] == x:
 *                 found = 1             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
      __pyx_v_found = 1;

      /* "pyrobuf_list.pyx":1262
 *             if self._data[i] == x:
 *                 found = 1
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         if found == 0:
 */
      goto __pyx_L4_break;

      /* "pyrobuf_list.pyx":1260
 * 
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 found = 1
 *                 break
 */
    }
  }
  __pyx_L4_break:;

  /* "pyrobuf_list.pyx":1264
 *                 break
 * 
 *         if found == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("x not in list")
 * 
 */
  __pyx_t_10 = ((__pyx_v_found == 0) != 0);
  if (unlikely(__pyx_t_10)) {

    /* "pyrobuf_list.pyx":1265
 * 
 *         if found == 0:
 *             raise ValueError("x not in list")             # <<<<<<<<<<<<<<
 * 
 *         for j in range(i, self._n_items):
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__57, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1265, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1265, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1264
 *                 break
 * 
 *         if found == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("x not in list")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1267
 *             raise ValueError("x not in list")
 * 
 *         for j in range(i, self._n_items):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j + 1]
 * 
 */
  __pyx_t_7 = __pyx_v_self->_n_items;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = __pyx_v_i; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_j = __pyx_t_9;

    /* "pyrobuf_list.pyx":1268
 * 
 *         for j in range(i, self._n_items):
 *             self._data[j] = self._data[j + 1]             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j + 1)]);
  }

  /* "pyrobuf_list.pyx":1270
 *             self._data[j] = self._data[j + 1]
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":1271
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1272
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = ((!__pyx_t_10) != 0);
  if (__pyx_t_11) {

    /* "pyrobuf_list.pyx":1273
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cdef void _append(self, uint32_t x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1273, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1273, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1273, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1272
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1254
 *         return self._data[self._n_items]
 * 
 *     cpdef remove(self, uint32_t x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef size_t j
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  uint32_t __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_uint32_t(__pyx_arg_x); if (unlikely((__pyx_v_x == ((uint32_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1254, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List_39remove(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self), ((uint32_t)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_39remove(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, uint32_t __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("remove", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_10Uint32List_remove(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1275
 *             listener._Modified()
 * 
 *     cdef void _append(self, uint32_t x):             # <<<<<<<<<<<<<<
 *         cdef uint32_t *mem
 * 
 */

static void __pyx_f_12pyrobuf_list_10Uint32List__append(struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, uint32_t __pyx_v_x) {
  uint32_t *__pyx_v_mem;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_append", 0);

  /* "pyrobuf_list.pyx":1278
 *         cdef uint32_t *mem
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <uint32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint32_t))
 *             if not mem:
 */
  __pyx_t_1 = ((__pyx_v_self->_n_items == __pyx_v_self->_size) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":1279
 * 
 *         if self._n_items == self._size:
 *             mem = <uint32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint32_t))             # <<<<<<<<<<<<<<
 *             if not mem:
 *                 raise MemoryError()
 */
    __pyx_v_mem = ((uint32_t *)PyMem_Realloc(__pyx_v_self->_data, ((2 * __pyx_v_self->_size) * (sizeof(uint32_t)))));

    /* "pyrobuf_list.pyx":1280
 *         if self._n_items == self._size:
 *             mem = <uint32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint32_t))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    __pyx_t_1 = ((!(__pyx_v_mem != 0)) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "pyrobuf_list.pyx":1281
 *             mem = <uint32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint32_t))
 *             if not mem:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *             self._data = mem
 */
      PyErr_NoMemory(); __PYX_ERR(0, 1281, __pyx_L1_error)

      /* "pyrobuf_list.pyx":1280
 *         if self._n_items == self._size:
 *             mem = <uint32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint32_t))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    }

    /* "pyrobuf_list.pyx":1283
 *                 raise MemoryError()
 * 
 *             self._data = mem             # <<<<<<<<<<<<<<
 *             self._size *= 2
 * 
 */
    __pyx_v_self->_data = __pyx_v_mem;

    /* "pyrobuf_list.pyx":1284
 * 
 *             self._data = mem
 *             self._size *= 2             # <<<<<<<<<<<<<<
 * 
 *         self._data[self._n_items] = x
 */
    __pyx_v_self->_size = (__pyx_v_self->_size * 2);

    /* "pyrobuf_list.pyx":1278
 *         cdef uint32_t *mem
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <uint32_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint32_t))
 *             if not mem:
 */
  }

  /* "pyrobuf_list.pyx":1286
 *             self._size *= 2
 * 
 *         self._data[self._n_items] = x             # <<<<<<<<<<<<<<
 *         self._n_items += 1
 * 
 */
  (__pyx_v_self->_data[__pyx_v_self->_n_items]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":1287
 * 
 *         self._data[self._n_items] = x
 *         self._n_items += 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items + 1);

  /* "pyrobuf_list.pyx":1275
 *             listener._Modified()
 * 
 *     cdef void _append(self, uint32_t x):             # <<<<<<<<<<<<<<
 *         cdef uint32_t *mem
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("pyrobuf_list.Uint32List._append", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_42__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_42__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List_41__reduce_cython__(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_41__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__58, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_44__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint32List_44__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint32List_43__setstate_cython__(((struct __pyx_obj_12pyrobuf_list_Uint32List *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint32List_43__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Uint32List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__59, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Uint32List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1293
 * cdef class Int64List:
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):             # <<<<<<<<<<<<<<
 *         self._data = <int64_t *>PyMem_Malloc(size * sizeof(int64_t))
 *         if not self._data:
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_9Int64List_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12pyrobuf_list_9Int64List_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_size;
  PyObject *__pyx_v_listener = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_size,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_listener);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 1293, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_size = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_size == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1293, __pyx_L3_error)
    } else {
      __pyx_v_size = ((size_t)16);
    }
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1293, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int64List.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List___cinit__(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self), __pyx_v_size, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_9Int64List___cinit__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, size_t __pyx_v_size, PyObject *__pyx_v_listener) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "pyrobuf_list.pyx":1294
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <int64_t *>PyMem_Malloc(size * sizeof(int64_t))             # <<<<<<<<<<<<<<
 *         if not self._data:
 *             raise MemoryError()
 */
  __pyx_v_self->_data = ((int64_t *)PyMem_Malloc((__pyx_v_size * (sizeof(int64_t)))));

  /* "pyrobuf_list.pyx":1295
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <int64_t *>PyMem_Malloc(size * sizeof(int64_t))
 *         if not self._data:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_self->_data != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":1296
 *         self._data = <int64_t *>PyMem_Malloc(size * sizeof(int64_t))
 *         if not self._data:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         self._n_items = 0
 */
    PyErr_NoMemory(); __PYX_ERR(0, 1296, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1295
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <int64_t *>PyMem_Malloc(size * sizeof(int64_t))
 *         if not self._data:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1298
 *             raise MemoryError()
 * 
 *         self._n_items = 0             # <<<<<<<<<<<<<<
 *         self._size = size
 *         if listener is None:
 */
  __pyx_v_self->_n_items = 0;

  /* "pyrobuf_list.pyx":1299
 * 
 *         self._n_items = 0
 *         self._size = size             # <<<<<<<<<<<<<<
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener
 */
  __pyx_v_self->_size = __pyx_v_size;

  /* "pyrobuf_list.pyx":1300
 *         self._n_items = 0
 *         self._size = size
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
  __pyx_t_1 = (__pyx_v_listener == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":1301
 *         self._size = size
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener             # <<<<<<<<<<<<<<
 *         else:
 *             self._listener = <PyObject *>listener
 */
    __pyx_v_self->_listener = ((PyObject *)null_listener);

    /* "pyrobuf_list.pyx":1300
 *         self._n_items = 0
 *         self._size = size
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
    goto __pyx_L4;
  }

  /* "pyrobuf_list.pyx":1303
 *             self._listener = <PyObject *>null_listener
 *         else:
 *             self._listener = <PyObject *>listener             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  /*else*/ {
    __pyx_v_self->_listener = ((PyObject *)__pyx_v_listener);
  }
  __pyx_L4:;

  /* "pyrobuf_list.pyx":1293
 * cdef class Int64List:
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):             # <<<<<<<<<<<<<<
 *         self._data = <int64_t *>PyMem_Malloc(size * sizeof(int64_t))
 *         if not self._data:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int64List.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1305
 *             self._listener = <PyObject *>listener
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self._data)
 * 
 */

/* Python wrapper */
static void __pyx_pw_12pyrobuf_list_9Int64List_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12pyrobuf_list_9Int64List_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_12pyrobuf_list_9Int64List_2__dealloc__(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12pyrobuf_list_9Int64List_2__dealloc__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "pyrobuf_list.pyx":1306
 * 
 *     def __dealloc__(self):
 *         PyMem_Free(self._data)             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, int64_t x):
 */
  PyMem_Free(__pyx_v_self->_data);

  /* "pyrobuf_list.pyx":1305
 *             self._listener = <PyObject *>listener
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self._data)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "pyrobuf_list.pyx":1308
 *         PyMem_Free(self._data)
 * 
 *     def __contains__(self, int64_t x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self._n_items):
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_9Int64List_5__contains__(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static int __pyx_pw_12pyrobuf_list_9Int64List_5__contains__(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  int64_t __pyx_v_x;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_int64_t(__pyx_arg_x); if (unlikely((__pyx_v_x == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1308, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int64List.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List_4__contains__(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self), ((int64_t)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_9Int64List_4__contains__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int64_t __pyx_v_x) {
  size_t __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "pyrobuf_list.pyx":1310
 *     def __contains__(self, int64_t x):
 *         cdef size_t i
 *         for i in range(self._n_items):             # <<<<<<<<<<<<<<
 *             if self._data[i] == x:
 *                 return True
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "pyrobuf_list.pyx":1311
 *         cdef size_t i
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */
    __pyx_t_4 = (((__pyx_v_self->_data[__pyx_v_i]) == __pyx_v_x) != 0);
    if (__pyx_t_4) {

      /* "pyrobuf_list.pyx":1312
 *         for i in range(self._n_items):
 *             if self._data[i] == x:
 *                 return True             # <<<<<<<<<<<<<<
 * 
 *         return False
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "pyrobuf_list.pyx":1311
 *         cdef size_t i
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */
    }
  }

  /* "pyrobuf_list.pyx":1314
 *                 return True
 * 
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     def __delitem__(self, int i):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1308
 *         PyMem_Free(self._data)
 * 
 *     def __contains__(self, int64_t x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self._n_items):
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1316
 *         return False
 * 
 *     def __delitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_9Int64List_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static int __pyx_pw_12pyrobuf_list_9Int64List_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1316, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int64List.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List_6__delitem__(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self), ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_9Int64List_6__delitem__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int __pyx_v_i) {
  size_t __pyx_v_j;
  PyObject *__pyx_v_listener = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__delitem__", 0);

  /* "pyrobuf_list.pyx":1317
 * 
 *     def __delitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":1318
 *     def __delitem__(self, int i):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":1317
 * 
 *     def __delitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":1320
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":1321
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t j
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__60, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1321, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1320
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1324
 * 
 *         cdef size_t j
 *         for j in range(i, self._n_items):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j + 1]
 * 
 */
  __pyx_t_4 = __pyx_v_self->_n_items;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = __pyx_v_i; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_j = __pyx_t_6;

    /* "pyrobuf_list.pyx":1325
 *         cdef size_t j
 *         for j in range(i, self._n_items):
 *             self._data[j] = self._data[j + 1]             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j + 1)]);
  }

  /* "pyrobuf_list.pyx":1327
 *             self._data[j] = self._data[j + 1]
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":1328
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":1329
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1329, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":1330
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, int i):
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    if (__pyx_t_8) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1330, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1330, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":1329
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1316
 *         return False
 * 
 *     def __delitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyrobuf_list.Int64List.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1332
 *             listener._Modified()
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1332, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int64List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List_8__getitem__(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self), ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_8__getitem__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int __pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pyrobuf_list.pyx":1333
 * 
 *     def __getitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":1334
 *     def __getitem__(self, int i):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":1333
 * 
 *     def __getitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":1336
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":1337
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         return self._data[i]
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__61, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1337, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1336
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1339
 *             raise IndexError("list index out of range")
 * 
 *         return self._data[i]             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_int64_t((__pyx_v_self->_data[__pyx_v_i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1332
 *             listener._Modified()
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyrobuf_list.Int64List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_12pyrobuf_list_9Int64List_12generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "pyrobuf_list.pyx":1341
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_11__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_11__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List_10__iter__(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_10__iter__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self) {
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_5___iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_5___iter__ *)__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_5___iter__(__pyx_ptype_12pyrobuf_list___pyx_scope_struct_5___iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_5___iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1341, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_12pyrobuf_list_9Int64List_12generator5, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_Int64List___iter, __pyx_n_s_pyrobuf_list); if (unlikely(!gen)) __PYX_ERR(0, 1341, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int64List.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_12pyrobuf_list_9Int64List_12generator5(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_5___iter__ *__pyx_cur_scope = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_5___iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1341, __pyx_L1_error)

  /* "pyrobuf_list.pyx":1342
 * 
 *     def __iter__(self):
 *         cdef size_t idx = 0             # <<<<<<<<<<<<<<
 *         while idx < self._n_items:
 *             yield self._data[idx]
 */
  __pyx_cur_scope->__pyx_v_idx = 0;

  /* "pyrobuf_list.pyx":1343
 *     def __iter__(self):
 *         cdef size_t idx = 0
 *         while idx < self._n_items:             # <<<<<<<<<<<<<<
 *             yield self._data[idx]
 *             idx += 1
 */
  while (1) {
    __pyx_t_1 = ((__pyx_cur_scope->__pyx_v_idx < __pyx_cur_scope->__pyx_v_self->_n_items) != 0);
    if (!__pyx_t_1) break;

    /* "pyrobuf_list.pyx":1344
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 *             yield self._data[idx]             # <<<<<<<<<<<<<<
 *             idx += 1
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_int64_t((__pyx_cur_scope->__pyx_v_self->_data[__pyx_cur_scope->__pyx_v_idx])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1344, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1344, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1345
 *         while idx < self._n_items:
 *             yield self._data[idx]
 *             idx += 1             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
    __pyx_cur_scope->__pyx_v_idx = (__pyx_cur_scope->__pyx_v_idx + 1);
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "pyrobuf_list.pyx":1341
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1347
 *             idx += 1
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._n_items
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_12pyrobuf_list_9Int64List_14__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_12pyrobuf_list_9Int64List_14__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List_13__len__(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_12pyrobuf_list_9Int64List_13__len__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pyrobuf_list.pyx":1348
 * 
 *     def __len__(self):
 *         return self._n_items             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = __pyx_v_self->_n_items;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1347
 *             idx += 1
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._n_items
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1350
 *         return self._n_items
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(list(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_16__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_16__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List_15__repr__(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_15__repr__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pyrobuf_list.pyx":1351
 * 
 *     def __repr__(self):
 *         return repr(list(self))             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, int i, int64_t x):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1350
 *         return self._n_items
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(list(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.Int64List.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1353
 *         return repr(list(self))
 * 
 *     def __setitem__(self, int i, int64_t x):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_9Int64List_18__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i, PyObject *__pyx_arg_x); /*proto*/
static int __pyx_pw_12pyrobuf_list_9Int64List_18__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i, PyObject *__pyx_arg_x) {
  int __pyx_v_i;
  int64_t __pyx_v_x;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1353, __pyx_L3_error)
  }
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_int64_t(__pyx_arg_x); if (unlikely((__pyx_v_x == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1353, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int64List.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List_17__setitem__(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self), ((int)__pyx_v_i), ((int64_t)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_9Int64List_17__setitem__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int __pyx_v_i, int64_t __pyx_v_x) {
  PyObject *__pyx_v_listener = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "pyrobuf_list.pyx":1354
 * 
 *     def __setitem__(self, int i, int64_t x):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":1355
 *     def __setitem__(self, int i, int64_t x):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":1354
 * 
 *     def __setitem__(self, int i, int64_t x):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":1357
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":1358
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__62, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1358, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1357
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1360
 *             raise IndexError("list index out of range")
 * 
 *         self._data[i] = x             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  (__pyx_v_self->_data[__pyx_v_i]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":1361
 * 
 *         self._data[i] = x
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":1362
 *         self._data[i] = x
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1362, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":1363
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1363, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1363, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1363, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":1362
 *         self._data[i] = x
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1353
 *         return repr(list(self))
 * 
 *     def __setitem__(self, int i, int64_t x):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.Int64List.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1365
 *             listener._Modified()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(list(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_20__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_20__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List_19__str__(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_19__str__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pyrobuf_list.pyx":1366
 * 
 *     def __str__(self):
 *         return str(list(self))             # <<<<<<<<<<<<<<
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1365
 *             listener._Modified()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(list(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.Int64List.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1368
 *         return str(list(self))
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t itemsize = sizeof(int64_t)
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_12pyrobuf_list_9Int64List_22__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_12pyrobuf_list_9Int64List_22__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List_21__getbuffer__(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_9Int64List_21__getbuffer__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags) {
  Py_ssize_t __pyx_v_itemsize;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  if (__pyx_v_buffer == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_buffer->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_buffer->obj);

  /* "pyrobuf_list.pyx":1370
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 * 
 *         cdef Py_ssize_t itemsize = sizeof(int64_t)             # <<<<<<<<<<<<<<
 *         self.shape[0] = self._n_items
 *         self.strides[0] = 1
 */
  __pyx_v_itemsize = (sizeof(int64_t));

  /* "pyrobuf_list.pyx":1371
 * 
 *         cdef Py_ssize_t itemsize = sizeof(int64_t)
 *         self.shape[0] = self._n_items             # <<<<<<<<<<<<<<
 *         self.strides[0] = 1
 * 
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  (__pyx_v_self->shape[0]) = __pyx_t_1;

  /* "pyrobuf_list.pyx":1372
 *         cdef Py_ssize_t itemsize = sizeof(int64_t)
 *         self.shape[0] = self._n_items
 *         self.strides[0] = 1             # <<<<<<<<<<<<<<
 * 
 *         buffer.buf = <char *>(self._data)
 */
  (__pyx_v_self->strides[0]) = 1;

  /* "pyrobuf_list.pyx":1374
 *         self.strides[0] = 1
 * 
 *         buffer.buf = <char *>(self._data)             # <<<<<<<<<<<<<<
 *         buffer.format = 'L'
 *         buffer.internal = NULL                  # see References
 */
  __pyx_v_buffer->buf = ((char *)__pyx_v_self->_data);

  /* "pyrobuf_list.pyx":1375
 * 
 *         buffer.buf = <char *>(self._data)
 *         buffer.format = 'L'             # <<<<<<<<<<<<<<
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 */
  __pyx_v_buffer->format = ((char *)"L");

  /* "pyrobuf_list.pyx":1376
 *         buffer.buf = <char *>(self._data)
 *         buffer.format = 'L'
 *         buffer.internal = NULL                  # see References             # <<<<<<<<<<<<<<
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items
 */
  __pyx_v_buffer->internal = NULL;

  /* "pyrobuf_list.pyx":1377
 *         buffer.format = 'L'
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize             # <<<<<<<<<<<<<<
 *         buffer.len = self._n_items
 *         buffer.ndim = 1
 */
  __pyx_v_buffer->itemsize = __pyx_v_itemsize;

  /* "pyrobuf_list.pyx":1378
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items             # <<<<<<<<<<<<<<
 *         buffer.ndim = 1
 *         buffer.obj = self
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  __pyx_v_buffer->len = __pyx_t_1;

  /* "pyrobuf_list.pyx":1379
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items
 *         buffer.ndim = 1             # <<<<<<<<<<<<<<
 *         buffer.obj = self
 *         buffer.readonly = 1
 */
  __pyx_v_buffer->ndim = 1;

  /* "pyrobuf_list.pyx":1380
 *         buffer.len = self._n_items
 *         buffer.ndim = 1
 *         buffer.obj = self             # <<<<<<<<<<<<<<
 *         buffer.readonly = 1
 *         buffer.shape = self.shape
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_buffer->obj);
  __Pyx_DECREF(__pyx_v_buffer->obj);
  __pyx_v_buffer->obj = ((PyObject *)__pyx_v_self);

  /* "pyrobuf_list.pyx":1381
 *         buffer.ndim = 1
 *         buffer.obj = self
 *         buffer.readonly = 1             # <<<<<<<<<<<<<<
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides
 */
  __pyx_v_buffer->readonly = 1;

  /* "pyrobuf_list.pyx":1382
 *         buffer.obj = self
 *         buffer.readonly = 1
 *         buffer.shape = self.shape             # <<<<<<<<<<<<<<
 *         buffer.strides = self.strides
 *         buffer.suboffsets = NULL                # for pointer arrays only
 */
  __pyx_t_2 = __pyx_v_self->shape;
  __pyx_v_buffer->shape = __pyx_t_2;

  /* "pyrobuf_list.pyx":1383
 *         buffer.readonly = 1
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides             # <<<<<<<<<<<<<<
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 */
  __pyx_t_2 = __pyx_v_self->strides;
  __pyx_v_buffer->strides = __pyx_t_2;

  /* "pyrobuf_list.pyx":1384
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides
 *         buffer.suboffsets = NULL                # for pointer arrays only             # <<<<<<<<<<<<<<
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):
 */
  __pyx_v_buffer->suboffsets = NULL;

  /* "pyrobuf_list.pyx":1368
 *         return str(list(self))
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t itemsize = sizeof(int64_t)
 */

  /* function exit code */
  __pyx_r = 0;
  if (__pyx_v_buffer->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_buffer->obj);
    __Pyx_DECREF(__pyx_v_buffer->obj); __pyx_v_buffer->obj = 0;
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1386
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_12pyrobuf_list_9Int64List_24__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer); /*proto*/
static CYTHON_UNUSED void __pyx_pw_12pyrobuf_list_9Int64List_24__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_12pyrobuf_list_9Int64List_23__releasebuffer__(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12pyrobuf_list_9Int64List_23__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "pyrobuf_list.pyx":1389
 *         pass
 * 
 *     def __richcmp__(self, other, op):             # <<<<<<<<<<<<<<
 *         if op == Py_EQ:
 *             return self._equal(other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_26__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_26__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op) {
  PyObject *__pyx_v_op = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_v_op = __Pyx_PyInt_From_int(__pyx_arg_op); if (unlikely(!__pyx_v_op)) __PYX_ERR(0, 1389, __pyx_L3_error)
  __Pyx_GOTREF(__pyx_v_op);
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int64List.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List_25__richcmp__(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((PyObject *)__pyx_v_op));

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_op);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_25__richcmp__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "pyrobuf_list.pyx":1390
 * 
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:             # <<<<<<<<<<<<<<
 *             return self._equal(other)
 *         elif op == Py_NE:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(Py_EQ); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_op, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1390, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1390, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "pyrobuf_list.pyx":1391
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:
 *             return self._equal(other)             # <<<<<<<<<<<<<<
 *         elif op == Py_NE:
 *             return not self._equal(other)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = ((struct __pyx_vtabstruct_12pyrobuf_list_Int64List *)__pyx_v_self->__pyx_vtab)->_equal(__pyx_v_self, __pyx_v_other, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pyrobuf_list.pyx":1390
 * 
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:             # <<<<<<<<<<<<<<
 *             return self._equal(other)
 *         elif op == Py_NE:
 */
  }

  /* "pyrobuf_list.pyx":1392
 *         if op == Py_EQ:
 *             return self._equal(other)
 *         elif op == Py_NE:             # <<<<<<<<<<<<<<
 *             return not self._equal(other)
 *         else:
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(Py_NE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_op, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1392, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1392, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(__pyx_t_3)) {

    /* "pyrobuf_list.pyx":1393
 *             return self._equal(other)
 *         elif op == Py_NE:
 *             return not self._equal(other)             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError('Operation not supported for lists')
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((struct __pyx_vtabstruct_12pyrobuf_list_Int64List *)__pyx_v_self->__pyx_vtab)->_equal(__pyx_v_self, __pyx_v_other, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1393, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1393, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_3)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1393, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pyrobuf_list.pyx":1392
 *         if op == Py_EQ:
 *             return self._equal(other)
 *         elif op == Py_NE:             # <<<<<<<<<<<<<<
 *             return not self._equal(other)
 *         else:
 */
  }

  /* "pyrobuf_list.pyx":1395
 *             return not self._equal(other)
 *         else:
 *             raise TypeError('Operation not supported for lists')             # <<<<<<<<<<<<<<
 * 
 *     def __neq__(self, other):
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__63, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1395, __pyx_L1_error)
  }

  /* "pyrobuf_list.pyx":1389
 *         pass
 * 
 *     def __richcmp__(self, other, op):             # <<<<<<<<<<<<<<
 *         if op == Py_EQ:
 *             return self._equal(other)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.Int64List.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1397
 *             raise TypeError('Operation not supported for lists')
 * 
 *     def __neq__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_28__neq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_28__neq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__neq__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List_27__neq__(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_27__neq__(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__neq__", 0);

  /* "pyrobuf_list.pyx":1398
 * 
 *     def __neq__(self, other):
 *         return not self == other             # <<<<<<<<<<<<<<
 * 
 *     cpdef _equal(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_self), __pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1398, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1398, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1397
 *             raise TypeError('Operation not supported for lists')
 * 
 *     def __neq__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Int64List.__neq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1400
 *         return not self == other
 * 
 *     cpdef _equal(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             if len(self) != len(other):
 */

static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int64List__equal(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_v_y = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *(*__pyx_t_14)(PyObject *);
  __Pyx_RefNannySetupContext("_equal", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_equal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_9Int64List_30_equal)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_other); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1400, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_other};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1400, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_other};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1400, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1400, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_other);
          __Pyx_GIVEREF(__pyx_v_other);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_other);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1400, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1401
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    /*try:*/ {

      /* "pyrobuf_list.pyx":1402
 *     cpdef _equal(self, other):
 *         try:
 *             if len(self) != len(other):             # <<<<<<<<<<<<<<
 *                 return False
 *         except TypeError:
 */
      __pyx_t_9 = PyObject_Length(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1402, __pyx_L3_error)
      __pyx_t_10 = PyObject_Length(__pyx_v_other); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1402, __pyx_L3_error)
      __pyx_t_11 = ((__pyx_t_9 != __pyx_t_10) != 0);
      if (__pyx_t_11) {

        /* "pyrobuf_list.pyx":1403
 *         try:
 *             if len(self) != len(other):
 *                 return False             # <<<<<<<<<<<<<<
 *         except TypeError:
 *             return False
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_False);
        __pyx_r = Py_False;
        goto __pyx_L7_try_return;

        /* "pyrobuf_list.pyx":1402
 *     cpdef _equal(self, other):
 *         try:
 *             if len(self) != len(other):             # <<<<<<<<<<<<<<
 *                 return False
 *         except TypeError:
 */
      }

      /* "pyrobuf_list.pyx":1401
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
    }
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1404
 *             if len(self) != len(other):
 *                 return False
 *         except TypeError:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
    __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_12) {
      __Pyx_AddTraceback("pyrobuf_list.Int64List._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 1404, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_3);

      /* "pyrobuf_list.pyx":1405
 *                 return False
 *         except TypeError:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         for x, y in zip(self, other):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "pyrobuf_list.pyx":1401
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "pyrobuf_list.pyx":1407
 *             return False
 * 
 *         for x, y in zip(self, other):             # <<<<<<<<<<<<<<
 *             if x != y:
 *                 return False
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));
  __Pyx_INCREF(__pyx_v_other);
  __Pyx_GIVEREF(__pyx_v_other);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_other);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_10 = 0;
    __pyx_t_13 = NULL;
  } else {
    __pyx_t_10 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_13 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1407, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_13)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 1407, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1407, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 1407, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1407, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_13(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1407, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1407, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1407, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1407, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1407, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_14 = Py_TYPE(__pyx_t_4)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_5 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_5)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_4), 2) < 0) __PYX_ERR(0, 1407, __pyx_L1_error)
      __pyx_t_14 = NULL;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L15_unpacking_done;
      __pyx_L14_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_14 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1407, __pyx_L1_error)
      __pyx_L15_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_x, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_y, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "pyrobuf_list.pyx":1408
 * 
 *         for x, y in zip(self, other):
 *             if x != y:             # <<<<<<<<<<<<<<
 *                 return False
 * 
 */
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_x, __pyx_v_y, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1408, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1408, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_11) {

      /* "pyrobuf_list.pyx":1409
 *         for x, y in zip(self, other):
 *             if x != y:
 *                 return False             # <<<<<<<<<<<<<<
 * 
 *         return True
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "pyrobuf_list.pyx":1408
 * 
 *         for x, y in zip(self, other):
 *             if x != y:             # <<<<<<<<<<<<<<
 *                 return False
 * 
 */
    }

    /* "pyrobuf_list.pyx":1407
 *             return False
 * 
 *         for x, y in zip(self, other):             # <<<<<<<<<<<<<<
 *             if x != y:
 *                 return False
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":1411
 *                 return False
 * 
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     cpdef append(self, int64_t x):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1400
 *         return not self == other
 * 
 *     cpdef _equal(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             if len(self) != len(other):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.Int64List._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_y);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_equal (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List_29_equal(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_29_equal(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_equal", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_9Int64List__equal(__pyx_v_self, __pyx_v_other, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Int64List._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1413
 *         return True
 * 
 *     cpdef append(self, int64_t x):             # <<<<<<<<<<<<<<
 *         self._append(x)
 *         listener = <object>self._listener
 */

static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int64List_append(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int64_t __pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("append", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_append); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_9Int64List_32append)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1413, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1413, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1413, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1413, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1413, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1414
 * 
 *     cpdef append(self, int64_t x):
 *         self._append(x)             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  ((struct __pyx_vtabstruct_12pyrobuf_list_Int64List *)__pyx_v_self->__pyx_vtab)->_append(__pyx_v_self, __pyx_v_x);

  /* "pyrobuf_list.pyx":1415
 *     cpdef append(self, int64_t x):
 *         self._append(x)
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1416
 *         self._append(x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1416, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = ((!__pyx_t_7) != 0);
  if (__pyx_t_8) {

    /* "pyrobuf_list.pyx":1417
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef extend(self, x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1417, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1417, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1417, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1416
 *         self._append(x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1413
 *         return True
 * 
 *     cpdef append(self, int64_t x):             # <<<<<<<<<<<<<<
 *         self._append(x)
 *         listener = <object>self._listener
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.Int64List.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  int64_t __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_int64_t(__pyx_arg_x); if (unlikely((__pyx_v_x == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1413, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int64List.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List_31append(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self), ((int64_t)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_31append(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int64_t __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("append", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_9Int64List_append(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Int64List.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1419
 *             listener._Modified()
 * 
 *     cpdef extend(self, x):             # <<<<<<<<<<<<<<
 *         for i in x:
 *             self._append(i)
 */

static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int64List_extend(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, PyObject *__pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int64_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("extend", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extend); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1419, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_9Int64List_34extend)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1419, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_x};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1419, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_x};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1419, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1419, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_x);
          __Pyx_GIVEREF(__pyx_v_x);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_x);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1419, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1420
 * 
 *     cpdef extend(self, x):
 *         for i in x:             # <<<<<<<<<<<<<<
 *             self._append(i)
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_x)) || PyTuple_CheckExact(__pyx_v_x)) {
    __pyx_t_1 = __pyx_v_x; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1420, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1420, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1420, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1420, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1420, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_7(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1420, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyrobuf_list.pyx":1421
 *     cpdef extend(self, x):
 *         for i in x:
 *             self._append(i)             # <<<<<<<<<<<<<<
 * 
 *         listener = <object>self._listener
 */
    __pyx_t_8 = __Pyx_PyInt_As_int64_t(__pyx_v_i); if (unlikely((__pyx_t_8 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1421, __pyx_L1_error)
    ((struct __pyx_vtabstruct_12pyrobuf_list_Int64List *)__pyx_v_self->__pyx_vtab)->_append(__pyx_v_self, __pyx_t_8);

    /* "pyrobuf_list.pyx":1420
 * 
 *     cpdef extend(self, x):
 *         for i in x:             # <<<<<<<<<<<<<<
 *             self._append(i)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1423
 *             self._append(i)
 * 
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1424
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1424, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = ((!__pyx_t_9) != 0);
  if (__pyx_t_10) {

    /* "pyrobuf_list.pyx":1425
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef insert(self, int i, int64_t x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1425, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1425, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1424
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1419
 *             listener._Modified()
 * 
 *     cpdef extend(self, x):             # <<<<<<<<<<<<<<
 *         for i in x:
 *             self._append(i)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.Int64List.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extend (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List_33extend(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_33extend(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("extend", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_9Int64List_extend(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Int64List.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1427
 *             listener._Modified()
 * 
 *     cpdef insert(self, int i, int64_t x):             # <<<<<<<<<<<<<<
 *         cdef int64_t *mem
 *         cdef size_t j
 */

static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int64List_insert(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int __pyx_v_i, int64_t __pyx_v_x, int __pyx_skip_dispatch) {
  int64_t *__pyx_v_mem;
  size_t __pyx_v_j;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  size_t __pyx_t_11;
  size_t __pyx_t_12;
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("insert", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_insert); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1427, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_9Int64List_36insert)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyInt_From_int64_t(__pyx_v_x); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1427, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1427, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1427, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
        __pyx_t_3 = 0;
        __pyx_t_4 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1427, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1431
 *         cdef size_t j
 * 
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_9 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_9) {

    /* "pyrobuf_list.pyx":1432
 * 
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":1431
 *         cdef size_t j
 * 
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":1434
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i > self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_10 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_10) {
  } else {
    __pyx_t_9 = __pyx_t_10;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_10 = ((((size_t)__pyx_v_i) > __pyx_v_self->_n_items) != 0);
  __pyx_t_9 = __pyx_t_10;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_9)) {

    /* "pyrobuf_list.pyx":1435
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         if self._n_items == self._size:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__64, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1435, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1435, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1434
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i > self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1437
 *             raise IndexError("list index out of range")
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <int64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int64_t))
 *             if not mem:
 */
  __pyx_t_9 = ((__pyx_v_self->_n_items == __pyx_v_self->_size) != 0);
  if (__pyx_t_9) {

    /* "pyrobuf_list.pyx":1438
 * 
 *         if self._n_items == self._size:
 *             mem = <int64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int64_t))             # <<<<<<<<<<<<<<
 *             if not mem:
 *                 raise MemoryError()
 */
    __pyx_v_mem = ((int64_t *)PyMem_Realloc(__pyx_v_self->_data, ((2 * __pyx_v_self->_size) * (sizeof(int64_t)))));

    /* "pyrobuf_list.pyx":1439
 *         if self._n_items == self._size:
 *             mem = <int64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int64_t))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    __pyx_t_9 = ((!(__pyx_v_mem != 0)) != 0);
    if (unlikely(__pyx_t_9)) {

      /* "pyrobuf_list.pyx":1440
 *             mem = <int64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int64_t))
 *             if not mem:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *             self._data = mem
 */
      PyErr_NoMemory(); __PYX_ERR(0, 1440, __pyx_L1_error)

      /* "pyrobuf_list.pyx":1439
 *         if self._n_items == self._size:
 *             mem = <int64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int64_t))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    }

    /* "pyrobuf_list.pyx":1442
 *                 raise MemoryError()
 * 
 *             self._data = mem             # <<<<<<<<<<<<<<
 *             self._size *= 2
 * 
 */
    __pyx_v_self->_data = __pyx_v_mem;

    /* "pyrobuf_list.pyx":1443
 * 
 *             self._data = mem
 *             self._size *= 2             # <<<<<<<<<<<<<<
 * 
 *         for j in range(self._n_items, <size_t>i, -1):
 */
    __pyx_v_self->_size = (__pyx_v_self->_size * 2);

    /* "pyrobuf_list.pyx":1437
 *             raise IndexError("list index out of range")
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <int64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int64_t))
 *             if not mem:
 */
  }

  /* "pyrobuf_list.pyx":1445
 *             self._size *= 2
 * 
 *         for j in range(self._n_items, <size_t>i, -1):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j - 1]
 * 
 */
  __pyx_t_11 = ((size_t)__pyx_v_i);
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = __pyx_v_self->_n_items + 1; __pyx_t_13 > __pyx_t_12 + 1; ) { __pyx_t_13-=1;
    __pyx_v_j = __pyx_t_13;

    /* "pyrobuf_list.pyx":1446
 * 
 *         for j in range(self._n_items, <size_t>i, -1):
 *             self._data[j] = self._data[j - 1]             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j - 1)]);
  }

  /* "pyrobuf_list.pyx":1448
 *             self._data[j] = self._data[j - 1]
 * 
 *         self._data[i] = x             # <<<<<<<<<<<<<<
 *         self._n_items += 1
 *         listener = <object>self._listener
 */
  (__pyx_v_self->_data[__pyx_v_i]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":1449
 * 
 *         self._data[i] = x
 *         self._n_items += 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items + 1);

  /* "pyrobuf_list.pyx":1450
 *         self._data[i] = x
 *         self._n_items += 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1451
 *         self._n_items += 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1451, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = ((!__pyx_t_9) != 0);
  if (__pyx_t_10) {

    /* "pyrobuf_list.pyx":1452
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef pop(self):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1452, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1452, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1452, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1451
 *         self._n_items += 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1427
 *             listener._Modified()
 * 
 *     cpdef insert(self, int i, int64_t x):             # <<<<<<<<<<<<<<
 *         cdef int64_t *mem
 *         cdef size_t j
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyrobuf_list.Int64List.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_i;
  int64_t __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("insert (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_i,&__pyx_n_s_x,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, 1); __PYX_ERR(0, 1427, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "insert") < 0)) __PYX_ERR(0, 1427, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_i = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1427, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_int64_t(values[1]); if (unlikely((__pyx_v_x == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1427, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1427, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int64List.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List_35insert(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self), __pyx_v_i, __pyx_v_x);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_35insert(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int __pyx_v_i, int64_t __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("insert", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_9Int64List_insert(__pyx_v_self, __pyx_v_i, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Int64List.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1454
 *             listener._Modified()
 * 
 *     cpdef pop(self):             # <<<<<<<<<<<<<<
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")
 */

static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int64List_pop(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("pop", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_9Int64List_38pop)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1454, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1454, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1455
 * 
 *     cpdef pop(self):
 *         if self._n_items == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("pop from empty list")
 * 
 */
  __pyx_t_5 = ((__pyx_v_self->_n_items == 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "pyrobuf_list.pyx":1456
 *     cpdef pop(self):
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__65, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1456, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1456, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1455
 * 
 *     cpdef pop(self):
 *         if self._n_items == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("pop from empty list")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1458
 *             raise IndexError("pop from empty list")
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":1459
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1460
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 *         return self._data[self._n_items]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1460, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = ((!__pyx_t_5) != 0);
  if (__pyx_t_6) {

    /* "pyrobuf_list.pyx":1461
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 *         return self._data[self._n_items]
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1461, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1461, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1461, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1460
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 *         return self._data[self._n_items]
 */
  }

  /* "pyrobuf_list.pyx":1462
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 *         return self._data[self._n_items]             # <<<<<<<<<<<<<<
 * 
 *     cpdef remove(self, int64_t x):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t((__pyx_v_self->_data[__pyx_v_self->_n_items])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1454
 *             listener._Modified()
 * 
 *     cpdef pop(self):             # <<<<<<<<<<<<<<
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyrobuf_list.Int64List.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pop (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List_37pop(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_37pop(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("pop", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_9Int64List_pop(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Int64List.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1464
 *         return self._data[self._n_items]
 * 
 *     cpdef remove(self, int64_t x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef size_t j
 */

static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_9Int64List_remove(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int64_t __pyx_v_x, int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  int __pyx_v_found;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("remove", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_remove); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1464, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_9Int64List_40remove)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1464, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1464, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1464, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1464, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1464, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1464, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1467
 *         cdef size_t i
 *         cdef size_t j
 *         cdef bint found = 0             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self._n_items):
 */
  __pyx_v_found = 0;

  /* "pyrobuf_list.pyx":1469
 *         cdef bint found = 0
 * 
 *         for i in range(self._n_items):             # <<<<<<<<<<<<<<
 *             if self._data[i] == x:
 *                 found = 1
 */
  __pyx_t_7 = __pyx_v_self->_n_items;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "pyrobuf_list.pyx":1470
 * 
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 found = 1
 *                 break
 */
    __pyx_t_10 = (((__pyx_v_self->_data[__pyx_v_i]) == __pyx_v_x) != 0);
    if (__pyx_t_10) {

      /* "pyrobuf_list.pyx":1471
 *         for i in range(self._n_items):
 *             if self._data[i] == x:
 *                 found = 1             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
      __pyx_v_found = 1;

      /* "pyrobuf_list.pyx":1472
 *             if self._data[i] == x:
 *                 found = 1
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         if found == 0:
 */
      goto __pyx_L4_break;

      /* "pyrobuf_list.pyx":1470
 * 
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 found = 1
 *                 break
 */
    }
  }
  __pyx_L4_break:;

  /* "pyrobuf_list.pyx":1474
 *                 break
 * 
 *         if found == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("x not in list")
 * 
 */
  __pyx_t_10 = ((__pyx_v_found == 0) != 0);
  if (unlikely(__pyx_t_10)) {

    /* "pyrobuf_list.pyx":1475
 * 
 *         if found == 0:
 *             raise ValueError("x not in list")             # <<<<<<<<<<<<<<
 * 
 *         for j in range(i, self._n_items):
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__66, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1475, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1474
 *                 break
 * 
 *         if found == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("x not in list")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1477
 *             raise ValueError("x not in list")
 * 
 *         for j in range(i, self._n_items):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j + 1]
 * 
 */
  __pyx_t_7 = __pyx_v_self->_n_items;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = __pyx_v_i; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_j = __pyx_t_9;

    /* "pyrobuf_list.pyx":1478
 * 
 *         for j in range(i, self._n_items):
 *             self._data[j] = self._data[j + 1]             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j + 1)]);
  }

  /* "pyrobuf_list.pyx":1480
 *             self._data[j] = self._data[j + 1]
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":1481
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1482
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1482, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = ((!__pyx_t_10) != 0);
  if (__pyx_t_11) {

    /* "pyrobuf_list.pyx":1483
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cdef void _append(self, int64_t x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1483, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1483, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1482
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1464
 *         return self._data[self._n_items]
 * 
 *     cpdef remove(self, int64_t x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef size_t j
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.Int64List.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  int64_t __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_int64_t(__pyx_arg_x); if (unlikely((__pyx_v_x == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1464, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Int64List.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List_39remove(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self), ((int64_t)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_39remove(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int64_t __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("remove", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_9Int64List_remove(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Int64List.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1485
 *             listener._Modified()
 * 
 *     cdef void _append(self, int64_t x):             # <<<<<<<<<<<<<<
 *         cdef int64_t *mem
 * 
 */

static void __pyx_f_12pyrobuf_list_9Int64List__append(struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, int64_t __pyx_v_x) {
  int64_t *__pyx_v_mem;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_append", 0);

  /* "pyrobuf_list.pyx":1488
 *         cdef int64_t *mem
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <int64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int64_t))
 *             if not mem:
 */
  __pyx_t_1 = ((__pyx_v_self->_n_items == __pyx_v_self->_size) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":1489
 * 
 *         if self._n_items == self._size:
 *             mem = <int64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int64_t))             # <<<<<<<<<<<<<<
 *             if not mem:
 *                 raise MemoryError()
 */
    __pyx_v_mem = ((int64_t *)PyMem_Realloc(__pyx_v_self->_data, ((2 * __pyx_v_self->_size) * (sizeof(int64_t)))));

    /* "pyrobuf_list.pyx":1490
 *         if self._n_items == self._size:
 *             mem = <int64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int64_t))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    __pyx_t_1 = ((!(__pyx_v_mem != 0)) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "pyrobuf_list.pyx":1491
 *             mem = <int64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int64_t))
 *             if not mem:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *             self._data = mem
 */
      PyErr_NoMemory(); __PYX_ERR(0, 1491, __pyx_L1_error)

      /* "pyrobuf_list.pyx":1490
 *         if self._n_items == self._size:
 *             mem = <int64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int64_t))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    }

    /* "pyrobuf_list.pyx":1493
 *                 raise MemoryError()
 * 
 *             self._data = mem             # <<<<<<<<<<<<<<
 *             self._size *= 2
 * 
 */
    __pyx_v_self->_data = __pyx_v_mem;

    /* "pyrobuf_list.pyx":1494
 * 
 *             self._data = mem
 *             self._size *= 2             # <<<<<<<<<<<<<<
 * 
 *         self._data[self._n_items] = x
 */
    __pyx_v_self->_size = (__pyx_v_self->_size * 2);

    /* "pyrobuf_list.pyx":1488
 *         cdef int64_t *mem
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <int64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(int64_t))
 *             if not mem:
 */
  }

  /* "pyrobuf_list.pyx":1496
 *             self._size *= 2
 * 
 *         self._data[self._n_items] = x             # <<<<<<<<<<<<<<
 *         self._n_items += 1
 * 
 */
  (__pyx_v_self->_data[__pyx_v_self->_n_items]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":1497
 * 
 *         self._data[self._n_items] = x
 *         self._n_items += 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items + 1);

  /* "pyrobuf_list.pyx":1485
 *             listener._Modified()
 * 
 *     cdef void _append(self, int64_t x):             # <<<<<<<<<<<<<<
 *         cdef int64_t *mem
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("pyrobuf_list.Int64List._append", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_42__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_42__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List_41__reduce_cython__(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_41__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__67, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Int64List.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_44__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_9Int64List_44__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_9Int64List_43__setstate_cython__(((struct __pyx_obj_12pyrobuf_list_Int64List *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_9Int64List_43__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Int64List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__68, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Int64List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1503
 * cdef class Uint64List:
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):             # <<<<<<<<<<<<<<
 *         self._data = <uint64_t *>PyMem_Malloc(size * sizeof(uint64_t))
 *         if not self._data:
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_10Uint64List_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12pyrobuf_list_10Uint64List_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_size;
  PyObject *__pyx_v_listener = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_size,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_listener);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 1503, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_size = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_size == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1503, __pyx_L3_error)
    } else {
      __pyx_v_size = ((size_t)16);
    }
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1503, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List___cinit__(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self), __pyx_v_size, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_10Uint64List___cinit__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, size_t __pyx_v_size, PyObject *__pyx_v_listener) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "pyrobuf_list.pyx":1504
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <uint64_t *>PyMem_Malloc(size * sizeof(uint64_t))             # <<<<<<<<<<<<<<
 *         if not self._data:
 *             raise MemoryError()
 */
  __pyx_v_self->_data = ((uint64_t *)PyMem_Malloc((__pyx_v_size * (sizeof(uint64_t)))));

  /* "pyrobuf_list.pyx":1505
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <uint64_t *>PyMem_Malloc(size * sizeof(uint64_t))
 *         if not self._data:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_self->_data != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":1506
 *         self._data = <uint64_t *>PyMem_Malloc(size * sizeof(uint64_t))
 *         if not self._data:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         self._n_items = 0
 */
    PyErr_NoMemory(); __PYX_ERR(0, 1506, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1505
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <uint64_t *>PyMem_Malloc(size * sizeof(uint64_t))
 *         if not self._data:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1508
 *             raise MemoryError()
 * 
 *         self._n_items = 0             # <<<<<<<<<<<<<<
 *         self._size = size
 *         if listener is None:
 */
  __pyx_v_self->_n_items = 0;

  /* "pyrobuf_list.pyx":1509
 * 
 *         self._n_items = 0
 *         self._size = size             # <<<<<<<<<<<<<<
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener
 */
  __pyx_v_self->_size = __pyx_v_size;

  /* "pyrobuf_list.pyx":1510
 *         self._n_items = 0
 *         self._size = size
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
  __pyx_t_1 = (__pyx_v_listener == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":1511
 *         self._size = size
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener             # <<<<<<<<<<<<<<
 *         else:
 *             self._listener = <PyObject *>listener
 */
    __pyx_v_self->_listener = ((PyObject *)null_listener);

    /* "pyrobuf_list.pyx":1510
 *         self._n_items = 0
 *         self._size = size
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
    goto __pyx_L4;
  }

  /* "pyrobuf_list.pyx":1513
 *             self._listener = <PyObject *>null_listener
 *         else:
 *             self._listener = <PyObject *>listener             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  /*else*/ {
    __pyx_v_self->_listener = ((PyObject *)__pyx_v_listener);
  }
  __pyx_L4:;

  /* "pyrobuf_list.pyx":1503
 * cdef class Uint64List:
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):             # <<<<<<<<<<<<<<
 *         self._data = <uint64_t *>PyMem_Malloc(size * sizeof(uint64_t))
 *         if not self._data:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1515
 *             self._listener = <PyObject *>listener
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self._data)
 * 
 */

/* Python wrapper */
static void __pyx_pw_12pyrobuf_list_10Uint64List_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12pyrobuf_list_10Uint64List_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_12pyrobuf_list_10Uint64List_2__dealloc__(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12pyrobuf_list_10Uint64List_2__dealloc__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "pyrobuf_list.pyx":1516
 * 
 *     def __dealloc__(self):
 *         PyMem_Free(self._data)             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, uint64_t x):
 */
  PyMem_Free(__pyx_v_self->_data);

  /* "pyrobuf_list.pyx":1515
 *             self._listener = <PyObject *>listener
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self._data)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "pyrobuf_list.pyx":1518
 *         PyMem_Free(self._data)
 * 
 *     def __contains__(self, uint64_t x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self._n_items):
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_10Uint64List_5__contains__(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static int __pyx_pw_12pyrobuf_list_10Uint64List_5__contains__(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  uint64_t __pyx_v_x;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_uint64_t(__pyx_arg_x); if (unlikely((__pyx_v_x == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1518, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List_4__contains__(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self), ((uint64_t)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_10Uint64List_4__contains__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, uint64_t __pyx_v_x) {
  size_t __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "pyrobuf_list.pyx":1520
 *     def __contains__(self, uint64_t x):
 *         cdef size_t i
 *         for i in range(self._n_items):             # <<<<<<<<<<<<<<
 *             if self._data[i] == x:
 *                 return True
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "pyrobuf_list.pyx":1521
 *         cdef size_t i
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */
    __pyx_t_4 = (((__pyx_v_self->_data[__pyx_v_i]) == __pyx_v_x) != 0);
    if (__pyx_t_4) {

      /* "pyrobuf_list.pyx":1522
 *         for i in range(self._n_items):
 *             if self._data[i] == x:
 *                 return True             # <<<<<<<<<<<<<<
 * 
 *         return False
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "pyrobuf_list.pyx":1521
 *         cdef size_t i
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */
    }
  }

  /* "pyrobuf_list.pyx":1524
 *                 return True
 * 
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     def __delitem__(self, int i):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1518
 *         PyMem_Free(self._data)
 * 
 *     def __contains__(self, uint64_t x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self._n_items):
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1526
 *         return False
 * 
 *     def __delitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_10Uint64List_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static int __pyx_pw_12pyrobuf_list_10Uint64List_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1526, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List_6__delitem__(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self), ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_10Uint64List_6__delitem__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, int __pyx_v_i) {
  size_t __pyx_v_j;
  PyObject *__pyx_v_listener = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__delitem__", 0);

  /* "pyrobuf_list.pyx":1527
 * 
 *     def __delitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":1528
 *     def __delitem__(self, int i):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":1527
 * 
 *     def __delitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":1530
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":1531
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t j
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__69, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1531, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1530
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1534
 * 
 *         cdef size_t j
 *         for j in range(i, self._n_items):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j + 1]
 * 
 */
  __pyx_t_4 = __pyx_v_self->_n_items;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = __pyx_v_i; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_j = __pyx_t_6;

    /* "pyrobuf_list.pyx":1535
 *         cdef size_t j
 *         for j in range(i, self._n_items):
 *             self._data[j] = self._data[j + 1]             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j + 1)]);
  }

  /* "pyrobuf_list.pyx":1537
 *             self._data[j] = self._data[j + 1]
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":1538
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":1539
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1539, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1539, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":1540
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, int i):
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    if (__pyx_t_8) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1540, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1540, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":1539
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1526
 *         return False
 * 
 *     def __delitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1542
 *             listener._Modified()
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1542, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List_8__getitem__(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self), ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_8__getitem__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, int __pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pyrobuf_list.pyx":1543
 * 
 *     def __getitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":1544
 *     def __getitem__(self, int i):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":1543
 * 
 *     def __getitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":1546
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":1547
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         return self._data[i]
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__70, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1547, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1547, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1546
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1549
 *             raise IndexError("list index out of range")
 * 
 *         return self._data[i]             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_uint64_t((__pyx_v_self->_data[__pyx_v_i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1542
 *             listener._Modified()
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_12pyrobuf_list_10Uint64List_12generator6(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "pyrobuf_list.pyx":1551
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_11__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_11__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List_10__iter__(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_10__iter__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self) {
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_6___iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_6___iter__ *)__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_6___iter__(__pyx_ptype_12pyrobuf_list___pyx_scope_struct_6___iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_6___iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1551, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_12pyrobuf_list_10Uint64List_12generator6, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_Uint64List___iter, __pyx_n_s_pyrobuf_list); if (unlikely(!gen)) __PYX_ERR(0, 1551, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_12pyrobuf_list_10Uint64List_12generator6(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_6___iter__ *__pyx_cur_scope = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_6___iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1551, __pyx_L1_error)

  /* "pyrobuf_list.pyx":1552
 * 
 *     def __iter__(self):
 *         cdef size_t idx = 0             # <<<<<<<<<<<<<<
 *         while idx < self._n_items:
 *             yield self._data[idx]
 */
  __pyx_cur_scope->__pyx_v_idx = 0;

  /* "pyrobuf_list.pyx":1553
 *     def __iter__(self):
 *         cdef size_t idx = 0
 *         while idx < self._n_items:             # <<<<<<<<<<<<<<
 *             yield self._data[idx]
 *             idx += 1
 */
  while (1) {
    __pyx_t_1 = ((__pyx_cur_scope->__pyx_v_idx < __pyx_cur_scope->__pyx_v_self->_n_items) != 0);
    if (!__pyx_t_1) break;

    /* "pyrobuf_list.pyx":1554
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 *             yield self._data[idx]             # <<<<<<<<<<<<<<
 *             idx += 1
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_uint64_t((__pyx_cur_scope->__pyx_v_self->_data[__pyx_cur_scope->__pyx_v_idx])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1554, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1555
 *         while idx < self._n_items:
 *             yield self._data[idx]
 *             idx += 1             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
    __pyx_cur_scope->__pyx_v_idx = (__pyx_cur_scope->__pyx_v_idx + 1);
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "pyrobuf_list.pyx":1551
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1557
 *             idx += 1
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._n_items
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_12pyrobuf_list_10Uint64List_14__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_12pyrobuf_list_10Uint64List_14__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List_13__len__(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_12pyrobuf_list_10Uint64List_13__len__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pyrobuf_list.pyx":1558
 * 
 *     def __len__(self):
 *         return self._n_items             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = __pyx_v_self->_n_items;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1557
 *             idx += 1
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._n_items
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1560
 *         return self._n_items
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(list(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_16__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_16__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List_15__repr__(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_15__repr__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pyrobuf_list.pyx":1561
 * 
 *     def __repr__(self):
 *         return repr(list(self))             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, int i, uint64_t x):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1560
 *         return self._n_items
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(list(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1563
 *         return repr(list(self))
 * 
 *     def __setitem__(self, int i, uint64_t x):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_10Uint64List_18__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i, PyObject *__pyx_arg_x); /*proto*/
static int __pyx_pw_12pyrobuf_list_10Uint64List_18__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i, PyObject *__pyx_arg_x) {
  int __pyx_v_i;
  uint64_t __pyx_v_x;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1563, __pyx_L3_error)
  }
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_uint64_t(__pyx_arg_x); if (unlikely((__pyx_v_x == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1563, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List_17__setitem__(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self), ((int)__pyx_v_i), ((uint64_t)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_10Uint64List_17__setitem__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, int __pyx_v_i, uint64_t __pyx_v_x) {
  PyObject *__pyx_v_listener = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "pyrobuf_list.pyx":1564
 * 
 *     def __setitem__(self, int i, uint64_t x):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":1565
 *     def __setitem__(self, int i, uint64_t x):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":1564
 * 
 *     def __setitem__(self, int i, uint64_t x):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":1567
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":1568
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__71, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1568, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1567
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1570
 *             raise IndexError("list index out of range")
 * 
 *         self._data[i] = x             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  (__pyx_v_self->_data[__pyx_v_i]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":1571
 * 
 *         self._data[i] = x
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":1572
 *         self._data[i] = x
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1572, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":1573
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1573, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1573, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":1572
 *         self._data[i] = x
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1563
 *         return repr(list(self))
 * 
 *     def __setitem__(self, int i, uint64_t x):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1575
 *             listener._Modified()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(list(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_20__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_20__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List_19__str__(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_19__str__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pyrobuf_list.pyx":1576
 * 
 *     def __str__(self):
 *         return str(list(self))             # <<<<<<<<<<<<<<
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1575
 *             listener._Modified()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(list(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1578
 *         return str(list(self))
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t itemsize = sizeof(uint64_t)
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_12pyrobuf_list_10Uint64List_22__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_12pyrobuf_list_10Uint64List_22__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List_21__getbuffer__(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_10Uint64List_21__getbuffer__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags) {
  Py_ssize_t __pyx_v_itemsize;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  if (__pyx_v_buffer == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_buffer->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_buffer->obj);

  /* "pyrobuf_list.pyx":1580
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 * 
 *         cdef Py_ssize_t itemsize = sizeof(uint64_t)             # <<<<<<<<<<<<<<
 *         self.shape[0] = self._n_items
 *         self.strides[0] = 1
 */
  __pyx_v_itemsize = (sizeof(uint64_t));

  /* "pyrobuf_list.pyx":1581
 * 
 *         cdef Py_ssize_t itemsize = sizeof(uint64_t)
 *         self.shape[0] = self._n_items             # <<<<<<<<<<<<<<
 *         self.strides[0] = 1
 * 
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  (__pyx_v_self->shape[0]) = __pyx_t_1;

  /* "pyrobuf_list.pyx":1582
 *         cdef Py_ssize_t itemsize = sizeof(uint64_t)
 *         self.shape[0] = self._n_items
 *         self.strides[0] = 1             # <<<<<<<<<<<<<<
 * 
 *         buffer.buf = <char *>(self._data)
 */
  (__pyx_v_self->strides[0]) = 1;

  /* "pyrobuf_list.pyx":1584
 *         self.strides[0] = 1
 * 
 *         buffer.buf = <char *>(self._data)             # <<<<<<<<<<<<<<
 *         buffer.format = 'K'
 *         buffer.internal = NULL                  # see References
 */
  __pyx_v_buffer->buf = ((char *)__pyx_v_self->_data);

  /* "pyrobuf_list.pyx":1585
 * 
 *         buffer.buf = <char *>(self._data)
 *         buffer.format = 'K'             # <<<<<<<<<<<<<<
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 */
  __pyx_v_buffer->format = ((char *)"K");

  /* "pyrobuf_list.pyx":1586
 *         buffer.buf = <char *>(self._data)
 *         buffer.format = 'K'
 *         buffer.internal = NULL                  # see References             # <<<<<<<<<<<<<<
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items
 */
  __pyx_v_buffer->internal = NULL;

  /* "pyrobuf_list.pyx":1587
 *         buffer.format = 'K'
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize             # <<<<<<<<<<<<<<
 *         buffer.len = self._n_items
 *         buffer.ndim = 1
 */
  __pyx_v_buffer->itemsize = __pyx_v_itemsize;

  /* "pyrobuf_list.pyx":1588
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items             # <<<<<<<<<<<<<<
 *         buffer.ndim = 1
 *         buffer.obj = self
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  __pyx_v_buffer->len = __pyx_t_1;

  /* "pyrobuf_list.pyx":1589
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items
 *         buffer.ndim = 1             # <<<<<<<<<<<<<<
 *         buffer.obj = self
 *         buffer.readonly = 1
 */
  __pyx_v_buffer->ndim = 1;

  /* "pyrobuf_list.pyx":1590
 *         buffer.len = self._n_items
 *         buffer.ndim = 1
 *         buffer.obj = self             # <<<<<<<<<<<<<<
 *         buffer.readonly = 1
 *         buffer.shape = self.shape
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_buffer->obj);
  __Pyx_DECREF(__pyx_v_buffer->obj);
  __pyx_v_buffer->obj = ((PyObject *)__pyx_v_self);

  /* "pyrobuf_list.pyx":1591
 *         buffer.ndim = 1
 *         buffer.obj = self
 *         buffer.readonly = 1             # <<<<<<<<<<<<<<
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides
 */
  __pyx_v_buffer->readonly = 1;

  /* "pyrobuf_list.pyx":1592
 *         buffer.obj = self
 *         buffer.readonly = 1
 *         buffer.shape = self.shape             # <<<<<<<<<<<<<<
 *         buffer.strides = self.strides
 *         buffer.suboffsets = NULL                # for pointer arrays only
 */
  __pyx_t_2 = __pyx_v_self->shape;
  __pyx_v_buffer->shape = __pyx_t_2;

  /* "pyrobuf_list.pyx":1593
 *         buffer.readonly = 1
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides             # <<<<<<<<<<<<<<
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 */
  __pyx_t_2 = __pyx_v_self->strides;
  __pyx_v_buffer->strides = __pyx_t_2;

  /* "pyrobuf_list.pyx":1594
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides
 *         buffer.suboffsets = NULL                # for pointer arrays only             # <<<<<<<<<<<<<<
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):
 */
  __pyx_v_buffer->suboffsets = NULL;

  /* "pyrobuf_list.pyx":1578
 *         return str(list(self))
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t itemsize = sizeof(uint64_t)
 */

  /* function exit code */
  __pyx_r = 0;
  if (__pyx_v_buffer->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_buffer->obj);
    __Pyx_DECREF(__pyx_v_buffer->obj); __pyx_v_buffer->obj = 0;
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1596
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_12pyrobuf_list_10Uint64List_24__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer); /*proto*/
static CYTHON_UNUSED void __pyx_pw_12pyrobuf_list_10Uint64List_24__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_12pyrobuf_list_10Uint64List_23__releasebuffer__(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12pyrobuf_list_10Uint64List_23__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "pyrobuf_list.pyx":1599
 *         pass
 * 
 *     def __richcmp__(self, other, op):             # <<<<<<<<<<<<<<
 *         if op == Py_EQ:
 *             return self._equal(other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_26__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_26__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op) {
  PyObject *__pyx_v_op = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_v_op = __Pyx_PyInt_From_int(__pyx_arg_op); if (unlikely(!__pyx_v_op)) __PYX_ERR(0, 1599, __pyx_L3_error)
  __Pyx_GOTREF(__pyx_v_op);
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List_25__richcmp__(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((PyObject *)__pyx_v_op));

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_op);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_25__richcmp__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "pyrobuf_list.pyx":1600
 * 
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:             # <<<<<<<<<<<<<<
 *             return self._equal(other)
 *         elif op == Py_NE:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(Py_EQ); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_op, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1600, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1600, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "pyrobuf_list.pyx":1601
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:
 *             return self._equal(other)             # <<<<<<<<<<<<<<
 *         elif op == Py_NE:
 *             return not self._equal(other)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = ((struct __pyx_vtabstruct_12pyrobuf_list_Uint64List *)__pyx_v_self->__pyx_vtab)->_equal(__pyx_v_self, __pyx_v_other, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pyrobuf_list.pyx":1600
 * 
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:             # <<<<<<<<<<<<<<
 *             return self._equal(other)
 *         elif op == Py_NE:
 */
  }

  /* "pyrobuf_list.pyx":1602
 *         if op == Py_EQ:
 *             return self._equal(other)
 *         elif op == Py_NE:             # <<<<<<<<<<<<<<
 *             return not self._equal(other)
 *         else:
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(Py_NE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_op, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1602, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1602, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(__pyx_t_3)) {

    /* "pyrobuf_list.pyx":1603
 *             return self._equal(other)
 *         elif op == Py_NE:
 *             return not self._equal(other)             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError('Operation not supported for lists')
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((struct __pyx_vtabstruct_12pyrobuf_list_Uint64List *)__pyx_v_self->__pyx_vtab)->_equal(__pyx_v_self, __pyx_v_other, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1603, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1603, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_3)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1603, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pyrobuf_list.pyx":1602
 *         if op == Py_EQ:
 *             return self._equal(other)
 *         elif op == Py_NE:             # <<<<<<<<<<<<<<
 *             return not self._equal(other)
 *         else:
 */
  }

  /* "pyrobuf_list.pyx":1605
 *             return not self._equal(other)
 *         else:
 *             raise TypeError('Operation not supported for lists')             # <<<<<<<<<<<<<<
 * 
 *     def __neq__(self, other):
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__72, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1605, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1605, __pyx_L1_error)
  }

  /* "pyrobuf_list.pyx":1599
 *         pass
 * 
 *     def __richcmp__(self, other, op):             # <<<<<<<<<<<<<<
 *         if op == Py_EQ:
 *             return self._equal(other)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1607
 *             raise TypeError('Operation not supported for lists')
 * 
 *     def __neq__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_28__neq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_28__neq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__neq__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List_27__neq__(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_27__neq__(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__neq__", 0);

  /* "pyrobuf_list.pyx":1608
 * 
 *     def __neq__(self, other):
 *         return not self == other             # <<<<<<<<<<<<<<
 * 
 *     cpdef _equal(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_self), __pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1608, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1608, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1607
 *             raise TypeError('Operation not supported for lists')
 * 
 *     def __neq__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.__neq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1610
 *         return not self == other
 * 
 *     cpdef _equal(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             if len(self) != len(other):
 */

static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint64List__equal(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_v_y = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *(*__pyx_t_14)(PyObject *);
  __Pyx_RefNannySetupContext("_equal", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_equal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1610, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint64List_30_equal)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_other); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1610, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_other};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1610, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_other};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1610, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1610, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_other);
          __Pyx_GIVEREF(__pyx_v_other);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_other);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1610, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1611
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    /*try:*/ {

      /* "pyrobuf_list.pyx":1612
 *     cpdef _equal(self, other):
 *         try:
 *             if len(self) != len(other):             # <<<<<<<<<<<<<<
 *                 return False
 *         except TypeError:
 */
      __pyx_t_9 = PyObject_Length(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1612, __pyx_L3_error)
      __pyx_t_10 = PyObject_Length(__pyx_v_other); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1612, __pyx_L3_error)
      __pyx_t_11 = ((__pyx_t_9 != __pyx_t_10) != 0);
      if (__pyx_t_11) {

        /* "pyrobuf_list.pyx":1613
 *         try:
 *             if len(self) != len(other):
 *                 return False             # <<<<<<<<<<<<<<
 *         except TypeError:
 *             return False
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_False);
        __pyx_r = Py_False;
        goto __pyx_L7_try_return;

        /* "pyrobuf_list.pyx":1612
 *     cpdef _equal(self, other):
 *         try:
 *             if len(self) != len(other):             # <<<<<<<<<<<<<<
 *                 return False
 *         except TypeError:
 */
      }

      /* "pyrobuf_list.pyx":1611
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
    }
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1614
 *             if len(self) != len(other):
 *                 return False
 *         except TypeError:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
    __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_12) {
      __Pyx_AddTraceback("pyrobuf_list.Uint64List._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 1614, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_3);

      /* "pyrobuf_list.pyx":1615
 *                 return False
 *         except TypeError:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         for x, y in zip(self, other):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "pyrobuf_list.pyx":1611
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "pyrobuf_list.pyx":1617
 *             return False
 * 
 *         for x, y in zip(self, other):             # <<<<<<<<<<<<<<
 *             if x != y:
 *                 return False
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));
  __Pyx_INCREF(__pyx_v_other);
  __Pyx_GIVEREF(__pyx_v_other);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_other);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_10 = 0;
    __pyx_t_13 = NULL;
  } else {
    __pyx_t_10 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_13 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1617, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_13)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 1617, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1617, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 1617, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1617, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_13(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1617, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1617, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1617, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1617, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1617, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_14 = Py_TYPE(__pyx_t_4)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_5 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_5)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_4), 2) < 0) __PYX_ERR(0, 1617, __pyx_L1_error)
      __pyx_t_14 = NULL;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L15_unpacking_done;
      __pyx_L14_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_14 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1617, __pyx_L1_error)
      __pyx_L15_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_x, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_y, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "pyrobuf_list.pyx":1618
 * 
 *         for x, y in zip(self, other):
 *             if x != y:             # <<<<<<<<<<<<<<
 *                 return False
 * 
 */
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_x, __pyx_v_y, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1618, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1618, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_11) {

      /* "pyrobuf_list.pyx":1619
 *         for x, y in zip(self, other):
 *             if x != y:
 *                 return False             # <<<<<<<<<<<<<<
 * 
 *         return True
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "pyrobuf_list.pyx":1618
 * 
 *         for x, y in zip(self, other):
 *             if x != y:             # <<<<<<<<<<<<<<
 *                 return False
 * 
 */
    }

    /* "pyrobuf_list.pyx":1617
 *             return False
 * 
 *         for x, y in zip(self, other):             # <<<<<<<<<<<<<<
 *             if x != y:
 *                 return False
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":1621
 *                 return False
 * 
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     cpdef append(self, uint64_t x):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1610
 *         return not self == other
 * 
 *     cpdef _equal(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             if len(self) != len(other):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_y);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_equal (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List_29_equal(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_29_equal(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_equal", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_10Uint64List__equal(__pyx_v_self, __pyx_v_other, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1623
 *         return True
 * 
 *     cpdef append(self, uint64_t x):             # <<<<<<<<<<<<<<
 *         self._append(x)
 *         listener = <object>self._listener
 */

static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint64List_append(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, uint64_t __pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("append", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_append); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1623, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint64List_32append)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_uint64_t(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1623, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1623, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1623, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1623, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1623, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1623, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1624
 * 
 *     cpdef append(self, uint64_t x):
 *         self._append(x)             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  ((struct __pyx_vtabstruct_12pyrobuf_list_Uint64List *)__pyx_v_self->__pyx_vtab)->_append(__pyx_v_self, __pyx_v_x);

  /* "pyrobuf_list.pyx":1625
 *     cpdef append(self, uint64_t x):
 *         self._append(x)
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1626
 *         self._append(x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1626, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = ((!__pyx_t_7) != 0);
  if (__pyx_t_8) {

    /* "pyrobuf_list.pyx":1627
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef extend(self, x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1627, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1627, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1627, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1626
 *         self._append(x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1623
 *         return True
 * 
 *     cpdef append(self, uint64_t x):             # <<<<<<<<<<<<<<
 *         self._append(x)
 *         listener = <object>self._listener
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  uint64_t __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_uint64_t(__pyx_arg_x); if (unlikely((__pyx_v_x == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1623, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List_31append(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self), ((uint64_t)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_31append(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, uint64_t __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("append", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_10Uint64List_append(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1629
 *             listener._Modified()
 * 
 *     cpdef extend(self, x):             # <<<<<<<<<<<<<<
 *         for i in x:
 *             self._append(i)
 */

static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint64List_extend(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, PyObject *__pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  uint64_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("extend", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extend); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1629, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint64List_34extend)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1629, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_x};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1629, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_x};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1629, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1629, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_x);
          __Pyx_GIVEREF(__pyx_v_x);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_x);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1629, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1630
 * 
 *     cpdef extend(self, x):
 *         for i in x:             # <<<<<<<<<<<<<<
 *             self._append(i)
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_x)) || PyTuple_CheckExact(__pyx_v_x)) {
    __pyx_t_1 = __pyx_v_x; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1630, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1630, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1630, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1630, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1630, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1630, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_7(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1630, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyrobuf_list.pyx":1631
 *     cpdef extend(self, x):
 *         for i in x:
 *             self._append(i)             # <<<<<<<<<<<<<<
 * 
 *         listener = <object>self._listener
 */
    __pyx_t_8 = __Pyx_PyInt_As_uint64_t(__pyx_v_i); if (unlikely((__pyx_t_8 == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1631, __pyx_L1_error)
    ((struct __pyx_vtabstruct_12pyrobuf_list_Uint64List *)__pyx_v_self->__pyx_vtab)->_append(__pyx_v_self, __pyx_t_8);

    /* "pyrobuf_list.pyx":1630
 * 
 *     cpdef extend(self, x):
 *         for i in x:             # <<<<<<<<<<<<<<
 *             self._append(i)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1633
 *             self._append(i)
 * 
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1634
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1634, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = ((!__pyx_t_9) != 0);
  if (__pyx_t_10) {

    /* "pyrobuf_list.pyx":1635
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef insert(self, int i, uint64_t x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1635, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1635, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1634
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1629
 *             listener._Modified()
 * 
 *     cpdef extend(self, x):             # <<<<<<<<<<<<<<
 *         for i in x:
 *             self._append(i)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extend (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List_33extend(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_33extend(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("extend", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_10Uint64List_extend(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1637
 *             listener._Modified()
 * 
 *     cpdef insert(self, int i, uint64_t x):             # <<<<<<<<<<<<<<
 *         cdef uint64_t *mem
 *         cdef size_t j
 */

static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint64List_insert(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, int __pyx_v_i, uint64_t __pyx_v_x, int __pyx_skip_dispatch) {
  uint64_t *__pyx_v_mem;
  size_t __pyx_v_j;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  size_t __pyx_t_11;
  size_t __pyx_t_12;
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("insert", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_insert); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1637, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint64List_36insert)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1637, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyInt_From_uint64_t(__pyx_v_x); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1637, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1637, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1637, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1637, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
        __pyx_t_3 = 0;
        __pyx_t_4 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1637, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1641
 *         cdef size_t j
 * 
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_9 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_9) {

    /* "pyrobuf_list.pyx":1642
 * 
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":1641
 *         cdef size_t j
 * 
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":1644
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i > self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_10 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_10) {
  } else {
    __pyx_t_9 = __pyx_t_10;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_10 = ((((size_t)__pyx_v_i) > __pyx_v_self->_n_items) != 0);
  __pyx_t_9 = __pyx_t_10;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_9)) {

    /* "pyrobuf_list.pyx":1645
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         if self._n_items == self._size:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__73, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1645, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1645, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1644
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i > self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1647
 *             raise IndexError("list index out of range")
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <uint64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint64_t))
 *             if not mem:
 */
  __pyx_t_9 = ((__pyx_v_self->_n_items == __pyx_v_self->_size) != 0);
  if (__pyx_t_9) {

    /* "pyrobuf_list.pyx":1648
 * 
 *         if self._n_items == self._size:
 *             mem = <uint64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint64_t))             # <<<<<<<<<<<<<<
 *             if not mem:
 *                 raise MemoryError()
 */
    __pyx_v_mem = ((uint64_t *)PyMem_Realloc(__pyx_v_self->_data, ((2 * __pyx_v_self->_size) * (sizeof(uint64_t)))));

    /* "pyrobuf_list.pyx":1649
 *         if self._n_items == self._size:
 *             mem = <uint64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint64_t))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    __pyx_t_9 = ((!(__pyx_v_mem != 0)) != 0);
    if (unlikely(__pyx_t_9)) {

      /* "pyrobuf_list.pyx":1650
 *             mem = <uint64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint64_t))
 *             if not mem:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *             self._data = mem
 */
      PyErr_NoMemory(); __PYX_ERR(0, 1650, __pyx_L1_error)

      /* "pyrobuf_list.pyx":1649
 *         if self._n_items == self._size:
 *             mem = <uint64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint64_t))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    }

    /* "pyrobuf_list.pyx":1652
 *                 raise MemoryError()
 * 
 *             self._data = mem             # <<<<<<<<<<<<<<
 *             self._size *= 2
 * 
 */
    __pyx_v_self->_data = __pyx_v_mem;

    /* "pyrobuf_list.pyx":1653
 * 
 *             self._data = mem
 *             self._size *= 2             # <<<<<<<<<<<<<<
 * 
 *         for j in range(self._n_items, <size_t>i, -1):
 */
    __pyx_v_self->_size = (__pyx_v_self->_size * 2);

    /* "pyrobuf_list.pyx":1647
 *             raise IndexError("list index out of range")
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <uint64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint64_t))
 *             if not mem:
 */
  }

  /* "pyrobuf_list.pyx":1655
 *             self._size *= 2
 * 
 *         for j in range(self._n_items, <size_t>i, -1):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j - 1]
 * 
 */
  __pyx_t_11 = ((size_t)__pyx_v_i);
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = __pyx_v_self->_n_items + 1; __pyx_t_13 > __pyx_t_12 + 1; ) { __pyx_t_13-=1;
    __pyx_v_j = __pyx_t_13;

    /* "pyrobuf_list.pyx":1656
 * 
 *         for j in range(self._n_items, <size_t>i, -1):
 *             self._data[j] = self._data[j - 1]             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j - 1)]);
  }

  /* "pyrobuf_list.pyx":1658
 *             self._data[j] = self._data[j - 1]
 * 
 *         self._data[i] = x             # <<<<<<<<<<<<<<
 *         self._n_items += 1
 *         listener = <object>self._listener
 */
  (__pyx_v_self->_data[__pyx_v_i]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":1659
 * 
 *         self._data[i] = x
 *         self._n_items += 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items + 1);

  /* "pyrobuf_list.pyx":1660
 *         self._data[i] = x
 *         self._n_items += 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1661
 *         self._n_items += 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1661, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = ((!__pyx_t_9) != 0);
  if (__pyx_t_10) {

    /* "pyrobuf_list.pyx":1662
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef pop(self):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1662, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1662, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1662, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1661
 *         self._n_items += 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1637
 *             listener._Modified()
 * 
 *     cpdef insert(self, int i, uint64_t x):             # <<<<<<<<<<<<<<
 *         cdef uint64_t *mem
 *         cdef size_t j
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_i;
  uint64_t __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("insert (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_i,&__pyx_n_s_x,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, 1); __PYX_ERR(0, 1637, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "insert") < 0)) __PYX_ERR(0, 1637, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_i = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1637, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_uint64_t(values[1]); if (unlikely((__pyx_v_x == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1637, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1637, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List_35insert(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self), __pyx_v_i, __pyx_v_x);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_35insert(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, int __pyx_v_i, uint64_t __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("insert", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_10Uint64List_insert(__pyx_v_self, __pyx_v_i, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1664
 *             listener._Modified()
 * 
 *     cpdef pop(self):             # <<<<<<<<<<<<<<
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")
 */

static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint64List_pop(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("pop", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1664, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint64List_38pop)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1664, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1664, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1665
 * 
 *     cpdef pop(self):
 *         if self._n_items == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("pop from empty list")
 * 
 */
  __pyx_t_5 = ((__pyx_v_self->_n_items == 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "pyrobuf_list.pyx":1666
 *     cpdef pop(self):
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__74, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1666, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1666, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1665
 * 
 *     cpdef pop(self):
 *         if self._n_items == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("pop from empty list")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1668
 *             raise IndexError("pop from empty list")
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":1669
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1670
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 *         return self._data[self._n_items]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1670, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = ((!__pyx_t_5) != 0);
  if (__pyx_t_6) {

    /* "pyrobuf_list.pyx":1671
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 *         return self._data[self._n_items]
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1671, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1671, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1671, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1670
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 *         return self._data[self._n_items]
 */
  }

  /* "pyrobuf_list.pyx":1672
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 *         return self._data[self._n_items]             # <<<<<<<<<<<<<<
 * 
 *     cpdef remove(self, uint64_t x):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_uint64_t((__pyx_v_self->_data[__pyx_v_self->_n_items])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1664
 *             listener._Modified()
 * 
 *     cpdef pop(self):             # <<<<<<<<<<<<<<
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pop (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List_37pop(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_37pop(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("pop", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_10Uint64List_pop(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1674
 *         return self._data[self._n_items]
 * 
 *     cpdef remove(self, uint64_t x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef size_t j
 */

static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_10Uint64List_remove(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, uint64_t __pyx_v_x, int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  int __pyx_v_found;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("remove", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_remove); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint64List_40remove)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_uint64_t(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1674, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1674, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1677
 *         cdef size_t i
 *         cdef size_t j
 *         cdef bint found = 0             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self._n_items):
 */
  __pyx_v_found = 0;

  /* "pyrobuf_list.pyx":1679
 *         cdef bint found = 0
 * 
 *         for i in range(self._n_items):             # <<<<<<<<<<<<<<
 *             if self._data[i] == x:
 *                 found = 1
 */
  __pyx_t_7 = __pyx_v_self->_n_items;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "pyrobuf_list.pyx":1680
 * 
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 found = 1
 *                 break
 */
    __pyx_t_10 = (((__pyx_v_self->_data[__pyx_v_i]) == __pyx_v_x) != 0);
    if (__pyx_t_10) {

      /* "pyrobuf_list.pyx":1681
 *         for i in range(self._n_items):
 *             if self._data[i] == x:
 *                 found = 1             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
      __pyx_v_found = 1;

      /* "pyrobuf_list.pyx":1682
 *             if self._data[i] == x:
 *                 found = 1
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         if found == 0:
 */
      goto __pyx_L4_break;

      /* "pyrobuf_list.pyx":1680
 * 
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 found = 1
 *                 break
 */
    }
  }
  __pyx_L4_break:;

  /* "pyrobuf_list.pyx":1684
 *                 break
 * 
 *         if found == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("x not in list")
 * 
 */
  __pyx_t_10 = ((__pyx_v_found == 0) != 0);
  if (unlikely(__pyx_t_10)) {

    /* "pyrobuf_list.pyx":1685
 * 
 *         if found == 0:
 *             raise ValueError("x not in list")             # <<<<<<<<<<<<<<
 * 
 *         for j in range(i, self._n_items):
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__75, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1685, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1685, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1684
 *                 break
 * 
 *         if found == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("x not in list")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1687
 *             raise ValueError("x not in list")
 * 
 *         for j in range(i, self._n_items):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j + 1]
 * 
 */
  __pyx_t_7 = __pyx_v_self->_n_items;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = __pyx_v_i; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_j = __pyx_t_9;

    /* "pyrobuf_list.pyx":1688
 * 
 *         for j in range(i, self._n_items):
 *             self._data[j] = self._data[j + 1]             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j + 1)]);
  }

  /* "pyrobuf_list.pyx":1690
 *             self._data[j] = self._data[j + 1]
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":1691
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1692
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1692, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = ((!__pyx_t_10) != 0);
  if (__pyx_t_11) {

    /* "pyrobuf_list.pyx":1693
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cdef void _append(self, uint64_t x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1693, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1693, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1693, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1692
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1674
 *         return self._data[self._n_items]
 * 
 *     cpdef remove(self, uint64_t x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef size_t j
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  uint64_t __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_uint64_t(__pyx_arg_x); if (unlikely((__pyx_v_x == ((uint64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1674, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List_39remove(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self), ((uint64_t)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_39remove(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, uint64_t __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("remove", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_10Uint64List_remove(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1695
 *             listener._Modified()
 * 
 *     cdef void _append(self, uint64_t x):             # <<<<<<<<<<<<<<
 *         cdef uint64_t *mem
 * 
 */

static void __pyx_f_12pyrobuf_list_10Uint64List__append(struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, uint64_t __pyx_v_x) {
  uint64_t *__pyx_v_mem;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_append", 0);

  /* "pyrobuf_list.pyx":1698
 *         cdef uint64_t *mem
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <uint64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint64_t))
 *             if not mem:
 */
  __pyx_t_1 = ((__pyx_v_self->_n_items == __pyx_v_self->_size) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":1699
 * 
 *         if self._n_items == self._size:
 *             mem = <uint64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint64_t))             # <<<<<<<<<<<<<<
 *             if not mem:
 *                 raise MemoryError()
 */
    __pyx_v_mem = ((uint64_t *)PyMem_Realloc(__pyx_v_self->_data, ((2 * __pyx_v_self->_size) * (sizeof(uint64_t)))));

    /* "pyrobuf_list.pyx":1700
 *         if self._n_items == self._size:
 *             mem = <uint64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint64_t))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    __pyx_t_1 = ((!(__pyx_v_mem != 0)) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "pyrobuf_list.pyx":1701
 *             mem = <uint64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint64_t))
 *             if not mem:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *             self._data = mem
 */
      PyErr_NoMemory(); __PYX_ERR(0, 1701, __pyx_L1_error)

      /* "pyrobuf_list.pyx":1700
 *         if self._n_items == self._size:
 *             mem = <uint64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint64_t))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    }

    /* "pyrobuf_list.pyx":1703
 *                 raise MemoryError()
 * 
 *             self._data = mem             # <<<<<<<<<<<<<<
 *             self._size *= 2
 * 
 */
    __pyx_v_self->_data = __pyx_v_mem;

    /* "pyrobuf_list.pyx":1704
 * 
 *             self._data = mem
 *             self._size *= 2             # <<<<<<<<<<<<<<
 * 
 *         self._data[self._n_items] = x
 */
    __pyx_v_self->_size = (__pyx_v_self->_size * 2);

    /* "pyrobuf_list.pyx":1698
 *         cdef uint64_t *mem
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <uint64_t *>PyMem_Realloc(self._data, 2 * self._size * sizeof(uint64_t))
 *             if not mem:
 */
  }

  /* "pyrobuf_list.pyx":1706
 *             self._size *= 2
 * 
 *         self._data[self._n_items] = x             # <<<<<<<<<<<<<<
 *         self._n_items += 1
 * 
 */
  (__pyx_v_self->_data[__pyx_v_self->_n_items]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":1707
 * 
 *         self._data[self._n_items] = x
 *         self._n_items += 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items + 1);

  /* "pyrobuf_list.pyx":1695
 *             listener._Modified()
 * 
 *     cdef void _append(self, uint64_t x):             # <<<<<<<<<<<<<<
 *         cdef uint64_t *mem
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("pyrobuf_list.Uint64List._append", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_42__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_42__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List_41__reduce_cython__(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_41__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__76, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_44__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_10Uint64List_44__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_10Uint64List_43__setstate_cython__(((struct __pyx_obj_12pyrobuf_list_Uint64List *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_10Uint64List_43__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_Uint64List *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__77, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.Uint64List.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1713
 * cdef class CharList:
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):             # <<<<<<<<<<<<<<
 *         self._data = <char *>PyMem_Malloc(size * sizeof(char))
 *         if not self._data:
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_8CharList_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_12pyrobuf_list_8CharList_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  size_t __pyx_v_size;
  PyObject *__pyx_v_listener = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_size,&__pyx_n_s_listener,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_listener);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 1713, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    if (values[0]) {
      __pyx_v_size = __Pyx_PyInt_As_size_t(values[0]); if (unlikely((__pyx_v_size == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1713, __pyx_L3_error)
    } else {
      __pyx_v_size = ((size_t)16);
    }
    __pyx_v_listener = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1713, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.CharList.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList___cinit__(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self), __pyx_v_size, __pyx_v_listener);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_8CharList___cinit__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, size_t __pyx_v_size, PyObject *__pyx_v_listener) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "pyrobuf_list.pyx":1714
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <char *>PyMem_Malloc(size * sizeof(char))             # <<<<<<<<<<<<<<
 *         if not self._data:
 *             raise MemoryError()
 */
  __pyx_v_self->_data = ((char *)PyMem_Malloc((__pyx_v_size * (sizeof(char)))));

  /* "pyrobuf_list.pyx":1715
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <char *>PyMem_Malloc(size * sizeof(char))
 *         if not self._data:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = ((!(__pyx_v_self->_data != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":1716
 *         self._data = <char *>PyMem_Malloc(size * sizeof(char))
 *         if not self._data:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         self._n_items = 0
 */
    PyErr_NoMemory(); __PYX_ERR(0, 1716, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1715
 *     def __cinit__(self, size_t size=16, object listener=None):
 *         self._data = <char *>PyMem_Malloc(size * sizeof(char))
 *         if not self._data:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1718
 *             raise MemoryError()
 * 
 *         self._n_items = 0             # <<<<<<<<<<<<<<
 *         self._size = size
 *         if listener is None:
 */
  __pyx_v_self->_n_items = 0;

  /* "pyrobuf_list.pyx":1719
 * 
 *         self._n_items = 0
 *         self._size = size             # <<<<<<<<<<<<<<
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener
 */
  __pyx_v_self->_size = __pyx_v_size;

  /* "pyrobuf_list.pyx":1720
 *         self._n_items = 0
 *         self._size = size
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
  __pyx_t_1 = (__pyx_v_listener == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":1721
 *         self._size = size
 *         if listener is None:
 *             self._listener = <PyObject *>null_listener             # <<<<<<<<<<<<<<
 *         else:
 *             self._listener = <PyObject *>listener
 */
    __pyx_v_self->_listener = ((PyObject *)null_listener);

    /* "pyrobuf_list.pyx":1720
 *         self._n_items = 0
 *         self._size = size
 *         if listener is None:             # <<<<<<<<<<<<<<
 *             self._listener = <PyObject *>null_listener
 *         else:
 */
    goto __pyx_L4;
  }

  /* "pyrobuf_list.pyx":1723
 *             self._listener = <PyObject *>null_listener
 *         else:
 *             self._listener = <PyObject *>listener             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  /*else*/ {
    __pyx_v_self->_listener = ((PyObject *)__pyx_v_listener);
  }
  __pyx_L4:;

  /* "pyrobuf_list.pyx":1713
 * cdef class CharList:
 * 
 *     def __cinit__(self, size_t size=16, object listener=None):             # <<<<<<<<<<<<<<
 *         self._data = <char *>PyMem_Malloc(size * sizeof(char))
 *         if not self._data:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyrobuf_list.CharList.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1725
 *             self._listener = <PyObject *>listener
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self._data)
 * 
 */

/* Python wrapper */
static void __pyx_pw_12pyrobuf_list_8CharList_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_12pyrobuf_list_8CharList_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_12pyrobuf_list_8CharList_2__dealloc__(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12pyrobuf_list_8CharList_2__dealloc__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "pyrobuf_list.pyx":1726
 * 
 *     def __dealloc__(self):
 *         PyMem_Free(self._data)             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, char x):
 */
  PyMem_Free(__pyx_v_self->_data);

  /* "pyrobuf_list.pyx":1725
 *             self._listener = <PyObject *>listener
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         PyMem_Free(self._data)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "pyrobuf_list.pyx":1728
 *         PyMem_Free(self._data)
 * 
 *     def __contains__(self, char x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self._n_items):
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_8CharList_5__contains__(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static int __pyx_pw_12pyrobuf_list_8CharList_5__contains__(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  char __pyx_v_x;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_char(__pyx_arg_x); if (unlikely((__pyx_v_x == (char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1728, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.CharList.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList_4__contains__(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self), ((char)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_8CharList_4__contains__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, char __pyx_v_x) {
  size_t __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "pyrobuf_list.pyx":1730
 *     def __contains__(self, char x):
 *         cdef size_t i
 *         for i in range(self._n_items):             # <<<<<<<<<<<<<<
 *             if self._data[i] == x:
 *                 return True
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "pyrobuf_list.pyx":1731
 *         cdef size_t i
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */
    __pyx_t_4 = (((__pyx_v_self->_data[__pyx_v_i]) == __pyx_v_x) != 0);
    if (__pyx_t_4) {

      /* "pyrobuf_list.pyx":1732
 *         for i in range(self._n_items):
 *             if self._data[i] == x:
 *                 return True             # <<<<<<<<<<<<<<
 * 
 *         return False
 */
      __pyx_r = 1;
      goto __pyx_L0;

      /* "pyrobuf_list.pyx":1731
 *         cdef size_t i
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 return True
 * 
 */
    }
  }

  /* "pyrobuf_list.pyx":1734
 *                 return True
 * 
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     def __delitem__(self, int i):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1728
 *         PyMem_Free(self._data)
 * 
 *     def __contains__(self, char x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(self._n_items):
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1736
 *         return False
 * 
 *     def __delitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_8CharList_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static int __pyx_pw_12pyrobuf_list_8CharList_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1736, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.CharList.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList_6__delitem__(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self), ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_8CharList_6__delitem__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, int __pyx_v_i) {
  size_t __pyx_v_j;
  PyObject *__pyx_v_listener = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  size_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__delitem__", 0);

  /* "pyrobuf_list.pyx":1737
 * 
 *     def __delitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":1738
 *     def __delitem__(self, int i):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":1737
 * 
 *     def __delitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":1740
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":1741
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t j
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__78, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1741, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1741, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1740
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1744
 * 
 *         cdef size_t j
 *         for j in range(i, self._n_items):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j + 1]
 * 
 */
  __pyx_t_4 = __pyx_v_self->_n_items;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = __pyx_v_i; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_j = __pyx_t_6;

    /* "pyrobuf_list.pyx":1745
 *         cdef size_t j
 *         for j in range(i, self._n_items):
 *             self._data[j] = self._data[j + 1]             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j + 1)]);
  }

  /* "pyrobuf_list.pyx":1747
 *             self._data[j] = self._data[j + 1]
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":1748
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":1749
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1749, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":1750
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, int i):
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1750, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    if (__pyx_t_8) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1750, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1750, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":1749
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1736
 *         return False
 * 
 *     def __delitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyrobuf_list.CharList.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1752
 *             listener._Modified()
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1752, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.CharList.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList_8__getitem__(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self), ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_8CharList_8__getitem__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, int __pyx_v_i) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "pyrobuf_list.pyx":1753
 * 
 *     def __getitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":1754
 *     def __getitem__(self, int i):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":1753
 * 
 *     def __getitem__(self, int i):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":1756
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":1757
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         return self._data[i]
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__79, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1757, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1757, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1756
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1759
 *             raise IndexError("list index out of range")
 * 
 *         return self._data[i]             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_char((__pyx_v_self->_data[__pyx_v_i])); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1752
 *             listener._Modified()
 * 
 *     def __getitem__(self, int i):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("pyrobuf_list.CharList.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_12pyrobuf_list_8CharList_12generator7(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "pyrobuf_list.pyx":1761
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_11__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_11__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList_10__iter__(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_8CharList_10__iter__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self) {
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_7___iter__ *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  __pyx_cur_scope = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_7___iter__ *)__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_7___iter__(__pyx_ptype_12pyrobuf_list___pyx_scope_struct_7___iter__, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_7___iter__ *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1761, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_12pyrobuf_list_8CharList_12generator7, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_iter, __pyx_n_s_CharList___iter, __pyx_n_s_pyrobuf_list); if (unlikely(!gen)) __PYX_ERR(0, 1761, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("pyrobuf_list.CharList.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_12pyrobuf_list_8CharList_12generator7(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_7___iter__ *__pyx_cur_scope = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_7___iter__ *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1761, __pyx_L1_error)

  /* "pyrobuf_list.pyx":1762
 * 
 *     def __iter__(self):
 *         cdef size_t idx = 0             # <<<<<<<<<<<<<<
 *         while idx < self._n_items:
 *             yield self._data[idx]
 */
  __pyx_cur_scope->__pyx_v_idx = 0;

  /* "pyrobuf_list.pyx":1763
 *     def __iter__(self):
 *         cdef size_t idx = 0
 *         while idx < self._n_items:             # <<<<<<<<<<<<<<
 *             yield self._data[idx]
 *             idx += 1
 */
  while (1) {
    __pyx_t_1 = ((__pyx_cur_scope->__pyx_v_idx < __pyx_cur_scope->__pyx_v_self->_n_items) != 0);
    if (!__pyx_t_1) break;

    /* "pyrobuf_list.pyx":1764
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 *             yield self._data[idx]             # <<<<<<<<<<<<<<
 *             idx += 1
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_char((__pyx_cur_scope->__pyx_v_self->_data[__pyx_cur_scope->__pyx_v_idx])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1764, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1764, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1765
 *         while idx < self._n_items:
 *             yield self._data[idx]
 *             idx += 1             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
    __pyx_cur_scope->__pyx_v_idx = (__pyx_cur_scope->__pyx_v_idx + 1);
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "pyrobuf_list.pyx":1761
 *         return self._data[i]
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         cdef size_t idx = 0
 *         while idx < self._n_items:
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1767
 *             idx += 1
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._n_items
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_12pyrobuf_list_8CharList_14__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_12pyrobuf_list_8CharList_14__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList_13__len__(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_12pyrobuf_list_8CharList_13__len__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "pyrobuf_list.pyx":1768
 * 
 *     def __len__(self):
 *         return self._n_items             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = __pyx_v_self->_n_items;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1767
 *             idx += 1
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._n_items
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1770
 *         return self._n_items
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(list(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_16__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_16__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList_15__repr__(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_8CharList_15__repr__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "pyrobuf_list.pyx":1771
 * 
 *     def __repr__(self):
 *         return repr(list(self))             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, int i, char x):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1770
 *         return self._n_items
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(list(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.CharList.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1773
 *         return repr(list(self))
 * 
 *     def __setitem__(self, int i, char x):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

/* Python wrapper */
static int __pyx_pw_12pyrobuf_list_8CharList_18__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i, PyObject *__pyx_arg_x); /*proto*/
static int __pyx_pw_12pyrobuf_list_8CharList_18__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_arg_i, PyObject *__pyx_arg_x) {
  int __pyx_v_i;
  char __pyx_v_x;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1773, __pyx_L3_error)
  }
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_char(__pyx_arg_x); if (unlikely((__pyx_v_x == (char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1773, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.CharList.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList_17__setitem__(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self), ((int)__pyx_v_i), ((char)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_8CharList_17__setitem__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, int __pyx_v_i, char __pyx_v_x) {
  PyObject *__pyx_v_listener = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "pyrobuf_list.pyx":1774
 * 
 *     def __setitem__(self, int i, char x):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_1 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":1775
 *     def __setitem__(self, int i, char x):
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":1774
 * 
 *     def __setitem__(self, int i, char x):
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":1777
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_2 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((((size_t)__pyx_v_i) >= __pyx_v_self->_n_items) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "pyrobuf_list.pyx":1778
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__80, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1778, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1778, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1777
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i >= self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1780
 *             raise IndexError("list index out of range")
 * 
 *         self._data[i] = x             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  (__pyx_v_self->_data[__pyx_v_i]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":1781
 * 
 *         self._data[i] = x
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":1782
 *         self._data[i] = x
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1782, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = ((!__pyx_t_1) != 0);
  if (__pyx_t_2) {

    /* "pyrobuf_list.pyx":1783
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1783, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1783, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1783, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyrobuf_list.pyx":1782
 *         self._data[i] = x
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1773
 *         return repr(list(self))
 * 
 *     def __setitem__(self, int i, char x):             # <<<<<<<<<<<<<<
 *         if i < 0:
 *             i += <int>self._n_items
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.CharList.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1785
 *             listener._Modified()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(list(self))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_20__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_20__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList_19__str__(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_8CharList_19__str__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "pyrobuf_list.pyx":1786
 * 
 *     def __str__(self):
 *         return str(list(self))             # <<<<<<<<<<<<<<
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PySequence_List(((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1785
 *             listener._Modified()
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return str(list(self))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.CharList.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1788
 *         return str(list(self))
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t itemsize = sizeof(char)
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_12pyrobuf_list_8CharList_22__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_12pyrobuf_list_8CharList_22__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList_21__getbuffer__(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_12pyrobuf_list_8CharList_21__getbuffer__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, Py_buffer *__pyx_v_buffer, CYTHON_UNUSED int __pyx_v_flags) {
  Py_ssize_t __pyx_v_itemsize;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  if (__pyx_v_buffer == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_buffer->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_buffer->obj);

  /* "pyrobuf_list.pyx":1790
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):
 * 
 *         cdef Py_ssize_t itemsize = sizeof(char)             # <<<<<<<<<<<<<<
 *         self.shape[0] = self._n_items
 *         self.strides[0] = 1
 */
  __pyx_v_itemsize = (sizeof(char));

  /* "pyrobuf_list.pyx":1791
 * 
 *         cdef Py_ssize_t itemsize = sizeof(char)
 *         self.shape[0] = self._n_items             # <<<<<<<<<<<<<<
 *         self.strides[0] = 1
 * 
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  (__pyx_v_self->shape[0]) = __pyx_t_1;

  /* "pyrobuf_list.pyx":1792
 *         cdef Py_ssize_t itemsize = sizeof(char)
 *         self.shape[0] = self._n_items
 *         self.strides[0] = 1             # <<<<<<<<<<<<<<
 * 
 *         buffer.buf = <char *>(self._data)
 */
  (__pyx_v_self->strides[0]) = 1;

  /* "pyrobuf_list.pyx":1794
 *         self.strides[0] = 1
 * 
 *         buffer.buf = <char *>(self._data)             # <<<<<<<<<<<<<<
 *         buffer.format = 'c'
 *         buffer.internal = NULL                  # see References
 */
  __pyx_v_buffer->buf = ((char *)__pyx_v_self->_data);

  /* "pyrobuf_list.pyx":1795
 * 
 *         buffer.buf = <char *>(self._data)
 *         buffer.format = 'c'             # <<<<<<<<<<<<<<
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 */
  __pyx_v_buffer->format = ((char *)"c");

  /* "pyrobuf_list.pyx":1796
 *         buffer.buf = <char *>(self._data)
 *         buffer.format = 'c'
 *         buffer.internal = NULL                  # see References             # <<<<<<<<<<<<<<
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items
 */
  __pyx_v_buffer->internal = NULL;

  /* "pyrobuf_list.pyx":1797
 *         buffer.format = 'c'
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize             # <<<<<<<<<<<<<<
 *         buffer.len = self._n_items
 *         buffer.ndim = 1
 */
  __pyx_v_buffer->itemsize = __pyx_v_itemsize;

  /* "pyrobuf_list.pyx":1798
 *         buffer.internal = NULL                  # see References
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items             # <<<<<<<<<<<<<<
 *         buffer.ndim = 1
 *         buffer.obj = self
 */
  __pyx_t_1 = __pyx_v_self->_n_items;
  __pyx_v_buffer->len = __pyx_t_1;

  /* "pyrobuf_list.pyx":1799
 *         buffer.itemsize = itemsize
 *         buffer.len = self._n_items
 *         buffer.ndim = 1             # <<<<<<<<<<<<<<
 *         buffer.obj = self
 *         buffer.readonly = 1
 */
  __pyx_v_buffer->ndim = 1;

  /* "pyrobuf_list.pyx":1800
 *         buffer.len = self._n_items
 *         buffer.ndim = 1
 *         buffer.obj = self             # <<<<<<<<<<<<<<
 *         buffer.readonly = 1
 *         buffer.shape = self.shape
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_buffer->obj);
  __Pyx_DECREF(__pyx_v_buffer->obj);
  __pyx_v_buffer->obj = ((PyObject *)__pyx_v_self);

  /* "pyrobuf_list.pyx":1801
 *         buffer.ndim = 1
 *         buffer.obj = self
 *         buffer.readonly = 1             # <<<<<<<<<<<<<<
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides
 */
  __pyx_v_buffer->readonly = 1;

  /* "pyrobuf_list.pyx":1802
 *         buffer.obj = self
 *         buffer.readonly = 1
 *         buffer.shape = self.shape             # <<<<<<<<<<<<<<
 *         buffer.strides = self.strides
 *         buffer.suboffsets = NULL                # for pointer arrays only
 */
  __pyx_t_2 = __pyx_v_self->shape;
  __pyx_v_buffer->shape = __pyx_t_2;

  /* "pyrobuf_list.pyx":1803
 *         buffer.readonly = 1
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides             # <<<<<<<<<<<<<<
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 */
  __pyx_t_2 = __pyx_v_self->strides;
  __pyx_v_buffer->strides = __pyx_t_2;

  /* "pyrobuf_list.pyx":1804
 *         buffer.shape = self.shape
 *         buffer.strides = self.strides
 *         buffer.suboffsets = NULL                # for pointer arrays only             # <<<<<<<<<<<<<<
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):
 */
  __pyx_v_buffer->suboffsets = NULL;

  /* "pyrobuf_list.pyx":1788
 *         return str(list(self))
 * 
 *     def __getbuffer__(self, Py_buffer *buffer, int flags):             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t itemsize = sizeof(char)
 */

  /* function exit code */
  __pyx_r = 0;
  if (__pyx_v_buffer->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_buffer->obj);
    __Pyx_DECREF(__pyx_v_buffer->obj); __pyx_v_buffer->obj = 0;
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1806
 *         buffer.suboffsets = NULL                # for pointer arrays only
 * 
 *     def __releasebuffer__(self, Py_buffer *buffer):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_12pyrobuf_list_8CharList_24__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer); /*proto*/
static CYTHON_UNUSED void __pyx_pw_12pyrobuf_list_8CharList_24__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_12pyrobuf_list_8CharList_23__releasebuffer__(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self), ((Py_buffer *)__pyx_v_buffer));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_12pyrobuf_list_8CharList_23__releasebuffer__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, CYTHON_UNUSED Py_buffer *__pyx_v_buffer) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "pyrobuf_list.pyx":1809
 *         pass
 * 
 *     def __richcmp__(self, other, op):             # <<<<<<<<<<<<<<
 *         if op == Py_EQ:
 *             return self._equal(other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_26__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_26__richcmp__(PyObject *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_arg_op) {
  PyObject *__pyx_v_op = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__richcmp__ (wrapper)", 0);
  __pyx_v_op = __Pyx_PyInt_From_int(__pyx_arg_op); if (unlikely(!__pyx_v_op)) __PYX_ERR(0, 1809, __pyx_L3_error)
  __Pyx_GOTREF(__pyx_v_op);
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.CharList.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList_25__richcmp__(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self), ((PyObject *)__pyx_v_other), ((PyObject *)__pyx_v_op));

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_op);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_8CharList_25__richcmp__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, PyObject *__pyx_v_other, PyObject *__pyx_v_op) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("__richcmp__", 0);

  /* "pyrobuf_list.pyx":1810
 * 
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:             # <<<<<<<<<<<<<<
 *             return self._equal(other)
 *         elif op == Py_NE:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(Py_EQ); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_op, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1810, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1810, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "pyrobuf_list.pyx":1811
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:
 *             return self._equal(other)             # <<<<<<<<<<<<<<
 *         elif op == Py_NE:
 *             return not self._equal(other)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = ((struct __pyx_vtabstruct_12pyrobuf_list_CharList *)__pyx_v_self->__pyx_vtab)->_equal(__pyx_v_self, __pyx_v_other, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1811, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "pyrobuf_list.pyx":1810
 * 
 *     def __richcmp__(self, other, op):
 *         if op == Py_EQ:             # <<<<<<<<<<<<<<
 *             return self._equal(other)
 *         elif op == Py_NE:
 */
  }

  /* "pyrobuf_list.pyx":1812
 *         if op == Py_EQ:
 *             return self._equal(other)
 *         elif op == Py_NE:             # <<<<<<<<<<<<<<
 *             return not self._equal(other)
 *         else:
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(Py_NE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_op, __pyx_t_2, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1812, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1812, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(__pyx_t_3)) {

    /* "pyrobuf_list.pyx":1813
 *             return self._equal(other)
 *         elif op == Py_NE:
 *             return not self._equal(other)             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError('Operation not supported for lists')
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((struct __pyx_vtabstruct_12pyrobuf_list_CharList *)__pyx_v_self->__pyx_vtab)->_equal(__pyx_v_self, __pyx_v_other, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1813, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_3)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "pyrobuf_list.pyx":1812
 *         if op == Py_EQ:
 *             return self._equal(other)
 *         elif op == Py_NE:             # <<<<<<<<<<<<<<
 *             return not self._equal(other)
 *         else:
 */
  }

  /* "pyrobuf_list.pyx":1815
 *             return not self._equal(other)
 *         else:
 *             raise TypeError('Operation not supported for lists')             # <<<<<<<<<<<<<<
 * 
 *     def __neq__(self, other):
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__81, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1815, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1815, __pyx_L1_error)
  }

  /* "pyrobuf_list.pyx":1809
 *         pass
 * 
 *     def __richcmp__(self, other, op):             # <<<<<<<<<<<<<<
 *         if op == Py_EQ:
 *             return self._equal(other)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("pyrobuf_list.CharList.__richcmp__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1817
 *             raise TypeError('Operation not supported for lists')
 * 
 *     def __neq__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_28__neq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_28__neq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__neq__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList_27__neq__(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_8CharList_27__neq__(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__neq__", 0);

  /* "pyrobuf_list.pyx":1818
 * 
 *     def __neq__(self, other):
 *         return not self == other             # <<<<<<<<<<<<<<
 * 
 *     cpdef _equal(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_self), __pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1818, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1818, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1818, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1817
 *             raise TypeError('Operation not supported for lists')
 * 
 *     def __neq__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.CharList.__neq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1820
 *         return not self == other
 * 
 *     cpdef _equal(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             if len(self) != len(other):
 */

static PyObject *__pyx_pw_12pyrobuf_list_8CharList_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_8CharList__equal(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_v_y = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *(*__pyx_t_14)(PyObject *);
  __Pyx_RefNannySetupContext("_equal", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_equal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1820, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_8CharList_30_equal)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_other); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1820, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_other};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1820, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_other};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1820, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1820, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_other);
          __Pyx_GIVEREF(__pyx_v_other);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_other);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1820, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1821
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_8);
    /*try:*/ {

      /* "pyrobuf_list.pyx":1822
 *     cpdef _equal(self, other):
 *         try:
 *             if len(self) != len(other):             # <<<<<<<<<<<<<<
 *                 return False
 *         except TypeError:
 */
      __pyx_t_9 = PyObject_Length(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1822, __pyx_L3_error)
      __pyx_t_10 = PyObject_Length(__pyx_v_other); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1822, __pyx_L3_error)
      __pyx_t_11 = ((__pyx_t_9 != __pyx_t_10) != 0);
      if (__pyx_t_11) {

        /* "pyrobuf_list.pyx":1823
 *         try:
 *             if len(self) != len(other):
 *                 return False             # <<<<<<<<<<<<<<
 *         except TypeError:
 *             return False
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(Py_False);
        __pyx_r = Py_False;
        goto __pyx_L7_try_return;

        /* "pyrobuf_list.pyx":1822
 *     cpdef _equal(self, other):
 *         try:
 *             if len(self) != len(other):             # <<<<<<<<<<<<<<
 *                 return False
 *         except TypeError:
 */
      }

      /* "pyrobuf_list.pyx":1821
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
    }
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1824
 *             if len(self) != len(other):
 *                 return False
 *         except TypeError:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
    __pyx_t_12 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_12) {
      __Pyx_AddTraceback("pyrobuf_list.CharList._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3) < 0) __PYX_ERR(0, 1824, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_t_3);

      /* "pyrobuf_list.pyx":1825
 *                 return False
 *         except TypeError:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *         for x, y in zip(self, other):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "pyrobuf_list.pyx":1821
 * 
 *     cpdef _equal(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             if len(self) != len(other):
 *                 return False
 */
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_6);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
    goto __pyx_L0;
    __pyx_L8_try_end:;
  }

  /* "pyrobuf_list.pyx":1827
 *             return False
 * 
 *         for x, y in zip(self, other):             # <<<<<<<<<<<<<<
 *             if x != y:
 *                 return False
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));
  __Pyx_INCREF(__pyx_v_other);
  __Pyx_GIVEREF(__pyx_v_other);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_other);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1827, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_10 = 0;
    __pyx_t_13 = NULL;
  } else {
    __pyx_t_10 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_13 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1827, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_13)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 1827, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1827, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 1827, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1827, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_13(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1827, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1827, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1827, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1827, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1827, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_14 = Py_TYPE(__pyx_t_4)->tp_iternext;
      index = 0; __pyx_t_1 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_1)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_5 = __pyx_t_14(__pyx_t_4); if (unlikely(!__pyx_t_5)) goto __pyx_L14_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_14(__pyx_t_4), 2) < 0) __PYX_ERR(0, 1827, __pyx_L1_error)
      __pyx_t_14 = NULL;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L15_unpacking_done;
      __pyx_L14_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_14 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1827, __pyx_L1_error)
      __pyx_L15_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_x, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_y, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "pyrobuf_list.pyx":1828
 * 
 *         for x, y in zip(self, other):
 *             if x != y:             # <<<<<<<<<<<<<<
 *                 return False
 * 
 */
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_x, __pyx_v_y, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1828, __pyx_L1_error)
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1828, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_11) {

      /* "pyrobuf_list.pyx":1829
 *         for x, y in zip(self, other):
 *             if x != y:
 *                 return False             # <<<<<<<<<<<<<<
 * 
 *         return True
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      goto __pyx_L0;

      /* "pyrobuf_list.pyx":1828
 * 
 *         for x, y in zip(self, other):
 *             if x != y:             # <<<<<<<<<<<<<<
 *                 return False
 * 
 */
    }

    /* "pyrobuf_list.pyx":1827
 *             return False
 * 
 *         for x, y in zip(self, other):             # <<<<<<<<<<<<<<
 *             if x != y:
 *                 return False
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyrobuf_list.pyx":1831
 *                 return False
 * 
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     cpdef append(self, char x):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_True);
  __pyx_r = Py_True;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1820
 *         return not self == other
 * 
 *     cpdef _equal(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             if len(self) != len(other):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.CharList._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_y);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_30_equal(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_equal (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList_29_equal(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_8CharList_29_equal(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_equal", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_8CharList__equal(__pyx_v_self, __pyx_v_other, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.CharList._equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1833
 *         return True
 * 
 *     cpdef append(self, char x):             # <<<<<<<<<<<<<<
 *         self._append(x)
 *         listener = <object>self._listener
 */

static PyObject *__pyx_pw_12pyrobuf_list_8CharList_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_8CharList_append(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, char __pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("append", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_append); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1833, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_8CharList_32append)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_char(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1833, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1833, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1833, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1833, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1833, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1833, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1834
 * 
 *     cpdef append(self, char x):
 *         self._append(x)             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  ((struct __pyx_vtabstruct_12pyrobuf_list_CharList *)__pyx_v_self->__pyx_vtab)->_append(__pyx_v_self, __pyx_v_x);

  /* "pyrobuf_list.pyx":1835
 *     cpdef append(self, char x):
 *         self._append(x)
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1836
 *         self._append(x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1836, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1836, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = ((!__pyx_t_7) != 0);
  if (__pyx_t_8) {

    /* "pyrobuf_list.pyx":1837
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef extend(self, x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1837, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1837, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1837, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1836
 *         self._append(x)
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1833
 *         return True
 * 
 *     cpdef append(self, char x):             # <<<<<<<<<<<<<<
 *         self._append(x)
 *         listener = <object>self._listener
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.CharList.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_32append(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  char __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("append (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_char(__pyx_arg_x); if (unlikely((__pyx_v_x == (char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1833, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.CharList.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList_31append(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self), ((char)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_8CharList_31append(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, char __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("append", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_8CharList_append(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.CharList.append", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1839
 *             listener._Modified()
 * 
 *     cpdef extend(self, x):             # <<<<<<<<<<<<<<
 *         for i in x:
 *             self._append(i)
 */

static PyObject *__pyx_pw_12pyrobuf_list_8CharList_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_8CharList_extend(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, PyObject *__pyx_v_x, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  char __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("extend", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_extend); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1839, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_8CharList_34extend)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (!__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1839, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_x};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1839, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_v_x};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1839, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1839, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
          __Pyx_INCREF(__pyx_v_x);
          __Pyx_GIVEREF(__pyx_v_x);
          PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_x);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1839, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1840
 * 
 *     cpdef extend(self, x):
 *         for i in x:             # <<<<<<<<<<<<<<
 *             self._append(i)
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_x)) || PyTuple_CheckExact(__pyx_v_x)) {
    __pyx_t_1 = __pyx_v_x; __Pyx_INCREF(__pyx_t_1); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1840, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1840, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1840, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1840, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1840, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1840, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_7(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1840, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyrobuf_list.pyx":1841
 *     cpdef extend(self, x):
 *         for i in x:
 *             self._append(i)             # <<<<<<<<<<<<<<
 * 
 *         listener = <object>self._listener
 */
    __pyx_t_8 = __Pyx_PyInt_As_char(__pyx_v_i); if (unlikely((__pyx_t_8 == (char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1841, __pyx_L1_error)
    ((struct __pyx_vtabstruct_12pyrobuf_list_CharList *)__pyx_v_self->__pyx_vtab)->_append(__pyx_v_self, __pyx_t_8);

    /* "pyrobuf_list.pyx":1840
 * 
 *     cpdef extend(self, x):
 *         for i in x:             # <<<<<<<<<<<<<<
 *             self._append(i)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1843
 *             self._append(i)
 * 
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1844
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1844, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = ((!__pyx_t_9) != 0);
  if (__pyx_t_10) {

    /* "pyrobuf_list.pyx":1845
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef insert(self, int i, char x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1845, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1845, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1845, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1844
 * 
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1839
 *             listener._Modified()
 * 
 *     cpdef extend(self, x):             # <<<<<<<<<<<<<<
 *         for i in x:
 *             self._append(i)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("pyrobuf_list.CharList.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_34extend(PyObject *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("extend (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList_33extend(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self), ((PyObject *)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_8CharList_33extend(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("extend", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_8CharList_extend(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.CharList.extend", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1847
 *             listener._Modified()
 * 
 *     cpdef insert(self, int i, char x):             # <<<<<<<<<<<<<<
 *         cdef char *mem
 *         cdef size_t j
 */

static PyObject *__pyx_pw_12pyrobuf_list_8CharList_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_8CharList_insert(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, int __pyx_v_i, char __pyx_v_x, int __pyx_skip_dispatch) {
  char *__pyx_v_mem;
  size_t __pyx_v_j;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  size_t __pyx_t_11;
  size_t __pyx_t_12;
  size_t __pyx_t_13;
  __Pyx_RefNannySetupContext("insert", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_insert); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1847, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_8CharList_36insert)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1847, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyInt_From_char(__pyx_v_x); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1847, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_5 = __pyx_t_1; __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1847, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_3, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1847, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1847, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_6) {
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
        __pyx_t_3 = 0;
        __pyx_t_4 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1847, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1851
 *         cdef size_t j
 * 
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  __pyx_t_9 = ((__pyx_v_i < 0) != 0);
  if (__pyx_t_9) {

    /* "pyrobuf_list.pyx":1852
 * 
 *         if i < 0:
 *             i += <int>self._n_items             # <<<<<<<<<<<<<<
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 */
    __pyx_v_i = (__pyx_v_i + ((int)__pyx_v_self->_n_items));

    /* "pyrobuf_list.pyx":1851
 *         cdef size_t j
 * 
 *         if i < 0:             # <<<<<<<<<<<<<<
 *             i += <int>self._n_items
 * 
 */
  }

  /* "pyrobuf_list.pyx":1854
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i > self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  __pyx_t_10 = ((__pyx_v_i < 0) != 0);
  if (!__pyx_t_10) {
  } else {
    __pyx_t_9 = __pyx_t_10;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_10 = ((((size_t)__pyx_v_i) > __pyx_v_self->_n_items) != 0);
  __pyx_t_9 = __pyx_t_10;
  __pyx_L5_bool_binop_done:;
  if (unlikely(__pyx_t_9)) {

    /* "pyrobuf_list.pyx":1855
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         if self._n_items == self._size:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__82, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1855, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1855, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1854
 *             i += <int>self._n_items
 * 
 *         if i < 0 or <size_t>i > self._n_items:             # <<<<<<<<<<<<<<
 *             raise IndexError("list index out of range")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1857
 *             raise IndexError("list index out of range")
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <char *>PyMem_Realloc(self._data, 2 * self._size * sizeof(char))
 *             if not mem:
 */
  __pyx_t_9 = ((__pyx_v_self->_n_items == __pyx_v_self->_size) != 0);
  if (__pyx_t_9) {

    /* "pyrobuf_list.pyx":1858
 * 
 *         if self._n_items == self._size:
 *             mem = <char *>PyMem_Realloc(self._data, 2 * self._size * sizeof(char))             # <<<<<<<<<<<<<<
 *             if not mem:
 *                 raise MemoryError()
 */
    __pyx_v_mem = ((char *)PyMem_Realloc(__pyx_v_self->_data, ((2 * __pyx_v_self->_size) * (sizeof(char)))));

    /* "pyrobuf_list.pyx":1859
 *         if self._n_items == self._size:
 *             mem = <char *>PyMem_Realloc(self._data, 2 * self._size * sizeof(char))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    __pyx_t_9 = ((!(__pyx_v_mem != 0)) != 0);
    if (unlikely(__pyx_t_9)) {

      /* "pyrobuf_list.pyx":1860
 *             mem = <char *>PyMem_Realloc(self._data, 2 * self._size * sizeof(char))
 *             if not mem:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *             self._data = mem
 */
      PyErr_NoMemory(); __PYX_ERR(0, 1860, __pyx_L1_error)

      /* "pyrobuf_list.pyx":1859
 *         if self._n_items == self._size:
 *             mem = <char *>PyMem_Realloc(self._data, 2 * self._size * sizeof(char))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    }

    /* "pyrobuf_list.pyx":1862
 *                 raise MemoryError()
 * 
 *             self._data = mem             # <<<<<<<<<<<<<<
 *             self._size *= 2
 * 
 */
    __pyx_v_self->_data = __pyx_v_mem;

    /* "pyrobuf_list.pyx":1863
 * 
 *             self._data = mem
 *             self._size *= 2             # <<<<<<<<<<<<<<
 * 
 *         for j in range(self._n_items, <size_t>i, -1):
 */
    __pyx_v_self->_size = (__pyx_v_self->_size * 2);

    /* "pyrobuf_list.pyx":1857
 *             raise IndexError("list index out of range")
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <char *>PyMem_Realloc(self._data, 2 * self._size * sizeof(char))
 *             if not mem:
 */
  }

  /* "pyrobuf_list.pyx":1865
 *             self._size *= 2
 * 
 *         for j in range(self._n_items, <size_t>i, -1):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j - 1]
 * 
 */
  __pyx_t_11 = ((size_t)__pyx_v_i);
  __pyx_t_12 = __pyx_t_11;
  for (__pyx_t_13 = __pyx_v_self->_n_items + 1; __pyx_t_13 > __pyx_t_12 + 1; ) { __pyx_t_13-=1;
    __pyx_v_j = __pyx_t_13;

    /* "pyrobuf_list.pyx":1866
 * 
 *         for j in range(self._n_items, <size_t>i, -1):
 *             self._data[j] = self._data[j - 1]             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j - 1)]);
  }

  /* "pyrobuf_list.pyx":1868
 *             self._data[j] = self._data[j - 1]
 * 
 *         self._data[i] = x             # <<<<<<<<<<<<<<
 *         self._n_items += 1
 *         listener = <object>self._listener
 */
  (__pyx_v_self->_data[__pyx_v_i]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":1869
 * 
 *         self._data[i] = x
 *         self._n_items += 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items + 1);

  /* "pyrobuf_list.pyx":1870
 *         self._data[i] = x
 *         self._n_items += 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1871
 *         self._n_items += 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 1871, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_10 = ((!__pyx_t_9) != 0);
  if (__pyx_t_10) {

    /* "pyrobuf_list.pyx":1872
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cpdef pop(self):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1872, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1872, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1872, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1871
 *         self._n_items += 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1847
 *             listener._Modified()
 * 
 *     cpdef insert(self, int i, char x):             # <<<<<<<<<<<<<<
 *         cdef char *mem
 *         cdef size_t j
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("pyrobuf_list.CharList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_36insert(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_i;
  char __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("insert (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_i,&__pyx_n_s_x,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_i)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, 1); __PYX_ERR(0, 1847, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "insert") < 0)) __PYX_ERR(0, 1847, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_i = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1847, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_char(values[1]); if (unlikely((__pyx_v_x == (char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1847, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("insert", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1847, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.CharList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList_35insert(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self), __pyx_v_i, __pyx_v_x);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_8CharList_35insert(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, int __pyx_v_i, char __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("insert", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_8CharList_insert(__pyx_v_self, __pyx_v_i, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.CharList.insert", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1874
 *             listener._Modified()
 * 
 *     cpdef pop(self):             # <<<<<<<<<<<<<<
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")
 */

static PyObject *__pyx_pw_12pyrobuf_list_8CharList_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_8CharList_pop(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("pop", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1874, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_8CharList_38pop)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1874, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1874, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1875
 * 
 *     cpdef pop(self):
 *         if self._n_items == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("pop from empty list")
 * 
 */
  __pyx_t_5 = ((__pyx_v_self->_n_items == 0) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "pyrobuf_list.pyx":1876
 *     cpdef pop(self):
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__83, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1876, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1876, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1875
 * 
 *     cpdef pop(self):
 *         if self._n_items == 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("pop from empty list")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1878
 *             raise IndexError("pop from empty list")
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":1879
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1880
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 *         return self._data[self._n_items]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1880, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_6 = ((!__pyx_t_5) != 0);
  if (__pyx_t_6) {

    /* "pyrobuf_list.pyx":1881
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 *         return self._data[self._n_items]
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1881, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1881, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1881, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1880
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 *         return self._data[self._n_items]
 */
  }

  /* "pyrobuf_list.pyx":1882
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 *         return self._data[self._n_items]             # <<<<<<<<<<<<<<
 * 
 *     cpdef remove(self, char x):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_char((__pyx_v_self->_data[__pyx_v_self->_n_items])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "pyrobuf_list.pyx":1874
 *             listener._Modified()
 * 
 *     cpdef pop(self):             # <<<<<<<<<<<<<<
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("pyrobuf_list.CharList.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_38pop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pop (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList_37pop(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_8CharList_37pop(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("pop", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_8CharList_pop(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1874, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.CharList.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1884
 *         return self._data[self._n_items]
 * 
 *     cpdef remove(self, char x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef size_t j
 */

static PyObject *__pyx_pw_12pyrobuf_list_8CharList_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_f_12pyrobuf_list_8CharList_remove(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, char __pyx_v_x, int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  int __pyx_v_found;
  PyObject *__pyx_v_listener = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  __Pyx_RefNannySetupContext("remove", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_remove); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1884, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_12pyrobuf_list_8CharList_40remove)) {
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_3 = __Pyx_PyInt_From_char(__pyx_v_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1884, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1; __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      if (!__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1884, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_3};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1884, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
          __pyx_t_3 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1884, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyrobuf_list.pyx":1887
 *         cdef size_t i
 *         cdef size_t j
 *         cdef bint found = 0             # <<<<<<<<<<<<<<
 * 
 *         for i in range(self._n_items):
 */
  __pyx_v_found = 0;

  /* "pyrobuf_list.pyx":1889
 *         cdef bint found = 0
 * 
 *         for i in range(self._n_items):             # <<<<<<<<<<<<<<
 *             if self._data[i] == x:
 *                 found = 1
 */
  __pyx_t_7 = __pyx_v_self->_n_items;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_i = __pyx_t_9;

    /* "pyrobuf_list.pyx":1890
 * 
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 found = 1
 *                 break
 */
    __pyx_t_10 = (((__pyx_v_self->_data[__pyx_v_i]) == __pyx_v_x) != 0);
    if (__pyx_t_10) {

      /* "pyrobuf_list.pyx":1891
 *         for i in range(self._n_items):
 *             if self._data[i] == x:
 *                 found = 1             # <<<<<<<<<<<<<<
 *                 break
 * 
 */
      __pyx_v_found = 1;

      /* "pyrobuf_list.pyx":1892
 *             if self._data[i] == x:
 *                 found = 1
 *                 break             # <<<<<<<<<<<<<<
 * 
 *         if found == 0:
 */
      goto __pyx_L4_break;

      /* "pyrobuf_list.pyx":1890
 * 
 *         for i in range(self._n_items):
 *             if self._data[i] == x:             # <<<<<<<<<<<<<<
 *                 found = 1
 *                 break
 */
    }
  }
  __pyx_L4_break:;

  /* "pyrobuf_list.pyx":1894
 *                 break
 * 
 *         if found == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("x not in list")
 * 
 */
  __pyx_t_10 = ((__pyx_v_found == 0) != 0);
  if (unlikely(__pyx_t_10)) {

    /* "pyrobuf_list.pyx":1895
 * 
 *         if found == 0:
 *             raise ValueError("x not in list")             # <<<<<<<<<<<<<<
 * 
 *         for j in range(i, self._n_items):
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__84, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1895, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1895, __pyx_L1_error)

    /* "pyrobuf_list.pyx":1894
 *                 break
 * 
 *         if found == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("x not in list")
 * 
 */
  }

  /* "pyrobuf_list.pyx":1897
 *             raise ValueError("x not in list")
 * 
 *         for j in range(i, self._n_items):             # <<<<<<<<<<<<<<
 *             self._data[j] = self._data[j + 1]
 * 
 */
  __pyx_t_7 = __pyx_v_self->_n_items;
  __pyx_t_8 = __pyx_t_7;
  for (__pyx_t_9 = __pyx_v_i; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
    __pyx_v_j = __pyx_t_9;

    /* "pyrobuf_list.pyx":1898
 * 
 *         for j in range(i, self._n_items):
 *             self._data[j] = self._data[j + 1]             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
    (__pyx_v_self->_data[__pyx_v_j]) = (__pyx_v_self->_data[(__pyx_v_j + 1)]);
  }

  /* "pyrobuf_list.pyx":1900
 *             self._data[j] = self._data[j + 1]
 * 
 *         self._n_items -= 1             # <<<<<<<<<<<<<<
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items - 1);

  /* "pyrobuf_list.pyx":1901
 * 
 *         self._n_items -= 1
 *         listener = <object>self._listener             # <<<<<<<<<<<<<<
 *         if not listener._is_present_in_parent:
 *             listener._Modified()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->_listener);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_listener = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1902
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_is_present_in_parent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_10 < 0)) __PYX_ERR(0, 1902, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = ((!__pyx_t_10) != 0);
  if (__pyx_t_11) {

    /* "pyrobuf_list.pyx":1903
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:
 *             listener._Modified()             # <<<<<<<<<<<<<<
 * 
 *     cdef void _append(self, char x):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_listener, __pyx_n_s_Modified); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1903, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1903, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1903, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyrobuf_list.pyx":1902
 *         self._n_items -= 1
 *         listener = <object>self._listener
 *         if not listener._is_present_in_parent:             # <<<<<<<<<<<<<<
 *             listener._Modified()
 * 
 */
  }

  /* "pyrobuf_list.pyx":1884
 *         return self._data[self._n_items]
 * 
 *     cpdef remove(self, char x):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef size_t j
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.CharList.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_listener);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_40remove(PyObject *__pyx_v_self, PyObject *__pyx_arg_x) {
  char __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove (wrapper)", 0);
  assert(__pyx_arg_x); {
    __pyx_v_x = __Pyx_PyInt_As_char(__pyx_arg_x); if (unlikely((__pyx_v_x == (char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1884, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.CharList.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList_39remove(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self), ((char)__pyx_v_x));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_8CharList_39remove(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, char __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("remove", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_12pyrobuf_list_8CharList_remove(__pyx_v_self, __pyx_v_x, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.CharList.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyrobuf_list.pyx":1905
 *             listener._Modified()
 * 
 *     cdef void _append(self, char x):             # <<<<<<<<<<<<<<
 *         cdef char *mem
 * 
 */

static void __pyx_f_12pyrobuf_list_8CharList__append(struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, char __pyx_v_x) {
  char *__pyx_v_mem;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_append", 0);

  /* "pyrobuf_list.pyx":1908
 *         cdef char *mem
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <char *>PyMem_Realloc(self._data, 2 * self._size * sizeof(char))
 *             if not mem:
 */
  __pyx_t_1 = ((__pyx_v_self->_n_items == __pyx_v_self->_size) != 0);
  if (__pyx_t_1) {

    /* "pyrobuf_list.pyx":1909
 * 
 *         if self._n_items == self._size:
 *             mem = <char *>PyMem_Realloc(self._data, 2 * self._size * sizeof(char))             # <<<<<<<<<<<<<<
 *             if not mem:
 *                 raise MemoryError()
 */
    __pyx_v_mem = ((char *)PyMem_Realloc(__pyx_v_self->_data, ((2 * __pyx_v_self->_size) * (sizeof(char)))));

    /* "pyrobuf_list.pyx":1910
 *         if self._n_items == self._size:
 *             mem = <char *>PyMem_Realloc(self._data, 2 * self._size * sizeof(char))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    __pyx_t_1 = ((!(__pyx_v_mem != 0)) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "pyrobuf_list.pyx":1911
 *             mem = <char *>PyMem_Realloc(self._data, 2 * self._size * sizeof(char))
 *             if not mem:
 *                 raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *             self._data = mem
 */
      PyErr_NoMemory(); __PYX_ERR(0, 1911, __pyx_L1_error)

      /* "pyrobuf_list.pyx":1910
 *         if self._n_items == self._size:
 *             mem = <char *>PyMem_Realloc(self._data, 2 * self._size * sizeof(char))
 *             if not mem:             # <<<<<<<<<<<<<<
 *                 raise MemoryError()
 * 
 */
    }

    /* "pyrobuf_list.pyx":1913
 *                 raise MemoryError()
 * 
 *             self._data = mem             # <<<<<<<<<<<<<<
 *             self._size *= 2
 * 
 */
    __pyx_v_self->_data = __pyx_v_mem;

    /* "pyrobuf_list.pyx":1914
 * 
 *             self._data = mem
 *             self._size *= 2             # <<<<<<<<<<<<<<
 * 
 *         self._data[self._n_items] = x
 */
    __pyx_v_self->_size = (__pyx_v_self->_size * 2);

    /* "pyrobuf_list.pyx":1908
 *         cdef char *mem
 * 
 *         if self._n_items == self._size:             # <<<<<<<<<<<<<<
 *             mem = <char *>PyMem_Realloc(self._data, 2 * self._size * sizeof(char))
 *             if not mem:
 */
  }

  /* "pyrobuf_list.pyx":1916
 *             self._size *= 2
 * 
 *         self._data[self._n_items] = x             # <<<<<<<<<<<<<<
 *         self._n_items += 1
 * 
 */
  (__pyx_v_self->_data[__pyx_v_self->_n_items]) = __pyx_v_x;

  /* "pyrobuf_list.pyx":1917
 * 
 *         self._data[self._n_items] = x
 *         self._n_items += 1             # <<<<<<<<<<<<<<
 * 
 */
  __pyx_v_self->_n_items = (__pyx_v_self->_n_items + 1);

  /* "pyrobuf_list.pyx":1905
 *             listener._Modified()
 * 
 *     cdef void _append(self, char x):             # <<<<<<<<<<<<<<
 *         cdef char *mem
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("pyrobuf_list.CharList._append", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_42__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_42__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList_41__reduce_cython__(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_8CharList_41__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__85, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.CharList.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_44__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_12pyrobuf_list_8CharList_44__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_12pyrobuf_list_8CharList_43__setstate_cython__(((struct __pyx_obj_12pyrobuf_list_CharList *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list_8CharList_43__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_12pyrobuf_list_CharList *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__86, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("pyrobuf_list.CharList.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_NullListener(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 */

/* Python wrapper */
static PyObject *__pyx_pw_12pyrobuf_list_1__pyx_unpickle_NullListener(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_12pyrobuf_list_1__pyx_unpickle_NullListener = {"__pyx_unpickle_NullListener", (PyCFunction)__pyx_pw_12pyrobuf_list_1__pyx_unpickle_NullListener, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_12pyrobuf_list_1__pyx_unpickle_NullListener(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_NullListener (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NullListener", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NullListener", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_NullListener") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NullListener", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pyrobuf_list.__pyx_unpickle_NullListener", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_12pyrobuf_list___pyx_unpickle_NullListener(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_12pyrobuf_list___pyx_unpickle_NullListener(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = NULL;
  PyObject *__pyx_v___pyx_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("__pyx_unpickle_NullListener", 0);

  /* "(tree fragment)":2
 * def __pyx_unpickle_NullListener(__pyx_type, long __pyx_checksum, __pyx_state):
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xd41d8cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":3
 * def __pyx_unpickle_NullListener(__pyx_type, long __pyx_checksum, __pyx_state):
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = NullListener.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":4
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = NullListener.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 4, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 4, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 4, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 4, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 4, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 4, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 4, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 4, __pyx_L1_error)

    /* "(tree fragment)":2
 * def __pyx_unpickle_NullListener(__pyx_type, long __pyx_checksum, __pyx_state):
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":5
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = NullListener.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NullListener__set_state(<NullListener> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_12pyrobuf_list_NullListener), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_6) {
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v___pyx_type};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
      PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v___pyx_type};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6); __pyx_t_6 = NULL;
      __Pyx_INCREF(__pyx_v___pyx_type);
      __Pyx_GIVEREF(__pyx_v___pyx_type);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v___pyx_type);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = NullListener.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_NullListener__set_state(<NullListener> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_7 = (__pyx_t_1 != 0);
  if (__pyx_t_7) {

    /* "(tree fragment)":7
 *     __pyx_result = NullListener.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NullListener__set_state(<NullListener> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_NullListener__set_state(NullListener __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 7, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_12pyrobuf_list___pyx_unpickle_NullListener__set_state(((struct __pyx_obj_12pyrobuf_list_NullListener *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = NullListener.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_NullListener__set_state(<NullListener> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":8
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NullListener__set_state(<NullListener> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_NullListener__set_state(NullListener __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_NullListener(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("pyrobuf_list.__pyx_unpickle_NullListener", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":9
 *         __pyx_unpickle_NullListener__set_state(<NullListener> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_NullListener__set_state(NullListener __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_12pyrobuf_list___pyx_unpickle_NullListener__set_state(struct __pyx_obj_12pyrobuf_list_NullListener *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_NullListener__set_state", 0);

  /* "(tree fragment)":10
 *     return __pyx_result
 * cdef __pyx_unpickle_NullListener__set_state(NullListener __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 10, __pyx_L1_error)
  }
  __pyx_t_2 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 10, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 10, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":11
 * cdef __pyx_unpickle_NullListener__set_state(NullListener __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 11, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    if (!__pyx_t_8) {
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 11, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_6};
        __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 11, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
        PyObject *__pyx_temp[2] = {__pyx_t_8, __pyx_t_6};
        __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 11, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(1+1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 11, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8); __pyx_t_8 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_9, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_9, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 11, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":10
 *     return __pyx_result
 * cdef __pyx_unpickle_NullListener__set_state(NullListener __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":9
 *         __pyx_unpickle_NullListener__set_state(<NullListener> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_NullListener__set_state(NullListener __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("pyrobuf_list.__pyx_unpickle_NullListener__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_12pyrobuf_list_NullListener __pyx_vtable_12pyrobuf_list_NullListener;

static PyObject *__pyx_tp_new_12pyrobuf_list_NullListener(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_12pyrobuf_list_NullListener *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12pyrobuf_list_NullListener *)o);
  p->__pyx_vtab = __pyx_vtabptr_12pyrobuf_list_NullListener;
  return o;
}

static void __pyx_tp_dealloc_12pyrobuf_list_NullListener(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyObject *__pyx_getprop_12pyrobuf_list_12NullListener__is_present_in_parent(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12pyrobuf_list_12NullListener_21_is_present_in_parent_1__get__(o);
}

static PyMethodDef __pyx_methods_12pyrobuf_list_NullListener[] = {
  {"_Modified", (PyCFunction)__pyx_pw_12pyrobuf_list_12NullListener_1_Modified, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_12NullListener_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_12NullListener_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12pyrobuf_list_NullListener[] = {
  {(char *)"_is_present_in_parent", __pyx_getprop_12pyrobuf_list_12NullListener__is_present_in_parent, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_12pyrobuf_list_NullListener = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.NullListener", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list_NullListener), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list_NullListener, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12pyrobuf_list_NullListener, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12pyrobuf_list_NullListener, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list_NullListener, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_12pyrobuf_list_TypedList(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_12pyrobuf_list_TypedList *p;
  PyObject *o = (&PyList_Type)->tp_new(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12pyrobuf_list_TypedList *)o);
  p->_list_type = ((PyTypeObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_12pyrobuf_list_TypedList(PyObject *o) {
  struct __pyx_obj_12pyrobuf_list_TypedList *p = (struct __pyx_obj_12pyrobuf_list_TypedList *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_list_type);
  PyObject_GC_Track(o);
  (&PyList_Type)->tp_dealloc(o);
}

static int __pyx_tp_traverse_12pyrobuf_list_TypedList(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12pyrobuf_list_TypedList *p = (struct __pyx_obj_12pyrobuf_list_TypedList *)o;
  if (!(&PyList_Type)->tp_traverse); else { e = (&PyList_Type)->tp_traverse(o,v,a); if (e) return e; }
  if (p->_list_type) {
    e = (*v)(((PyObject *)p->_list_type), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_12pyrobuf_list_TypedList(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_12pyrobuf_list_TypedList *p = (struct __pyx_obj_12pyrobuf_list_TypedList *)o;
  if (!(&PyList_Type)->tp_clear); else (&PyList_Type)->tp_clear(o);
  tmp = ((PyObject*)p->_list_type);
  p->_list_type = ((PyTypeObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static int __pyx_mp_ass_subscript_12pyrobuf_list_TypedList(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_12pyrobuf_list_9TypedList_3__setitem__(o, i, v);
  }
  else {
    if ((&PyList_Type)->tp_as_mapping && (&PyList_Type)->tp_as_mapping->mp_ass_subscript)
      return (&PyList_Type)->tp_as_mapping->mp_ass_subscript(o, i, v);
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_getprop_12pyrobuf_list_9TypedList_list_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_12pyrobuf_list_9TypedList_9list_type_1__get__(o);
}

static PyMethodDef __pyx_methods_12pyrobuf_list_TypedList[] = {
  {"add", (PyCFunction)__pyx_pw_12pyrobuf_list_9TypedList_5add, METH_VARARGS|METH_KEYWORDS, 0},
  {"append", (PyCFunction)__pyx_pw_12pyrobuf_list_9TypedList_7append, METH_O, 0},
  {"extend", (PyCFunction)__pyx_pw_12pyrobuf_list_9TypedList_9extend, METH_O, 0},
  {"insert", (PyCFunction)__pyx_pw_12pyrobuf_list_9TypedList_11insert, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_9TypedList_13__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_9TypedList_15__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_12pyrobuf_list_TypedList[] = {
  {(char *)"list_type", __pyx_getprop_12pyrobuf_list_9TypedList_list_type, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyMappingMethods __pyx_tp_as_mapping_TypedList = {
  0, /*mp_length*/
  0, /*mp_subscript*/
  __pyx_mp_ass_subscript_12pyrobuf_list_TypedList, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_12pyrobuf_list_TypedList = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.TypedList", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list_TypedList), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list_TypedList, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_TypedList, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_12pyrobuf_list_TypedList, /*tp_traverse*/
  __pyx_tp_clear_12pyrobuf_list_TypedList, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12pyrobuf_list_TypedList, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_12pyrobuf_list_TypedList, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_12pyrobuf_list_9TypedList_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list_TypedList, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_12pyrobuf_list_BytesList(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o = (&PyList_Type)->tp_new(t, a, k);
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_12pyrobuf_list_BytesList(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  PyObject_GC_Track(o);
  (&PyList_Type)->tp_dealloc(o);
}

static int __pyx_tp_traverse_12pyrobuf_list_BytesList(PyObject *o, visitproc v, void *a) {
  int e;
  if (!(&PyList_Type)->tp_traverse); else { e = (&PyList_Type)->tp_traverse(o,v,a); if (e) return e; }
  return 0;
}

static int __pyx_tp_clear_12pyrobuf_list_BytesList(PyObject *o) {
  if (!(&PyList_Type)->tp_clear); else (&PyList_Type)->tp_clear(o);
  return 0;
}

static int __pyx_mp_ass_subscript_12pyrobuf_list_BytesList(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_12pyrobuf_list_9BytesList_3__setitem__(o, i, v);
  }
  else {
    if ((&PyList_Type)->tp_as_mapping && (&PyList_Type)->tp_as_mapping->mp_ass_subscript)
      return (&PyList_Type)->tp_as_mapping->mp_ass_subscript(o, i, v);
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyMethodDef __pyx_methods_12pyrobuf_list_BytesList[] = {
  {"append", (PyCFunction)__pyx_pw_12pyrobuf_list_9BytesList_5append, METH_O, 0},
  {"extend", (PyCFunction)__pyx_pw_12pyrobuf_list_9BytesList_7extend, METH_O, 0},
  {"insert", (PyCFunction)__pyx_pw_12pyrobuf_list_9BytesList_9insert, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_9BytesList_11__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_9BytesList_13__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyMappingMethods __pyx_tp_as_mapping_BytesList = {
  0, /*mp_length*/
  0, /*mp_subscript*/
  __pyx_mp_ass_subscript_12pyrobuf_list_BytesList, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_12pyrobuf_list_BytesList = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.BytesList", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list_BytesList), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list_BytesList, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_BytesList, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_12pyrobuf_list_BytesList, /*tp_traverse*/
  __pyx_tp_clear_12pyrobuf_list_BytesList, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12pyrobuf_list_BytesList, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_12pyrobuf_list_9BytesList_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list_BytesList, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_12pyrobuf_list_StringList(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o = (&PyList_Type)->tp_new(t, a, k);
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_12pyrobuf_list_StringList(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  PyObject_GC_Track(o);
  (&PyList_Type)->tp_dealloc(o);
}

static int __pyx_tp_traverse_12pyrobuf_list_StringList(PyObject *o, visitproc v, void *a) {
  int e;
  if (!(&PyList_Type)->tp_traverse); else { e = (&PyList_Type)->tp_traverse(o,v,a); if (e) return e; }
  return 0;
}

static int __pyx_tp_clear_12pyrobuf_list_StringList(PyObject *o) {
  if (!(&PyList_Type)->tp_clear); else (&PyList_Type)->tp_clear(o);
  return 0;
}

static int __pyx_mp_ass_subscript_12pyrobuf_list_StringList(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_12pyrobuf_list_10StringList_3__setitem__(o, i, v);
  }
  else {
    if ((&PyList_Type)->tp_as_mapping && (&PyList_Type)->tp_as_mapping->mp_ass_subscript)
      return (&PyList_Type)->tp_as_mapping->mp_ass_subscript(o, i, v);
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyMethodDef __pyx_methods_12pyrobuf_list_StringList[] = {
  {"append", (PyCFunction)__pyx_pw_12pyrobuf_list_10StringList_5append, METH_O, 0},
  {"extend", (PyCFunction)__pyx_pw_12pyrobuf_list_10StringList_7extend, METH_O, 0},
  {"insert", (PyCFunction)__pyx_pw_12pyrobuf_list_10StringList_9insert, METH_VARARGS|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_10StringList_11__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_10StringList_13__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyMappingMethods __pyx_tp_as_mapping_StringList = {
  0, /*mp_length*/
  0, /*mp_subscript*/
  __pyx_mp_ass_subscript_12pyrobuf_list_StringList, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_12pyrobuf_list_StringList = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.StringList", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list_StringList), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list_StringList, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_StringList, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_12pyrobuf_list_StringList, /*tp_traverse*/
  __pyx_tp_clear_12pyrobuf_list_StringList, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12pyrobuf_list_StringList, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_12pyrobuf_list_10StringList_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list_StringList, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_12pyrobuf_list_DoubleList __pyx_vtable_12pyrobuf_list_DoubleList;

static PyObject *__pyx_tp_new_12pyrobuf_list_DoubleList(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_12pyrobuf_list_DoubleList *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12pyrobuf_list_DoubleList *)o);
  p->__pyx_vtab = __pyx_vtabptr_12pyrobuf_list_DoubleList;
  if (unlikely(__pyx_pw_12pyrobuf_list_10DoubleList_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12pyrobuf_list_DoubleList(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_12pyrobuf_list_10DoubleList_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_12pyrobuf_list_DoubleList(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_12pyrobuf_list_DoubleList(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_12pyrobuf_list_10DoubleList_18__setitem__(o, i, v);
  }
  else {
    return __pyx_pw_12pyrobuf_list_10DoubleList_7__delitem__(o, i);
  }
}

static PyMethodDef __pyx_methods_12pyrobuf_list_DoubleList[] = {
  {"__neq__", (PyCFunction)__pyx_pw_12pyrobuf_list_10DoubleList_28__neq__, METH_O, 0},
  {"_equal", (PyCFunction)__pyx_pw_12pyrobuf_list_10DoubleList_30_equal, METH_O, 0},
  {"append", (PyCFunction)__pyx_pw_12pyrobuf_list_10DoubleList_32append, METH_O, 0},
  {"extend", (PyCFunction)__pyx_pw_12pyrobuf_list_10DoubleList_34extend, METH_O, 0},
  {"insert", (PyCFunction)__pyx_pw_12pyrobuf_list_10DoubleList_36insert, METH_VARARGS|METH_KEYWORDS, 0},
  {"pop", (PyCFunction)__pyx_pw_12pyrobuf_list_10DoubleList_38pop, METH_NOARGS, 0},
  {"remove", (PyCFunction)__pyx_pw_12pyrobuf_list_10DoubleList_40remove, METH_O, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_10DoubleList_42__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_10DoubleList_44__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_DoubleList = {
  __pyx_pw_12pyrobuf_list_10DoubleList_14__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_12pyrobuf_list_DoubleList, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_12pyrobuf_list_10DoubleList_5__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_DoubleList = {
  __pyx_pw_12pyrobuf_list_10DoubleList_14__len__, /*mp_length*/
  __pyx_pw_12pyrobuf_list_10DoubleList_9__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_12pyrobuf_list_DoubleList, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_DoubleList = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_pw_12pyrobuf_list_10DoubleList_22__getbuffer__, /*bf_getbuffer*/
  __pyx_pw_12pyrobuf_list_10DoubleList_24__releasebuffer__, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type_12pyrobuf_list_DoubleList = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.DoubleList", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list_DoubleList), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list_DoubleList, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_12pyrobuf_list_10DoubleList_16__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_DoubleList, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_DoubleList, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_12pyrobuf_list_10DoubleList_20__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_DoubleList, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_12pyrobuf_list_10DoubleList_26__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_12pyrobuf_list_10DoubleList_11__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12pyrobuf_list_DoubleList, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list_DoubleList, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_12pyrobuf_list_FloatList __pyx_vtable_12pyrobuf_list_FloatList;

static PyObject *__pyx_tp_new_12pyrobuf_list_FloatList(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_12pyrobuf_list_FloatList *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12pyrobuf_list_FloatList *)o);
  p->__pyx_vtab = __pyx_vtabptr_12pyrobuf_list_FloatList;
  if (unlikely(__pyx_pw_12pyrobuf_list_9FloatList_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12pyrobuf_list_FloatList(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_12pyrobuf_list_9FloatList_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_12pyrobuf_list_FloatList(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_12pyrobuf_list_FloatList(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_12pyrobuf_list_9FloatList_18__setitem__(o, i, v);
  }
  else {
    return __pyx_pw_12pyrobuf_list_9FloatList_7__delitem__(o, i);
  }
}

static PyMethodDef __pyx_methods_12pyrobuf_list_FloatList[] = {
  {"__neq__", (PyCFunction)__pyx_pw_12pyrobuf_list_9FloatList_28__neq__, METH_O, 0},
  {"_equal", (PyCFunction)__pyx_pw_12pyrobuf_list_9FloatList_30_equal, METH_O, 0},
  {"append", (PyCFunction)__pyx_pw_12pyrobuf_list_9FloatList_32append, METH_O, 0},
  {"extend", (PyCFunction)__pyx_pw_12pyrobuf_list_9FloatList_34extend, METH_O, 0},
  {"insert", (PyCFunction)__pyx_pw_12pyrobuf_list_9FloatList_36insert, METH_VARARGS|METH_KEYWORDS, 0},
  {"pop", (PyCFunction)__pyx_pw_12pyrobuf_list_9FloatList_38pop, METH_NOARGS, 0},
  {"remove", (PyCFunction)__pyx_pw_12pyrobuf_list_9FloatList_40remove, METH_O, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_9FloatList_42__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_9FloatList_44__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_FloatList = {
  __pyx_pw_12pyrobuf_list_9FloatList_14__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_12pyrobuf_list_FloatList, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_12pyrobuf_list_9FloatList_5__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_FloatList = {
  __pyx_pw_12pyrobuf_list_9FloatList_14__len__, /*mp_length*/
  __pyx_pw_12pyrobuf_list_9FloatList_9__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_12pyrobuf_list_FloatList, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_FloatList = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_pw_12pyrobuf_list_9FloatList_22__getbuffer__, /*bf_getbuffer*/
  __pyx_pw_12pyrobuf_list_9FloatList_24__releasebuffer__, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type_12pyrobuf_list_FloatList = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.FloatList", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list_FloatList), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list_FloatList, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_12pyrobuf_list_9FloatList_16__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_FloatList, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_FloatList, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_12pyrobuf_list_9FloatList_20__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_FloatList, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_12pyrobuf_list_9FloatList_26__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_12pyrobuf_list_9FloatList_11__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12pyrobuf_list_FloatList, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list_FloatList, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_12pyrobuf_list_IntList __pyx_vtable_12pyrobuf_list_IntList;

static PyObject *__pyx_tp_new_12pyrobuf_list_IntList(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_12pyrobuf_list_IntList *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12pyrobuf_list_IntList *)o);
  p->__pyx_vtab = __pyx_vtabptr_12pyrobuf_list_IntList;
  if (unlikely(__pyx_pw_12pyrobuf_list_7IntList_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12pyrobuf_list_IntList(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_12pyrobuf_list_7IntList_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_12pyrobuf_list_IntList(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_12pyrobuf_list_IntList(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_12pyrobuf_list_7IntList_18__setitem__(o, i, v);
  }
  else {
    return __pyx_pw_12pyrobuf_list_7IntList_7__delitem__(o, i);
  }
}

static PyMethodDef __pyx_methods_12pyrobuf_list_IntList[] = {
  {"__neq__", (PyCFunction)__pyx_pw_12pyrobuf_list_7IntList_28__neq__, METH_O, 0},
  {"_equal", (PyCFunction)__pyx_pw_12pyrobuf_list_7IntList_30_equal, METH_O, 0},
  {"append", (PyCFunction)__pyx_pw_12pyrobuf_list_7IntList_32append, METH_O, 0},
  {"extend", (PyCFunction)__pyx_pw_12pyrobuf_list_7IntList_34extend, METH_O, 0},
  {"insert", (PyCFunction)__pyx_pw_12pyrobuf_list_7IntList_36insert, METH_VARARGS|METH_KEYWORDS, 0},
  {"pop", (PyCFunction)__pyx_pw_12pyrobuf_list_7IntList_38pop, METH_NOARGS, 0},
  {"remove", (PyCFunction)__pyx_pw_12pyrobuf_list_7IntList_40remove, METH_O, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_7IntList_42__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_7IntList_44__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_IntList = {
  __pyx_pw_12pyrobuf_list_7IntList_14__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_12pyrobuf_list_IntList, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_12pyrobuf_list_7IntList_5__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_IntList = {
  __pyx_pw_12pyrobuf_list_7IntList_14__len__, /*mp_length*/
  __pyx_pw_12pyrobuf_list_7IntList_9__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_12pyrobuf_list_IntList, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_IntList = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_pw_12pyrobuf_list_7IntList_22__getbuffer__, /*bf_getbuffer*/
  __pyx_pw_12pyrobuf_list_7IntList_24__releasebuffer__, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type_12pyrobuf_list_IntList = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.IntList", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list_IntList), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list_IntList, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_12pyrobuf_list_7IntList_16__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_IntList, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_IntList, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_12pyrobuf_list_7IntList_20__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_IntList, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_12pyrobuf_list_7IntList_26__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_12pyrobuf_list_7IntList_11__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12pyrobuf_list_IntList, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list_IntList, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_12pyrobuf_list_Int32List __pyx_vtable_12pyrobuf_list_Int32List;

static PyObject *__pyx_tp_new_12pyrobuf_list_Int32List(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_12pyrobuf_list_Int32List *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12pyrobuf_list_Int32List *)o);
  p->__pyx_vtab = __pyx_vtabptr_12pyrobuf_list_Int32List;
  if (unlikely(__pyx_pw_12pyrobuf_list_9Int32List_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12pyrobuf_list_Int32List(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_12pyrobuf_list_9Int32List_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_12pyrobuf_list_Int32List(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_12pyrobuf_list_Int32List(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_12pyrobuf_list_9Int32List_18__setitem__(o, i, v);
  }
  else {
    return __pyx_pw_12pyrobuf_list_9Int32List_7__delitem__(o, i);
  }
}

static PyMethodDef __pyx_methods_12pyrobuf_list_Int32List[] = {
  {"__neq__", (PyCFunction)__pyx_pw_12pyrobuf_list_9Int32List_28__neq__, METH_O, 0},
  {"_equal", (PyCFunction)__pyx_pw_12pyrobuf_list_9Int32List_30_equal, METH_O, 0},
  {"append", (PyCFunction)__pyx_pw_12pyrobuf_list_9Int32List_32append, METH_O, 0},
  {"extend", (PyCFunction)__pyx_pw_12pyrobuf_list_9Int32List_34extend, METH_O, 0},
  {"insert", (PyCFunction)__pyx_pw_12pyrobuf_list_9Int32List_36insert, METH_VARARGS|METH_KEYWORDS, 0},
  {"pop", (PyCFunction)__pyx_pw_12pyrobuf_list_9Int32List_38pop, METH_NOARGS, 0},
  {"remove", (PyCFunction)__pyx_pw_12pyrobuf_list_9Int32List_40remove, METH_O, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_9Int32List_42__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_9Int32List_44__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_Int32List = {
  __pyx_pw_12pyrobuf_list_9Int32List_14__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_12pyrobuf_list_Int32List, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_12pyrobuf_list_9Int32List_5__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Int32List = {
  __pyx_pw_12pyrobuf_list_9Int32List_14__len__, /*mp_length*/
  __pyx_pw_12pyrobuf_list_9Int32List_9__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_12pyrobuf_list_Int32List, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_Int32List = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_pw_12pyrobuf_list_9Int32List_22__getbuffer__, /*bf_getbuffer*/
  __pyx_pw_12pyrobuf_list_9Int32List_24__releasebuffer__, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type_12pyrobuf_list_Int32List = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.Int32List", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list_Int32List), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list_Int32List, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_12pyrobuf_list_9Int32List_16__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Int32List, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Int32List, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_12pyrobuf_list_9Int32List_20__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_Int32List, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_12pyrobuf_list_9Int32List_26__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_12pyrobuf_list_9Int32List_11__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12pyrobuf_list_Int32List, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list_Int32List, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_12pyrobuf_list_Uint32List __pyx_vtable_12pyrobuf_list_Uint32List;

static PyObject *__pyx_tp_new_12pyrobuf_list_Uint32List(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_12pyrobuf_list_Uint32List *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12pyrobuf_list_Uint32List *)o);
  p->__pyx_vtab = __pyx_vtabptr_12pyrobuf_list_Uint32List;
  if (unlikely(__pyx_pw_12pyrobuf_list_10Uint32List_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12pyrobuf_list_Uint32List(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_12pyrobuf_list_10Uint32List_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_12pyrobuf_list_Uint32List(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_12pyrobuf_list_Uint32List(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_12pyrobuf_list_10Uint32List_18__setitem__(o, i, v);
  }
  else {
    return __pyx_pw_12pyrobuf_list_10Uint32List_7__delitem__(o, i);
  }
}

static PyMethodDef __pyx_methods_12pyrobuf_list_Uint32List[] = {
  {"__neq__", (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint32List_28__neq__, METH_O, 0},
  {"_equal", (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint32List_30_equal, METH_O, 0},
  {"append", (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint32List_32append, METH_O, 0},
  {"extend", (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint32List_34extend, METH_O, 0},
  {"insert", (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint32List_36insert, METH_VARARGS|METH_KEYWORDS, 0},
  {"pop", (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint32List_38pop, METH_NOARGS, 0},
  {"remove", (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint32List_40remove, METH_O, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint32List_42__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint32List_44__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_Uint32List = {
  __pyx_pw_12pyrobuf_list_10Uint32List_14__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_12pyrobuf_list_Uint32List, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_12pyrobuf_list_10Uint32List_5__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Uint32List = {
  __pyx_pw_12pyrobuf_list_10Uint32List_14__len__, /*mp_length*/
  __pyx_pw_12pyrobuf_list_10Uint32List_9__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_12pyrobuf_list_Uint32List, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_Uint32List = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_pw_12pyrobuf_list_10Uint32List_22__getbuffer__, /*bf_getbuffer*/
  __pyx_pw_12pyrobuf_list_10Uint32List_24__releasebuffer__, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type_12pyrobuf_list_Uint32List = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.Uint32List", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list_Uint32List), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list_Uint32List, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_12pyrobuf_list_10Uint32List_16__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Uint32List, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Uint32List, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_12pyrobuf_list_10Uint32List_20__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_Uint32List, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_12pyrobuf_list_10Uint32List_26__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_12pyrobuf_list_10Uint32List_11__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12pyrobuf_list_Uint32List, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list_Uint32List, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_12pyrobuf_list_Int64List __pyx_vtable_12pyrobuf_list_Int64List;

static PyObject *__pyx_tp_new_12pyrobuf_list_Int64List(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_12pyrobuf_list_Int64List *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12pyrobuf_list_Int64List *)o);
  p->__pyx_vtab = __pyx_vtabptr_12pyrobuf_list_Int64List;
  if (unlikely(__pyx_pw_12pyrobuf_list_9Int64List_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12pyrobuf_list_Int64List(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_12pyrobuf_list_9Int64List_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_12pyrobuf_list_Int64List(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_12pyrobuf_list_Int64List(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_12pyrobuf_list_9Int64List_18__setitem__(o, i, v);
  }
  else {
    return __pyx_pw_12pyrobuf_list_9Int64List_7__delitem__(o, i);
  }
}

static PyMethodDef __pyx_methods_12pyrobuf_list_Int64List[] = {
  {"__neq__", (PyCFunction)__pyx_pw_12pyrobuf_list_9Int64List_28__neq__, METH_O, 0},
  {"_equal", (PyCFunction)__pyx_pw_12pyrobuf_list_9Int64List_30_equal, METH_O, 0},
  {"append", (PyCFunction)__pyx_pw_12pyrobuf_list_9Int64List_32append, METH_O, 0},
  {"extend", (PyCFunction)__pyx_pw_12pyrobuf_list_9Int64List_34extend, METH_O, 0},
  {"insert", (PyCFunction)__pyx_pw_12pyrobuf_list_9Int64List_36insert, METH_VARARGS|METH_KEYWORDS, 0},
  {"pop", (PyCFunction)__pyx_pw_12pyrobuf_list_9Int64List_38pop, METH_NOARGS, 0},
  {"remove", (PyCFunction)__pyx_pw_12pyrobuf_list_9Int64List_40remove, METH_O, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_9Int64List_42__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_9Int64List_44__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_Int64List = {
  __pyx_pw_12pyrobuf_list_9Int64List_14__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_12pyrobuf_list_Int64List, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_12pyrobuf_list_9Int64List_5__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Int64List = {
  __pyx_pw_12pyrobuf_list_9Int64List_14__len__, /*mp_length*/
  __pyx_pw_12pyrobuf_list_9Int64List_9__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_12pyrobuf_list_Int64List, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_Int64List = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_pw_12pyrobuf_list_9Int64List_22__getbuffer__, /*bf_getbuffer*/
  __pyx_pw_12pyrobuf_list_9Int64List_24__releasebuffer__, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type_12pyrobuf_list_Int64List = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.Int64List", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list_Int64List), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list_Int64List, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_12pyrobuf_list_9Int64List_16__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Int64List, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Int64List, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_12pyrobuf_list_9Int64List_20__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_Int64List, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_12pyrobuf_list_9Int64List_26__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_12pyrobuf_list_9Int64List_11__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12pyrobuf_list_Int64List, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list_Int64List, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_12pyrobuf_list_Uint64List __pyx_vtable_12pyrobuf_list_Uint64List;

static PyObject *__pyx_tp_new_12pyrobuf_list_Uint64List(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_12pyrobuf_list_Uint64List *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12pyrobuf_list_Uint64List *)o);
  p->__pyx_vtab = __pyx_vtabptr_12pyrobuf_list_Uint64List;
  if (unlikely(__pyx_pw_12pyrobuf_list_10Uint64List_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12pyrobuf_list_Uint64List(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_12pyrobuf_list_10Uint64List_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_12pyrobuf_list_Uint64List(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_12pyrobuf_list_Uint64List(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_12pyrobuf_list_10Uint64List_18__setitem__(o, i, v);
  }
  else {
    return __pyx_pw_12pyrobuf_list_10Uint64List_7__delitem__(o, i);
  }
}

static PyMethodDef __pyx_methods_12pyrobuf_list_Uint64List[] = {
  {"__neq__", (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint64List_28__neq__, METH_O, 0},
  {"_equal", (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint64List_30_equal, METH_O, 0},
  {"append", (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint64List_32append, METH_O, 0},
  {"extend", (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint64List_34extend, METH_O, 0},
  {"insert", (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint64List_36insert, METH_VARARGS|METH_KEYWORDS, 0},
  {"pop", (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint64List_38pop, METH_NOARGS, 0},
  {"remove", (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint64List_40remove, METH_O, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint64List_42__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_10Uint64List_44__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_Uint64List = {
  __pyx_pw_12pyrobuf_list_10Uint64List_14__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_12pyrobuf_list_Uint64List, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_12pyrobuf_list_10Uint64List_5__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Uint64List = {
  __pyx_pw_12pyrobuf_list_10Uint64List_14__len__, /*mp_length*/
  __pyx_pw_12pyrobuf_list_10Uint64List_9__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_12pyrobuf_list_Uint64List, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_Uint64List = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_pw_12pyrobuf_list_10Uint64List_22__getbuffer__, /*bf_getbuffer*/
  __pyx_pw_12pyrobuf_list_10Uint64List_24__releasebuffer__, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type_12pyrobuf_list_Uint64List = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.Uint64List", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list_Uint64List), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list_Uint64List, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_12pyrobuf_list_10Uint64List_16__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Uint64List, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Uint64List, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_12pyrobuf_list_10Uint64List_20__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_Uint64List, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_12pyrobuf_list_10Uint64List_26__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_12pyrobuf_list_10Uint64List_11__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12pyrobuf_list_Uint64List, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list_Uint64List, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_12pyrobuf_list_CharList __pyx_vtable_12pyrobuf_list_CharList;

static PyObject *__pyx_tp_new_12pyrobuf_list_CharList(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_12pyrobuf_list_CharList *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_12pyrobuf_list_CharList *)o);
  p->__pyx_vtab = __pyx_vtabptr_12pyrobuf_list_CharList;
  if (unlikely(__pyx_pw_12pyrobuf_list_8CharList_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_12pyrobuf_list_CharList(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_12pyrobuf_list_8CharList_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_12pyrobuf_list_CharList(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_12pyrobuf_list_CharList(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_12pyrobuf_list_8CharList_18__setitem__(o, i, v);
  }
  else {
    return __pyx_pw_12pyrobuf_list_8CharList_7__delitem__(o, i);
  }
}

static PyMethodDef __pyx_methods_12pyrobuf_list_CharList[] = {
  {"__neq__", (PyCFunction)__pyx_pw_12pyrobuf_list_8CharList_28__neq__, METH_O, 0},
  {"_equal", (PyCFunction)__pyx_pw_12pyrobuf_list_8CharList_30_equal, METH_O, 0},
  {"append", (PyCFunction)__pyx_pw_12pyrobuf_list_8CharList_32append, METH_O, 0},
  {"extend", (PyCFunction)__pyx_pw_12pyrobuf_list_8CharList_34extend, METH_O, 0},
  {"insert", (PyCFunction)__pyx_pw_12pyrobuf_list_8CharList_36insert, METH_VARARGS|METH_KEYWORDS, 0},
  {"pop", (PyCFunction)__pyx_pw_12pyrobuf_list_8CharList_38pop, METH_NOARGS, 0},
  {"remove", (PyCFunction)__pyx_pw_12pyrobuf_list_8CharList_40remove, METH_O, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_8CharList_42__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_12pyrobuf_list_8CharList_44__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_CharList = {
  __pyx_pw_12pyrobuf_list_8CharList_14__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_12pyrobuf_list_CharList, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_12pyrobuf_list_8CharList_5__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_CharList = {
  __pyx_pw_12pyrobuf_list_8CharList_14__len__, /*mp_length*/
  __pyx_pw_12pyrobuf_list_8CharList_9__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_12pyrobuf_list_CharList, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_CharList = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_pw_12pyrobuf_list_8CharList_22__getbuffer__, /*bf_getbuffer*/
  __pyx_pw_12pyrobuf_list_8CharList_24__releasebuffer__, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type_12pyrobuf_list_CharList = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.CharList", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list_CharList), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list_CharList, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_12pyrobuf_list_8CharList_16__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_CharList, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_CharList, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_12pyrobuf_list_8CharList_20__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_CharList, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_pw_12pyrobuf_list_8CharList_26__richcmp__, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_12pyrobuf_list_8CharList_11__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_12pyrobuf_list_CharList, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list_CharList, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_12pyrobuf_list___pyx_scope_struct____iter__ *__pyx_freelist_12pyrobuf_list___pyx_scope_struct____iter__[8];
static int __pyx_freecount_12pyrobuf_list___pyx_scope_struct____iter__ = 0;

static PyObject *__pyx_tp_new_12pyrobuf_list___pyx_scope_struct____iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_12pyrobuf_list___pyx_scope_struct____iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct____iter__)))) {
    o = (PyObject*)__pyx_freelist_12pyrobuf_list___pyx_scope_struct____iter__[--__pyx_freecount_12pyrobuf_list___pyx_scope_struct____iter__];
    memset(o, 0, sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct____iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_12pyrobuf_list___pyx_scope_struct____iter__(PyObject *o) {
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct____iter__ *p = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct____iter__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_12pyrobuf_list___pyx_scope_struct____iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct____iter__)))) {
    __pyx_freelist_12pyrobuf_list___pyx_scope_struct____iter__[__pyx_freecount_12pyrobuf_list___pyx_scope_struct____iter__++] = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct____iter__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_12pyrobuf_list___pyx_scope_struct____iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct____iter__ *p = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct____iter__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_12pyrobuf_list___pyx_scope_struct____iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.__pyx_scope_struct____iter__", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct____iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list___pyx_scope_struct____iter__, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_12pyrobuf_list___pyx_scope_struct____iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list___pyx_scope_struct____iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_1___iter__ *__pyx_freelist_12pyrobuf_list___pyx_scope_struct_1___iter__[8];
static int __pyx_freecount_12pyrobuf_list___pyx_scope_struct_1___iter__ = 0;

static PyObject *__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_1___iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_12pyrobuf_list___pyx_scope_struct_1___iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_1___iter__)))) {
    o = (PyObject*)__pyx_freelist_12pyrobuf_list___pyx_scope_struct_1___iter__[--__pyx_freecount_12pyrobuf_list___pyx_scope_struct_1___iter__];
    memset(o, 0, sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_1___iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_12pyrobuf_list___pyx_scope_struct_1___iter__(PyObject *o) {
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_1___iter__ *p = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_1___iter__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_12pyrobuf_list___pyx_scope_struct_1___iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_1___iter__)))) {
    __pyx_freelist_12pyrobuf_list___pyx_scope_struct_1___iter__[__pyx_freecount_12pyrobuf_list___pyx_scope_struct_1___iter__++] = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_1___iter__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_12pyrobuf_list___pyx_scope_struct_1___iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_1___iter__ *p = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_1___iter__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_12pyrobuf_list___pyx_scope_struct_1___iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.__pyx_scope_struct_1___iter__", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_1___iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list___pyx_scope_struct_1___iter__, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_12pyrobuf_list___pyx_scope_struct_1___iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list___pyx_scope_struct_1___iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_2___iter__ *__pyx_freelist_12pyrobuf_list___pyx_scope_struct_2___iter__[8];
static int __pyx_freecount_12pyrobuf_list___pyx_scope_struct_2___iter__ = 0;

static PyObject *__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_2___iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_12pyrobuf_list___pyx_scope_struct_2___iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_2___iter__)))) {
    o = (PyObject*)__pyx_freelist_12pyrobuf_list___pyx_scope_struct_2___iter__[--__pyx_freecount_12pyrobuf_list___pyx_scope_struct_2___iter__];
    memset(o, 0, sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_2___iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_12pyrobuf_list___pyx_scope_struct_2___iter__(PyObject *o) {
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_2___iter__ *p = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_2___iter__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_12pyrobuf_list___pyx_scope_struct_2___iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_2___iter__)))) {
    __pyx_freelist_12pyrobuf_list___pyx_scope_struct_2___iter__[__pyx_freecount_12pyrobuf_list___pyx_scope_struct_2___iter__++] = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_2___iter__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_12pyrobuf_list___pyx_scope_struct_2___iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_2___iter__ *p = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_2___iter__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_12pyrobuf_list___pyx_scope_struct_2___iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.__pyx_scope_struct_2___iter__", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_2___iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list___pyx_scope_struct_2___iter__, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_12pyrobuf_list___pyx_scope_struct_2___iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list___pyx_scope_struct_2___iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_3___iter__ *__pyx_freelist_12pyrobuf_list___pyx_scope_struct_3___iter__[8];
static int __pyx_freecount_12pyrobuf_list___pyx_scope_struct_3___iter__ = 0;

static PyObject *__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_3___iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_12pyrobuf_list___pyx_scope_struct_3___iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_3___iter__)))) {
    o = (PyObject*)__pyx_freelist_12pyrobuf_list___pyx_scope_struct_3___iter__[--__pyx_freecount_12pyrobuf_list___pyx_scope_struct_3___iter__];
    memset(o, 0, sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_3___iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_12pyrobuf_list___pyx_scope_struct_3___iter__(PyObject *o) {
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_3___iter__ *p = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_3___iter__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_12pyrobuf_list___pyx_scope_struct_3___iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_3___iter__)))) {
    __pyx_freelist_12pyrobuf_list___pyx_scope_struct_3___iter__[__pyx_freecount_12pyrobuf_list___pyx_scope_struct_3___iter__++] = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_3___iter__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_12pyrobuf_list___pyx_scope_struct_3___iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_3___iter__ *p = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_3___iter__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_12pyrobuf_list___pyx_scope_struct_3___iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.__pyx_scope_struct_3___iter__", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_3___iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list___pyx_scope_struct_3___iter__, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_12pyrobuf_list___pyx_scope_struct_3___iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list___pyx_scope_struct_3___iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_4___iter__ *__pyx_freelist_12pyrobuf_list___pyx_scope_struct_4___iter__[8];
static int __pyx_freecount_12pyrobuf_list___pyx_scope_struct_4___iter__ = 0;

static PyObject *__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_4___iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_12pyrobuf_list___pyx_scope_struct_4___iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_4___iter__)))) {
    o = (PyObject*)__pyx_freelist_12pyrobuf_list___pyx_scope_struct_4___iter__[--__pyx_freecount_12pyrobuf_list___pyx_scope_struct_4___iter__];
    memset(o, 0, sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_4___iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_12pyrobuf_list___pyx_scope_struct_4___iter__(PyObject *o) {
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_4___iter__ *p = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_4___iter__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_12pyrobuf_list___pyx_scope_struct_4___iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_4___iter__)))) {
    __pyx_freelist_12pyrobuf_list___pyx_scope_struct_4___iter__[__pyx_freecount_12pyrobuf_list___pyx_scope_struct_4___iter__++] = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_4___iter__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_12pyrobuf_list___pyx_scope_struct_4___iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_4___iter__ *p = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_4___iter__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_12pyrobuf_list___pyx_scope_struct_4___iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.__pyx_scope_struct_4___iter__", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_4___iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list___pyx_scope_struct_4___iter__, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_12pyrobuf_list___pyx_scope_struct_4___iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list___pyx_scope_struct_4___iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_5___iter__ *__pyx_freelist_12pyrobuf_list___pyx_scope_struct_5___iter__[8];
static int __pyx_freecount_12pyrobuf_list___pyx_scope_struct_5___iter__ = 0;

static PyObject *__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_5___iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_12pyrobuf_list___pyx_scope_struct_5___iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_5___iter__)))) {
    o = (PyObject*)__pyx_freelist_12pyrobuf_list___pyx_scope_struct_5___iter__[--__pyx_freecount_12pyrobuf_list___pyx_scope_struct_5___iter__];
    memset(o, 0, sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_5___iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_12pyrobuf_list___pyx_scope_struct_5___iter__(PyObject *o) {
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_5___iter__ *p = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_5___iter__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_12pyrobuf_list___pyx_scope_struct_5___iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_5___iter__)))) {
    __pyx_freelist_12pyrobuf_list___pyx_scope_struct_5___iter__[__pyx_freecount_12pyrobuf_list___pyx_scope_struct_5___iter__++] = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_5___iter__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_12pyrobuf_list___pyx_scope_struct_5___iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_5___iter__ *p = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_5___iter__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_12pyrobuf_list___pyx_scope_struct_5___iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.__pyx_scope_struct_5___iter__", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_5___iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list___pyx_scope_struct_5___iter__, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_12pyrobuf_list___pyx_scope_struct_5___iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list___pyx_scope_struct_5___iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_6___iter__ *__pyx_freelist_12pyrobuf_list___pyx_scope_struct_6___iter__[8];
static int __pyx_freecount_12pyrobuf_list___pyx_scope_struct_6___iter__ = 0;

static PyObject *__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_6___iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_12pyrobuf_list___pyx_scope_struct_6___iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_6___iter__)))) {
    o = (PyObject*)__pyx_freelist_12pyrobuf_list___pyx_scope_struct_6___iter__[--__pyx_freecount_12pyrobuf_list___pyx_scope_struct_6___iter__];
    memset(o, 0, sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_6___iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_12pyrobuf_list___pyx_scope_struct_6___iter__(PyObject *o) {
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_6___iter__ *p = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_6___iter__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_12pyrobuf_list___pyx_scope_struct_6___iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_6___iter__)))) {
    __pyx_freelist_12pyrobuf_list___pyx_scope_struct_6___iter__[__pyx_freecount_12pyrobuf_list___pyx_scope_struct_6___iter__++] = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_6___iter__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_12pyrobuf_list___pyx_scope_struct_6___iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_6___iter__ *p = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_6___iter__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_12pyrobuf_list___pyx_scope_struct_6___iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.__pyx_scope_struct_6___iter__", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_6___iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list___pyx_scope_struct_6___iter__, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_12pyrobuf_list___pyx_scope_struct_6___iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list___pyx_scope_struct_6___iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_7___iter__ *__pyx_freelist_12pyrobuf_list___pyx_scope_struct_7___iter__[8];
static int __pyx_freecount_12pyrobuf_list___pyx_scope_struct_7___iter__ = 0;

static PyObject *__pyx_tp_new_12pyrobuf_list___pyx_scope_struct_7___iter__(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_12pyrobuf_list___pyx_scope_struct_7___iter__ > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_7___iter__)))) {
    o = (PyObject*)__pyx_freelist_12pyrobuf_list___pyx_scope_struct_7___iter__[--__pyx_freecount_12pyrobuf_list___pyx_scope_struct_7___iter__];
    memset(o, 0, sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_7___iter__));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_12pyrobuf_list___pyx_scope_struct_7___iter__(PyObject *o) {
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_7___iter__ *p = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_7___iter__ *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_12pyrobuf_list___pyx_scope_struct_7___iter__ < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_7___iter__)))) {
    __pyx_freelist_12pyrobuf_list___pyx_scope_struct_7___iter__[__pyx_freecount_12pyrobuf_list___pyx_scope_struct_7___iter__++] = ((struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_7___iter__ *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_12pyrobuf_list___pyx_scope_struct_7___iter__(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_7___iter__ *p = (struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_7___iter__ *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_12pyrobuf_list___pyx_scope_struct_7___iter__ = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyrobuf_list.__pyx_scope_struct_7___iter__", /*tp_name*/
  sizeof(struct __pyx_obj_12pyrobuf_list___pyx_scope_struct_7___iter__), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_12pyrobuf_list___pyx_scope_struct_7___iter__, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_12pyrobuf_list___pyx_scope_struct_7___iter__, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_12pyrobuf_list___pyx_scope_struct_7___iter__, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_pyrobuf_list(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_pyrobuf_list},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "pyrobuf_list",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
  {&__pyx_n_s_CharList___iter, __pyx_k_CharList___iter, sizeof(__pyx_k_CharList___iter), 0, 0, 1, 1},
  {&__pyx_n_s_DoubleList___iter, __pyx_k_DoubleList___iter, sizeof(__pyx_k_DoubleList___iter), 0, 0, 1, 1},
  {&__pyx_n_s_FloatList___iter, __pyx_k_FloatList___iter, sizeof(__pyx_k_FloatList___iter), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_k_Incompatible_checksums_s_vs_0xd4, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xd4), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_n_s_Int32List___iter, __pyx_k_Int32List___iter, sizeof(__pyx_k_Int32List___iter), 0, 0, 1, 1},
  {&__pyx_n_s_Int64List___iter, __pyx_k_Int64List___iter, sizeof(__pyx_k_Int64List___iter), 0, 0, 1, 1},
  {&__pyx_n_s_IntList___iter, __pyx_k_IntList___iter, sizeof(__pyx_k_IntList___iter), 0, 0, 1, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_n_s_MergeFrom, __pyx_k_MergeFrom, sizeof(__pyx_k_MergeFrom), 0, 0, 1, 1},
  {&__pyx_n_s_Modified, __pyx_k_Modified, sizeof(__pyx_k_Modified), 0, 0, 1, 1},
  {&__pyx_kp_s_Operation_not_supported_for_list, __pyx_k_Operation_not_supported_for_list, sizeof(__pyx_k_Operation_not_supported_for_list), 0, 0, 1, 0},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_Uint32List___iter, __pyx_k_Uint32List___iter, sizeof(__pyx_k_Uint32List___iter), 0, 0, 1, 1},
  {&__pyx_n_s_Uint64List___iter, __pyx_k_Uint64List___iter, sizeof(__pyx_k_Uint64List___iter), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_equal, __pyx_k_equal, sizeof(__pyx_k_equal), 0, 0, 1, 1},
  {&__pyx_n_s_extend, __pyx_k_extend, sizeof(__pyx_k_extend), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_insert, __pyx_k_insert, sizeof(__pyx_k_insert), 0, 0, 1, 1},
  {&__pyx_n_s_is_present_in_parent, __pyx_k_is_present_in_parent, sizeof(__pyx_k_is_present_in_parent), 0, 0, 1, 1},
  {&__pyx_n_s_iter, __pyx_k_iter, sizeof(__pyx_k_iter), 0, 0, 1, 1},
  {&__pyx_kp_s_list_index_out_of_range, __pyx_k_list_index_out_of_range, sizeof(__pyx_k_list_index_out_of_range), 0, 0, 1, 0},
  {&__pyx_kp_s_list_indices_must_be_integers_no, __pyx_k_list_indices_must_be_integers_no, sizeof(__pyx_k_list_indices_must_be_integers_no), 0, 0, 1, 0},
  {&__pyx_n_s_list_type, __pyx_k_list_type, sizeof(__pyx_k_list_type), 0, 0, 1, 1},
  {&__pyx_n_s_listener, __pyx_k_listener, sizeof(__pyx_k_listener), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_null_listener, __pyx_k_null_listener, sizeof(__pyx_k_null_listener), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_kp_s_pop_from_empty_list, __pyx_k_pop_from_empty_list, sizeof(__pyx_k_pop_from_empty_list), 0, 0, 1, 0},
  {&__pyx_n_s_pyrobuf_list, __pyx_k_pyrobuf_list, sizeof(__pyx_k_pyrobuf_list), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_capi, __pyx_k_pyx_capi, sizeof(__pyx_k_pyx_capi), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_NullListener, __pyx_k_pyx_unpickle_NullListener, sizeof(__pyx_k_pyx_unpickle_NullListener), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_kp_s_r_has_type_s_but_expected_one_o, __pyx_k_r_has_type_s_but_expected_one_o, sizeof(__pyx_k_r_has_type_s_but_expected_one_o), 0, 0, 1, 0},
  {&__pyx_kp_s_r_has_type_s_but_expected_one_o_2, __pyx_k_r_has_type_s_but_expected_one_o_2, sizeof(__pyx_k_r_has_type_s_but_expected_one_o_2), 0, 0, 1, 0},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_remove, __pyx_k_remove, sizeof(__pyx_k_remove), 0, 0, 1, 1},
  {&__pyx_kp_s_self__listener_cannot_be_convert, __pyx_k_self__listener_cannot_be_convert, sizeof(__pyx_k_self__listener_cannot_be_convert), 0, 0, 1, 0},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_setitem, __pyx_k_setitem, sizeof(__pyx_k_setitem), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_kp_s_type_mismatch, __pyx_k_type_mismatch, sizeof(__pyx_k_type_mismatch), 0, 0, 1, 0},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_kp_s_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 0, 1, 0},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_kp_s_x_not_in_list, __pyx_k_x_not_in_list, sizeof(__pyx_k_x_not_in_list), 0, 0, 1, 0},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(0, 34, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 37, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(0, 246, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 260, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 271, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 357, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 425, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "pyrobuf_list.pyx":35
 *             assert isinstance(x, self._list_type)
 *         except AssertionError:
 *             raise Exception("type mismatch")             # <<<<<<<<<<<<<<
 * 
 *         super(TypedList, self).__setitem__(i, x)
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_type_mismatch); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "pyrobuf_list.pyx":56
 *             assert isinstance(x, self._list_type)
 *         except AssertionError:
 *             raise Exception("type mismatch")             # <<<<<<<<<<<<<<
 * 
 *         super(TypedList, self).append(x)
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_type_mismatch); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "pyrobuf_list.pyx":77
 *             assert isinstance(x, self._list_type)
 *         except AssertionError:
 *             raise Exception("type mismatch")             # <<<<<<<<<<<<<<
 * 
 *         super(TypedList, self).insert(i, x)
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_type_mismatch); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_self__listener_cannot_be_convert); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "(tree fragment)":4
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_self__listener_cannot_be_convert); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_self__listener_cannot_be_convert); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "(tree fragment)":4
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_self__listener_cannot_be_convert); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "pyrobuf_list.pyx":166
 *                 value = x
 *             elif isinstance(x, bytes):
 *                 value = x.decode('utf-8')             # <<<<<<<<<<<<<<
 *             else:
 *                 raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (x, type(x), bytes, str))
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "pyrobuf_list.pyx":178
 *                     values.append(val)
 *                 elif isinstance(val, bytes):
 *                     values.append(val.decode('utf-8'))             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (val, type(val), bytes, str))
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "pyrobuf_list.pyx":197
 *             value = x
 *         elif isinstance(x, bytes):
 *             value = x.decode('utf-8')             # <<<<<<<<<<<<<<
 *         else:
 *             raise TypeError("%r has type %s, but expected one of: (%s, %s)" % (x, type(x), bytes, str))
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "pyrobuf_list.pyx":211
 *         for val in x:
 *             if isinstance(val, bytes):
 *                 value = val.decode('utf-8')             # <<<<<<<<<<<<<<
 *             elif isinstance(val, str):
 *                 value = val
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "pyrobuf_list.pyx":225
 *         cdef str value
 *         if isinstance(x, bytes):
 *             value = x.decode('utf-8')             # <<<<<<<<<<<<<<
 *         elif isinstance(x, str):
 *             value = x
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_self__listener_cannot_be_convert); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "(tree fragment)":4
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self._listener cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_self__listener_cannot_be_convert); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "pyrobuf_list.pyx":271
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t j
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "pyrobuf_list.pyx":287
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         return self._data[i]
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "pyrobuf_list.pyx":308
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "pyrobuf_list.pyx":345
 *             return not self._equal(other)
 *         else:
 *             raise TypeError('Operation not supported for lists')             # <<<<<<<<<<<<<<
 * 
 *     def __neq__(self, other):
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_Operation_not_supported_for_list); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "pyrobuf_list.pyx":385
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         if self._n_items == self._size:
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "pyrobuf_list.pyx":406
 *     cpdef pop(self):
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_s_pop_from_empty_list); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "pyrobuf_list.pyx":425
 * 
 *         if found == 0:
 *             raise ValueError("x not in list")             # <<<<<<<<<<<<<<
 * 
 *         for j in range(i, self._n_items):
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_s_x_not_in_list); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "pyrobuf_list.pyx":481
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t j
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "pyrobuf_list.pyx":497
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         return self._data[i]
 */
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "pyrobuf_list.pyx":518
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "pyrobuf_list.pyx":555
 *             return not self._equal(other)
 *         else:
 *             raise TypeError('Operation not supported for lists')             # <<<<<<<<<<<<<<
 * 
 *     def __neq__(self, other):
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_s_Operation_not_supported_for_list); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "pyrobuf_list.pyx":595
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         if self._n_items == self._size:
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "pyrobuf_list.pyx":616
 *     cpdef pop(self):
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_s_pop_from_empty_list); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "pyrobuf_list.pyx":635
 * 
 *         if found == 0:
 *             raise ValueError("x not in list")             # <<<<<<<<<<<<<<
 * 
 *         for j in range(i, self._n_items):
 */
  __pyx_tuple__30 = PyTuple_Pack(1, __pyx_kp_s_x_not_in_list); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__31 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "pyrobuf_list.pyx":691
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t j
 */
  __pyx_tuple__33 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "pyrobuf_list.pyx":707
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         return self._data[i]
 */
  __pyx_tuple__34 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "pyrobuf_list.pyx":728
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
  __pyx_tuple__35 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(0, 728, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "pyrobuf_list.pyx":765
 *             return not self._equal(other)
 *         else:
 *             raise TypeError('Operation not supported for lists')             # <<<<<<<<<<<<<<
 * 
 *     def __neq__(self, other):
 */
  __pyx_tuple__36 = PyTuple_Pack(1, __pyx_kp_s_Operation_not_supported_for_list); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "pyrobuf_list.pyx":805
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         if self._n_items == self._size:
 */
  __pyx_tuple__37 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);

  /* "pyrobuf_list.pyx":826
 *     cpdef pop(self):
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
  __pyx_tuple__38 = PyTuple_Pack(1, __pyx_kp_s_pop_from_empty_list); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "pyrobuf_list.pyx":845
 * 
 *         if found == 0:
 *             raise ValueError("x not in list")             # <<<<<<<<<<<<<<
 * 
 *         for j in range(i, self._n_items):
 */
  __pyx_tuple__39 = PyTuple_Pack(1, __pyx_kp_s_x_not_in_list); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__40 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__41 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);

  /* "pyrobuf_list.pyx":901
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t j
 */
  __pyx_tuple__42 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);

  /* "pyrobuf_list.pyx":917
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         return self._data[i]
 */
  __pyx_tuple__43 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);

  /* "pyrobuf_list.pyx":938
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
  __pyx_tuple__44 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(0, 938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);

  /* "pyrobuf_list.pyx":975
 *             return not self._equal(other)
 *         else:
 *             raise TypeError('Operation not supported for lists')             # <<<<<<<<<<<<<<
 * 
 *     def __neq__(self, other):
 */
  __pyx_tuple__45 = PyTuple_Pack(1, __pyx_kp_s_Operation_not_supported_for_list); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);

  /* "pyrobuf_list.pyx":1015
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         if self._n_items == self._size:
 */
  __pyx_tuple__46 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);

  /* "pyrobuf_list.pyx":1036
 *     cpdef pop(self):
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
  __pyx_tuple__47 = PyTuple_Pack(1, __pyx_kp_s_pop_from_empty_list); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(0, 1036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);

  /* "pyrobuf_list.pyx":1055
 * 
 *         if found == 0:
 *             raise ValueError("x not in list")             # <<<<<<<<<<<<<<
 * 
 *         for j in range(i, self._n_items):
 */
  __pyx_tuple__48 = PyTuple_Pack(1, __pyx_kp_s_x_not_in_list); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(0, 1055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__49 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__50 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);

  /* "pyrobuf_list.pyx":1111
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t j
 */
  __pyx_tuple__51 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(0, 1111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);

  /* "pyrobuf_list.pyx":1127
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         return self._data[i]
 */
  __pyx_tuple__52 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(0, 1127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);

  /* "pyrobuf_list.pyx":1148
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
  __pyx_tuple__53 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(0, 1148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);

  /* "pyrobuf_list.pyx":1185
 *             return not self._equal(other)
 *         else:
 *             raise TypeError('Operation not supported for lists')             # <<<<<<<<<<<<<<
 * 
 *     def __neq__(self, other):
 */
  __pyx_tuple__54 = PyTuple_Pack(1, __pyx_kp_s_Operation_not_supported_for_list); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(0, 1185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);

  /* "pyrobuf_list.pyx":1225
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         if self._n_items == self._size:
 */
  __pyx_tuple__55 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(0, 1225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);

  /* "pyrobuf_list.pyx":1246
 *     cpdef pop(self):
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
  __pyx_tuple__56 = PyTuple_Pack(1, __pyx_kp_s_pop_from_empty_list); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(0, 1246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);

  /* "pyrobuf_list.pyx":1265
 * 
 *         if found == 0:
 *             raise ValueError("x not in list")             # <<<<<<<<<<<<<<
 * 
 *         for j in range(i, self._n_items):
 */
  __pyx_tuple__57 = PyTuple_Pack(1, __pyx_kp_s_x_not_in_list); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__58 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__59 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);

  /* "pyrobuf_list.pyx":1321
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t j
 */
  __pyx_tuple__60 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__60)) __PYX_ERR(0, 1321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__60);
  __Pyx_GIVEREF(__pyx_tuple__60);

  /* "pyrobuf_list.pyx":1337
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         return self._data[i]
 */
  __pyx_tuple__61 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__61)) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__61);
  __Pyx_GIVEREF(__pyx_tuple__61);

  /* "pyrobuf_list.pyx":1358
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
  __pyx_tuple__62 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__62)) __PYX_ERR(0, 1358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__62);
  __Pyx_GIVEREF(__pyx_tuple__62);

  /* "pyrobuf_list.pyx":1395
 *             return not self._equal(other)
 *         else:
 *             raise TypeError('Operation not supported for lists')             # <<<<<<<<<<<<<<
 * 
 *     def __neq__(self, other):
 */
  __pyx_tuple__63 = PyTuple_Pack(1, __pyx_kp_s_Operation_not_supported_for_list); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(0, 1395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);

  /* "pyrobuf_list.pyx":1435
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         if self._n_items == self._size:
 */
  __pyx_tuple__64 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__64)) __PYX_ERR(0, 1435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);

  /* "pyrobuf_list.pyx":1456
 *     cpdef pop(self):
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
  __pyx_tuple__65 = PyTuple_Pack(1, __pyx_kp_s_pop_from_empty_list); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(0, 1456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);

  /* "pyrobuf_list.pyx":1475
 * 
 *         if found == 0:
 *             raise ValueError("x not in list")             # <<<<<<<<<<<<<<
 * 
 *         for j in range(i, self._n_items):
 */
  __pyx_tuple__66 = PyTuple_Pack(1, __pyx_kp_s_x_not_in_list); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(0, 1475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__67 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__68 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);

  /* "pyrobuf_list.pyx":1531
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t j
 */
  __pyx_tuple__69 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__69)) __PYX_ERR(0, 1531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__69);
  __Pyx_GIVEREF(__pyx_tuple__69);

  /* "pyrobuf_list.pyx":1547
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         return self._data[i]
 */
  __pyx_tuple__70 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__70)) __PYX_ERR(0, 1547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__70);
  __Pyx_GIVEREF(__pyx_tuple__70);

  /* "pyrobuf_list.pyx":1568
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
  __pyx_tuple__71 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__71)) __PYX_ERR(0, 1568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);

  /* "pyrobuf_list.pyx":1605
 *             return not self._equal(other)
 *         else:
 *             raise TypeError('Operation not supported for lists')             # <<<<<<<<<<<<<<
 * 
 *     def __neq__(self, other):
 */
  __pyx_tuple__72 = PyTuple_Pack(1, __pyx_kp_s_Operation_not_supported_for_list); if (unlikely(!__pyx_tuple__72)) __PYX_ERR(0, 1605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__72);
  __Pyx_GIVEREF(__pyx_tuple__72);

  /* "pyrobuf_list.pyx":1645
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         if self._n_items == self._size:
 */
  __pyx_tuple__73 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__73)) __PYX_ERR(0, 1645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);

  /* "pyrobuf_list.pyx":1666
 *     cpdef pop(self):
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
  __pyx_tuple__74 = PyTuple_Pack(1, __pyx_kp_s_pop_from_empty_list); if (unlikely(!__pyx_tuple__74)) __PYX_ERR(0, 1666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__74);
  __Pyx_GIVEREF(__pyx_tuple__74);

  /* "pyrobuf_list.pyx":1685
 * 
 *         if found == 0:
 *             raise ValueError("x not in list")             # <<<<<<<<<<<<<<
 * 
 *         for j in range(i, self._n_items):
 */
  __pyx_tuple__75 = PyTuple_Pack(1, __pyx_kp_s_x_not_in_list); if (unlikely(!__pyx_tuple__75)) __PYX_ERR(0, 1685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__75);
  __Pyx_GIVEREF(__pyx_tuple__75);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__76 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__76)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__76);
  __Pyx_GIVEREF(__pyx_tuple__76);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__77 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__77)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__77);
  __Pyx_GIVEREF(__pyx_tuple__77);

  /* "pyrobuf_list.pyx":1741
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         cdef size_t j
 */
  __pyx_tuple__78 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__78)) __PYX_ERR(0, 1741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__78);
  __Pyx_GIVEREF(__pyx_tuple__78);

  /* "pyrobuf_list.pyx":1757
 * 
 *         if i < 0 or <size_t>i >= self._n_items or i < 0:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         return self._data[i]
 */
  __pyx_tuple__79 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__79)) __PYX_ERR(0, 1757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__79);
  __Pyx_GIVEREF(__pyx_tuple__79);

  /* "pyrobuf_list.pyx":1778
 * 
 *         if i < 0 or <size_t>i >= self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         self._data[i] = x
 */
  __pyx_tuple__80 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__80)) __PYX_ERR(0, 1778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__80);
  __Pyx_GIVEREF(__pyx_tuple__80);

  /* "pyrobuf_list.pyx":1815
 *             return not self._equal(other)
 *         else:
 *             raise TypeError('Operation not supported for lists')             # <<<<<<<<<<<<<<
 * 
 *     def __neq__(self, other):
 */
  __pyx_tuple__81 = PyTuple_Pack(1, __pyx_kp_s_Operation_not_supported_for_list); if (unlikely(!__pyx_tuple__81)) __PYX_ERR(0, 1815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__81);
  __Pyx_GIVEREF(__pyx_tuple__81);

  /* "pyrobuf_list.pyx":1855
 * 
 *         if i < 0 or <size_t>i > self._n_items:
 *             raise IndexError("list index out of range")             # <<<<<<<<<<<<<<
 * 
 *         if self._n_items == self._size:
 */
  __pyx_tuple__82 = PyTuple_Pack(1, __pyx_kp_s_list_index_out_of_range); if (unlikely(!__pyx_tuple__82)) __PYX_ERR(0, 1855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__82);
  __Pyx_GIVEREF(__pyx_tuple__82);

  /* "pyrobuf_list.pyx":1876
 *     cpdef pop(self):
 *         if self._n_items == 0:
 *             raise IndexError("pop from empty list")             # <<<<<<<<<<<<<<
 * 
 *         self._n_items -= 1
 */
  __pyx_tuple__83 = PyTuple_Pack(1, __pyx_kp_s_pop_from_empty_list); if (unlikely(!__pyx_tuple__83)) __PYX_ERR(0, 1876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__83);
  __Pyx_GIVEREF(__pyx_tuple__83);

  /* "pyrobuf_list.pyx":1895
 * 
 *         if found == 0:
 *             raise ValueError("x not in list")             # <<<<<<<<<<<<<<
 * 
 *         for j in range(i, self._n_items):
 */
  __pyx_tuple__84 = PyTuple_Pack(1, __pyx_kp_s_x_not_in_list); if (unlikely(!__pyx_tuple__84)) __PYX_ERR(0, 1895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__84);
  __Pyx_GIVEREF(__pyx_tuple__84);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__85 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__85)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__85);
  __Pyx_GIVEREF(__pyx_tuple__85);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__86 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__86)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__86);
  __Pyx_GIVEREF(__pyx_tuple__86);

  /* "(tree fragment)":1
 * def __pyx_unpickle_NullListener(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 */
  __pyx_tuple__87 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__87)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__87);
  __Pyx_GIVEREF(__pyx_tuple__87);
  __pyx_codeobj__88 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__87, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_NullListener, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__88)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_222419149 = PyInt_FromLong(222419149L); if (unlikely(!__pyx_int_222419149)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_modinit_global_init_code(void); /*proto*/
static int __Pyx_modinit_variable_export_code(void); /*proto*/
static int __Pyx_modinit_function_export_code(void); /*proto*/
static int __Pyx_modinit_type_init_code(void); /*proto*/
static int __Pyx_modinit_type_import_code(void); /*proto*/
static int __Pyx_modinit_variable_import_code(void); /*proto*/
static int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  null_listener = ((struct __pyx_obj_12pyrobuf_list_NullListener *)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  if (__Pyx_ExportVoidPtr(__pyx_n_s_null_listener, (void *)&null_listener, "struct __pyx_obj_12pyrobuf_list_NullListener *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_12pyrobuf_list_NullListener = &__pyx_vtable_12pyrobuf_list_NullListener;
  __pyx_vtable_12pyrobuf_list_NullListener._Modified = (void (*)(struct __pyx_obj_12pyrobuf_list_NullListener *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_12NullListener__Modified;
  if (PyType_Ready(&__pyx_type_12pyrobuf_list_NullListener) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __pyx_type_12pyrobuf_list_NullListener.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list_NullListener.tp_dictoffset && __pyx_type_12pyrobuf_list_NullListener.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list_NullListener.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12pyrobuf_list_NullListener.tp_dict, __pyx_vtabptr_12pyrobuf_list_NullListener) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "NullListener", (PyObject *)&__pyx_type_12pyrobuf_list_NullListener) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12pyrobuf_list_NullListener) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __pyx_ptype_12pyrobuf_list_NullListener = &__pyx_type_12pyrobuf_list_NullListener;
  __pyx_type_12pyrobuf_list_TypedList.tp_base = (&PyList_Type);
  if (PyType_Ready(&__pyx_type_12pyrobuf_list_TypedList) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __pyx_type_12pyrobuf_list_TypedList.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list_TypedList.tp_dictoffset && __pyx_type_12pyrobuf_list_TypedList.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list_TypedList.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttrString(__pyx_m, "TypedList", (PyObject *)&__pyx_type_12pyrobuf_list_TypedList) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12pyrobuf_list_TypedList) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __pyx_ptype_12pyrobuf_list_TypedList = &__pyx_type_12pyrobuf_list_TypedList;
  __pyx_type_12pyrobuf_list_BytesList.tp_base = (&PyList_Type);
  if (PyType_Ready(&__pyx_type_12pyrobuf_list_BytesList) < 0) __PYX_ERR(0, 85, __pyx_L1_error)
  __pyx_type_12pyrobuf_list_BytesList.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list_BytesList.tp_dictoffset && __pyx_type_12pyrobuf_list_BytesList.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list_BytesList.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttrString(__pyx_m, "BytesList", (PyObject *)&__pyx_type_12pyrobuf_list_BytesList) < 0) __PYX_ERR(0, 85, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12pyrobuf_list_BytesList) < 0) __PYX_ERR(0, 85, __pyx_L1_error)
  __pyx_ptype_12pyrobuf_list_BytesList = &__pyx_type_12pyrobuf_list_BytesList;
  __pyx_type_12pyrobuf_list_StringList.tp_base = (&PyList_Type);
  if (PyType_Ready(&__pyx_type_12pyrobuf_list_StringList) < 0) __PYX_ERR(0, 150, __pyx_L1_error)
  __pyx_type_12pyrobuf_list_StringList.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list_StringList.tp_dictoffset && __pyx_type_12pyrobuf_list_StringList.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list_StringList.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttrString(__pyx_m, "StringList", (PyObject *)&__pyx_type_12pyrobuf_list_StringList) < 0) __PYX_ERR(0, 150, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12pyrobuf_list_StringList) < 0) __PYX_ERR(0, 150, __pyx_L1_error)
  __pyx_ptype_12pyrobuf_list_StringList = &__pyx_type_12pyrobuf_list_StringList;
  __pyx_vtabptr_12pyrobuf_list_DoubleList = &__pyx_vtable_12pyrobuf_list_DoubleList;
  __pyx_vtable_12pyrobuf_list_DoubleList._append = (void (*)(struct __pyx_obj_12pyrobuf_list_DoubleList *, double))__pyx_f_12pyrobuf_list_10DoubleList__append;
  __pyx_vtable_12pyrobuf_list_DoubleList._equal = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_DoubleList *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_10DoubleList__equal;
  __pyx_vtable_12pyrobuf_list_DoubleList.append = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_DoubleList *, double, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_10DoubleList_append;
  __pyx_vtable_12pyrobuf_list_DoubleList.extend = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_DoubleList *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_10DoubleList_extend;
  __pyx_vtable_12pyrobuf_list_DoubleList.insert = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_DoubleList *, int, double, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_10DoubleList_insert;
  __pyx_vtable_12pyrobuf_list_DoubleList.pop = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_DoubleList *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_10DoubleList_pop;
  __pyx_vtable_12pyrobuf_list_DoubleList.remove = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_DoubleList *, double, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_10DoubleList_remove;
  if (PyType_Ready(&__pyx_type_12pyrobuf_list_DoubleList) < 0) __PYX_ERR(0, 241, __pyx_L1_error)
  __pyx_type_12pyrobuf_list_DoubleList.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list_DoubleList.tp_dictoffset && __pyx_type_12pyrobuf_list_DoubleList.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list_DoubleList.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12pyrobuf_list_DoubleList.tp_dict, __pyx_vtabptr_12pyrobuf_list_DoubleList) < 0) __PYX_ERR(0, 241, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "DoubleList", (PyObject *)&__pyx_type_12pyrobuf_list_DoubleList) < 0) __PYX_ERR(0, 241, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12pyrobuf_list_DoubleList) < 0) __PYX_ERR(0, 241, __pyx_L1_error)
  __pyx_ptype_12pyrobuf_list_DoubleList = &__pyx_type_12pyrobuf_list_DoubleList;
  __pyx_vtabptr_12pyrobuf_list_FloatList = &__pyx_vtable_12pyrobuf_list_FloatList;
  __pyx_vtable_12pyrobuf_list_FloatList._append = (void (*)(struct __pyx_obj_12pyrobuf_list_FloatList *, float))__pyx_f_12pyrobuf_list_9FloatList__append;
  __pyx_vtable_12pyrobuf_list_FloatList._equal = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_FloatList *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_9FloatList__equal;
  __pyx_vtable_12pyrobuf_list_FloatList.append = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_FloatList *, float, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_9FloatList_append;
  __pyx_vtable_12pyrobuf_list_FloatList.extend = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_FloatList *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_9FloatList_extend;
  __pyx_vtable_12pyrobuf_list_FloatList.insert = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_FloatList *, int, float, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_9FloatList_insert;
  __pyx_vtable_12pyrobuf_list_FloatList.pop = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_FloatList *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_9FloatList_pop;
  __pyx_vtable_12pyrobuf_list_FloatList.remove = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_FloatList *, float, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_9FloatList_remove;
  if (PyType_Ready(&__pyx_type_12pyrobuf_list_FloatList) < 0) __PYX_ERR(0, 451, __pyx_L1_error)
  __pyx_type_12pyrobuf_list_FloatList.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list_FloatList.tp_dictoffset && __pyx_type_12pyrobuf_list_FloatList.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list_FloatList.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12pyrobuf_list_FloatList.tp_dict, __pyx_vtabptr_12pyrobuf_list_FloatList) < 0) __PYX_ERR(0, 451, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "FloatList", (PyObject *)&__pyx_type_12pyrobuf_list_FloatList) < 0) __PYX_ERR(0, 451, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12pyrobuf_list_FloatList) < 0) __PYX_ERR(0, 451, __pyx_L1_error)
  __pyx_ptype_12pyrobuf_list_FloatList = &__pyx_type_12pyrobuf_list_FloatList;
  __pyx_vtabptr_12pyrobuf_list_IntList = &__pyx_vtable_12pyrobuf_list_IntList;
  __pyx_vtable_12pyrobuf_list_IntList._append = (void (*)(struct __pyx_obj_12pyrobuf_list_IntList *, int))__pyx_f_12pyrobuf_list_7IntList__append;
  __pyx_vtable_12pyrobuf_list_IntList._equal = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_IntList *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_7IntList__equal;
  __pyx_vtable_12pyrobuf_list_IntList.append = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_IntList *, int, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_7IntList_append;
  __pyx_vtable_12pyrobuf_list_IntList.extend = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_IntList *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_7IntList_extend;
  __pyx_vtable_12pyrobuf_list_IntList.insert = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_IntList *, int, int, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_7IntList_insert;
  __pyx_vtable_12pyrobuf_list_IntList.pop = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_IntList *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_7IntList_pop;
  __pyx_vtable_12pyrobuf_list_IntList.remove = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_IntList *, int, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_7IntList_remove;
  if (PyType_Ready(&__pyx_type_12pyrobuf_list_IntList) < 0) __PYX_ERR(0, 661, __pyx_L1_error)
  __pyx_type_12pyrobuf_list_IntList.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list_IntList.tp_dictoffset && __pyx_type_12pyrobuf_list_IntList.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list_IntList.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12pyrobuf_list_IntList.tp_dict, __pyx_vtabptr_12pyrobuf_list_IntList) < 0) __PYX_ERR(0, 661, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "IntList", (PyObject *)&__pyx_type_12pyrobuf_list_IntList) < 0) __PYX_ERR(0, 661, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12pyrobuf_list_IntList) < 0) __PYX_ERR(0, 661, __pyx_L1_error)
  __pyx_ptype_12pyrobuf_list_IntList = &__pyx_type_12pyrobuf_list_IntList;
  __pyx_vtabptr_12pyrobuf_list_Int32List = &__pyx_vtable_12pyrobuf_list_Int32List;
  __pyx_vtable_12pyrobuf_list_Int32List._append = (void (*)(struct __pyx_obj_12pyrobuf_list_Int32List *, int32_t))__pyx_f_12pyrobuf_list_9Int32List__append;
  __pyx_vtable_12pyrobuf_list_Int32List._equal = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Int32List *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_9Int32List__equal;
  __pyx_vtable_12pyrobuf_list_Int32List.append = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Int32List *, int32_t, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_9Int32List_append;
  __pyx_vtable_12pyrobuf_list_Int32List.extend = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Int32List *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_9Int32List_extend;
  __pyx_vtable_12pyrobuf_list_Int32List.insert = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Int32List *, int, int32_t, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_9Int32List_insert;
  __pyx_vtable_12pyrobuf_list_Int32List.pop = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Int32List *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_9Int32List_pop;
  __pyx_vtable_12pyrobuf_list_Int32List.remove = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Int32List *, int32_t, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_9Int32List_remove;
  if (PyType_Ready(&__pyx_type_12pyrobuf_list_Int32List) < 0) __PYX_ERR(0, 871, __pyx_L1_error)
  __pyx_type_12pyrobuf_list_Int32List.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list_Int32List.tp_dictoffset && __pyx_type_12pyrobuf_list_Int32List.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list_Int32List.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12pyrobuf_list_Int32List.tp_dict, __pyx_vtabptr_12pyrobuf_list_Int32List) < 0) __PYX_ERR(0, 871, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "Int32List", (PyObject *)&__pyx_type_12pyrobuf_list_Int32List) < 0) __PYX_ERR(0, 871, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12pyrobuf_list_Int32List) < 0) __PYX_ERR(0, 871, __pyx_L1_error)
  __pyx_ptype_12pyrobuf_list_Int32List = &__pyx_type_12pyrobuf_list_Int32List;
  __pyx_vtabptr_12pyrobuf_list_Uint32List = &__pyx_vtable_12pyrobuf_list_Uint32List;
  __pyx_vtable_12pyrobuf_list_Uint32List._append = (void (*)(struct __pyx_obj_12pyrobuf_list_Uint32List *, uint32_t))__pyx_f_12pyrobuf_list_10Uint32List__append;
  __pyx_vtable_12pyrobuf_list_Uint32List._equal = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Uint32List *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_10Uint32List__equal;
  __pyx_vtable_12pyrobuf_list_Uint32List.append = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Uint32List *, uint32_t, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_10Uint32List_append;
  __pyx_vtable_12pyrobuf_list_Uint32List.extend = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Uint32List *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_10Uint32List_extend;
  __pyx_vtable_12pyrobuf_list_Uint32List.insert = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Uint32List *, int, uint32_t, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_10Uint32List_insert;
  __pyx_vtable_12pyrobuf_list_Uint32List.pop = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Uint32List *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_10Uint32List_pop;
  __pyx_vtable_12pyrobuf_list_Uint32List.remove = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Uint32List *, uint32_t, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_10Uint32List_remove;
  if (PyType_Ready(&__pyx_type_12pyrobuf_list_Uint32List) < 0) __PYX_ERR(0, 1081, __pyx_L1_error)
  __pyx_type_12pyrobuf_list_Uint32List.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list_Uint32List.tp_dictoffset && __pyx_type_12pyrobuf_list_Uint32List.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list_Uint32List.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12pyrobuf_list_Uint32List.tp_dict, __pyx_vtabptr_12pyrobuf_list_Uint32List) < 0) __PYX_ERR(0, 1081, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "Uint32List", (PyObject *)&__pyx_type_12pyrobuf_list_Uint32List) < 0) __PYX_ERR(0, 1081, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12pyrobuf_list_Uint32List) < 0) __PYX_ERR(0, 1081, __pyx_L1_error)
  __pyx_ptype_12pyrobuf_list_Uint32List = &__pyx_type_12pyrobuf_list_Uint32List;
  __pyx_vtabptr_12pyrobuf_list_Int64List = &__pyx_vtable_12pyrobuf_list_Int64List;
  __pyx_vtable_12pyrobuf_list_Int64List._append = (void (*)(struct __pyx_obj_12pyrobuf_list_Int64List *, int64_t))__pyx_f_12pyrobuf_list_9Int64List__append;
  __pyx_vtable_12pyrobuf_list_Int64List._equal = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Int64List *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_9Int64List__equal;
  __pyx_vtable_12pyrobuf_list_Int64List.append = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Int64List *, int64_t, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_9Int64List_append;
  __pyx_vtable_12pyrobuf_list_Int64List.extend = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Int64List *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_9Int64List_extend;
  __pyx_vtable_12pyrobuf_list_Int64List.insert = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Int64List *, int, int64_t, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_9Int64List_insert;
  __pyx_vtable_12pyrobuf_list_Int64List.pop = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Int64List *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_9Int64List_pop;
  __pyx_vtable_12pyrobuf_list_Int64List.remove = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Int64List *, int64_t, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_9Int64List_remove;
  if (PyType_Ready(&__pyx_type_12pyrobuf_list_Int64List) < 0) __PYX_ERR(0, 1291, __pyx_L1_error)
  __pyx_type_12pyrobuf_list_Int64List.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list_Int64List.tp_dictoffset && __pyx_type_12pyrobuf_list_Int64List.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list_Int64List.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12pyrobuf_list_Int64List.tp_dict, __pyx_vtabptr_12pyrobuf_list_Int64List) < 0) __PYX_ERR(0, 1291, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "Int64List", (PyObject *)&__pyx_type_12pyrobuf_list_Int64List) < 0) __PYX_ERR(0, 1291, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12pyrobuf_list_Int64List) < 0) __PYX_ERR(0, 1291, __pyx_L1_error)
  __pyx_ptype_12pyrobuf_list_Int64List = &__pyx_type_12pyrobuf_list_Int64List;
  __pyx_vtabptr_12pyrobuf_list_Uint64List = &__pyx_vtable_12pyrobuf_list_Uint64List;
  __pyx_vtable_12pyrobuf_list_Uint64List._append = (void (*)(struct __pyx_obj_12pyrobuf_list_Uint64List *, uint64_t))__pyx_f_12pyrobuf_list_10Uint64List__append;
  __pyx_vtable_12pyrobuf_list_Uint64List._equal = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Uint64List *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_10Uint64List__equal;
  __pyx_vtable_12pyrobuf_list_Uint64List.append = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Uint64List *, uint64_t, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_10Uint64List_append;
  __pyx_vtable_12pyrobuf_list_Uint64List.extend = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Uint64List *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_10Uint64List_extend;
  __pyx_vtable_12pyrobuf_list_Uint64List.insert = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Uint64List *, int, uint64_t, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_10Uint64List_insert;
  __pyx_vtable_12pyrobuf_list_Uint64List.pop = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Uint64List *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_10Uint64List_pop;
  __pyx_vtable_12pyrobuf_list_Uint64List.remove = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_Uint64List *, uint64_t, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_10Uint64List_remove;
  if (PyType_Ready(&__pyx_type_12pyrobuf_list_Uint64List) < 0) __PYX_ERR(0, 1501, __pyx_L1_error)
  __pyx_type_12pyrobuf_list_Uint64List.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list_Uint64List.tp_dictoffset && __pyx_type_12pyrobuf_list_Uint64List.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list_Uint64List.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12pyrobuf_list_Uint64List.tp_dict, __pyx_vtabptr_12pyrobuf_list_Uint64List) < 0) __PYX_ERR(0, 1501, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "Uint64List", (PyObject *)&__pyx_type_12pyrobuf_list_Uint64List) < 0) __PYX_ERR(0, 1501, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12pyrobuf_list_Uint64List) < 0) __PYX_ERR(0, 1501, __pyx_L1_error)
  __pyx_ptype_12pyrobuf_list_Uint64List = &__pyx_type_12pyrobuf_list_Uint64List;
  __pyx_vtabptr_12pyrobuf_list_CharList = &__pyx_vtable_12pyrobuf_list_CharList;
  __pyx_vtable_12pyrobuf_list_CharList._append = (void (*)(struct __pyx_obj_12pyrobuf_list_CharList *, char))__pyx_f_12pyrobuf_list_8CharList__append;
  __pyx_vtable_12pyrobuf_list_CharList._equal = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_CharList *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_8CharList__equal;
  __pyx_vtable_12pyrobuf_list_CharList.append = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_CharList *, char, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_8CharList_append;
  __pyx_vtable_12pyrobuf_list_CharList.extend = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_CharList *, PyObject *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_8CharList_extend;
  __pyx_vtable_12pyrobuf_list_CharList.insert = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_CharList *, int, char, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_8CharList_insert;
  __pyx_vtable_12pyrobuf_list_CharList.pop = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_CharList *, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_8CharList_pop;
  __pyx_vtable_12pyrobuf_list_CharList.remove = (PyObject *(*)(struct __pyx_obj_12pyrobuf_list_CharList *, char, int __pyx_skip_dispatch))__pyx_f_12pyrobuf_list_8CharList_remove;
  if (PyType_Ready(&__pyx_type_12pyrobuf_list_CharList) < 0) __PYX_ERR(0, 1711, __pyx_L1_error)
  __pyx_type_12pyrobuf_list_CharList.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list_CharList.tp_dictoffset && __pyx_type_12pyrobuf_list_CharList.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list_CharList.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_12pyrobuf_list_CharList.tp_dict, __pyx_vtabptr_12pyrobuf_list_CharList) < 0) __PYX_ERR(0, 1711, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "CharList", (PyObject *)&__pyx_type_12pyrobuf_list_CharList) < 0) __PYX_ERR(0, 1711, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_12pyrobuf_list_CharList) < 0) __PYX_ERR(0, 1711, __pyx_L1_error)
  __pyx_ptype_12pyrobuf_list_CharList = &__pyx_type_12pyrobuf_list_CharList;
  if (PyType_Ready(&__pyx_type_12pyrobuf_list___pyx_scope_struct____iter__) < 0) __PYX_ERR(0, 291, __pyx_L1_error)
  __pyx_type_12pyrobuf_list___pyx_scope_struct____iter__.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list___pyx_scope_struct____iter__.tp_dictoffset && __pyx_type_12pyrobuf_list___pyx_scope_struct____iter__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list___pyx_scope_struct____iter__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_12pyrobuf_list___pyx_scope_struct____iter__ = &__pyx_type_12pyrobuf_list___pyx_scope_struct____iter__;
  if (PyType_Ready(&__pyx_type_12pyrobuf_list___pyx_scope_struct_1___iter__) < 0) __PYX_ERR(0, 501, __pyx_L1_error)
  __pyx_type_12pyrobuf_list___pyx_scope_struct_1___iter__.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list___pyx_scope_struct_1___iter__.tp_dictoffset && __pyx_type_12pyrobuf_list___pyx_scope_struct_1___iter__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list___pyx_scope_struct_1___iter__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_12pyrobuf_list___pyx_scope_struct_1___iter__ = &__pyx_type_12pyrobuf_list___pyx_scope_struct_1___iter__;
  if (PyType_Ready(&__pyx_type_12pyrobuf_list___pyx_scope_struct_2___iter__) < 0) __PYX_ERR(0, 711, __pyx_L1_error)
  __pyx_type_12pyrobuf_list___pyx_scope_struct_2___iter__.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list___pyx_scope_struct_2___iter__.tp_dictoffset && __pyx_type_12pyrobuf_list___pyx_scope_struct_2___iter__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list___pyx_scope_struct_2___iter__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_12pyrobuf_list___pyx_scope_struct_2___iter__ = &__pyx_type_12pyrobuf_list___pyx_scope_struct_2___iter__;
  if (PyType_Ready(&__pyx_type_12pyrobuf_list___pyx_scope_struct_3___iter__) < 0) __PYX_ERR(0, 921, __pyx_L1_error)
  __pyx_type_12pyrobuf_list___pyx_scope_struct_3___iter__.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list___pyx_scope_struct_3___iter__.tp_dictoffset && __pyx_type_12pyrobuf_list___pyx_scope_struct_3___iter__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list___pyx_scope_struct_3___iter__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_12pyrobuf_list___pyx_scope_struct_3___iter__ = &__pyx_type_12pyrobuf_list___pyx_scope_struct_3___iter__;
  if (PyType_Ready(&__pyx_type_12pyrobuf_list___pyx_scope_struct_4___iter__) < 0) __PYX_ERR(0, 1131, __pyx_L1_error)
  __pyx_type_12pyrobuf_list___pyx_scope_struct_4___iter__.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list___pyx_scope_struct_4___iter__.tp_dictoffset && __pyx_type_12pyrobuf_list___pyx_scope_struct_4___iter__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list___pyx_scope_struct_4___iter__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_12pyrobuf_list___pyx_scope_struct_4___iter__ = &__pyx_type_12pyrobuf_list___pyx_scope_struct_4___iter__;
  if (PyType_Ready(&__pyx_type_12pyrobuf_list___pyx_scope_struct_5___iter__) < 0) __PYX_ERR(0, 1341, __pyx_L1_error)
  __pyx_type_12pyrobuf_list___pyx_scope_struct_5___iter__.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list___pyx_scope_struct_5___iter__.tp_dictoffset && __pyx_type_12pyrobuf_list___pyx_scope_struct_5___iter__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list___pyx_scope_struct_5___iter__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_12pyrobuf_list___pyx_scope_struct_5___iter__ = &__pyx_type_12pyrobuf_list___pyx_scope_struct_5___iter__;
  if (PyType_Ready(&__pyx_type_12pyrobuf_list___pyx_scope_struct_6___iter__) < 0) __PYX_ERR(0, 1551, __pyx_L1_error)
  __pyx_type_12pyrobuf_list___pyx_scope_struct_6___iter__.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list___pyx_scope_struct_6___iter__.tp_dictoffset && __pyx_type_12pyrobuf_list___pyx_scope_struct_6___iter__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list___pyx_scope_struct_6___iter__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_12pyrobuf_list___pyx_scope_struct_6___iter__ = &__pyx_type_12pyrobuf_list___pyx_scope_struct_6___iter__;
  if (PyType_Ready(&__pyx_type_12pyrobuf_list___pyx_scope_struct_7___iter__) < 0) __PYX_ERR(0, 1761, __pyx_L1_error)
  __pyx_type_12pyrobuf_list___pyx_scope_struct_7___iter__.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_12pyrobuf_list___pyx_scope_struct_7___iter__.tp_dictoffset && __pyx_type_12pyrobuf_list___pyx_scope_struct_7___iter__.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_12pyrobuf_list___pyx_scope_struct_7___iter__.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_12pyrobuf_list___pyx_scope_struct_7___iter__ = &__pyx_type_12pyrobuf_list___pyx_scope_struct_7___iter__;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  0); if (unlikely(!__pyx_ptype_7cpython_4type_type)) __PYX_ERR(2, 9, __pyx_L1_error)
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), 0); if (unlikely(!__pyx_ptype_7cpython_4bool_bool)) __PYX_ERR(3, 8, __pyx_L1_error)
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), 0); if (unlikely(!__pyx_ptype_7cpython_7complex_complex)) __PYX_ERR(4, 15, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (!(defined(__cplusplus)) || (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ > 4)))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initpyrobuf_list(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initpyrobuf_list(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_pyrobuf_list(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_pyrobuf_list(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        result = PyDict_SetItemString(moddict, to_name, value);
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__") < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static int __pyx_pymod_exec_pyrobuf_list(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m && __pyx_m == __pyx_pyinit_module) return 0;
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_pyrobuf_list(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("pyrobuf_list", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_pyrobuf_list) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pyrobuf_list")) {
      if (unlikely(PyDict_SetItemString(modules, "pyrobuf_list", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  if (unlikely(__Pyx_modinit_variable_export_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "pyrobuf_list.pyx":15
 *         return
 * 
 * cdef public NullListener null_listener = NullListener()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_12pyrobuf_list_NullListener)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XGOTREF(((PyObject *)null_listener));
  __Pyx_DECREF_SET(null_listener, ((struct __pyx_obj_12pyrobuf_list_NullListener *)__pyx_t_1));
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_NullListener(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_12pyrobuf_list_1__pyx_unpickle_NullListener, NULL, __pyx_n_s_pyrobuf_list); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_NullListener, __pyx_t_1) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyrobuf_list.pyx":1
 * from cpython.mem cimport PyMem_Malloc, PyMem_Realloc, PyMem_Free             # <<<<<<<<<<<<<<
 * from libc.stdint cimport *
 * from cpython.ref cimport PyObject
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init pyrobuf_list", 0, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pyrobuf_list");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
#include "frameobject.h"
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = f->f_localsplus;
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyErrFetchRestore */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
  static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyErrExceptionMatches */
  #if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr */
  static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
  static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* GetModuleGlobalName */
  static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    if (likely(result)) {
        Py_INCREF(result);
    } else if (unlikely(PyErr_Occurred())) {
        result = NULL;
    } else {
#else
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* RaiseDoubleKeywords */
      static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
      static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
      static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* ArgTypeTest */
      static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* SaveResetException */
      #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if PY_VERSION_HEX >= 0x030700A3
    *type = tstate->exc_state.exc_type;
    *value = tstate->exc_state.exc_value;
    *tb = tstate->exc_state.exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if PY_VERSION_HEX >= 0x030700A3
    tmp_type = tstate->exc_state.exc_type;
    tmp_value = tstate->exc_state.exc_value;
    tmp_tb = tstate->exc_state.exc_traceback;
    tstate->exc_state.exc_type = type;
    tstate->exc_state.exc_value = value;
    tstate->exc_state.exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
      #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb) {
#endif
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if PY_VERSION_HEX >= 0x030700A3
    tmp_type = tstate->exc_state.exc_type;
    tmp_value = tstate->exc_state.exc_value;
    tmp_tb = tstate->exc_state.exc_traceback;
    tstate->exc_state.exc_type = local_type;
    tstate->exc_state.exc_value = local_value;
    tstate->exc_state.exc_traceback = local_tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* RaiseException */
        #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* KeywordStringCheck */
        static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* PyObjectCallMethod1 */
        static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = NULL;
#if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(method))) {
        PyObject *self = PyMethod_GET_SELF(method);
        if (likely(self)) {
            PyObject *args;
            PyObject *function = PyMethod_GET_FUNCTION(method);
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(function)) {
                PyObject *args[2] = {self, arg};
                result = __Pyx_PyFunction_FastCall(function, args, 2);
                goto done;
            }
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(function)) {
                PyObject *args[2] = {self, arg};
                result = __Pyx_PyCFunction_FastCall(function, args, 2);
                goto done;
            }
            #endif
            args = PyTuple_New(2);
            if (unlikely(!args)) goto done;
            Py_INCREF(self);
            PyTuple_SET_ITEM(args, 0, self);
            Py_INCREF(arg);
            PyTuple_SET_ITEM(args, 1, arg);
            Py_INCREF(function);
            result = __Pyx_PyObject_Call(function, args, NULL);
            Py_DECREF(args);
            Py_DECREF(function);
            return result;
        }
    }
#endif
    result = __Pyx_PyObject_CallOneArg(method, arg);
    goto done;
done:
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method, *result;
    method = __Pyx_PyObject_GetAttrStr(obj, method_name);
    if (unlikely(!method)) return NULL;
    result = __Pyx__PyObject_CallMethod1(method, arg);
    Py_DECREF(method);
    return result;
}

/* append */
        static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* RaiseTooManyValuesToUnpack */
        static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
        static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
        static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
        static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* Import */
        static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
        static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
        static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* GetItemInt */
        static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely((0 <= wrapped_i) & (wrapped_i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely((0 <= wrapped_i) & (wrapped_i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely((n >= 0) & (n < PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyObject_GenericGetAttrNoDict */
        #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
        #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
        static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* SetupReduce */
        static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto GOOD;
BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* CLineInTraceback */
        #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
      use_cline = __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback);
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (PyObject_Not(use_cline) != 0) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
        static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
        #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
        #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int32_t(int32_t value) {
    const int32_t neg_one = (int32_t) -1, const_zero = (int32_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int32_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int32_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int32_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value) {
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint32_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint32_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint32_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value) {
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint64_t(uint64_t value) {
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(uint64_t),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_char(char value) {
    const char neg_one = (char) -1, const_zero = (char) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(char) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(char) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(char) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(char),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
        static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE int32_t __Pyx_PyInt_As_int32_t(PyObject *x) {
    const int32_t neg_one = (int32_t) -1, const_zero = (int32_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int32_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int32_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int32_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int32_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int32_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(int32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) >= 2 * PyLong_SHIFT) {
                            return (int32_t) (((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) >= 3 * PyLong_SHIFT) {
                            return (int32_t) (((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) >= 4 * PyLong_SHIFT) {
                            return (int32_t) (((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int32_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int32_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int32_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int32_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int32_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int32_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int32_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int32_t) (((int32_t)-1)*(((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int32_t) ((((((int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int32_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int32_t) (((int32_t)-1)*(((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int32_t) ((((((((int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int32_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int32_t) (((int32_t)-1)*(((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int32_t) ((((((((((int32_t)digits[3]) << PyLong_SHIFT) | (int32_t)digits[2]) << PyLong_SHIFT) | (int32_t)digits[1]) << PyLong_SHIFT) | (int32_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int32_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int32_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int32_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int32_t) -1;
        }
    } else {
        int32_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int32_t) -1;
        val = __Pyx_PyInt_As_int32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int32_t");
    return (int32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int32_t");
    return (int32_t) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *x) {
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint32_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint32_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint32_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint32_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 2 * PyLong_SHIFT) {
                            return (uint32_t) (((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 3 * PyLong_SHIFT) {
                            return (uint32_t) (((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) >= 4 * PyLong_SHIFT) {
                            return (uint32_t) (((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint32_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint32_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint32_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint32_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint32_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint32_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint32_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint32_t) ((((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint32_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint32_t) ((((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint32_t) (((uint32_t)-1)*(((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint32_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint32_t) ((((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint32_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint32_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint32_t) -1;
        }
    } else {
        uint32_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint32_t) -1;
        val = __Pyx_PyInt_As_uint32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint32_t");
    return (uint32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint32_t");
    return (uint32_t) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *x) {
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int64_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int64_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int64_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int64_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(int64_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) >= 2 * PyLong_SHIFT) {
                            return (int64_t) (((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int64_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) >= 3 * PyLong_SHIFT) {
                            return (int64_t) (((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int64_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) >= 4 * PyLong_SHIFT) {
                            return (int64_t) (((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int64_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int64_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int64_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int64_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int64_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int64_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int64_t) (((int64_t)-1)*(((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int64_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int64_t) ((((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int64_t) (((int64_t)-1)*(((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int64_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int64_t) ((((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int64_t) (((int64_t)-1)*(((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int64_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int64_t) ((((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int64_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int64_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int64_t) -1;
        }
    } else {
        int64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int64_t) -1;
        val = __Pyx_PyInt_As_int64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int64_t");
    return (int64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int64_t");
    return (int64_t) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE uint64_t __Pyx_PyInt_As_uint64_t(PyObject *x) {
    const uint64_t neg_one = (uint64_t) -1, const_zero = (uint64_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(uint64_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(uint64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint64_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint64_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(uint64_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(uint64_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) >= 2 * PyLong_SHIFT) {
                            return (uint64_t) (((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint64_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) >= 3 * PyLong_SHIFT) {
                            return (uint64_t) (((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint64_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) >= 4 * PyLong_SHIFT) {
                            return (uint64_t) (((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (uint64_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(uint64_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint64_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (uint64_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(uint64_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(uint64_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(uint64_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint64_t) (((uint64_t)-1)*(((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(uint64_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT) {
                            return (uint64_t) ((((((uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(uint64_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint64_t) (((uint64_t)-1)*(((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(uint64_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT) {
                            return (uint64_t) ((((((((uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(uint64_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint64_t) (((uint64_t)-1)*(((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(uint64_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(uint64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(uint64_t) - 1 > 4 * PyLong_SHIFT) {
                            return (uint64_t) ((((((((((uint64_t)digits[3]) << PyLong_SHIFT) | (uint64_t)digits[2]) << PyLong_SHIFT) | (uint64_t)digits[1]) << PyLong_SHIFT) | (uint64_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(uint64_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(uint64_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(uint64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            uint64_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (uint64_t) -1;
        }
    } else {
        uint64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint64_t) -1;
        val = __Pyx_PyInt_As_uint64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint64_t");
    return (uint64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint64_t");
    return (uint64_t) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *x) {
    const char neg_one = (char) -1, const_zero = (char) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(char, digit, digits[0])
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 2 * PyLong_SHIFT) {
                            return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 3 * PyLong_SHIFT) {
                            return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 4 * PyLong_SHIFT) {
                            return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(char, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (char) -1;
        }
    } else {
        char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyInt_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
        #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* FetchCommonType */
        static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* SwapException */
        #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if PY_VERSION_HEX >= 0x030700A3
    tmp_type = tstate->exc_state.exc_type;
    tmp_value = tstate->exc_state.exc_value;
    tmp_tb = tstate->exc_state.exc_traceback;
    tstate->exc_state.exc_type = *type;
    tstate->exc_state.exc_value = *value;
    tstate->exc_state.exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* CoroutineBase */
        #include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__pyx_CoroutineObject *self) {
    PyObject *exc_type = self->exc_type;
    PyObject *exc_value = self->exc_value;
    PyObject *exc_traceback = self->exc_traceback;
    self->exc_type = NULL;
    self->exc_value = NULL;
    self->exc_traceback = NULL;
    Py_XDECREF(exc_type);
    Py_XDECREF(exc_value);
    Py_XDECREF(exc_traceback);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if (0) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if (0) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    if (self->exc_type) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        if (self->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) self->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
#endif
        __Pyx_ExceptionSwap(&self->exc_type, &self->exc_value,
                            &self->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(self);
        __Pyx_ExceptionSave(&self->exc_type, &self->exc_value, &self->exc_traceback);
    }
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__pyx_CoroutineObject *self) {
    if (likely(self->exc_traceback)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) self->exc_traceback;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    Py_VISIT(gen->exc_type);
    Py_VISIT(gen->exc_value);
    Py_VISIT(gen->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    Py_CLEAR(gen->exc_type);
    Py_CLEAR(gen->exc_value);
    Py_CLEAR(gen->exc_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    self->ob_refcnt = 1;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        self->ob_refcnt = refcnt;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->exc_type = NULL;
    gen->exc_value = NULL;
    gen->exc_traceback = NULL;
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
            static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
            #ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
            static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
            static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* PyObjectSetAttrStr */
            #if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* VoidPtrExport */
            static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig) {
    PyObject *d;
    PyObject *cobj = 0;
    d = PyDict_GetItem(__pyx_d, __pyx_n_s_pyx_capi);
    Py_XINCREF(d);
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (__Pyx_PyObject_SetAttrStr(__pyx_m, __pyx_n_s_pyx_capi, d) < 0)
            goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItem(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* ModuleImport */
            #ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

/* TypeImport */
            #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s has the wrong size, try recompiling. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

/* InitStrings */
            static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
